/*
===============================================================================

  FILE:  geo_structures.cpp

  CONTENTS:

    Implementation of GMSH structures


  PROGRAMMERS:

    clement.courbet@ecp.fr

  COPYRIGHT:

    copyright (C) 2009 clement courbet

    This software is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.

  CHANGE HISTORY:

    23 November 2009 -- Created

===============================================================================
*/
#include <math.h>
#include <assert.h>
#include <stdio.h>
#include <string.h>

#include <iostream.h>
#include <list.h>

#include "geo_structures.h"

//#define PRINT_DEBUG_OUTPUT

/**
 * Curve functions
 *
 */
using namespace Geo;


inline Point* Curve::operator () ( int i ) const
{
	assert ( i<=subdiv );
	if ( i<1 ) return start;
	else if ( i>=subdiv ) return end;
	else return mesh[i-1];
}


inline void Curve::addSurface()
{
#ifdef PRINT_DEBUG_OUTPUT
	fprintf ( stderr, "adding surface to curve %i\n",  this->ind() );
#endif
	nSurfaces++;
}

/**
 * Line functions
 *
 */

void Line::generateMesh()
{
#ifdef PRINT_DEBUG_OUTPUT
	fprintf ( stderr, "\t\tBEGIN Line::generateMesh()\n" );
#endif
	//if we are already visited, no need to remesh.
	if ( !visited )
	{
		for ( int i=1; i<subdiv; ++i )
		{
			Point *p = new Point ( -1, Point::TYPE_ON_CURVE, nSurfaces );
			//there are two surfaces per block that use this curve, and each block has two hexas that reference this vertex, so the nomber of slots is nSurfaces.
			getAtI ( i, p->pos );
			mesh.push_back ( p );
		}
		visited=true;
	}
#ifdef PRINT_DEBUG_OUTPUT
	fprintf ( stderr, "\t\tEND Line::generateMesh()\n" );
#endif
}

inline void Line::getAtI ( int i, float *p )
{
	p[0] = ( ( subdiv-i ) *start->pos[0] + i*end->pos[0] ) /subdiv;
	p[1] = ( ( subdiv-i ) *start->pos[1] + i*end->pos[1] ) /subdiv;
	p[2] = ( ( subdiv-i ) *start->pos[2] + i*end->pos[2] ) /subdiv;
}

/**
 * Circle functions
 *
 */

Circle::Circle ( int index, Point *s, Point *c, Point *e ) : Curve ( index, s, e ), center ( c )
{
	//compute radius
	float t = start->pos[0]-center->pos[0];
	R2 = t*t;
	t = start->pos[1]-center->pos[1];
	R2 += t*t;
	t = start->pos[2]-center->pos[2];
	R2 += t*t;
	//compute interior angle
	theta = acosf ( ( ( start->pos[0]-center->pos[0] ) * ( end->pos[0]-center->pos[0] ) + ( start->pos[1]-center->pos[1] ) * ( end->pos[1]-center->pos[1] ) + ( start->pos[2]-center->pos[2] ) * ( end->pos[2]-center->pos[2] ) ) /R2 );
};

void Circle::generateMesh()
{
#ifdef PRINT_DEBUG_OUTPUT
	fprintf ( stderr, "\t\tBEGIN Circle::generateMesh()\n" );
#endif
	//if we are already visited, no need to remesh.
	if ( !visited )
	{
		//compute normal vector
		float v0 = ( start->pos[1]-center->pos[1] ) * ( end->pos[2]-center->pos[2] ) - ( start->pos[2]-center->pos[2] ) * ( end->pos[1]-center->pos[1] );
		float v1 = - ( start->pos[0]-center->pos[0] ) * ( end->pos[2]-center->pos[2] ) + ( start->pos[2]-center->pos[2] ) * ( end->pos[0]-center->pos[0] );
		float v2 = ( start->pos[0]-center->pos[0] ) * ( end->pos[1]-center->pos[1] ) - ( start->pos[1]-center->pos[1] ) * ( end->pos[0]-center->pos[0] );
		//normalize it
		float norm = sqrt ( v0*v0+v1*v1+v2*v2 );
		v0 /= norm;
		v1 /= norm;
		v2 /= norm;

		for ( int i=1; i<subdiv; ++i )
		{
			//compute rotation matrix
			float R[9];
			float c =  cosf ( theta* ( ( double ) i/subdiv ) );
			float t =  1 - c;
			float s =  sinf ( theta* ( ( double ) i/subdiv ) );
			R[0] = c + v0*v0*t; //0,0
			R[4] = c + v1*v1*t; //1,1
			R[8] = c + v2*v2*t; //2,2
			double t1 = v0*v0*t;
			double t2 = v2*s;
			R[3] = t1 + t2;
			R[1] = t1 - t2;
			t1 = v0*v2*t;
			t2 = v1*s;
			R[6] = t1 - t2;
			R[2] = t1 + t2;
			t1 = v1*v2*t;
			t2 = v0*s;
			R[7] = t1 + t2;
			R[5] = t1 - t2;
			float cs[3];
			cs[0] = start->pos[0]-center->pos[0];
			cs[1] = start->pos[1]-center->pos[1];
			cs[2] = start->pos[2]-center->pos[2];
			//rotate vector cs
			matMul ( R, cs );
			//retreive the resulting vector endpoint
			Point *p = new Point ( -1, Point::TYPE_ON_CURVE, nSurfaces );
			//there are two surfaces per block that use this curve, and each block has two hexas that reference this vertex, so the nomber of slots is nSurfaces.
			p->pos[0] = cs[0] + center->pos[0];
			p->pos[1] = cs[1] + center->pos[1];
			p->pos[2] = cs[2] + center->pos[2];
			mesh.push_back ( p );
		}
		visited=true;
	}
#ifdef PRINT_DEBUG_OUTPUT
	fprintf ( stderr, "\t\tEND Circle::generateMesh()\n" );
#endif
}


/**
 * Surface functions
 *
 */
Surface::Surface ( int i, Curve **cp, bool *r )
		: geo_index ( i ), curves ( cp ), reverse ( r ), visited ( false )
{
	blocks[0] = NULL;
	blocks[1] = NULL;
}

inline Block* Surface::getBlock ( int bNum )
{
	assert ( bNum==0 || bNum==1 );
	return blocks[bNum];
}

inline void Surface::addBlock ( Block* b )
{
#ifdef PRINT_DEBUG_OUTPUT
	fprintf ( stderr, "adding block %i to surface %i\n", b->ind(), this->ind() );
#endif
	assert ( blocks[1]==NULL );
	if ( blocks[0] )
		blocks[1] = b;
	else
		blocks[0] = b;

	for ( int i=0; i<4; ++i )
		curves[i]->addSurface();
}

Surface::~Surface()
{
	delete[] reverse;
	delete[] curves;
}

Point* Surface::operator () ( int x, int y )
{
	int subdivX = curves[0]->getSubdiv();
	int subdivY = curves[1]->getSubdiv();
	assert ( x<=subdivX );
	assert ( y<=subdivY );
	if ( x<1 && y<1 )
		return reverse[0] ? curves[0]->endPoint() : curves[0]->startPoint();
	else if ( x>=subdivX && y<1 )
		return reverse[1] ? curves[1]->endPoint() : curves[1]->startPoint();
	else if ( x>=subdivX && y>=subdivY )
		return reverse[2] ? curves[2]->endPoint() : curves[2]->startPoint();
	else if ( x<1 && y>=subdivY )
		return reverse[3] ? curves[3]->endPoint() : curves[3]->startPoint();
	else if ( y<1 )
		return ( *curves[0] ) ( reverse[0]? subdivX-x:x );
	else if ( x>=subdivX )
		return ( *curves[1] ) ( reverse[1]? subdivY-y:y );
	else if ( y>=subdivY )
		return ( *curves[2] ) ( reverse[2]? x: subdivX-x );
	else if ( x<1 )
		return ( *curves[3] ) ( reverse[3]? y: subdivY-y );
	else return mesh[ ( x-1 ) * ( subdivY-1 ) + y-1];
}

void Surface::getCorners ( CyclicPointArray4 &s_idx, bool reverseNormal )
{
	if ( reverseNormal )
	{
		s_idx[0] = reverse[3] ? curves[3]->startPoint() : curves[3]->endPoint();
		s_idx[1] = reverse[2] ? curves[2]->startPoint() : curves[2]->endPoint();
		s_idx[2] = reverse[1] ? curves[1]->startPoint() : curves[1]->endPoint();
		s_idx[3] = reverse[0] ? curves[0]->startPoint() : curves[0]->endPoint();
	}
	else
	{
		s_idx[0] = reverse[0] ? curves[0]->endPoint() : curves[0]->startPoint();
		s_idx[1] = reverse[1] ? curves[1]->endPoint() : curves[1]->startPoint();
		s_idx[2] = reverse[2] ? curves[2]->endPoint() : curves[2]->startPoint();
		s_idx[3] = reverse[3] ? curves[3]->endPoint() : curves[3]->startPoint();
	}
}


void Surface::getCorners ( Point** s_idx, bool reverseNormal )
{
	if ( reverseNormal )
	{
		s_idx[0] = reverse[3] ? curves[3]->startPoint() : curves[3]->endPoint();
		s_idx[1] = reverse[2] ? curves[2]->startPoint() : curves[2]->endPoint();
		s_idx[2] = reverse[1] ? curves[1]->startPoint() : curves[1]->endPoint();
		s_idx[3] = reverse[0] ? curves[0]->startPoint() : curves[0]->endPoint();
	}
	else
	{
		s_idx[0] = reverse[0] ? curves[0]->endPoint() : curves[0]->startPoint();
		s_idx[1] = reverse[1] ? curves[1]->endPoint() : curves[1]->startPoint();
		s_idx[2] = reverse[2] ? curves[2]->endPoint() : curves[2]->startPoint();
		s_idx[3] = reverse[3] ? curves[3]->endPoint() : curves[3]->startPoint();
	}
}

void Surface::getCenterAndNormal ( float *center, float * normal )
{
	Point* corners[4];
	this->getCorners ( corners, false );

	center[0] = 0.25f* ( corners[0]->pos[0]+corners[1]->pos[0]+corners[2]->pos[0]+corners[3]->pos[0] );
	center[1] = 0.25f* ( corners[0]->pos[1]+corners[1]->pos[1]+corners[2]->pos[1]+corners[3]->pos[1] );
	center[2] = 0.25f* ( corners[0]->pos[2]+corners[1]->pos[2]+corners[2]->pos[2]+corners[3]->pos[2] );

	normal[0] = ( corners[0]->pos[1] - center[1] ) * ( corners[1]->pos[2] - center[2] ) - ( corners[0]->pos[2] - center[2] ) * ( corners[1]->pos[1] - center[1] );
	normal[1] = - ( corners[0]->pos[0] - center[0] ) * ( corners[1]->pos[2] - center[2] ) + ( corners[0]->pos[2] - center[2] ) * ( corners[1]->pos[0] - center[0] );
	normal[2] = ( corners[0]->pos[0] - center[0] ) * ( corners[1]->pos[1] - center[1] ) - ( corners[0]->pos[1] - center[1] ) * ( corners[1]->pos[0] - center[0] );
}

void Surface::generateMesh()
{
#ifdef PRINT_DEBUG_OUTPUT
	fprintf ( stderr, "\tBEGIN Surface(%i)::generateMesh()\n", this->ind() );
#endif
	if ( !visited )
	{
		//make sure that all boundary curves are meshed, and that opposing curves have the same subdiv number.
		if ( curves[0]->getSubdiv() != curves[2]->getSubdiv() )
		{
			fprintf ( stderr, "ERROR: Surface %i is not transfinite (curves %i and %i are not equally subdivided)\n", geo_index, curves[0]->ind(), curves[2]->ind() );
			exit ( 1 );
		}
		if ( curves[1]->getSubdiv() != curves[3]->getSubdiv() )
		{
			fprintf ( stderr, "ERROR: Surface %i is not transfinite (curves %i and %i are not equally subdivided)\n", geo_index, curves[1]->ind(), curves[3]->ind() );
			exit ( 1 );
		}

		curves[0]->generateMesh();
		curves[1]->generateMesh();
		curves[2]->generateMesh();
		curves[3]->generateMesh();

		Point* corners[4];
		this->getCorners ( corners, false );

		//now mesh the surface
		int subdivX = curves[0]->getSubdiv();
		int subdivY = curves[1]->getSubdiv();
		//The vertices have 4 slots per block.
		int nSlots = blocks[1] ? 8 : 4;

		for ( int x=1; x<subdivX; ++x )
		{
			for ( int y=1; y<subdivY; ++y )
			{
				Point *p = new Point ( -1, Point::TYPE_ON_SURFACE, nSlots );
				float u = ( float ) x/subdivX;
				float v = ( float ) y/subdivY;
				const Point *e0 = ( *curves[0] ) ( reverse[0]? subdivX-x:x );
				const Point *e1 = ( *curves[1] ) ( reverse[1]? subdivY-y:y );
				const Point *e2 = ( *curves[2] ) ( reverse[2]? x:subdivX-x );
				const Point *e3 = ( *curves[3] ) ( reverse[3]? y:subdivY-y );
				for ( int k=0; k<3; ++k )
				{
					p->pos[k] =
					    u*e1->pos[k] + ( 1-u ) *e3->pos[k] + v*e2->pos[k] + ( 1-v ) *e0->pos[k]
					    -
					    ( ( 1-u ) * ( 1-v ) *corners[0]->pos[k] + ( 1-u ) *v*corners[3]->pos[k] + u* ( 1-v ) *corners[1]->pos[k] + u*v*corners[2]->pos[k] );
				}
				mesh.push_back ( p );
			}
		}
	}
#ifdef PRINT_DEBUG_OUTPUT
	fprintf ( stderr, "\tEND Surface(%i)::generateMesh()\n", this->ind() );
#endif
}

inline void Surface::getSubdiv ( int &subdivX, int &subdivY )
{
	subdivX = curves[0]->getSubdiv();
	subdivY = curves[1]->getSubdiv();
}

/**
 * Block functions
 *
 */

Block::Block ( int id, Surface **sp )
		: geo_index ( id ), meshing ( false )
{
	//reorder the faces. a face will be picked that will be f0, f1 the face that shares v3 and v0, f2 v0 and v1, f3 v1 and v2, f4 v2 and v3, and f5 the remaining face.
	//If possible, we should take the surface with smallest area as f0 to minimize meory footprint. If the transfinite subdivision is not defined yet, just pick the first surface.
	int f0Pick = 0;
	unsigned int pickedArea = 0xffffffff;
	for(int i=0; i<6; ++i)
	{
		int subdivX, subdivY;
		sp[i]->getSubdiv(subdivX, subdivY);
		unsigned int area = subdivX*subdivY;
		if(!area)
		{
#ifdef PRINT_DEBUG_OUTPUT
			fprintf(stderr, "Block %i: cannot use cache-friendly layout.\n", id);
#endif
			goto pickFace0; //one of the faces subdivision is not defined yet
		}
		else if(area<pickedArea)
		{
			f0Pick = i;
			pickedArea = area;
		}
	}
	
pickFace0:
	surfaces[0] = sp[f0Pick];

#ifdef PRINT_DEBUG_OUTPUT
	fprintf(stderr, "Block %i: Picked surface %i as face 0.\n", id, f0Pick);
#endif

	//f0 is initially not reversed
	bool f0Reversed=false;
redoWithReversedF0:
	Point* f0Points[4];
	Point* fsPoints[4]={NULL,NULL,NULL,NULL};
	surfaces[0]->getCorners ( f0Points, f0Reversed );
	bool pointsInCommon[4];
	for ( int si=0; si<6; ++si )
	{
		if(surfaces[0]==sp[si]) continue; //skip f0

		for ( int i=0; i<4;++i ) pointsInCommon[i] = false;
		sp[si]->getCorners ( fsPoints, false );
		for ( int i=0; i<4; ++i )
		{
			for ( int j=0; j<4; ++j )
			{
				if ( f0Points[i]==fsPoints[j] )
				{
					pointsInCommon[i] = true;
				}
			}
		}
		if ( pointsInCommon[3] && pointsInCommon[0] )
		{
			surfaces[1] = sp[si];
		}
		else if ( pointsInCommon[0] && pointsInCommon[1] )
		{
			surfaces[2] = sp[si];
		}
		else if ( pointsInCommon[1] && pointsInCommon[2] )
		{
			surfaces[3] = sp[si];
		}
		else if ( pointsInCommon[2] && pointsInCommon[3] )
		{
			surfaces[4] = sp[si];
		}
		else
		{
			surfaces[5] = sp[si];
		}
	}

	//compute orientation:
	//let A and B be the centerpoint of each opposite face, then if dot(AB, n) > 0 then n points inside
	float nA[3], nB[3], A[3], B[3], AB[3], n_AB;

	//f0, f5
	surfaces[0]->getCenterAndNormal ( A, nA );
	surfaces[5]->getCenterAndNormal ( B, nB );
	AB[0] = B[0]-A[0]; AB[1] = B[1]-A[1]; AB[2] = B[2]-A[2];
	n_AB = dot ( nA, AB );
	if ( n_AB > 0 )
		reverse[0]=false;
	else
		reverse[0]=true;
	n_AB = dot ( nB, AB );
	if ( n_AB < 0 )
		reverse[5]=false;
	else
		reverse[5]=true;

	//f1, f3
	surfaces[1]->getCenterAndNormal ( A, nA );
	surfaces[3]->getCenterAndNormal ( B, nB );
	AB[0] = B[0]-A[0]; AB[1] = B[1]-A[1]; AB[2] = B[2]-A[2];
	n_AB = dot ( nA, AB );
	if ( n_AB > 0 )
		reverse[1]=false;
	else
		reverse[1]=true;
	n_AB = dot ( nB, AB );
	if ( n_AB < 0 )
		reverse[3]=false;
	else
		reverse[3]=true;

	//f2, f4
	surfaces[2]->getCenterAndNormal ( A, nA );
	surfaces[4]->getCenterAndNormal ( B, nB );
	AB[0] = B[0]-A[0]; AB[1] = B[1]-A[1]; AB[2] = B[2]-A[2];
	n_AB = dot ( nA, AB );
	if ( n_AB > 0 )
		reverse[2]=false;
	else
		reverse[2]=true;
	n_AB = dot ( nB, AB );
	if ( n_AB < 0 )
		reverse[4]=false;
	else
		reverse[4]=true;

	//if f0 reversed, then we have a contradiction, and must compute the faces again...
	if ( !f0Reversed && reverse[0] )
	{
		f0Reversed = true;
		goto redoWithReversedF0;
	}
	

	//now compute rotation:
	Point* corners[8];
	this->getCornersSlow ( corners );
	surfaces[0]->getCorners ( fsPoints, reverse[0] );
	for ( int i=0; i<4;++i )
		if ( fsPoints[i]==corners[0] )
		{
			sRotation[0] = i;
			break;
		}
	surfaces[1]->getCorners ( fsPoints, reverse[1] );
	for ( int i=0; i<4;++i )
		if ( fsPoints[i]==corners[0] )
		{
			sRotation[1] = i;
			break;
		}
	surfaces[2]->getCorners ( fsPoints, reverse[2] );
	for ( int i=0; i<4;++i )
		if ( fsPoints[i]==corners[0] )
		{
			sRotation[2] = i;
			break;
		}
	surfaces[3]->getCorners ( fsPoints, !reverse[3] ); //NOTE: we want this one pointing outwards
	for ( int i=0; i<4;++i )
		if ( fsPoints[i]==corners[1] )
		{
			sRotation[3] = i;
			break;
		}
	surfaces[4]->getCorners ( fsPoints, !reverse[4] );//NOTE: we want this one pointing outwards
	for ( int i=0; i<4;++i )
		if ( fsPoints[i]==corners[3] )
		{
			sRotation[4] = i;
			break;
		}
	surfaces[5]->getCorners ( fsPoints, !reverse[5] );//NOTE: we want this one pointing outwards
	for ( int i=0; i<4;++i )
		if ( fsPoints[i]==corners[4] )
		{
			sRotation[5] = i;
			break;
		}
#ifdef PRINT_DEBUG_OUTPUT
	fprintf ( stderr, "block corners are:\n" );
	for ( int i=0; i<8; ++i )
		fprintf ( stderr, "\t%i\n", corners[i]->ind() );
#endif

	//add the block to each surface
	for ( int i=0; i<6; ++i )
	{
		surfaces[i]->addBlock ( this );
	}

	//add a slot to each corner point
	for ( int i=0; i<8; ++i )
		corners[i]->addSlot();

}

Block::~Block()
{
	delete[] reverse;
	delete[] surfaces;
}

//get the eight corner vertices of the block in the CUBIT order.
void Block::getCorners ( Point **corners )
{
	CyclicPointArray4 f0Corners;
	CyclicPointArray4 rf5Corners;
	surfaces[0]->getCorners ( f0Corners, reverse[0] );
	surfaces[5]->getCorners ( rf5Corners, !reverse[5] );//NOTE: we want this one pointing outwards

	for ( int i=0; i<4;++i )
	{
		corners[i] = f0Corners[i+sRotation[0]];
		corners[i+4] = rf5Corners[i+sRotation[5]];
	}

}

//get the eight corner vertices of the block in the CUBIT order (slow version that does not need sRotation)
void Block::getCornersSlow ( Point **corners )
{
	CyclicPointArray4 f0Corners;
	CyclicPointArray4  rf1Corners;
	CyclicPointArray4  rf5Corners;
	surfaces[0]->getCorners ( f0Corners, reverse[0] );
	surfaces[1]->getCorners ( rf1Corners, !reverse[1] );
	surfaces[5]->getCorners ( rf5Corners, !reverse[5] );

	//V0 is the second of the two consecutive vertices shared by f0 and f1
	for ( int i=0; i<4; ++i )
	{
		for ( int j=0; j<4; ++j )
		{
			if ( f0Corners[i]==rf1Corners[j] )
			{
				if ( f0Corners[i+1]==rf1Corners[j+1] )
				{
					//v0 is f0Corners[i+1]
					for ( int k=0; k<4; ++k )
						corners[k] = f0Corners[i+1+k];
					//v4 is rf1Corners[j+2]
					corners[4] = rf1Corners[j+2];
					goto v0v4Found;
				}
				else
				{
					//v0 is f0Corners[i]
					for ( int k=0; k<4; ++k )
						corners[k] = f0Corners[i+k];
					//v4 is rf1Corners[j+1]
					corners[4] = rf1Corners[j+1];
					goto v0v4Found;
				}
			}
		}
	}

	fprintf ( stderr, "ERROR: In block %i, surface f0 (%i) and f1 (%i) do not share V0\n", geo_index, surfaces[0]->ind(), surfaces[1]->ind() );
	exit ( 1 );

v0v4Found:

	//then V5, V6, V7 are the corners that come after V4 in the reversed f5
	for ( int i=0; i<4; ++i )
	{
		if ( corners[4]==rf5Corners[i] )
		{
			corners[5] = rf5Corners[i+1];
			corners[6] = rf5Corners[i+2];
			corners[7] = rf5Corners[i+3];
			return;
		}
	}

	fprintf ( stderr, "ERROR: In block %i, surface f1 (%i) and f5 (%i) do not share V4 (%i)\n", geo_index, surfaces[1]->ind(), surfaces[5]->ind(), corners[4]->ind() );
	exit ( 1 );
}

void Block::beginMeshing()
{
#ifdef PRINT_DEBUG_OUTPUT
	fprintf ( stderr, "BEGIN Block(%i)::beginMeshing()\n", this->ind() );
#endif
	//mesh all the surfaces
	for ( int i=0; i<6; ++i )
		surfaces[i]->generateMesh();

	int dummy;
	surfaces[0]->getSubdiv ( subdivX, subdivY );
	if ( ( sRotation[0]&1 ) xor reverse[0] )
	{
		int t = subdivX;
		subdivX = subdivY;
		subdivY = t;
	}
	surfaces[1]->getSubdiv ( dummy, subdivZ );
	if ( ( sRotation[1]&1 ) xor reverse[1] )
	{
		subdivZ = dummy;
	}
	//add all the boundary vertices to the hash
	activeVerts = new CoordsPointHash();
	Coords coords;
	int iCoords = 0;
	coords.z = 0;
#ifdef PRINT_DEBUG_OUTPUT
	fprintf ( stderr, "\tf0=%i (rot:%i, rev:%i)\n", surfaces[0]->ind(), sRotation[0], reverse[0] );
#endif
	for ( coords.y=0; coords.y<subdivY+1; ++coords.y )
	{
		for ( coords.x=0; coords.x<subdivX+1; ++coords.x )
		{
			int localX, localY;
			getLocalSurfaceCoords ( 0, coords, localX, localY );
			activeVerts->insert ( CoordsPointHash::value_type ( iCoords, ( *surfaces[0] ) ( localX, localY ) ) );
#ifdef PRINT_DEBUG_OUTPUT
			fprintf ( stderr, "\t\tinserting vertex %i at coord %i (%i,%i,%i) (%i,%i)\n", ( *surfaces[0] ) ( localX, localY )->idx(), iCoords, coords.x, coords.y, coords.z, localX, localY );
#endif
			++iCoords;
		}
	}
	coords.z = subdivZ;
	iCoords = ( subdivX+1 ) * ( subdivY+1 ) *subdivZ;
#ifdef PRINT_DEBUG_OUTPUT
	fprintf ( stderr, "\tf5=%i (rot:%i, rev:%i)\n", surfaces[5]->ind(), sRotation[5], reverse[5] );
#endif
	for ( coords.y=0; coords.y<subdivY+1; ++coords.y )
	{
		for ( coords.x=0; coords.x<subdivX+1; ++coords.x )
		{
			int localX, localY;
			getLocalSurfaceCoords ( 5, coords, localX, localY );
			activeVerts->insert ( CoordsPointHash::value_type ( iCoords, ( *surfaces[5] ) ( localX, localY ) ) );
#ifdef PRINT_DEBUG_OUTPUT
			fprintf ( stderr, "\t\tinserting vertex %i at coord %i (%i,%i,%i) (%i,%i)\n", ( *surfaces[5] ) ( localX, localY )->idx(), iCoords, coords.x, coords.y, coords.z, localX, localY );
#endif
			++iCoords;
		}
	}
	coords.x = 0;
	iCoords = 0;
#ifdef PRINT_DEBUG_OUTPUT
	fprintf ( stderr, "\tf1=%i (rot:%i, rev:%i)\n", surfaces[1]->ind(), sRotation[1], reverse[1] );
#endif
	for ( coords.z=0; coords.z<subdivZ+1; ++coords.z )
	{
		for ( coords.y=0; coords.y<subdivY+1; ++coords.y )
		{
			int localX, localY;
			getLocalSurfaceCoords ( 1, coords, localX, localY );
			activeVerts->insert ( CoordsPointHash::value_type ( iCoords, ( *surfaces[1] ) ( localX, localY ) ) );
#ifdef PRINT_DEBUG_OUTPUT
			fprintf ( stderr, "\t\tinserting vertex %i at coord %i (%i,%i,%i) (%i,%i)\n", ( *surfaces[1] ) ( localX, localY )->idx(), iCoords, coords.x, coords.y, coords.z, localX, localY );
#endif
			iCoords += ( subdivX+1 );
		}
	}
	coords.x = subdivX;
	iCoords = subdivX;
#ifdef PRINT_DEBUG_OUTPUT
	fprintf ( stderr, "\tf3=%i (rot:%i, rev:%i)\n", surfaces[3]->ind(), sRotation[3], reverse[3] );
#endif
	for ( coords.z=0; coords.z<subdivZ+1; ++coords.z )
	{
		for ( coords.y=0; coords.y<subdivY+1; ++coords.y )
		{
			int localX, localY;
			getLocalSurfaceCoords ( 3, coords, localX, localY );
			activeVerts->insert ( CoordsPointHash::value_type ( iCoords, ( *surfaces[3] ) ( localX, localY ) ) );
#ifdef PRINT_DEBUG_OUTPUT
			fprintf ( stderr, "\t\tinserting vertex %i at coord %i (%i,%i,%i) (%i,%i)\n", ( *surfaces[3] ) ( localX, localY )->idx(), iCoords, coords.x, coords.y, coords.z, localX, localY );
#endif
			iCoords += ( subdivX+1 );
		}
	}
	coords.y = 0;
	iCoords = 0;
#ifdef PRINT_DEBUG_OUTPUT
	fprintf ( stderr, "\tf2=%i (rot:%i, rev:%i)\n", surfaces[2]->ind(), sRotation[2], reverse[2] );
#endif
	for ( coords.z=0; coords.z<subdivZ+1; ++coords.z )
	{
		for ( coords.x=0; coords.x<subdivX+1; ++coords.x )
		{
			int localX, localY;
			getLocalSurfaceCoords ( 2, coords, localX, localY );
			activeVerts->insert ( CoordsPointHash::value_type ( iCoords, ( *surfaces[2] ) ( localX, localY ) ) );
#ifdef PRINT_DEBUG_OUTPUT
			fprintf ( stderr, "\t\tinserting vertex %i at coord %i (%i,%i,%i) (%i,%i)\n", ( *surfaces[2] ) ( localX, localY )->idx(), iCoords, coords.x, coords.y, coords.z, localX, localY );
#endif
			++iCoords;
		}
		iCoords += ( subdivX+1 ) *subdivY;
	}
	coords.y = subdivY;
	iCoords = ( subdivX+1 ) *subdivY;
#ifdef PRINT_DEBUG_OUTPUT
	fprintf ( stderr, "\tf4=%i (rot:%i, rev:%i)\n", surfaces[4]->ind(), sRotation[4], reverse[4] );
#endif
	for ( coords.z=0; coords.z<subdivZ+1; ++coords.z )
	{
		for ( coords.x=0; coords.x<subdivX+1; ++coords.x )
		{
			int localX, localY;
			getLocalSurfaceCoords ( 4, coords, localX, localY );
			activeVerts->insert ( CoordsPointHash::value_type ( iCoords, ( *surfaces[4] ) ( localX, localY ) ) );
#ifdef PRINT_DEBUG_OUTPUT
			fprintf ( stderr, "\t\tinserting vertex %i at coord %i (%i,%i,%i) (%i,%i)\n", ( *surfaces[4] ) ( localX, localY )->idx(), iCoords, coords.x, coords.y, coords.z, localX, localY );
#endif
			++iCoords;
		}
		iCoords += ( subdivX+1 ) *subdivY;
	}

	//begin by first hex:
	currentHexCoords.x = 0;
	currentHexCoords.y = 0;
	currentHexCoords.z = 0;
	nHexas = 0;
#ifdef OPTIMIZE_CACHE_MISS
	traversalDirX = true;
	traversalDirY = true;
#endif
#ifdef PRINT_DEBUG_OUTPUT
	fprintf ( stderr, "END Block(%i)::beginMeshing()\n", this->ind() );
#endif
}

bool Block::getNextHex ( int *c_idx, StreamList &verticesToStream, bool *final, int &v_count )
{
	if ( nHexas++ == subdivX*subdivY*subdivZ )
		return false; //no more hexahedra to process

#ifdef PRINT_DEBUG_OUTPUT
	fprintf ( stderr, "current coords: %i, %i, %i\n", currentHexCoords.x, currentHexCoords.y, currentHexCoords.z );
#endif
	//for the moment, scanline order
	//TODO: zigzagging order
	CoordsPointHash::iterator vXit;
	//Coords<int> vXc;
	int vXc;
	int hexC = currentHexCoords.x + currentHexCoords.y* ( subdivX+1 ) + currentHexCoords.z* ( subdivX+1 ) * ( subdivY+1 );
	Point *vX;

	//vertex0
	getPointAtCoords ( hexC, verticesToStream, vXit, v_count );
	vX = vXit->second;
	c_idx[0] = vX->idx();
	vX->takeSlot();
	//finalize vertices with no more available slots
	if ( !vX->hasAvailableSlots() )
	{
#ifdef PRINT_DEBUG_OUTPUT
		fprintf ( stderr, "finalizing vertex %i\n", vX->idx() );
#endif
		final[0] = true;
		if ( vX->isInterior() )
		{
			delete vX;
			activeVerts->erase ( vXit );
		}
	}
	else
		final[0] = false;

	//vertex1
	vXc = hexC + 1;
	getPointAtCoords ( vXc, verticesToStream, vXit, v_count );
	vX = vXit->second;
	c_idx[1] = vX->idx();
	vX->takeSlot();
	//finalize vertices with no more available slots
	if ( !vX->hasAvailableSlots() )
	{
#ifdef PRINT_DEBUG_OUTPUT
		fprintf ( stderr, "finalizing vertex %i\n", vX->idx() );
#endif
		final[1] = true;
		if ( vX->isInterior() )
		{
			delete vX;
			activeVerts->erase ( vXit );
		}
	}
	else
		final[1] = false;

	//vertex2
	vXc = hexC + 1 + ( subdivX+1 );
	getPointAtCoords ( vXc, verticesToStream, vXit, v_count );
	vX = vXit->second;
	c_idx[2] = vX->idx();
	vX->takeSlot();
	//finalize vertices with no more available slots
	if ( !vX->hasAvailableSlots() )
	{
#ifdef PRINT_DEBUG_OUTPUT
		fprintf ( stderr, "finalizing vertex %i\n", vX->idx() );
#endif
		final[2] = true;
		if ( vX->isInterior() )
		{
			delete vX;
			activeVerts->erase ( vXit );
		}
	}
	else
		final[2] = false;

	//vertex3
	vXc = hexC + ( subdivX+1 );
	getPointAtCoords ( vXc, verticesToStream, vXit, v_count );
	vX = vXit->second;
	c_idx[3] = vX->idx();
	vX->takeSlot();
	//finalize vertices with no more available slots
	if ( !vX->hasAvailableSlots() )
	{
#ifdef PRINT_DEBUG_OUTPUT
		fprintf ( stderr, "finalizing vertex %i\n", vX->idx() );
#endif
		final[3] = true;
		if ( vX->isInterior() )
		{
			delete vX;
			activeVerts->erase ( vXit );
		}
	}
	else
		final[3] = false;

	//vertex4
	vXc = hexC + ( subdivX+1 ) * ( subdivY+1 );
	getPointAtCoords ( vXc, verticesToStream, vXit, v_count );
	vX = vXit->second;
	c_idx[4] = vX->idx();
	vX->takeSlot();
	//finalize vertices with no more available slots
	if ( !vX->hasAvailableSlots() )
	{
#ifdef PRINT_DEBUG_OUTPUT
		fprintf ( stderr, "finalizing vertex %i\n", vX->idx() );
#endif
		final[4] = true;
		if ( vX->isInterior() )
		{
			delete vX;
			activeVerts->erase ( vXit );
		}
	}
	else
		final[4] = false;

	//vertex5
	vXc = hexC + 1 + ( subdivX+1 ) * ( subdivY+1 );
	getPointAtCoords ( vXc, verticesToStream, vXit, v_count );
	vX = vXit->second;
	c_idx[5] = vX->idx();
	vX->takeSlot();
	//finalize vertices with no more available slots
	if ( !vX->hasAvailableSlots() )
	{
#ifdef PRINT_DEBUG_OUTPUT
		fprintf ( stderr, "finalizing vertex %i\n", vX->idx() );
#endif
		final[5] = true;
		if ( vX->isInterior() )
		{
			delete vX;
			activeVerts->erase ( vXit );
		}
	}
	else
		final[5] = false;

	//vertex6
	vXc = hexC + 1 + ( subdivX+1 ) + ( subdivX+1 ) * ( subdivY+1 );
	getPointAtCoords ( vXc, verticesToStream, vXit, v_count );
	vX = vXit->second;
	c_idx[6] = vX->idx();
	vX->takeSlot();
	//finalize vertices with no more available slots
	if ( !vX->hasAvailableSlots() )
	{
#ifdef PRINT_DEBUG_OUTPUT
		fprintf ( stderr, "finalizing vertex %i\n", vX->idx() );
#endif
		final[6] = true;
		if ( vX->isInterior() )
		{
			delete vX;
			activeVerts->erase ( vXit );
		}
	}
	else
		final[6] = false;

	//vertex7
	vXc = hexC + ( subdivX+1 ) + ( subdivX+1 ) * ( subdivY+1 );
	getPointAtCoords ( vXc, verticesToStream, vXit, v_count );
	vX = vXit->second;
	c_idx[7] = vX->idx();
	vX->takeSlot();
	//finalize vertices with no more available slots
	if ( !vX->hasAvailableSlots() )
	{
#ifdef PRINT_DEBUG_OUTPUT
		fprintf ( stderr, "finalizing vertex %i\n", vX->idx() );
#endif
		final[7] = true;
		if ( vX->isInterior() )
		{
			delete vX;
			activeVerts->erase ( vXit );
		}
	}
	else
		final[7] = false;

#ifdef OPTIMIZE_CACHE_MISS
	//begin with traversalDirX = true
	//begin with traversalDirY = true
	// - increase x if traversalDirX, decrease else.
	// - at the end of the row (i.e. when (nHexas%subdivX) is zero), increase y if traversalDirX, decrease y else, and change traversalDirX.
	// - at the end of the slice (i.e. when (i.e. when (nHexas%(subdivX*subdivY)) is zero), increase z and change traversalDirX and traversalDirY.

	//printf("coordinates before: (%i, %i, %i), nHexas=%i, subdiv=(%i,%i,%i)\n", currentHexCoords.x, currentHexCoords.y, currentHexCoords.z, nHexas, subdivX, subdivY, subdivZ);

	if((nHexas%(subdivX*subdivY))==0) //end of slice
	{
		++currentHexCoords.z;
		traversalDirX = ! traversalDirX;
		traversalDirY = ! traversalDirY;
	}
	else
	{
		if((nHexas%subdivX)==0) //end of row
		{
			if(traversalDirY) 
				++currentHexCoords.y;
			else
				--currentHexCoords.y;
			traversalDirX = ! traversalDirX;
		}
		else //neither end of slice nor row
		{
			if(traversalDirX) 
				++currentHexCoords.x;
			else
				--currentHexCoords.x;
		}
	}
#else
	//prepare the next coords (scanline order)
	if ( ++currentHexCoords.x == subdivX )
	{
		if ( ++currentHexCoords.y == subdivY )
		{
			++currentHexCoords.z;
			currentHexCoords.y = 0;
		}
		currentHexCoords.x = 0;
	}
#endif

	return true;
}

void Block::endMeshing()
{
	//delete all finalized vertices
	for ( CoordsPointHash::iterator it = activeVerts->begin(); it!= activeVerts->end(); ++it )
	{
		Point *p = it->second;
		if ( p->isInterior() )
		{
			delete p;
			fprintf ( stderr, "WARNING: found an unfinalized interior vertex. That should not happen.\n" );
		}
		else if ( !p->hasAvailableSlots() )
		{
			delete p; //this is an already finalized point that we could not delete sibce it was used for interpolation.
		}
	}
	delete activeVerts;
}

void Block::getPointAtCoords ( int vXc, StreamList &verticesToStream, CoordsPointHash::iterator &vXit, int &v_count )
{
#ifdef PRINT_DEBUG_OUTPUT
	fprintf ( stderr, "getPointAtCoords(%04i)", vXc );
#endif
	Point *vX;
	vXit = activeVerts->find ( vXc );
	if ( vXit == activeVerts->end() )
	{
		//if it does not exist, create it.
		vX = new Point ( -1, Point::TYPE_INTERIOR, 8 );
		//insert it in the hash
		std::pair <CoordsPointHash::iterator, bool> ret = activeVerts->insert ( CoordsPointHash::value_type ( vXc, vX ) );
		vXit = ret.first;
		//and add it to the list of vertices to be streamed
		vX->set_idx ( v_count++ );
		float *nvCoords = new float[3];
		//interpolate.
		transfiniteInterpolation ( vXc, vX->pos );
		memcpy ( nvCoords, vX->pos, 3*sizeof ( float ) );
		verticesToStream.push_back ( new std::pair<int, float*> ( vX->idx(), nvCoords ) );
#ifdef PRINT_DEBUG_OUTPUT
		fprintf ( stderr, " -> (N) %04i\n", vX->idx() );
#endif
	}
	else
	{
		vX = vXit->second;
		if ( !vX->streamed() )
		{
			vX->set_idx ( v_count++ );
			float *nvCoords = new float[3];
			memcpy ( nvCoords, vX->pos, 3*sizeof ( float ) );
			verticesToStream.push_back ( new std::pair<int, float*> ( vX->idx(), nvCoords ) );
		}
#ifdef PRINT_DEBUG_OUTPUT
		fprintf ( stderr, " -> (E) %04i\n", vX->idx() );
#endif
	}
}

//tranfinite interpolation for the given point at coords vXc
void Block::transfiniteInterpolation ( int vXc, float *pos )
{
	//find projections of the current point on all edges (e_.) and all faces (c_.)
	CoordsPointHash::iterator tIt;
	int vPc;
	int x = vXc % ( subdivX+1 );
	int y = ( vXc % ( ( subdivX+1 ) * ( subdivY+1 ) ) ) / ( subdivX+1 );
	int z = vXc / ( ( subdivX+1 ) * ( subdivY+1 ) );
#ifdef PRINT_DEBUG_OUTPUT
	fprintf ( stderr,"interpolation for vertex at (%i, %i, %i)\n", x, y, z );
#endif

	double u = ( double ) x/subdivX;
	double v = ( double ) y/subdivY;
	double w = ( double ) z/subdivZ;

	//corner points
	Point * cor[8];
	this->getCorners ( cor );
	const float *v0Pos = cor[0]->pos;
	const float *v1Pos = cor[1]->pos;
	const float *v2Pos = cor[2]->pos;
	const float *v3Pos = cor[3]->pos;
	const float *v4Pos = cor[4]->pos;
	const float *v5Pos = cor[5]->pos;
	const float *v6Pos = cor[6]->pos;
	const float *v7Pos = cor[7]->pos;

	//edge points
	vPc = x;
	tIt = activeVerts->find ( vPc );
	const float *e0Pos = tIt->second->pos;
	vPc = y* ( subdivX+1 );
	tIt = activeVerts->find ( vPc );
	const float *e1Pos = tIt->second->pos;
	vPc = z* ( subdivX+1 ) * ( subdivY+1 );
	tIt = activeVerts->find ( vPc );
	const float *e8Pos = tIt->second->pos;
	vPc = subdivY* ( subdivX+1 ) + x;
	tIt = activeVerts->find ( vPc );
	const float *e2Pos = tIt->second->pos;
	vPc = y* ( subdivX+1 ) + subdivX;
	tIt = activeVerts->find ( vPc );
	const float *e3Pos = tIt->second->pos;
	vPc = z* ( subdivX+1 ) * ( subdivY+1 ) + subdivY* ( subdivX+1 );
	tIt = activeVerts->find ( vPc );
	const float *e9Pos = tIt->second->pos;
	vPc = subdivZ* ( subdivX+1 ) * ( subdivY+1 ) + x;
	tIt = activeVerts->find ( vPc );
	const float *e4Pos = tIt->second->pos;
	vPc = subdivZ* ( subdivX+1 ) * ( subdivY+1 ) + y* ( subdivX+1 );
	tIt = activeVerts->find ( vPc );
	const float *e7Pos = tIt->second->pos;
	vPc = z* ( subdivX+1 ) * ( subdivY+1 ) + subdivX;
	tIt = activeVerts->find ( vPc );
	const float *e11Pos = tIt->second->pos;
	vPc = subdivZ* ( subdivX+1 ) * ( subdivY+1 ) + subdivY* ( subdivX+1 ) + x;
	tIt = activeVerts->find ( vPc );
	const float *e6Pos = tIt->second->pos;
	vPc = subdivZ* ( subdivX+1 ) * ( subdivY+1 ) + y* ( subdivX+1 ) + subdivX;
	tIt = activeVerts->find ( vPc );
	const float *e5Pos = tIt->second->pos;
	vPc = z* ( subdivX+1 ) * ( subdivY+1 ) + subdivY* ( subdivX+1 ) + subdivX;
	tIt = activeVerts->find ( vPc );
	const float *e10Pos = tIt->second->pos;

	//face points
	vPc = y* ( subdivX+1 ) + x;
	tIt = activeVerts->find ( vPc );
	const float *c0Pos = tIt->second->pos;
	vPc = z* ( subdivX+1 ) * ( subdivY+1 ) + y* ( subdivX+1 );
	tIt = activeVerts->find ( vPc );
	const float *c1Pos = tIt->second->pos;
	vPc = z* ( subdivX+1 ) * ( subdivY+1 ) + x;
	tIt = activeVerts->find ( vPc );
	const float *c2Pos = tIt->second->pos;
	vPc = subdivZ* ( subdivX+1 ) * ( subdivY+1 ) + y* ( subdivX+1 ) + x;
	tIt = activeVerts->find ( vPc );
	const float *c5Pos = tIt->second->pos;
	vPc = z* ( subdivX+1 ) * ( subdivY+1 ) + y* ( subdivX+1 ) + subdivX;
	tIt = activeVerts->find ( vPc );
	const float *c3Pos = tIt->second->pos;
	vPc = z* ( subdivX+1 ) * ( subdivY+1 ) + subdivY* ( subdivX+1 ) + x;
	tIt = activeVerts->find ( vPc );
	const float *c4Pos = tIt->second->pos;

	//now interpolate
	for ( int i=0; i<3; ++i )
	{
		pos[i] =
		    u*c3Pos[i] + ( 1-u ) *c1Pos[i] + v*c4Pos[i] + ( 1-v ) *c2Pos[i] + w*c5Pos[i] + ( 1-w ) *c0Pos[i]
		    - (
		        u*v*e10Pos[i] + ( 1-u ) *v*e9Pos[i] + u* ( 1-v ) *e11Pos[i] + ( 1-u ) * ( 1-v ) *e8Pos[i]
		        + v*w*e6Pos[i] + ( 1-v ) *w*e4Pos[i] + v* ( 1-w ) *e2Pos[i] + ( 1-v ) * ( 1-w ) *e0Pos[i]
		        + u*w*e5Pos[i] + ( 1-u ) *w*e7Pos[i] + u* ( 1-w ) *e3Pos[i] + ( 1-u ) * ( 1-w ) *e1Pos[i]
		    ) +
		    u*v*w*v6Pos[i] + ( 1-u ) *v*w*v7Pos[i] + u* ( 1-v ) *w*v5Pos[i] + u*v* ( 1-w ) *v2Pos[i] + ( 1-u ) * ( 1-v ) *w*v4Pos[i] + ( 1-u ) *v* ( 1-w ) *v3Pos[i] + u* ( 1-v ) * ( 1-w ) *v1Pos[i] + ( 1-u ) * ( 1-v ) * ( 1-w ) *v0Pos[i] ;
	}
}

void Block::getLocalSurfaceCoords ( int surfaceIndex, Coords &blockCoords, int &localX, int &localY )
{
	int lSubdivX, lSubdivY;
	surfaces[surfaceIndex]->getSubdiv ( lSubdivX, lSubdivY );
	switch ( surfaceIndex )
	{
		case 0:
			if ( reverse[surfaceIndex] )
			{
				switch ( sRotation[surfaceIndex] )
				{
					case 0:
						localX = blockCoords.y;
						localY = blockCoords.x;
						return;
					case 3:
						localX = lSubdivX-blockCoords.x;
						localY = blockCoords.y;
						return;
					case 2:
						localX = lSubdivX-blockCoords.y;
						localY = lSubdivY-blockCoords.x;
						return;
					case 1:
						localX = blockCoords.x;
						localY = lSubdivY-blockCoords.y;
						return;
				}
			}
			else
			{
				switch ( sRotation[surfaceIndex] )
				{
					case 0:
						localX = blockCoords.x;
						localY = blockCoords.y;
						return;
					case 3:
						localX = blockCoords.y;
						localY = lSubdivY-blockCoords.x;
						return;
					case 2:
						localX = lSubdivX-blockCoords.x;
						localY = lSubdivY-blockCoords.y;
						return;
					case 1:
						localX = lSubdivX-blockCoords.y;
						localY = blockCoords.x;
						return;
				}
			}
			return;
		case 1:
			if ( reverse[surfaceIndex] )
			{
				switch ( sRotation[surfaceIndex] )
				{
					case 0:
						localX = blockCoords.z;
						localY = blockCoords.y;
						return;
					case 3:
						localX = lSubdivX-blockCoords.y;
						localY = blockCoords.z;
						return;
					case 2:
						localX = lSubdivX-blockCoords.z;
						localY = lSubdivY-blockCoords.y;
						return;
					case 1:
						localX = blockCoords.y;
						localY = lSubdivY-blockCoords.z;
						return;
				}
			}
			else
			{
				switch ( sRotation[surfaceIndex] )
				{
					case 0:
						localX = blockCoords.y;
						localY = blockCoords.z;
						return;
					case 3:
						localX = blockCoords.z;
						localY = lSubdivY-blockCoords.y;
						return;
					case 2:
						localX = lSubdivX-blockCoords.y;
						localY = lSubdivY-blockCoords.z;
						return;
					case 1:
						localX = lSubdivX-blockCoords.z;
						localY = blockCoords.y;
						return;
				}
			}
			return;
		case 2:
			if ( reverse[surfaceIndex] )
			{
				switch ( sRotation[surfaceIndex] )
				{
					case 0:
						localX = blockCoords.x;
						localY = blockCoords.z;
						return;
					case 3:
						localX = lSubdivX-blockCoords.z;
						localY = blockCoords.x;
						return;
					case 2:
						localX = lSubdivX-blockCoords.x;
						localY = lSubdivY-blockCoords.z;
						return;
					case 1:
						localX = blockCoords.z;
						localY = lSubdivY-blockCoords.x;
						return;
				}
			}
			else
			{
				switch ( sRotation[surfaceIndex] )
				{
					case 0:
						localX = blockCoords.z;
						localY = blockCoords.x;
						return;
					case 3:
						localX = blockCoords.x;
						localY = lSubdivY-blockCoords.z;
						return;
					case 2:
						localX = lSubdivX-blockCoords.z;
						localY = lSubdivY-blockCoords.x;
						return;
					case 1:
						localX = lSubdivX-blockCoords.x;
						localY = blockCoords.z;
						return;
				}
			}
			return;
		case 3:
			if ( !reverse[surfaceIndex] ) //NOTE ! here
			{
				switch ( sRotation[surfaceIndex] )
				{
					case 0:
						localX = blockCoords.z;
						localY = blockCoords.y;
						return;
					case 3:
						localX = lSubdivX-blockCoords.y;
						localY = blockCoords.z;
						return;
					case 2:
						localX = lSubdivX-blockCoords.z;
						localY = lSubdivY-blockCoords.y;
						return;
					case 1:
						localX = blockCoords.y;
						localY = lSubdivY-blockCoords.z;
						return;
				}
			}
			else
			{
				switch ( sRotation[surfaceIndex] )
				{
					case 0:
						localX = blockCoords.y;
						localY = blockCoords.z;
						return;
					case 3:
						localX = blockCoords.z;
						localY = lSubdivY-blockCoords.y;
						return;
					case 2:
						localX = lSubdivX-blockCoords.y;
						localY = lSubdivY-blockCoords.z;
						return;
					case 1:
						localX = lSubdivX-blockCoords.z;
						localY = blockCoords.y;
						return;
				}
			}
			return;
		case 4:
			if ( !reverse[surfaceIndex] ) //NOTE ! here
			{
				switch ( sRotation[surfaceIndex] )
				{
					case 0:
						localX = blockCoords.x;
						localY = blockCoords.z;
						return;
					case 3:
						localX = lSubdivX-blockCoords.z;
						localY = blockCoords.x;
						return;
					case 2:
						localX = lSubdivX-blockCoords.x;
						localY = lSubdivY-blockCoords.z;
						return;
					case 1:
						localX = blockCoords.z;
						localY = lSubdivY-blockCoords.x;
						return;
				}
			}
			else
			{
				switch ( sRotation[surfaceIndex] )
				{
					case 0:
						localX = blockCoords.z;
						localY = blockCoords.x;
						return;
					case 3:
						localX = blockCoords.x;
						localY = lSubdivY-blockCoords.z;
						return;
					case 2:
						localX = lSubdivX-blockCoords.z;
						localY = lSubdivY-blockCoords.x;
						return;
					case 1:
						localX = lSubdivX-blockCoords.x;
						localY = blockCoords.z;
						return;
				}
			}
			return;
		case 5:
			if ( !reverse[surfaceIndex] ) //NOTE ! here
			{
				switch ( sRotation[surfaceIndex] )
				{
					case 0:
						localX = blockCoords.y;
						localY = blockCoords.x;
						return;
					case 3:
						localX = lSubdivX-blockCoords.x;
						localY = blockCoords.y;
						return;
					case 2:
						localX = lSubdivX-blockCoords.y;
						localY = lSubdivY-blockCoords.x;
						return;
					case 1:
						localX = blockCoords.x;
						localY = lSubdivY-blockCoords.y;
						return;
				}
			}
			else
			{
				switch ( sRotation[surfaceIndex] )
				{
					case 0:
						localX = blockCoords.x;
						localY = blockCoords.y;
						return;
					case 3:
						localX = blockCoords.y;
						localY = lSubdivY-blockCoords.x;
						return;
					case 2:
						localX = lSubdivX-blockCoords.x;
						localY = lSubdivY-blockCoords.y;
						return;
					case 1:
						localX = lSubdivX-blockCoords.y;
						localY = blockCoords.x;
						return;
				}
			}
			return;
	}
}

