/*
===============================================================================

  FILE:  svcommon_svc.cpp
  
  CONTENTS:

    code that is common to svreader_svh.cpp and svwriter_svh.cpp

  PROGRAMMERS:

    martin isenburg@cs.unc.edu

  COPYRIGHT:

    copyright (C) 2005  martin isenburg@cs.unc.edu

    This software is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.

  CHANGE HISTORY:

    25 January 2010 -- created behind trader joe's   

===============================================================================
*/

// shared defines

#define SVH_START 	0
#define SVH_HUT 	1
#define SVH_ROOF 	2
#define SVH_STEP 	3
#define SVH_CORNER 	4
#define SVH_BRIDGE 	5
#define SVH_TUNNEL 	6
#define SVH_GAP 	7
#define SVH_PIT 	8
#define SVH_DEN 	9

#define SVH_NUMBER_OF_OPERATIONS 10

#define SVH_PREDICT_NONE          0
#define SVH_PREDICT_SHOOT         1
#define SVH_PREDICT_PARALLELOGRAM 2
#define SVH_PREDICT_LORENZO       3

#define SVH_NUMBER_OF_PREDICTIONS  4

#define SVH_MAX_DEGREE_ONE 10
#define SVH_MAX_USE_COUNT 18

//the initial size of the face list. this one does *not* need to be shared by coder and decoder.
#define SVH_INITIAL_VERTEX_LIST_SIZE 10

//MAGIC_NUMBER_F0_AROUND_V1_RM is used in the coding of f0 when only v1 is cached.
//if v1 has more faces around it than MAGIC_NUMBER_F0_AROUND_V1_RM, we do not use a range model.
#define SVH_MAGIC_NUMBER_F0_AROUND_V1_RM 14

//seems like its better only for grids. not implemented in the decoder, so dont use it.
#define SVH_BETTER_SHOOT_PREDICTION
#undef SVH_BETTER_SHOOT_PREDICTION

#define SVH_FACECACHE_DEPENDS_ON_FINALIZATION
#undef SVH_FACECACHE_DEPENDS_ON_FINALIZATION

#define SVH_RMOP_DEPENDS_ON_FINALIZATION
#undef SVH_RMOP_DEPENDS_ON_FINALIZATION

//FIXME: this doesnt work. why ?
#define SVH_INITIALIZE_VISVERTICES
#undef SVH_INITIALIZE_VISVERTICES

#define SVH_DUMB_JOIN_CODING
//#undef SVH_DUMB_JOIN_CODING

//let the decoder figure out when a vertex is finalized (i.e. when its list of faces is empty)
#define SVH_IMPLICIT_FINALIZATION
#undef SVH_IMPLICIT_FINALIZATION

//do not undefine this
#define SVH_COMPRESS_GEOMETRY

#define SVH_COMPRESS_CELL_DATA
//#undef SVH_COMPRESS_CELL_DATA

//second order prediction only makes sense if cell data compression is enabled
#ifdef SVH_COMPRESS_CELL_DATA
	#define SVH_CELL_DATA_SECONDORDER_PREDICTION
	//#undef SVH_CELL_DATA_SECONDORDER_PREDICTION
#else
	#undef SVH_CELL_DATA_SECONDORDER_PREDICTION
#endif

// shared includes
#include "rangefloatcompressor.h"
#include "rangeintegercompressor.h"
#include "positionquantizer_new.h"
#include "dynamicvector.h"
#include "littlecache.h"
#include "vec3fv.h"
#include "vec3iv.h"
#include "vecnfv.h"
#include "vecniv.h"

// shared variables and functions

static DynamicVector* dv;
static LittleCache* lc;
static LittleCache* lc_f;

static RangeFloatCompressor* v_fc[3];
static PositionQuantizerNew* v_pq = NULL;
static RangeIntegerCompressor* v_ic[3];

#ifdef SVH_COMPRESS_CELL_DATA
static RangeFloatCompressor** c_fc;
static PositionQuantizerNew* c_quantizer = NULL;
static RangeIntegerCompressor** c_ic;
#endif

// what was the last operation
static int last_op;

// context about the last finalization
static int last_fin;

// is there more to encode
static RangeModel* rmDone;

// codes next operation
#ifdef SVH_RMOP_DEPENDS_ON_FINALIZATION
static RangeModel*** rmOpX;
#else
static RangeModel** rmOp;
#endif

// codes cache hits for start operations
static RangeModel** rmS_Cache;

//codes visited state of each vertex in the start operation
static RangeModel* rmStart_VisitedVertices;

// codes cache hits for Hut operation
#ifdef SVH_FACECACHE_DEPENDS_ON_FINALIZATION
static RangeModel*** rmHut_FaceCacheX;
#else
static RangeModel** rmHut_FaceCache;
#endif
static RangeModel** rmHut_EdgeCache;
static RangeModel** rmHut_VertexCache;

// codes cache hits for Roof operation
#ifdef SVH_FACECACHE_DEPENDS_ON_FINALIZATION
static RangeModel*** rmRoof_FaceCacheX;
#else
static RangeModel** rmRoof_FaceCache;
#endif
static RangeModel** rmRoof_EdgeCache;
static RangeModel** rmRoof_VertexCache;

// codes cache hits Step operations
#ifdef SVH_FACECACHE_DEPENDS_ON_FINALIZATION
static RangeModel*** rmSCGPD_FaceCacheX;
#else
static RangeModel** rmSCGPD_FaceCache;
#endif
static RangeModel** rmSCGPD_EdgeCache;
static RangeModel** rmSCGPD_VertexCache;

// codes cache hits Bridge operations
#ifdef SVH_FACECACHE_DEPENDS_ON_FINALIZATION
static RangeModel*** rmBT_FaceCacheX;
#else
static RangeModel** rmBT_FaceCache;
#endif
static RangeModel** rmBT_EdgeCache;
static RangeModel** rmBT_VertexCache;

static RangeModel* rmHut_HasJoin;
static RangeModel* rmStep_HasJoin;
static RangeModel* rmCorner_HasJoin;

#ifdef SVH_DUMB_JOIN_CODING
//range table for HUT/STEP/CORNER operations, to tell which verts are joined
static RangeModel* rmHut_JoinedVertices;
static RangeModel* rmStep_JoinedVertices;
#else
//from which vertex we can join v6 with an edge:
// 0 means 2, 1 means 5, 2 means 7, 3 means neither.
static RangeModel* rmCorner_EdgeJoinFrom;
#endif

static RangeModel* rm_ConfirmFaceMerge;

static RangeModel** rm_VertexCandi;
static RangeModel* rm_FaceCandi;

// codes vertex finalization
static RangeModel*** rmFinalized;

static void initModels(int compress)
{
  int i,j;
  unsigned int t[257];

  // range table for whether there is more
  rmDone = new RangeModel(2,0,compress);

  // range table for which operation it is
#ifdef SVH_RMOP_DEPENDS_ON_FINALIZATION
  rmOpX = new RangeModel** [20];
  for(i = 0; i < 20; i++)
  {
    rmOpX[i] = new RangeModel* [SVH_NUMBER_OF_OPERATIONS];
    for(j=0; j< SVH_NUMBER_OF_OPERATIONS; j++) //for each operation
      rmOpX[i][j] = new RangeModel(SVH_NUMBER_OF_OPERATIONS,0,compress);
  }
#else
  rmOp = new RangeModel* [SVH_NUMBER_OF_OPERATIONS];
  for(j=0; j< SVH_NUMBER_OF_OPERATIONS; j++) //for each operation
    rmOp[j] = new RangeModel(SVH_NUMBER_OF_OPERATIONS,0,compress);
#endif

  // range tables for start (S) operation
  rmS_Cache = new RangeModel*[8];
  for(j=0; j< 8; j++) //for each vertex
    rmS_Cache[j] = new RangeModel(9,0,compress);

#ifdef SVH_INITIALIZE_VISVERTICES
  //most of the time a START will not be connected at all with the mesh.
  bzero(t, 257*sizeof(unsigned int));
  t[0]=1; 
  rmStart_VisitedVertices = new RangeModel(257, t, compress);
#else
  rmStart_VisitedVertices = new RangeModel(257, NULL, compress);
#endif

  // range tables for Hut operation
#ifdef SVH_FACECACHE_DEPENDS_ON_FINALIZATION
  rmHut_FaceCacheX = new RangeModel** [20];
  for(i = 0; i < 20; i++)
  {
    rmHut_FaceCacheX[i] = new RangeModel* [SVH_NUMBER_OF_OPERATIONS];
    t[0] = 1; t[1] = 1; t[2] = 1; t[3] = 1; t[4] = 1; t[5] = 1; t[6] = 1;
    rmHut_FaceCacheX[i][SVH_START] = new RangeModel(7,t,compress); // after SVH_START
    t[0] = 0; t[1] = 1; t[2] = 1; t[3] = 1; t[4] = 1; t[5] = 1; t[6] = 1;
    rmHut_FaceCacheX[i][SVH_HUT] = new RangeModel(7,t,compress); // after SVH_HUT
    t[0] = 0; t[1] = 1; t[2] = 1; t[3] = 1; t[4] = 1; t[5] = 0; t[6] = 1;
    rmHut_FaceCacheX[i][SVH_ROOF] = new RangeModel(7,t,compress); // after SVH_ROOF
    t[0] = 0; t[1] = 0; t[2] = 1; t[3] = 1; t[4] = 1; t[5] = 1; t[6] = 1;
    rmHut_FaceCacheX[i][SVH_STEP] = new RangeModel(7,t,compress); // after SVH_STEP
    t[0] = 0; t[1] = 0; t[2] = 0; t[3] = 1; t[4] = 1; t[5] = 1; t[6] = 1;
    rmHut_FaceCacheX[i][SVH_CORNER] = new RangeModel(7,t,compress); // after SVH_CORNER
    t[0] = 0; t[1] = 0; t[2] = 1; t[3] = 0; t[4] = 1; t[5] = 1; t[6] = 1;
    rmHut_FaceCacheX[i][SVH_BRIDGE] = new RangeModel(7,t,compress); // after SVH_BRIDGE
    t[0] = 0; t[1] = 0; t[2] = 1; t[3] = 0; t[4] = 1; t[5] = 0; t[6] = 1;
    rmHut_FaceCacheX[i][SVH_TUNNEL] = new RangeModel(7,t,compress); // after SVH_TUNNEL
    t[0] = 0; t[1] = 0; t[2] = 0; t[3] = 0; t[4] = 1; t[5] = 1; t[6] = 1;
    rmHut_FaceCacheX[i][SVH_GAP] = new RangeModel(7,t,compress); // after SVH_GAP
    t[0] = 0; t[1] = 0; t[2] = 0; t[3] = 0; t[4] = 0; t[5] = 1; t[6] = 1;
    rmHut_FaceCacheX[i][SVH_PIT] = new RangeModel(7,t,compress); // after SVH_PIT
    t[0] = 0; t[1] = 0; t[2] = 0; t[3] = 0; t[4] = 0; t[5] = 0; t[6] = 1;
    rmHut_FaceCacheX[i][SVH_DEN] = new RangeModel(7,t,compress); // after SVH_DEN
  }
#else
  rmHut_FaceCache = new RangeModel* [SVH_NUMBER_OF_OPERATIONS];
  t[0] = 1; t[1] = 1; t[2] = 1; t[3] = 1; t[4] = 1; t[5] = 1; t[6] = 1;
  rmHut_FaceCache[SVH_START] = new RangeModel(7,t,compress); // after SVH_START
  t[0] = 0; t[1] = 1; t[2] = 1; t[3] = 1; t[4] = 1; t[5] = 1; t[6] = 1;
  rmHut_FaceCache[SVH_HUT] = new RangeModel(7,t,compress); // after SVH_HUT
  t[0] = 0; t[1] = 1; t[2] = 1; t[3] = 1; t[4] = 1; t[5] = 0; t[6] = 1;
  rmHut_FaceCache[SVH_ROOF] = new RangeModel(7,t,compress); // after SVH_ROOF
  t[0] = 0; t[1] = 0; t[2] = 1; t[3] = 1; t[4] = 1; t[5] = 1; t[6] = 1;
  rmHut_FaceCache[SVH_STEP] = new RangeModel(7,t,compress); // after SVH_STEP
  t[0] = 0; t[1] = 0; t[2] = 0; t[3] = 1; t[4] = 1; t[5] = 1; t[6] = 1;
  rmHut_FaceCache[SVH_CORNER] = new RangeModel(7,t,compress); // after SVH_CORNER
  t[0] = 0; t[1] = 0; t[2] = 1; t[3] = 0; t[4] = 1; t[5] = 1;
  rmHut_FaceCache[SVH_BRIDGE] = new RangeModel(7,t,compress); // after SVH_BRIDGE
  t[0] = 0; t[1] = 0; t[2] = 1; t[3] = 0; t[4] = 1; t[5] = 0; t[6] = 1;
  rmHut_FaceCache[SVH_TUNNEL] = new RangeModel(7,t,compress); // after SVH_TUNNEL
  t[0] = 0; t[1] = 0; t[2] = 0; t[3] = 0; t[4] = 1; t[5] = 1; t[6] = 1;
  rmHut_FaceCache[SVH_GAP] = new RangeModel(7,t,compress); // after SVH_GAP
  t[0] = 0; t[1] = 0; t[2] = 0; t[3] = 0; t[4] = 0; t[5] = 1; t[6] = 1;
  rmHut_FaceCache[SVH_PIT] = new RangeModel(7,t,compress); // after SVH_PIT
  t[0] = 0; t[1] = 0; t[2] = 0; t[3] = 0; t[4] = 0; t[5] = 0; t[6] = 1;
  rmHut_FaceCache[SVH_DEN] = new RangeModel(7,t,compress); // after SVH_DEN
#endif

  rmHut_EdgeCache = new RangeModel* [SVH_NUMBER_OF_OPERATIONS];
  for(j=0; j< SVH_NUMBER_OF_OPERATIONS; j++) //for each operation
    rmHut_EdgeCache[j] = new RangeModel(25,0,compress);

  rmHut_VertexCache = new RangeModel* [SVH_NUMBER_OF_OPERATIONS];
  for(j=0; j< SVH_NUMBER_OF_OPERATIONS; j++) //for each operation
    rmHut_VertexCache[j] = new RangeModel(9,0,compress);


  // range tables for Roof operation
#ifdef SVH_FACECACHE_DEPENDS_ON_FINALIZATION
  rmRoof_FaceCacheX = new RangeModel** [20];
  for(i = 0; i < 20; i++)
  {
    rmRoof_FaceCacheX[i] = new RangeModel* [SVH_NUMBER_OF_OPERATIONS];
    t[0] = 1; t[1] = 1; t[2] = 1; t[3] = 1; t[4] = 1; t[5] = 1; t[6] = 1;
    rmRoof_FaceCacheX[i][SVH_START] = new RangeModel(7,t,compress); // after SVH_START
    t[0] = 0; t[1] = 1; t[2] = 1; t[3] = 1; t[4] = 1; t[5] = 1; t[6] = 1;
    rmRoof_FaceCacheX[i][SVH_HUT] = new RangeModel(7,t,compress); // after SVH_HUT
    t[0] = 0; t[1] = 1; t[2] = 1; t[3] = 1; t[4] = 1; t[5] = 0; t[6] = 1;
    rmRoof_FaceCacheX[i][SVH_ROOF] = new RangeModel(7,t,compress); // after SVH_ROOF
    t[0] = 0; t[1] = 0; t[2] = 1; t[3] = 1; t[4] = 1; t[5] = 1; t[6] = 1;
    rmRoof_FaceCacheX[i][SVH_STEP] = new RangeModel(7,t,compress); // after SVH_STEP
    t[0] = 0; t[1] = 0; t[2] = 0; t[3] = 1; t[4] = 1; t[5] = 1; t[6] = 1;
    rmRoof_FaceCacheX[i][SVH_CORNER] = new RangeModel(7,t,compress); // after SVH_CORNER
    t[0] = 0; t[1] = 0; t[2] = 1; t[3] = 0; t[4] = 1; t[5] = 1; t[6] = 1;
    rmRoof_FaceCacheX[i][SVH_BRIDGE] = new RangeModel(7,t,compress); // after SVH_BRIDGE
    t[0] = 0; t[1] = 0; t[2] = 1; t[3] = 0; t[4] = 1; t[5] = 0; t[6] = 1;
    rmRoof_FaceCacheX[i][SVH_TUNNEL] = new RangeModel(7,t,compress); // after SVH_TUNNEL
    t[0] = 0; t[1] = 0; t[2] = 0; t[3] = 0; t[4] = 1; t[5] = 1; t[6] = 1;
    rmRoof_FaceCacheX[i][SVH_GAP] = new RangeModel(7,t,compress); // after SVH_GAP
    t[0] = 0; t[1] = 0; t[2] = 0; t[3] = 0; t[4] = 0; t[5] = 1; t[6] = 1;
    rmRoof_FaceCacheX[i][SVH_PIT] = new RangeModel(7,t,compress); // after SVH_PIT
    t[0] = 0; t[1] = 0; t[2] = 0; t[3] = 0; t[4] = 0; t[5] = 0; t[6] = 1;
    rmRoof_FaceCacheX[i][SVH_DEN] = new RangeModel(7,t,compress); // after SVH_DEN
  }
#else
  rmRoof_FaceCache = new RangeModel* [SVH_NUMBER_OF_OPERATIONS];
  t[0] = 1; t[1] = 1; t[2] = 1; t[3] = 1; t[4] = 1; t[5] = 1; t[6] = 1;
  rmRoof_FaceCache[SVH_START] = new RangeModel(7,t,compress); // after SVH_START
  t[0] = 0; t[1] = 1; t[2] = 1; t[3] = 1; t[4] = 1; t[5] = 1; t[6] = 1;
  rmRoof_FaceCache[SVH_HUT] = new RangeModel(7,t,compress); // after SVH_HUT
  t[0] = 0; t[1] = 1; t[2] = 1; t[3] = 1; t[4] = 1; t[5] = 0; t[6] = 1;
  rmRoof_FaceCache[SVH_ROOF] = new RangeModel(7,t,compress); // after SVH_ROOF
  t[0] = 0; t[1] = 0; t[2] = 1; t[3] = 1; t[4] = 1; t[5] = 1; t[6] = 1;
  rmRoof_FaceCache[SVH_STEP] = new RangeModel(7,t,compress); // after SVH_STEP
  t[0] = 0; t[1] = 0; t[2] = 0; t[3] = 1; t[4] = 1; t[5] = 1; t[6] = 1;
  rmRoof_FaceCache[SVH_CORNER] = new RangeModel(7,t,compress); // after SVH_CORNER
  t[0] = 0; t[1] = 0; t[2] = 1; t[3] = 0; t[4] = 1; t[5] = 1; t[6] = 1;
  rmRoof_FaceCache[SVH_BRIDGE] = new RangeModel(7,t,compress); // after SVH_BRIDGE
  t[0] = 0; t[1] = 0; t[2] = 1; t[3] = 0; t[4] = 1; t[5] = 0; t[6] = 1;
  rmRoof_FaceCache[SVH_TUNNEL] = new RangeModel(7,t,compress); // after SVH_TUNNEL
  t[0] = 0; t[1] = 0; t[2] = 0; t[3] = 0; t[4] = 1; t[5] = 1; t[6] = 1;
  rmRoof_FaceCache[SVH_GAP] = new RangeModel(7,t,compress); // after SVH_GAP
  t[0] = 0; t[1] = 0; t[2] = 0; t[3] = 0; t[4] = 0; t[5] = 1; t[6] = 1;
  rmRoof_FaceCache[SVH_PIT] = new RangeModel(7,t,compress); // after SVH_PIT
  t[0] = 0; t[1] = 0; t[2] = 0; t[3] = 0; t[4] = 0; t[5] = 0; t[6] = 1;
  rmRoof_FaceCache[SVH_DEN] = new RangeModel(7,t,compress); // after SVH_DEN
#endif

  rmRoof_EdgeCache = new RangeModel* [SVH_NUMBER_OF_OPERATIONS];
  for(j=0; j< SVH_NUMBER_OF_OPERATIONS; j++) //for each operation
    rmRoof_EdgeCache[j] = new RangeModel(25,0,compress);

  rmRoof_VertexCache = new RangeModel* [SVH_NUMBER_OF_OPERATIONS];
  for(j=0; j< SVH_NUMBER_OF_OPERATIONS; j++) //for each operation
    rmRoof_VertexCache[j] = new RangeModel(9,0,compress);

  // range tables for Step (Sp) operation
#ifdef SVH_FACECACHE_DEPENDS_ON_FINALIZATION
  rmSCGPD_FaceCacheX = new RangeModel** [20];
  for(i = 0; i < 20; i++)
  {
    rmSCGPD_FaceCacheX[i] = new RangeModel* [SVH_NUMBER_OF_OPERATIONS];
    t[0] = 1; t[1] = 1; t[2] = 1; t[3] = 1; t[4] = 1; t[5] = 1;
    t[6] = 1; t[7] = 1; t[8] = 1; t[9] = 1; t[10] = 1; t[11] = 1;
    t[12] = 1; t[13] = 1; t[14] = 1; t[15] = 1; t[16] = 1; t[17] = 1;
    t[18] = 1; t[19] = 1; t[20] = 1; t[21] = 1; t[22] = 1; t[23] = 1; t[24] = 1;
    rmSCGPD_FaceCacheX[i][SVH_START] = new RangeModel(25,t,compress); // after SVH_START
    t[0] = 0; t[1] = 1; t[2] = 1; t[3] = 1; t[4] = 1; t[5] = 1;
    t[6] = 0; t[7] = 1; t[8] = 1; t[9] = 1; t[10] = 1; t[11] = 1;
    t[12] = 0; t[13] = 1; t[14] = 1; t[15] = 1; t[16] = 1; t[17] = 1;
    t[18] = 0; t[19] = 1; t[20] = 1; t[21] = 1; t[22] = 1; t[23] = 1; t[24] = 1;
    rmSCGPD_FaceCacheX[i][SVH_HUT] = new RangeModel(25,t,compress); // after SVH_HUT
    t[0] = 0; t[1] = 1; t[2] = 1; t[3] = 1; t[4] = 1; t[5] = 0;
    t[6] = 0; t[7] = 1; t[8] = 1; t[9] = 1; t[10] = 1; t[11] = 0;
    t[12] = 0; t[13] = 1; t[14] = 1; t[15] = 1; t[16] = 1; t[17] = 0;
    t[18] = 0; t[19] = 1; t[20] = 1; t[21] = 1; t[22] = 1; t[23] = 0; t[24] = 1;
    rmSCGPD_FaceCacheX[i][SVH_ROOF] = new RangeModel(25,t,compress); // after SVH_ROOF
    t[0] = 0; t[1] = 0; t[2] = 1; t[3] = 1; t[4] = 1; t[5] = 1;
    t[6] = 0; t[7] = 0; t[8] = 1; t[9] = 1; t[10] = 1; t[11] = 1;
    t[12] = 0; t[13] = 0; t[14] = 1; t[15] = 1; t[16] = 1; t[17] = 1;
    t[18] = 0; t[19] = 0; t[20] = 1; t[21] = 1; t[22] = 1; t[23] = 1; t[24] = 1;
    rmSCGPD_FaceCacheX[i][SVH_STEP] = new RangeModel(25,t,compress); // after SVH_STEP
    t[0] = 0; t[1] = 0; t[2] = 0; t[3] = 1; t[4] = 1; t[5] = 1;
    t[6] = 0; t[7] = 0; t[8] = 0; t[9] = 1; t[10] = 1; t[11] = 1;
    t[12] = 0; t[13] = 0; t[14] = 0; t[15] = 1; t[16] = 1; t[17] = 1;
    t[18] = 0; t[19] = 0; t[20] = 0; t[21] = 1; t[22] = 1; t[23] = 1; t[24] = 1;
    rmSCGPD_FaceCacheX[i][SVH_CORNER] = new RangeModel(25,t,compress); // after SVH_CORNER
    t[0] = 0; t[1] = 0; t[2] = 1; t[3] = 0; t[4] = 1; t[5] = 1;
    t[6] = 0; t[7] = 0; t[8] = 1; t[9] = 0; t[10] = 1; t[11] = 1;
    t[12] = 0; t[13] = 0; t[14] = 1; t[15] = 0; t[16] = 1; t[17] = 1;
    t[18] = 0; t[19] = 0; t[20] = 1; t[21] = 0; t[22] = 1; t[23] = 1; t[24] = 1;
    rmSCGPD_FaceCacheX[i][SVH_BRIDGE] = new RangeModel(25,t,compress); // after SVH_BRIDGE
    t[0] = 0; t[1] = 0; t[2] = 1; t[3] = 0; t[4] = 1; t[5] = 0; 
    t[6] = 0; t[7] = 0; t[8] = 1; t[9] = 0; t[10] = 1; t[11] = 0;
    t[12] = 0; t[13] = 0; t[14] = 1; t[15] = 0; t[16] = 1; t[17] = 0;
    t[18] = 0; t[19] = 0; t[20] = 1; t[21] = 0; t[22] = 1; t[23] = 0; t[24] = 1;
    rmSCGPD_FaceCacheX[i][SVH_TUNNEL] = new RangeModel(25,t,compress); // after SVH_TUNNEL
    t[0] = 0; t[1] = 0; t[2] = 0; t[3] = 0; t[4] = 1; t[5] = 1; 
    t[6] = 0; t[7] = 0; t[8] = 0; t[9] = 0; t[10] = 1; t[11] = 1;
    t[12] = 0; t[13] = 0; t[14] = 0; t[15] = 0; t[16] = 1; t[17] = 1;
    t[18] = 0; t[19] = 0; t[20] = 0; t[21] = 0; t[22] = 1; t[23] = 1; t[24] = 1;
    rmSCGPD_FaceCacheX[i][SVH_GAP] = new RangeModel(25,t,compress); // after SVH_GAP
    t[0] = 0; t[1] = 0; t[2] = 0; t[3] = 0; t[4] = 0; t[5] = 1; 
    t[6] = 0; t[7] = 0; t[8] = 0; t[9] = 0; t[10] = 0; t[11] = 1;
    t[12] = 0; t[13] = 0; t[14] = 0; t[15] = 0; t[16] = 0; t[17] = 1;
    t[18] = 0; t[19] = 0; t[20] = 0; t[21] = 0; t[22] = 0; t[23] = 1; t[24] = 1;
    rmSCGPD_FaceCacheX[i][SVH_PIT] = new RangeModel(25,t,compress); // after SVH_PIT
    t[0] = 0; t[1] = 0; t[2] = 0; t[3] = 0; t[4] = 0; t[5] = 0; 
    t[6] = 0; t[7] = 0; t[8] = 0; t[9] = 0; t[10] = 0; t[11] = 0;
    t[12] = 0; t[13] = 0; t[14] = 0; t[15] = 0; t[16] = 0; t[17] = 0;
    t[18] = 0; t[19] = 0; t[20] = 0; t[21] = 0; t[22] = 0; t[23] = 0; t[24] = 1;
    rmSCGPD_FaceCacheX[i][SVH_DEN] = new RangeModel(25,t,compress); // after SVH_DEN
  }
#else
  rmSCGPD_FaceCache = new RangeModel* [SVH_NUMBER_OF_OPERATIONS];
  t[0] = 1; t[1] = 1; t[2] = 1; t[3] = 1; t[4] = 1; t[5] = 1;
  t[6] = 1; t[7] = 1; t[8] = 1; t[9] = 1; t[10] = 1; t[11] = 1;
  t[12] = 1; t[13] = 1; t[14] = 1; t[15] = 1; t[16] = 1; t[17] = 1;
  t[18] = 1; t[19] = 1; t[20] = 1; t[21] = 1; t[22] = 1; t[23] = 1; t[24] = 1;
  rmSCGPD_FaceCache[SVH_START] = new RangeModel(25,t,compress); // after SVH_START
  t[0] = 0; t[1] = 1; t[2] = 1; t[3] = 1; t[4] = 1; t[5] = 1;
  t[6] = 0; t[7] = 1; t[8] = 1; t[9] = 1; t[10] = 1; t[11] = 1;
  t[12] = 0; t[13] = 1; t[14] = 1; t[15] = 1; t[16] = 1; t[17] = 1;
  t[18] = 0; t[19] = 1; t[20] = 1; t[21] = 1; t[22] = 1; t[23] = 1; t[24] = 1;
  rmSCGPD_FaceCache[SVH_HUT] = new RangeModel(25,t,compress); // after SVH_HUT
  t[0] = 0; t[1] = 1; t[2] = 1; t[3] = 1; t[4] = 1; t[5] = 0;
  t[6] = 0; t[7] = 1; t[8] = 1; t[9] = 1; t[10] = 1; t[11] = 0;
  t[12] = 0; t[13] = 1; t[14] = 1; t[15] = 1; t[16] = 1; t[17] = 0;
  t[18] = 0; t[19] = 1; t[20] = 1; t[21] = 1; t[22] = 1; t[23] = 0; t[24] = 1;
  rmSCGPD_FaceCache[SVH_ROOF] = new RangeModel(25,t,compress); // after SVH_ROOF
  t[0] = 0; t[1] = 0; t[2] = 1; t[3] = 1; t[4] = 1; t[5] = 1;
  t[6] = 0; t[7] = 0; t[8] = 1; t[9] = 1; t[10] = 1; t[11] = 1;
  t[12] = 0; t[13] = 0; t[14] = 1; t[15] = 1; t[16] = 1; t[17] = 1;
  t[18] = 0; t[19] = 0; t[20] = 1; t[21] = 1; t[22] = 1; t[23] = 1; t[24] = 1;
  rmSCGPD_FaceCache[SVH_STEP] = new RangeModel(25,t,compress); // after SVH_STEP
  t[0] = 0; t[1] = 0; t[2] = 0; t[3] = 1; t[4] = 1; t[5] = 1;
  t[6] = 0; t[7] = 0; t[8] = 0; t[9] = 1; t[10] = 1; t[11] = 1;
  t[12] = 0; t[13] = 0; t[14] = 0; t[15] = 1; t[16] = 1; t[17] = 1;
  t[18] = 0; t[19] = 0; t[20] = 0; t[21] = 1; t[22] = 1; t[23] = 1; t[24] = 1;
  rmSCGPD_FaceCache[SVH_CORNER] = new RangeModel(25,t,compress); // after SVH_CORNER
  t[0] = 0; t[1] = 0; t[2] = 1; t[3] = 0; t[4] = 1; t[5] = 1;
  t[6] = 0; t[7] = 0; t[8] = 1; t[9] = 0; t[10] = 1; t[11] = 1;
  t[12] = 0; t[13] = 0; t[14] = 1; t[15] = 0; t[16] = 1; t[17] = 1;
  t[18] = 0; t[19] = 0; t[20] = 1; t[21] = 0; t[22] = 1; t[23] = 1; t[24] = 1;
  rmSCGPD_FaceCache[SVH_BRIDGE] = new RangeModel(25,t,compress); // after SVH_BRIDGE
  t[0] = 0; t[1] = 0; t[2] = 1; t[3] = 0; t[4] = 1; t[5] = 0; 
  t[6] = 0; t[7] = 0; t[8] = 1; t[9] = 0; t[10] = 1; t[11] = 0;
  t[12] = 0; t[13] = 0; t[14] = 1; t[15] = 0; t[16] = 1; t[17] = 0;
  t[18] = 0; t[19] = 0; t[20] = 1; t[21] = 0; t[22] = 1; t[23] = 0; t[24] = 1;
  rmSCGPD_FaceCache[SVH_TUNNEL] = new RangeModel(25,t,compress); // after SVH_TUNNEL
  t[0] = 0; t[1] = 0; t[2] = 0; t[3] = 0; t[4] = 1; t[5] = 1; 
  t[6] = 0; t[7] = 0; t[8] = 0; t[9] = 0; t[10] = 1; t[11] = 1;
  t[12] = 0; t[13] = 0; t[14] = 0; t[15] = 0; t[16] = 1; t[17] = 1;
  t[18] = 0; t[19] = 0; t[20] = 0; t[21] = 0; t[22] = 1; t[23] = 1; t[24] = 1;
  rmSCGPD_FaceCache[SVH_GAP] = new RangeModel(25,t,compress); // after SVH_GAP
  t[0] = 0; t[1] = 0; t[2] = 0; t[3] = 0; t[4] = 0; t[5] = 1; 
  t[6] = 0; t[7] = 0; t[8] = 0; t[9] = 0; t[10] = 0; t[11] = 1;
  t[12] = 0; t[13] = 0; t[14] = 0; t[15] = 0; t[16] = 0; t[17] = 1;
  t[18] = 0; t[19] = 0; t[20] = 0; t[21] = 0; t[22] = 0; t[23] = 1; t[24] = 1;
  rmSCGPD_FaceCache[SVH_PIT] = new RangeModel(25,t,compress); // after SVH_PIT
  t[0] = 0; t[1] = 0; t[2] = 0; t[3] = 0; t[4] = 0; t[5] = 0; 
  t[6] = 0; t[7] = 0; t[8] = 0; t[9] = 0; t[10] = 0; t[11] = 0;
  t[12] = 0; t[13] = 0; t[14] = 0; t[15] = 0; t[16] = 0; t[17] = 0;
  t[18] = 0; t[19] = 0; t[20] = 0; t[21] = 0; t[22] = 0; t[23] = 0; t[24] = 1;
  rmSCGPD_FaceCache[SVH_DEN] = new RangeModel(25,t,compress); // after SVH_DEN
#endif

  rmSCGPD_EdgeCache = new RangeModel* [SVH_NUMBER_OF_OPERATIONS];
  for(j=0; j< SVH_NUMBER_OF_OPERATIONS; j++) //for each operation
    rmSCGPD_EdgeCache[j] = new RangeModel(97,0,compress);

  rmSCGPD_VertexCache = new RangeModel* [SVH_NUMBER_OF_OPERATIONS];
  for(j=0; j< SVH_NUMBER_OF_OPERATIONS; j++) //for each operation
    rmSCGPD_VertexCache[j] = new RangeModel(33,0,compress);

// range tables for Bridge operation
#ifdef SVH_FACECACHE_DEPENDS_ON_FINALIZATION
  rmBT_FaceCacheX = new RangeModel** [20];
  for(i = 0; i < 20; i++)
  {
    rmBT_FaceCacheX[i] = new RangeModel* [SVH_NUMBER_OF_OPERATIONS];
    t[0] = 1; t[1] = 1; t[2] = 1; t[3] = 1; t[4] = 1; t[5] = 1;
    t[6] = 1; t[7] = 1; t[8] = 1; t[9] = 1; t[10] = 1; t[11] = 1; t[12] = 1;
    rmBT_FaceCacheX[i][SVH_START] = new RangeModel(13,t,compress); // after SVH_START
    t[0] = 0; t[1] = 1; t[2] = 1; t[3] = 1; t[4] = 1; t[5] = 1;
    t[6] = 0; t[7] = 1; t[8] = 1; t[9] = 1; t[10] = 1; t[11] = 1; t[12] = 1;
    rmBT_FaceCacheX[i][SVH_HUT] = new RangeModel(13,t,compress); // after SVH_HUT
    t[0] = 0; t[1] = 1; t[2] = 1; t[3] = 1; t[4] = 1; t[5] = 0;
    t[6] = 0; t[7] = 1; t[8] = 1; t[9] = 1; t[10] = 1; t[11] = 0; t[12] = 1;
    rmBT_FaceCacheX[i][SVH_ROOF] = new RangeModel(13,t,compress); // after SVH_ROOF
    t[0] = 0; t[1] = 0; t[2] = 1; t[3] = 1; t[4] = 1; t[5] = 1;
    t[6] = 0; t[7] = 0; t[8] = 1; t[9] = 1; t[10] = 1; t[11] = 1; t[12] = 1;
    rmBT_FaceCacheX[i][SVH_STEP] = new RangeModel(13,t,compress); // after SVH_STEP
    t[0] = 0; t[1] = 0; t[2] = 0; t[3] = 1; t[4] = 1; t[5] = 1;
    t[6] = 0; t[7] = 0; t[8] = 0; t[9] = 1; t[10] = 1; t[11] = 1; t[12] = 1;
    rmBT_FaceCacheX[i][SVH_CORNER] = new RangeModel(13,t,compress); // after SVH_CORNER
    t[0] = 0; t[1] = 0; t[2] = 1; t[3] = 0; t[4] = 1; t[5] = 1;
    t[6] = 0; t[7] = 0; t[8] = 1; t[9] = 0; t[10] = 1; t[11] = 1; t[12] = 1;
    rmBT_FaceCacheX[i][SVH_BRIDGE] = new RangeModel(13,t,compress); // after SVH_BRIDGE
    t[0] = 0; t[1] = 0; t[2] = 1; t[3] = 0; t[4] = 1; t[5] = 0; 
    t[6] = 0; t[7] = 0; t[8] = 1; t[9] = 0; t[10] = 1; t[11] = 0; t[12] = 1;
    rmBT_FaceCacheX[i][SVH_TUNNEL] = new RangeModel(13,t,compress); // after SVH_TUNNEL
    t[0] = 0; t[1] = 0; t[2] = 0; t[3] = 0; t[4] = 1; t[5] = 1; 
    t[6] = 0; t[7] = 0; t[8] = 0; t[9] = 0; t[10] = 1; t[11] = 1; t[12] = 1;
    rmBT_FaceCacheX[i][SVH_GAP] = new RangeModel(13,t,compress); // after SVH_GAP
    t[0] = 0; t[1] = 0; t[2] = 0; t[3] = 0; t[4] = 0; t[5] = 1; 
    t[6] = 0; t[7] = 0; t[8] = 0; t[9] = 0; t[10] = 0; t[11] = 1; t[12] = 1;
    rmBT_FaceCacheX[i][SVH_PIT] = new RangeModel(13,t,compress); // after SVH_PIT
    t[0] = 0; t[1] = 0; t[2] = 0; t[3] = 0; t[4] = 0; t[5] = 0; 
    t[6] = 0; t[7] = 0; t[8] = 0; t[9] = 0; t[10] = 0; t[11] = 0; t[12] = 1;
    rmBT_FaceCacheX[i][SVH_DEN] = new RangeModel(13,t,compress); // after SVH_DEN
  }
#else
  rmBT_FaceCache = new RangeModel* [SVH_NUMBER_OF_OPERATIONS];
  t[0] = 1; t[1] = 1; t[2] = 1; t[3] = 1; t[4] = 1; t[5] = 1;
  t[6] = 1; t[7] = 1; t[8] = 1; t[9] = 1; t[10] = 1; t[11] = 1; t[12] = 1;
  rmBT_FaceCache[SVH_START] = new RangeModel(13,t,compress); // after SVH_START
  t[0] = 0; t[1] = 1; t[2] = 1; t[3] = 1; t[4] = 1; t[5] = 1;
  t[6] = 0; t[7] = 1; t[8] = 1; t[9] = 1; t[10] = 1; t[11] = 1; t[12] = 1;
  rmBT_FaceCache[SVH_HUT] = new RangeModel(13,t,compress); // after SVH_HUT
  t[0] = 0; t[1] = 1; t[2] = 1; t[3] = 1; t[4] = 1; t[5] = 0;
  t[6] = 0; t[7] = 1; t[8] = 1; t[9] = 1; t[10] = 1; t[11] = 0; t[12] = 1;
  rmBT_FaceCache[SVH_ROOF] = new RangeModel(13,t,compress); // after SVH_ROOF
  t[0] = 0; t[1] = 0; t[2] = 1; t[3] = 1; t[4] = 1; t[5] = 1;
  t[6] = 0; t[7] = 0; t[8] = 1; t[9] = 1; t[10] = 1; t[11] = 1; t[12] = 1;
  rmBT_FaceCache[SVH_STEP] = new RangeModel(13,t,compress); // after SVH_STEP
  t[0] = 0; t[1] = 0; t[2] = 0; t[3] = 1; t[4] = 1; t[5] = 1;
  t[6] = 0; t[7] = 0; t[8] = 0; t[9] = 1; t[10] = 1; t[11] = 1; t[12] = 1;
  rmBT_FaceCache[SVH_CORNER] = new RangeModel(13,t,compress); // after SVH_CORNER
  t[0] = 0; t[1] = 0; t[2] = 1; t[3] = 0; t[4] = 1; t[5] = 1;
  t[6] = 0; t[7] = 0; t[8] = 1; t[9] = 0; t[10] = 1; t[11] = 1; t[12] = 1;
  rmBT_FaceCache[SVH_BRIDGE] = new RangeModel(13,t,compress); // after SVH_BRIDGE
  t[0] = 0; t[1] = 0; t[2] = 1; t[3] = 0; t[4] = 1; t[5] = 0; 
  t[6] = 0; t[7] = 0; t[8] = 1; t[9] = 0; t[10] = 1; t[11] = 0; t[12] = 1;
  rmBT_FaceCache[SVH_TUNNEL] = new RangeModel(13,t,compress); // after SVH_TUNNEL
  t[0] = 0; t[1] = 0; t[2] = 0; t[3] = 0; t[4] = 1; t[5] = 1; 
  t[6] = 0; t[7] = 0; t[8] = 0; t[9] = 0; t[10] = 1; t[11] = 1; t[12] = 1;
  rmBT_FaceCache[SVH_GAP] = new RangeModel(13,t,compress); // after SVH_GAP
  t[0] = 0; t[1] = 0; t[2] = 0; t[3] = 0; t[4] = 0; t[5] = 1; 
  t[6] = 0; t[7] = 0; t[8] = 0; t[9] = 0; t[10] = 0; t[11] = 1; t[12] = 1;
  rmBT_FaceCache[SVH_PIT] = new RangeModel(13,t,compress); // after SVH_PIT
  t[0] = 0; t[1] = 0; t[2] = 0; t[3] = 0; t[4] = 0; t[5] = 0; 
  t[6] = 0; t[7] = 0; t[8] = 0; t[9] = 0; t[10] = 0; t[11] = 0; t[12] = 1;
  rmBT_FaceCache[SVH_DEN] = new RangeModel(13,t,compress); // after SVH_DEN
#endif

  rmBT_EdgeCache = new RangeModel* [SVH_NUMBER_OF_OPERATIONS];
  for(j=0; j< SVH_NUMBER_OF_OPERATIONS; j++) //for each operation
    rmBT_EdgeCache[j] = new RangeModel(49,0,compress);

  rmBT_VertexCache = new RangeModel* [SVH_NUMBER_OF_OPERATIONS];
  for(j=0; j< SVH_NUMBER_OF_OPERATIONS; j++) //for each operation
    rmBT_VertexCache[j] = new RangeModel(17,0,compress);

  rm_VertexCandi = new RangeModel*[SVH_MAGIC_NUMBER_F0_AROUND_V1_RM];
  for(i = 2; i < SVH_MAGIC_NUMBER_F0_AROUND_V1_RM; i++)
  {
    rm_VertexCandi[i] = new RangeModel(i,0,compress);
  }

  rm_FaceCandi = new RangeModel(2,0,compress);

  rmFinalized = new RangeModel**[SVH_MAX_DEGREE_ONE];
  for(i = 0; i < SVH_MAX_DEGREE_ONE; i++)
  {
    rmFinalized[i] = new RangeModel*[SVH_MAX_USE_COUNT];
    for(j = 1; j < SVH_MAX_USE_COUNT; j++)
    {
      rmFinalized[i][j] = new RangeModel(2,0,compress);
    }
  }

  t[0]=0; t[1]=1;
  rm_ConfirmFaceMerge = new RangeModel(2,t,compress);

  //range table for HUT/STEP/CORNER operations, to tell whether or not there are joined verts.
  rmHut_HasJoin = new RangeModel(2,0,compress); 
  rmStep_HasJoin = new RangeModel(2,0,compress); 
  rmCorner_HasJoin = new RangeModel(2,0,compress); 

#ifdef SVH_DUMB_JOIN_CODING
  //range table for HUT/STEP/CORNER operations, to tell which verts are joined
  rmHut_JoinedVertices = new RangeModel(15,0,compress); 
  rmStep_JoinedVertices = new RangeModel(3,0,compress); 
#else
  //from which vertex we can join v6 with an edge:
  // 0 means 2, 1 means 5, 2 means 7, 3 means neither.
  rmCorner_EdgeJoinFrom = new RangeModel(4,0,compress);
#endif
}

static void finishModels()
{
  int i,j;

  // range table for whether there is more
  delete rmDone;

  // range table for which operation it is
#ifdef SVH_RMOP_DEPENDS_ON_FINALIZATION
  for(i=0; i<SVH_NUMBER_OF_OPERATIONS; i++)
  {
    for(j=0; j<SVH_NUMBER_OF_OPERATIONS; j++)
      delete rmOpX[i][j];
    delete[] rmOpX[i];
  }
  delete[] rmOpX;
#else
  for(j=0; j<SVH_NUMBER_OF_OPERATIONS; j++)
    delete rmOp[j];
  delete[] rmOp;
#endif

#ifdef SVH_DUMB_JOIN_CODING
  //range table for HUT(H)/STEP(Sp)/CORNER(C) operations, to tell whether or not there are joined verts.
  delete rmHut_HasJoin;
  delete rmStep_HasJoin;
  delete rmCorner_HasJoin;

  delete rmHut_JoinedVertices;
  delete rmStep_JoinedVertices;
#endif

  // range tables for start (S) operations
  for(j=0; j< 8; j++) //for each vertex
    delete rmS_Cache[j];
  delete[] rmS_Cache;

  delete rmStart_VisitedVertices;

  // range tables for Hut operations
#ifdef SVH_FACECACHE_DEPENDS_ON_FINALIZATION
  for(i=0; i< 20; i++)
  {
    for(j=0; j< SVH_NUMBER_OF_OPERATIONS; j++) //for each operation
      delete rmHut_FaceCacheX[i][j];
    delete[] rmHut_FaceCacheX[i];
  }
  delete[] rmHut_FaceCacheX;
#else
  for(j=0; j< SVH_NUMBER_OF_OPERATIONS; j++) //for each operation
    delete rmHut_FaceCache[j];
  delete[] rmHut_FaceCache;
#endif

  for(j=0; j< SVH_NUMBER_OF_OPERATIONS; j++) //for each operation
    delete rmHut_EdgeCache[j];
  delete[] rmHut_EdgeCache;

  for(j=0; j< SVH_NUMBER_OF_OPERATIONS; j++) //for each operation
    delete rmHut_VertexCache[j];
  delete[] rmHut_VertexCache;


  // range tables for Roof operations
#ifdef SVH_FACECACHE_DEPENDS_ON_FINALIZATION
  for(i=0; i< 20; i++)
  {
    for(j=0; j< SVH_NUMBER_OF_OPERATIONS; j++) //for each operation
      delete rmRoof_FaceCacheX[i][j];
    delete[] rmRoof_FaceCacheX[i];
  }
  delete[] rmRoof_FaceCacheX;
#else
  for(j=0; j< SVH_NUMBER_OF_OPERATIONS; j++) //for each operation
    delete rmRoof_FaceCache[j];
  delete[] rmRoof_FaceCache;
#endif

  for(j=0; j< SVH_NUMBER_OF_OPERATIONS; j++) //for each operation
    delete rmRoof_EdgeCache[j];
  delete[] rmRoof_EdgeCache;

  for(j=0; j< SVH_NUMBER_OF_OPERATIONS; j++) //for each operation
    delete rmRoof_VertexCache[j];
  delete[] rmRoof_VertexCache;


  // range tables for Step operations
#ifdef SVH_FACECACHE_DEPENDS_ON_FINALIZATION
  for(i=0; i< 20; i++)
  {
    for(j=0; j< SVH_NUMBER_OF_OPERATIONS; j++) //for each operation
      delete rmSCGPD_FaceCacheX[i][j];
    delete[] rmSCGPD_FaceCacheX[i];
  }
  delete[] rmSCGPD_FaceCacheX;
#else
  for(j=0; j< SVH_NUMBER_OF_OPERATIONS; j++) //for each operation
    delete rmSCGPD_FaceCache[j];
  delete[] rmSCGPD_FaceCache;
#endif

  for(j=0; j< SVH_NUMBER_OF_OPERATIONS; j++) //for each operation
    delete rmSCGPD_EdgeCache[j];
  delete[] rmSCGPD_EdgeCache;

  for(j=0; j< SVH_NUMBER_OF_OPERATIONS; j++) //for each operation
    delete rmSCGPD_VertexCache[j];
  delete[] rmSCGPD_VertexCache;

  // range tables for Bridge operations
#ifdef SVH_FACECACHE_DEPENDS_ON_FINALIZATION
  for(i=0; i< 20; i++)
  {
    for(j=0; j< SVH_NUMBER_OF_OPERATIONS; j++) //for each operation
      delete rmBT_FaceCacheX[i][j];
    delete[] rmBT_FaceCacheX[i];
  }
  delete[] rmBT_FaceCacheX;
#else
  for(j=0; j< SVH_NUMBER_OF_OPERATIONS; j++) //for each operation
    delete rmBT_FaceCache[j];
  delete[] rmBT_FaceCache;
#endif
  for(j=0; j< SVH_NUMBER_OF_OPERATIONS; j++) //for each operation
    delete rmBT_EdgeCache[j];
  delete[] rmBT_EdgeCache;

  for(j=0; j< SVH_NUMBER_OF_OPERATIONS; j++) //for each operation
    delete rmBT_VertexCache[j];
  delete[] rmBT_VertexCache;

  for(i = 0; i < SVH_MAX_DEGREE_ONE; i++)
  {
    for(j = 1; j < SVH_MAX_USE_COUNT; j++)
      delete rmFinalized[i][j];
    delete[] rmFinalized[i];
  }
  delete[] rmFinalized;

  for(i = 2; i < SVH_MAGIC_NUMBER_F0_AROUND_V1_RM; i++)
    delete rm_VertexCandi[i];
  delete[] rm_VertexCandi;

  delete rm_FaceCandi;

  delete rm_ConfirmFaceMerge;
}

// efficient memory allocation for vertices

struct SVedge;

typedef struct SVvertex
{
	SVvertex* buffer_next;    // used for efficient memory management and by dynamicvector data structure
	float* v_prop_f;
	I32 index;
	I32 use_count;
	I32 in_cache;
	I32 finalize;
	int list_size;
	int list_alloc;
	SVedge** list; //list of faces around this vertex
} SVvertex;

static int vertex_pnum_f = 0;
static SVvertex* vertex_buffer_next;
static int vertex_buffer_alloc;
static int vertex_buffer_increment;
static int vertex_buffer_number_chunks;
static SVvertex* vertex_buffer_chunks[16];
#ifdef PRINT_CONTROL_OUTPUT
static int vertex_buffer_size;
static int vertex_buffer_maxsize;
#endif

static void initVertexBuffer(int alloc, int pnum_f)
{
  vertex_pnum_f = pnum_f;
  vertex_buffer_next = 0;
  vertex_buffer_alloc = 0;
  vertex_buffer_increment = alloc;
  vertex_buffer_number_chunks = 0;
#ifdef PRINT_CONTROL_OUTPUT
  vertex_buffer_size = 0;
  vertex_buffer_maxsize = 0;
#endif
}

static SVvertex* allocVertex()
{
  if(vertex_buffer_next == 0)
  {
    int increment = vertex_buffer_increment;
    vertex_buffer_increment *= 2;
    vertex_buffer_next =(SVvertex*) malloc(sizeof(SVvertex) * increment);
    if(vertex_buffer_next == 0)
    {
      fprintf(stderr,"ERROR (SVreader_svh): malloc for vertex buffer %d failed\n", increment);
      return 0;
    }
    else
    {
      vertex_buffer_chunks[vertex_buffer_number_chunks] = vertex_buffer_next;
      vertex_buffer_number_chunks++;      
    }
    vertex_buffer_alloc += increment;
    if(vertex_pnum_f)
    {
      vertex_buffer_next[0].v_prop_f =(float*) malloc(sizeof(float) * increment * vertex_pnum_f);
      if(vertex_buffer_next[0].v_prop_f == 0)
      {
        fprintf(stderr,"ERROR (SVreader_svh): malloc for vertex_buffer_next->v_prop_f %d failed\n", increment);
        return 0;
      }
    }
    for(int i = 0; i < increment; i++)
    {
      vertex_buffer_next[i].buffer_next = &(vertex_buffer_next[i+1]);
      vertex_buffer_next[i].list = 0;
      if(vertex_pnum_f)
      {
        vertex_buffer_next[i].v_prop_f = &(vertex_buffer_next[0].v_prop_f[i*vertex_pnum_f]);
      }
    }
    vertex_buffer_next[increment-1].buffer_next = 0;
  }
  // get pointer to next available vertex
  SVvertex* vertex = vertex_buffer_next;
  vertex_buffer_next = vertex->buffer_next;

  if(vertex->list == 0)
  {
    vertex->list =(SVedge**) malloc(sizeof(SVedge*) *SVH_INITIAL_VERTEX_LIST_SIZE);
    vertex->list_alloc = SVH_INITIAL_VERTEX_LIST_SIZE;
  }
  vertex->list_size = 0;
  vertex->use_count = 0;

#ifdef PRINT_CONTROL_OUTPUT
  vertex_buffer_size++;
  if(vertex_buffer_size > vertex_buffer_maxsize) vertex_buffer_maxsize = vertex_buffer_size;
#endif

  return vertex;
}

static void deallocVertex(SVvertex* vertex)
{
  vertex->buffer_next = vertex_buffer_next;
  vertex_buffer_next = vertex;
#ifdef PRINT_CONTROL_OUTPUT
  vertex_buffer_size--;
#endif
}

static void destroyVertexBuffer()
{
  for (int i = vertex_buffer_number_chunks-1; i >= 0; i--)
  {
    vertex_buffer_increment/=2;
    for (int j = 0; j < vertex_buffer_increment; j++)
    {
      if (vertex_buffer_chunks[i][j].list) free(vertex_buffer_chunks[i][j].list);
    }
    if(vertex_pnum_f)
    {
      free(vertex_buffer_chunks[i][0].v_prop_f);
    }
    free(vertex_buffer_chunks[i]);
  }
  vertex_buffer_number_chunks = 0;
  vertex_buffer_next = 0;
  vertex_buffer_alloc = 0;
}

// efficient memory allocation for half-edges

typedef struct SVedge     // this is an oriented half-edge
{
	SVedge* buffer_next;    // used for efficient memory management
	SVvertex* target;       // target vertex of oriented half-edge
	SVedge* next;           // next half-edge in ccw-oriented chain of half-edges around the face
#ifdef SVH_COMPRESS_CELL_DATA
	float* c_prop_f;
#endif
} SVedge;

#ifdef SVH_COMPRESS_CELL_DATA
static int cell_pnum_f = 0;
static float *cellDataTmp = 0;
#ifdef SVH_CELL_DATA_SECONDORDER_PREDICTION
static float *tmpCellDataResidual = 0;
#endif
#endif
static SVedge* edge_buffer_next;
static int edge_buffer_alloc;
static int edge_buffer_increment;
static int edge_buffer_number_chunks;
static SVedge* edge_buffer_chunks[16];
static int edge_buffer_size;
#ifdef PRINT_CONTROL_OUTPUT
static int edge_buffer_maxsize;
#endif

#ifdef SVH_COMPRESS_CELL_DATA
static void initEdgeBuffer(int alloc, int c_pnum_f)
#else
static void initEdgeBuffer(int alloc)
#endif
{
#ifdef SVH_COMPRESS_CELL_DATA
  cell_pnum_f = c_pnum_f;
  cellDataTmp = new float[c_pnum_f];
#ifdef SVH_CELL_DATA_SECONDORDER_PREDICTION
  tmpCellDataResidual = new float[c_pnum_f];
#endif
#endif
  edge_buffer_next = 0;
  edge_buffer_alloc = 0;
  edge_buffer_increment = alloc;
  edge_buffer_number_chunks = 0;
  edge_buffer_size = 0;
#ifdef PRINT_CONTROL_OUTPUT
  edge_buffer_maxsize = 0;
#endif
}

static SVedge* allocEdge()
{
  if(edge_buffer_next == 0)
  {
    int increment = edge_buffer_increment;
    edge_buffer_increment *= 2;
    edge_buffer_next =(SVedge*) malloc(sizeof(SVedge) * increment);
    if(edge_buffer_next == 0)
    {
      fprintf(stderr,"malloc for edge buffer failed\n");
      return 0;
    }
    else
    {
      edge_buffer_chunks[edge_buffer_number_chunks] = edge_buffer_next;
      edge_buffer_number_chunks++;      
    }
    edge_buffer_alloc += increment;
#ifdef SVH_COMPRESS_CELL_DATA
    if(cell_pnum_f)
    {
#ifdef SVH_CELL_DATA_SECONDORDER_PREDICTION
      edge_buffer_next[0].c_prop_f =(float*) malloc(sizeof(float) *increment*cell_pnum_f * 2);
#else
      edge_buffer_next[0].c_prop_f =(float*) malloc(sizeof(float) *increment*cell_pnum_f);
#endif
      if(edge_buffer_next[0].c_prop_f == 0)
      {
        fprintf(stderr,"malloc for edge_buffer_next->c_prop_f failed\n");
        return 0;
      }
    }
#endif
    for(int i = 0; i < increment; i++)
    {
      edge_buffer_next[i].buffer_next = &(edge_buffer_next[i+1]);
#ifdef SVH_COMPRESS_CELL_DATA
      if(cell_pnum_f)
      {
#ifdef SVH_CELL_DATA_SECONDORDER_PREDICTION
        edge_buffer_next[i].c_prop_f = &(edge_buffer_next[0].c_prop_f[i*cell_pnum_f*2]);
#else
        edge_buffer_next[i].c_prop_f = &(edge_buffer_next[0].c_prop_f[i*cell_pnum_f]);
#endif
      }
#endif
    }
    edge_buffer_next[increment-1].buffer_next = 0;
  }
  // get index of next available edge
  SVedge* edge = edge_buffer_next;
  edge_buffer_next = edge->buffer_next;

  edge->buffer_next = 0;
  edge->target = 0;
  edge->next = 0;

  edge_buffer_size++;

#ifdef PRINT_CONTROL_OUTPUT
  if(edge_buffer_size > edge_buffer_maxsize) edge_buffer_maxsize = edge_buffer_size;
#endif

  return edge;
}

static void deallocEdge(SVedge* edge)
{
  edge->buffer_next = edge_buffer_next;
  edge_buffer_next = edge;
  edge_buffer_size--;
}

static void destroyEdgeBuffer()
{
  for (int i = edge_buffer_number_chunks-1; i >= 0; i--)
  {
    edge_buffer_increment/=2;
#ifdef SVH_COMPRESS_CELL_DATA
    if(cell_pnum_f)
    {
      free(edge_buffer_chunks[i][0].c_prop_f);
    }
#endif
    free(edge_buffer_chunks[i]);
  }
  edge_buffer_number_chunks = 0;
  edge_buffer_next = 0;
  edge_buffer_alloc = 0;
}

static void addEdgeToVertex(SVedge* edge, SVvertex* vertex)
{
  if(vertex->list_size == vertex->list_alloc)
  {
    vertex->list_alloc += 2;
    vertex->list =(SVedge**) realloc(vertex->list, sizeof(SVedge*) *vertex->list_alloc);
  }
  vertex->list[vertex->list_size] = edge;
  vertex->list_size++;
}

static void removeEdgeFromVertex(SVedge* edge, SVvertex* vertex)
{
  int i;
  for(i = 0; i < vertex->list_size; i++)
  {
    if(vertex->list[i] == edge)
    {
      vertex->list_size--;
      vertex->list[i] = vertex->list[vertex->list_size];
      return;
    }
  }
  fprintf(stderr,"FATAL ERROR: did not find edge in vertex list\n");
  exit(0);
}
