/*
===============================================================================

  FILE:  svreaddereferenced.cpp
  
  CONTENTS:
  
    see corresponding header file

  PROGRAMMERS:
  
    martin isenburg@cs.unc.edu
  
  COPYRIGHT:
  
    copyright (C) 2007  martin isenburg@cs.unc.edu
    
    This software is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
  
  CHANGE HISTORY:
  
    see corresponding header file
  
===============================================================================
*/
#include "svreaddereferenced.h"

#include <stdio.h>
#include <assert.h>

//this is for compatibility with the new g++
#ifdef __GNUC__
	#include <ext/hash_map>
	#include <deque>
	namespace std { using namespace __gnu_cxx; }
	using namespace std;
#else
	#include <hash_map.h>
	#include <deque.h>
#endif


#include "vecnfv.h"
#include "vecniv.h"

#define PRINT_CONTROL_OUTPUT
#undef PRINT_CONTROL_OUTPUT

struct SVcell;

typedef struct SVvertex
{
  SVvertex* buffer_next;    // used for efficient memory management
  float* prop_f;
  int index;
  short list_size;
  short list_alloc;
  SVcell** list;
} SVvertex;

typedef struct SVcell
{
  SVcell* buffer_next;  // used for efficient memory management
  float* prop_f;
  SVvertex* vertices[8];
  char number_vertices;
  char arrived_vertices;
} SVcell;

typedef hash_map<int, SVvertex*> my_vertex_hash;
typedef deque<SVcell*> my_cell_queue;

static my_vertex_hash* vertex_hash = 0;
static my_cell_queue* cell_queue = 0;

// efficient memory allocation

static SVvertex* vertex_buffer_next;
static int vertex_buffer_increment;
static int vertex_buffer_number_chunks;
static SVvertex* vertex_buffer_chunks[16];
#ifdef PRINT_CONTROL_OUTPUT
static int vertex_buffer_size;
static int vertex_buffer_maxsize;
#endif // PRINT_CONTROL_OUTPUT

static void initVertexBuffer(int increment)
{
  vertex_buffer_next = 0;
  vertex_buffer_increment = increment;
  vertex_buffer_number_chunks = 0;
#ifdef PRINT_CONTROL_OUTPUT
  vertex_buffer_size = 0;
  vertex_buffer_maxsize = 0;
#endif // PRINT_CONTROL_OUTPUT
}

static SVvertex* allocVertex()
{
  if (vertex_buffer_next == 0)
  {
    vertex_buffer_next = (SVvertex*)malloc(sizeof(SVvertex)*vertex_buffer_increment);
    if (vertex_buffer_next == 0)
    {
      fprintf(stderr,"FATAL ERROR: malloc for vertex buffer failed\n");
      return 0;
    }
    else
    {
      vertex_buffer_chunks[vertex_buffer_number_chunks] = vertex_buffer_next;
      vertex_buffer_number_chunks++;
    }
    for (int i = 0; i < vertex_buffer_increment; i++)
    {
      vertex_buffer_next[i].buffer_next = &(vertex_buffer_next[i+1]);
      vertex_buffer_next[i].prop_f = 0;
      vertex_buffer_next[i].list = 0;
    }
    vertex_buffer_next[vertex_buffer_increment-1].buffer_next = 0;
    vertex_buffer_increment *= 2;
  }
  // get pointer to next available vertex
  SVvertex* vertex = vertex_buffer_next;
  vertex_buffer_next = vertex->buffer_next;

  // initialize data fields
  vertex->list_size = 0;
  vertex->buffer_next = 0;

#ifdef PRINT_CONTROL_OUTPUT
  vertex_buffer_size++;
  if (vertex_buffer_size > vertex_buffer_maxsize) vertex_buffer_maxsize = vertex_buffer_size;
#endif // PRINT_CONTROL_OUTPUT

  return vertex;
}

static void deallocVertex(SVvertex* vertex)
{
  vertex->buffer_next = vertex_buffer_next;
  vertex_buffer_next = vertex;
#ifdef PRINT_CONTROL_OUTPUT
  vertex_buffer_size--;
#endif // PRINT_CONTROL_OUTPUT
}

static void destroyVertexBuffer()
{
  for (int i = 0; i < vertex_buffer_number_chunks; i++)
  {
    free(vertex_buffer_chunks[i]);
  }
  vertex_buffer_number_chunks = 0;
  vertex_buffer_next = 0;
}

static SVcell* cell_buffer_next;
static int cell_buffer_increment;
static int cell_buffer_number_chunks;
static SVcell* cell_buffer_chunks[16];
#ifdef PRINT_CONTROL_OUTPUT
static int cell_buffer_size;
static int cell_buffer_maxsize;
#endif // PRINT_CONTROL_OUTPUT

static void initCellBuffer(int increment)
{
  cell_buffer_next = 0;
  cell_buffer_increment = increment;
  cell_buffer_number_chunks = 0;
#ifdef PRINT_CONTROL_OUTPUT
  cell_buffer_size = 0;
  cell_buffer_maxsize = 0;
#endif // PRINT_CONTROL_OUTPUT
}

static SVcell* allocCell()
{
  if (cell_buffer_next == 0)
  {
    cell_buffer_next = (SVcell*)malloc(sizeof(SVcell)*cell_buffer_increment);
    if (cell_buffer_next == 0)
    {
      fprintf(stderr,"FATAL ERROR: malloc for cell buffer failed\n");
      return 0;
    }
    else
    {
      cell_buffer_chunks[cell_buffer_number_chunks] = cell_buffer_next;
      cell_buffer_number_chunks++;
    }
    for (int i = 0; i < cell_buffer_increment; i++)
    {
      cell_buffer_next[i].buffer_next = &(cell_buffer_next[i+1]);
    }
    cell_buffer_next[cell_buffer_increment-1].buffer_next = 0;
    cell_buffer_increment *= 2;
  }
  // get pointer to next available cell
  SVcell* cell = cell_buffer_next;
  cell_buffer_next = cell->buffer_next;
 
  // initialize data fields
  cell->arrived_vertices = 0;

#ifdef PRINT_CONTROL_OUTPUT
  cell_buffer_size++;
  if (cell_buffer_size > cell_buffer_maxsize) cell_buffer_maxsize = cell_buffer_size;
#endif // PRINT_CONTROL_OUTPUT

  return cell;
}

static void deallocCell(SVcell* cell)
{
  cell->buffer_next = cell_buffer_next;
  cell_buffer_next = cell;
#ifdef PRINT_CONTROL_OUTPUT
  cell_buffer_size--;
#endif // PRINT_CONTROL_OUTPUT
}

static void destroyCellBuffer()
{
  for (int i = 0; i < cell_buffer_number_chunks; i++)
  {
    free(cell_buffer_chunks[i]);
  }
  cell_buffer_number_chunks = 0;
  cell_buffer_next = 0;
}

bool SVreadDereferenced::open(SVreader* svreader)
{
  if (svreader == 0)
  {
    return false;
  }
  this->svreader = svreader;

  nverts = svreader->nverts;
  ncells = svreader->ncells;

  v_count = 0;
  c_count = 0;

  v_pnum_f = svreader->v_pnum_f;
  v_pmin_f = svreader->v_pmin_f;
  v_pmax_f = svreader->v_pmax_f;

  c_pnum_f = svreader->c_pnum_f;
  c_pmin_f = svreader->c_pmin_f;
  c_pmax_f = svreader->c_pmax_f;

  c_idx = new int[8];
  c_final = new bool[8];
  c_v_prop_f = new float*[8];

  vertex_hash = new my_vertex_hash;
  cell_queue = new my_cell_queue;

  initVertexBuffer(1024);
  initCellBuffer(2048);

  return true;
}

void SVreadDereferenced::close(bool close_file)
{
  // close of SVreadDereferenced
  svreader->close(close_file);
#ifdef PRINT_CONTROL_OUTPUT
  fprintf(stderr,"vertex_buffer_maxsize (size) %d (%d) cell_buffer_maxsize (size) t %d (%d)\n",vertex_buffer_maxsize,vertex_buffer_size,cell_buffer_maxsize,cell_buffer_size);
#endif
  delete vertex_hash;
  delete cell_queue;
  destroyVertexBuffer();
  destroyCellBuffer();
  if (c_v_prop_f) delete [] c_v_prop_f; c_v_prop_f = 0;

  // close of SVreader interface
  v_count = -1;
  c_count = -1;
  if (c_idx) delete [] c_idx; c_idx = 0;
  if (c_final) delete [] c_final; c_final = 0;
}

int SVreadDereferenced::read_input()
{
  int i;
  SVvertex* vertex;
  SVcell* cell;
  my_vertex_hash::iterator hash_elements[8];

  SVevent event = svreader->read_element();
    
  if (event == SV_TETRAHEDRON)
  {
    assert(svreader->c_nverts == 4);
    // create cell
    cell = allocCell();
    // this is a tet
    cell->number_vertices = 4;
    // lookup the cell's vertices
    for (i = 0; i < 4; i++)
    {
      // look for vertex in the hash
      hash_elements[i] = vertex_hash->find(svreader->c_idx[i]);
      // did we find the vertex in the hash  
      if (hash_elements[i] == vertex_hash->end())
      {
        // vertex has not arrived yet. so we create the vertex ...
        vertex = allocVertex();
	      if (vertex->prop_f == 0) vertex->prop_f = new float[v_pnum_f];
        // ... insert it into hash ...
        vertex_hash->insert(my_vertex_hash::value_type(svreader->c_idx[i], vertex));
        // ... and mark it as not having arrived
        vertex->index = -2;
      }
      else
      {
        // use vertex found in hash
        vertex = (*hash_elements[i]).second;
      }
      // add vertex to cell
      cell->vertices[i] = vertex;
      // has this vertex arrived already
      if (vertex->index == -2)
      {
        // no ... we store the cell in a list so we can tell it about the arrival later
        if (vertex->list)
        {
          if (vertex->list_size == vertex->list_alloc)
          {
            vertex->list_alloc += 4;
            vertex->list = (SVcell**)realloc(vertex->list,vertex->list_alloc*sizeof(SVcell*));
          }
        }
        else
        {
          vertex->list_alloc = 10;
          vertex->list = (SVcell**)malloc(vertex->list_alloc*sizeof(SVcell*));
          assert(vertex->list_size == 0);
        }
        vertex->list[vertex->list_size] = cell;
      }
      else
      {
        // yes ... we can tell the cell now
        cell->arrived_vertices++;
      }
      // vertex has one more cell
      vertex->list_size++;
    }
    // have all vertices of this cell already arrived
    if (cell->arrived_vertices == 4)
    {
      // then it is ready to be output
      cell_queue->push_back(cell);
    }
    // check for finalization of the cell's vertices
    for (i = 0; i < 4; i++)
    {
      // look for vertex in the hash
      if (svreader->c_final[i])
      {
        // mark vertex as finalized
        cell->vertices[i]->buffer_next = cell->vertices[i];
        // remove vertex from hash
        vertex_hash->erase(hash_elements[i]);
      }
    }
  }
  else if (event == SV_VERTEX)
  {
    // look for vertex in the vertex hash
    hash_elements[0] = vertex_hash->find(svreader->v_idx);
    // did we find the vertex in the hash  
    if (hash_elements[0] == vertex_hash->end())
    {
      // vertex had not been referenced yet. so we create the vertex ...
      vertex = allocVertex();
      if (vertex->prop_f == 0) vertex->prop_f = new float[v_pnum_f];
      // ... insert it into hash
      vertex_hash->insert(my_vertex_hash::value_type(svreader->v_idx, vertex));
    }
    else
    {
      // use vertex found in hash
      vertex = (*hash_elements[0]).second;
      // make sure it is not yet marked as arrived
      assert(vertex->index == -2);
      // make sure it has at least one cell in its list
      assert(vertex->list_size > 0);
    }
    // copy coordinates
    VecCopyNfv(vertex->prop_f, svreader->v_prop_f, v_pnum_f);
    // and mark the vertex as having arrived
    vertex->index = -1;
    // did we already have cells
    if (vertex->list_size)
    {
      // then loop over them to tell them about the arrival
      for (i = 0; i < vertex->list_size; i++)
      {
        // get the cell
        cell = vertex->list[i];
        // increment the arrival counter
        cell->arrived_vertices++;
        // is this the last vertex to arrive
        if (cell->arrived_vertices == cell->number_vertices)
        {
          // then we can output it
          cell_queue->push_back(cell);
        }
      }
    }
    // in post order the arrival of a vertex marks its finalization
    if (svreader->post_order)
    {
      // mark vertex as finalized
      vertex->buffer_next = vertex;
      // so we remove it from hash
      vertex_hash->erase(hash_elements[0]);
    }
  }
  else if (event == SV_HEXAHEDRON)
  {
    assert(svreader->c_nverts == 8);
    // create cell
    cell = allocCell();
    // this is a hex
    cell->number_vertices = 8;
    // lookup the cell's vertices
    for (i = 0; i < 8; i++)
    {
      // look for vertex in the hash
      hash_elements[i] = vertex_hash->find(svreader->c_idx[i]);
      // did we find the vertex in the hash  
      if (hash_elements[i] == vertex_hash->end())
      {
        // vertex has not arrived yet. so we create the vertex ...
        vertex = allocVertex();
	      if (vertex->prop_f == 0) vertex->prop_f = new float[v_pnum_f];
        // ... insert it into hash ...
        vertex_hash->insert(my_vertex_hash::value_type(svreader->c_idx[i], vertex));
        // ... and mark it as not having arrived
        vertex->index = -2;
      }
      else
      {
        // use vertex found in hash
        vertex = (*hash_elements[i]).second;
      }
      // add vertex to cell
      cell->vertices[i] = vertex;
      // has this vertex arrived already
      if (vertex->index == -2)
      {
        // no ... we store the cell in a list so we can tell it about the arrival later
        if (vertex->list)
        {
          if (vertex->list_size == vertex->list_alloc)
          {
            vertex->list_alloc += 4;
            vertex->list = (SVcell**)realloc(vertex->list,vertex->list_alloc*sizeof(SVcell*));
          }
        }
        else
        {
          vertex->list_alloc = 10;
          vertex->list = (SVcell**)malloc(vertex->list_alloc*sizeof(SVcell*));
          assert(vertex->list_size == 0);
        }
        vertex->list[vertex->list_size] = cell;
      }
      else
      {
        // yes ... we can tell the cell the cell now
        cell->arrived_vertices++;
      }
      // vertex has one more cell
      vertex->list_size++;
    }
    // have all vertices of this cell already arrived
    if (cell->arrived_vertices == 8)
    {
      // then it is ready to be output
      cell_queue->push_back(cell);
    }
    // check for finalization of the cell's vertices
    for (i = 0; i < 8; i++)
    {
      // look for vertex in the hash
      if (svreader->c_final[i])
      {
        // mark vertex as finalized
        cell->vertices[i]->buffer_next = cell->vertices[i];
        // remove vertex from hash
        vertex_hash->erase(hash_elements[i]);
      }
    }
  }
  else if (event == SV_FINALIZED)
  {
    // look for vertex in the vertex hash
    hash_elements[0] = vertex_hash->find(svreader->v_idx);
    // did we find the vertex in the hash  
    if (hash_elements[0] == vertex_hash->end())
    {
      // vertex was not found in hash
      fprintf(stderr, "FATAL ERROR: finalized vertex %d not in hash. corrupt streaming mesh.\n", svreader->v_idx);
      return -1;
    }
    else
    {
      // use vertex found in hash
      vertex = (*hash_elements[0]).second;
    }
    // mark vertex as finalized
    vertex->buffer_next = vertex;
    // so we remove it from hash
    vertex_hash->erase(hash_elements[0]);
  }
  else if (event == SV_EOF)
  {
    if (vertex_hash->size())
    {
      // get some hash element ... 
      hash_elements[0] = vertex_hash->begin();
      // ... or rather its unfinalized vertex ...
      vertex = (*hash_elements[0]).second;
      // ... mark this vertex as finalized ...
      vertex->buffer_next = vertex;
      // ... and remove it from hash
      vertex_hash->erase(hash_elements[0]);
    }
    else
    {
      return 0;
    }
  }
  return 1;
}

int SVreadDereferenced::read_cell()
{
  int i;
  SVvertex* vertex;
  SVcell* cell;

  // read from input until we have a cell for output (or fail)

  while (!cell_queue->size())
  {
    int ok = read_input();
    if (ok <= 0)
    {
      return ok;
    }
  }

  // get next output cell

  cell = cell_queue->front();
  cell_queue->pop_front();

  // make sure it has everything

  assert(cell->arrived_vertices == cell->number_vertices);
  
  // now we have a cell

  have_cell = cell->number_vertices;

  // look for new or finalized vertices and fill in indices, finalization, and positions

  for (i = 0; i < cell->number_vertices; i++)
  {
    vertex = cell->vertices[i];
    // is this vertex a new vertex?
    if (vertex->index == -1)
    {
      vertex->index = v_count + have_new;
      new_vertices[have_new++] = i;
    }
    assert(vertex->index >= 0);
    // is this vertex to be finalized 
    if (vertex->list_size == 1 && vertex->buffer_next) // if this is the last cell and the vertex was finalized
    {
      c_final[i] = true;
      finalized_vertices[have_finalized++] = i;
      deallocVertex(vertex); // the memory can still be used until the next vertex is allocated.
    }
    else
    {
      vertex->list_size--;
      c_final[i] = false;
    }
    // fill index array
    c_idx[i] = vertex->index;
    // set additional position array
    c_v_prop_f[i] = vertex->prop_f;
  }

  deallocCell(cell);

  return 1;
}

SVevent SVreadDereferenced::read_element()
{
  if ((have_new + have_cell) == 0)
  {
    next_new = 0;
    have_finalized = next_finalized = 0;
    
    int ok = read_cell();
    if (ok <= 0)
    {
      return (SVevent)ok;
    }
  }
  if (have_new)
  {
    v_idx = c_idx[new_vertices[next_new]];
    v_prop_f = c_v_prop_f[new_vertices[next_new]];
    have_new--; next_new++;
    v_count++;
    return SV_VERTEX;
  }
  else
  {
    c_nverts = have_cell;
    have_cell = 0;
    c_count++;
    if (c_nverts == 4)
      return SV_TETRAHEDRON;
    else
      return SV_HEXAHEDRON;
  }
}

SVevent SVreadDereferenced::read_event()
{
  if ((have_new + have_cell + have_finalized) == 0)
  {
    return read_element();
  }
  if (have_new)
  {
    v_idx = c_idx[new_vertices[next_new]];
    v_prop_f = c_v_prop_f[new_vertices[next_new]];
    have_new--; next_new++;
    v_count++;
    return SV_VERTEX;
  }
  else if (have_cell)
  {
    c_nverts = have_cell;
    have_cell = 0;
    c_count++;
    if (c_nverts == 4)
      return SV_TETRAHEDRON;
    else
      return SV_HEXAHEDRON;
  }
  else if (have_finalized)
  {
    final_idx = c_idx[finalized_vertices[next_finalized]];
    have_finalized--; next_finalized++;
    return SV_FINALIZED;
  }
  return SV_ERROR;
}

SVreadDereferenced::SVreadDereferenced()
{
  // init of SVreadDereferenced
  c_v_prop_f = 0;

  svreader = 0;

  have_new = 0; next_new = 0;
  have_cell = 0;
  have_finalized = 0; next_finalized = 0;
}

SVreadDereferenced::~SVreadDereferenced()
{
  // clean-up for SVreadDereferenced
  v_prop_f = 0;
  c_prop_f = 0;
  v_pnum_f = 0;
  v_pmin_f = 0;
  v_pmax_f = 0;
  c_pnum_f = 0;
  c_pmin_f = 0;
  c_pmax_f = 0;
  if (svreader) delete svreader;
}
