/*
===============================================================================

  FILE:  svreader_jrs.cpp
  
  CONTENTS:
  
    see corresponding header file
  
  PROGRAMMERS:
  
    martin isenburg@cs.unc.edu
  
  COPYRIGHT:
  
    copyright (C) 2008  martin isenburg@cs.unc.edu
    
    This software is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
  
  CHANGE HISTORY:
  
    see corresponding header file
  
===============================================================================
*/
#include "svreader_jrs.h"

#ifdef _WIN32
#include <fcntl.h>
#include <io.h>
#endif

#include <stdlib.h>
#include <string.h>

#include "vecnfv.h"

#define MAX_LINE_SIZE 1024

bool SVreader_jrs::open(FILE* file_node, FILE* file_ele)
{
  char dummy[MAX_LINE_SIZE];
  int num_coords;
  int num_attrs;

  if (file_node == 0 || file_ele == 0)
  {
    return false;
  }

  this->file_node = file_node;
  line_node = (char*)malloc(sizeof(char) * MAX_LINE_SIZE);
  if (fgets(line_node, sizeof(char) * MAX_LINE_SIZE, file_node) == 0)
  {
    free(line_node);
    line_node = 0;
    fprintf(stderr, "ERROR: running out of data when reading header of *.node file\n");
    return false;
  }

  // skip comments and empty lines of *.node file

  while (line_node[0] == '#' || line_node[1] == '#' || sscanf(line_node, "%s", dummy) < 1)
  {
    skipped_lines++;
    if (fgets(line_node, sizeof(char) * MAX_LINE_SIZE, file_node) == 0)
    {
      free(line_node);
      line_node = 0;
      fprintf(stderr, "ERROR: running out of data after skipping %d lines in *.node file\n", skipped_lines);
      return false;
    }
  }

  // read header of *.node file

  if (sscanf(line_node, "%d %d %d %s %s", &nverts, &num_coords, &num_attrs, dummy, dummy) != 4)
  {
    free(line_node);
    line_node = 0;
    fprintf(stderr, "ERROR: corrupt header in *.node file\n");
    return false;
  }
  
  ncoords = num_coords + num_attrs;

  // read the first line of actual data in *.node file

  if (fgets(line_node, sizeof(char) * MAX_LINE_SIZE, file_node) == 0)
  {
    free(line_node);
    line_node = 0;
    fprintf(stderr, "ERROR: running out of data in *.node file after reading 0 of %d points\n", nverts);
    return false;
  }

  // skip comments and empty lines of *.node file

  while (line_node[0] == '#' || line_node[1] == '#' || sscanf(line_node, "%s", dummy) < 1)
  {
    // we are skipping this line
    skipped_lines++;
    if (fgets(line_node, sizeof(char) * MAX_LINE_SIZE, file_node) == 0)
    {
      free(line_node);
      line_node = 0;
      fprintf(stderr, "ERROR: running out of data after skipping %d lines in *.node file\n", skipped_lines);
      return false;
    }
  }

  // check first index of the first vertex (this is how much we need to substract to go to zero)

  if (sscanf(line_node, "%d %s %s", &minus, dummy, dummy) != 3)
  {
    fprintf(stderr, "ERROR: the first line the *.node continaing points is corrupt:\n%s\n", line_node);
    return false;
  }

  // start processing the ele file

  this->file_ele = file_ele;
  line_ele = (char*)malloc(sizeof(char) * MAX_LINE_SIZE);
  if (fgets(line_ele, sizeof(char) * MAX_LINE_SIZE, file_ele) == 0)
  {
    free(line_ele);
    line_ele = 0;
    fprintf(stderr, "ERROR: running out of data when reading header of *.ele file\n");
    return false;
  }

  // skip comments and empty lines of *.ele file

  while (line_ele[0] == '#' || line_ele[1] == '#' || sscanf(line_ele, "%s", dummy) < 1)
  {
    skipped_lines++;
    if (fgets(line_ele, sizeof(char) * MAX_LINE_SIZE, file_ele) == 0)
    {
      free(line_ele);
      line_ele = 0;
      fprintf(stderr, "ERROR: running out of data while skipping lines in *.ele file\n");
      return false;
    }
  }

  // read header of *.ele file

  if (sscanf(line_ele, "%d %s %s %s", &ncells, dummy, dummy, dummy) != 3)
  {
    free(line_ele);
    line_ele = 0;
    fprintf(stderr, "ERROR: corrupt header in *.ele file\n");
    return false;
  }

  // read the first line of actual data in *.ele file

  if (fgets(line_ele, sizeof(char) * MAX_LINE_SIZE, file_ele) == 0)
  {
    free(line_ele);
    line_ele = 0;
    fprintf(stderr, "ERROR: running out of data in *.ele file after reading 0 of %d tetrahedra\n", ncells);
    return false;
  }

  // skip comments and empty lines after header in *.ele file

  while (line_ele[0] == '#' || line_ele[1] == '#' || sscanf(line_ele, "%s", dummy) < 1)
  {
    skipped_lines++;
    if (fgets(line_ele, sizeof(char) * MAX_LINE_SIZE, file_ele) == 0)
    {
      free(line_ele);
      line_ele = 0;
      fprintf(stderr, "ERROR: running out of data while skipping comments in *.ele file\n");
      return false;
    }
  }

  type = SV_TET;  // only TET meshes are supported
  post_order = false;

  v_pnum_f = 3;
  if (v_prop_f) delete [] v_prop_f;
  v_prop_f = new float[v_pnum_f];

  v_count = 0;
  c_count = 0;

  c_nverts = 4;
  c_idx = new I32[4];   // only TET meshes are supported
  c_final = new bool[4];

  c_final[0] = false; 
  c_final[1] = false;
  c_final[2] = false;
  c_final[3] = false;

  return true;
}

void SVreader_jrs::close(bool close_file)
{
  // close of SVreader_jrs
  if (skipped_lines) fprintf(stderr,"WARNING: skipped %d lines.\n",skipped_lines);
  skipped_lines = 0;
  if (close_file && file_node && file_node != stdin) fclose(file_node);
  file_node = 0;
  if (close_file && file_ele && file_ele != stdin) fclose(file_ele);
  file_ele = 0;
  if (line_node)
  {
    free(line_node);
    line_node = 0;
  }
  if (line_ele)
  {
    free(line_ele);
    line_ele = 0;
  }

  // close of SVreader interface
  v_count = -1;
  c_count = -1;
  if (v_prop_f) delete [] v_prop_f;
  v_prop_f = 0;
  if (c_idx) delete [] c_idx;
  c_idx = 0;
  if (c_final) delete [] c_final;
  c_final = 0; 
}

SVevent SVreader_jrs::read_element()
{
  if (line_node == 0 && line_ele == 0)
  {
    return SV_EOF;
  }

  if (line_node && v_count < nverts)
  {
    sscanf(line_node, "%d %f %f %f", &v_idx, &(v_prop_f[0]), &(v_prop_f[1]), &(v_prop_f[2]));
    v_idx = v_count;
    v_count++;
    if (fgets(line_node, sizeof(char) * MAX_LINE_SIZE, file_node) == 0)
    {
      if (v_count < nverts)
      {
        fprintf(stderr,"WARNING: early end of node file after %d of %d vertices\n", v_count, nverts);
      }
      free(line_node);
      line_node = 0;
    }
    return SV_VERTEX;
  }
  else if (line_ele && c_count < ncells)
  {
    I32 idx;

    sscanf(line_ele, "%d %d %d %d %d", &idx, &(c_idx[0]), &(c_idx[1]), &(c_idx[2]), &(c_idx[3]));
    
    if (minus)
    {
      c_idx[0] -= minus;
      c_idx[1] -= minus;
      c_idx[2] -= minus;
      c_idx[3] -= minus;
    }

    c_count++;

    if (fgets(line_ele, sizeof(char) * MAX_LINE_SIZE, file_ele) == 0)
    {
      if (c_count < ncells)
      {
        fprintf(stderr,"WARNING: early end of ele file after %d of %d cells\n", c_count, ncells);
      }
      free(line_ele);
      line_ele = 0;
    }
    return SV_TETRAHEDRON;
  }
  else
  {
    if (line_node)
    {
      free(line_node);
      line_node = 0;
    }
    if (line_ele)
    {
      free(line_ele);
      line_ele = 0;
    }
    return SV_EOF;
  }
}

SVevent SVreader_jrs::read_event()
{
  return read_element();
}

SVreader_jrs::SVreader_jrs()
{
  // init of SVreader_jrs
  ncoords = 0;
  minus = 0;
  skipped_lines = 0;
  file_node = 0;
  file_ele = 0;
  line_node = 0;
  line_ele = 0;
}

SVreader_jrs::~SVreader_jrs()
{
  // clean-up for SVreader_jrs
}
