/*
===============================================================================

  FILE:  SVreader_silo.cpp

  CONTENTS:

    see corresponding header file

  PROGRAMMERS:

    clement.courbet@ecp.fr

  COPYRIGHT:

    copyright (C) 2009  clement.courbet@ecp.fr

    This software is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.

  CHANGE HISTORY:

    see corresponding header file

===============================================================================
*/
#include "svreader_silo.h"

#ifdef _WIN32
#include <fcntl.h>
#include <io.h>
#endif

#include <stdlib.h>

#ifdef BUILD_SILO_READER

bool SVreader_silo::open ( const char *filename )
{
	if ( file == 0 )
	{
		return false;
	}

#ifdef _WIN32
	if ( file == stdin )
	{
		if ( _setmode ( _fileno ( stdin ), _O_BINARY ) == -1 )
		{
			fprintf ( stderr, "ERROR: cannot set stdin to binary (untranslated) mode\n" );
		}
	}
#endif

	//open the silo DB
	if( (dbfile = DBOpen(filename, DB_UNKNOWN, DB_READ)) == NULL);
	{
		fprintf(stderr, "ERROR: The silo library cannot open %s for reading.\n", filename);
		return false;
	}

	//get the TOC of the dataset:
	DBtoc *toc = DBGetToc (dbfile);
	if(!toc)
	{
		fprintf(stderr, "ERROR: Cannot get Table Of Contents for file %s.\n", filename);
		return false;
	}

       char tmpname[256];
       int i;
       DBucdmesh *mesh;

       if (DBInqVarExists(dbfile, "v"))
          printf("Found field \"v\"\n");

       if (DBInqMeshname(dbfile, "v", tmpname))
          printf("Found that field \"v\" lives on mesh %s\n", tmpname);

       mesh = DBGetUcdmesh(dbfile, "hex");
       if (mesh == NULL)
          printf("couldn't read the mesh\n");

       printf("zones nodelist\n");
       for (i = 0; i < mesh->zones->lnodelist; i++)
          printf("%d ", mesh->zones->nodelist[i]);
       printf("\n");

       printf("faces nodelist\n");
       for (i = 0; i < mesh->faces->lnodelist; i++)
          printf("%d ", mesh->faces->nodelist[i]);
       printf("\n");

       DBFreeUcdmesh(mesh);
    }
    DBClose(dbfile);

	read_header();

	v_count = 0;
	c_count = 0;

	if ( v_pnum_f ) v_prop_f = new float[v_pnum_f];
	if ( c_pnum_f ) c_prop_f = new float[c_pnum_f];

	c_idx = new int32_t[8];
	c_final = new bool[8];
	finalized_vertices = new int32_t[8];
	return true;
}

void SVreader_silo::close ( bool close_file )
{
	// close of SVreader_silo
	if ( close_file && file && file != stdin ) fclose ( file );
	file = 0;
	have_finalized = 0; next_finalized = 0;
	delete [] finalized_vertices;
	finalized_vertices = 0;

	// close of SVreader interface
	v_count = -1;
	c_count = -1;
	if ( v_prop_f ) delete [] v_prop_f;
	v_prop_f = 0;
	if ( c_idx ) delete [] c_idx;
	c_idx = 0;
	if ( c_final ) delete [] c_final;
	c_final = 0;
}

SVevent SVreader_silo::read_element()
{
	
	have_finalized = next_finalized = 0;
	//read element type
	if( (v_count<nverts) || (c_count<ncells))
	{
		unsigned char elementType;
		readUChar(&elementType);
		if ( elementType == SVG_VERTEX ) // next element is a vertex
		{
			if ( v_pnum_f )
			{
				readNFloats ( v_prop_f, v_pnum_f );
			}
			v_idx = v_count;
			v_count++;
			return SV_VERTEX;
		}
		else if ( elementType == SVG_TETRAHEDRON ) // next element is a tetrahedron
		{
			c_count++;
			for ( int i = 0; i < 4; i++ )
			{
				readInt32(&(c_idx[i])); //read the index
				if ( c_idx[i] < 0 )
				{
					c_idx[i] = v_count+c_idx[i];
					c_final[i] = true;
					finalized_vertices[have_finalized] = c_idx[i];
					have_finalized++;
				}
				else
				{
					c_idx[i] = c_idx[i]-1;
					c_final[i] = false;
				}
			}
			//read cell data
			if ( c_pnum_f )
			{
				readNFloats ( c_prop_f, c_pnum_f );
			}
      c_nverts = 4;
			return SV_TETRAHEDRON;
		}
		else if ( elementType == SVG_HEXAHEDRON ) // next element is a hexahedron
		{
			c_count++;
			for ( int i = 0; i < 8; i++ )
			{
				readInt32(&(c_idx[i])); //read the index
				if ( c_idx[i] < 0 )
				{
					c_idx[i] = v_count+c_idx[i];
					c_final[i] = true;
					finalized_vertices[have_finalized] = c_idx[i];
					have_finalized++;
				}
				else
				{
					c_idx[i] = c_idx[i]-1;
					c_final[i] = false;
				}
			}
			//read cell data
			if ( c_pnum_f )
			{
				readNFloats ( c_prop_f, c_pnum_f );
			}
      c_nverts = 8;
			return SV_HEXAHEDRON;
		}
		else
		{
			printf("ERROR: SVreader_silo: unsupported element type %i\n", elementType);
			exit(1);
		}
	}

	if ( nverts != -1 && v_count != nverts )
	{
		fprintf ( stderr,"WARNING: wrong vertex count: v_count (%d) != nverts (%d)\n", v_count, nverts );
	}
	nverts = v_count;
	if ( ncells != -1 && c_count != ncells )
	{
		fprintf ( stderr,"WARNING: wrong face count: c_count (%d) != ncells (%d)\n", c_count, ncells );
	}
	ncells = c_count;
	return SV_EOF;
}

SVevent SVreader_silo::read_event()
{
	if ( have_finalized )
	{
		final_idx = finalized_vertices[next_finalized];
		have_finalized--; next_finalized++;
		return SV_FINALIZED;
	}
	else
	{
		return read_element();
	}
}

#define SV_LITTLE_ENDIAN 0
#define SV_BIG_ENDIAN 1

void SVreader_silo::read_header()
{
	// read nverts
	readInt32(&nverts);
	// read ncells
	readInt32(&ncells);
	// read how many floating-point attributes per vertex
	readInt32(&v_pnum_f);
	// read how many floating-point attributes per cell
	readInt32(&c_pnum_f);

	// read bounding box of floating-point attributes (per vertex)
	if ( v_pnum_f )
	{
		if ( v_pmin_f ) delete [] v_pmin_f;
		if ( v_pmax_f ) delete [] v_pmax_f;
		v_pmin_f = new float[v_pnum_f];
		v_pmax_f = new float[v_pnum_f];
		readNFloats ( v_pmin_f, v_pnum_f );
		readNFloats ( v_pmax_f, v_pnum_f );
	}
	// read bounding box of floating-point attributes (per cell)
	if ( c_pnum_f )
	{
		if ( c_pmin_f ) delete [] c_pmin_f;
		if ( c_pmax_f ) delete [] c_pmax_f;
		c_pmin_f = new float[c_pnum_f];
		c_pmax_f = new float[c_pnum_f];
		readNFloats ( c_pmin_f, c_pnum_f );
		readNFloats ( c_pmax_f, c_pnum_f );
	}

	//reserved data
	int32_t dummy;
	readInt32(&dummy);
	if(dummy!=0)
	{
		fprintf(stderr, "Error: SVreader_silo: reserved bytes must be zero !\n");
		exit(1);
	}
	readInt32(&dummy);
	if(dummy!=0)
	{
		fprintf(stderr, "Error: SVreader_silo: reserved bytes must be zero !\n");
		exit(1);
	}
}

SVreader_silo::SVreader_silo()
{
	// init of SVreader_silo
	file = 0;
	have_finalized = 0; next_finalized = 0;
}

SVreader_silo::~SVreader_silo()
{
	// clean-up for SVreader_silo
}

#endif //BUILD_SILO_READER
