/*
===============================================================================

  FILE:  SVreader_svc.cpp

  CONTENTS:

    see corresponding header file

  PROGRAMMERS:

    clement.courbet@ecp.fr

  COPYRIGHT:

    copyright (C) 2009  clement courbet

    This software is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.

  CHANGE HISTORY:

    see corresponding header file

===============================================================================
*/
#include "svreader_svh.h"

#ifdef _WIN32
#include <fcntl.h>
#include <io.h>
#endif

#include <stdlib.h>
#include <string.h>
#include <assert.h>

#define PRINT_CONTROL_OUTPUT
#undef PRINT_CONTROL_OUTPUT

#define PRINT_DEBUG_OUTPUT
#undef PRINT_DEBUG_OUTPUT

#define PRINT_GEOM_DEBUG_OUTPUT
#undef PRINT_GEOM_DEBUG_OUTPUT

#define PRINT_CELLDATA_DEBUG_OUTPUT
#undef PRINT_CELLDATA_DEBUG_OUTPUT

//code that is common to svreader_svh.cpp and svwriter_svh.cpp
#include "svcommon_svh.cpp"

//DEBUG this is a temporary to determine LS weights for cell data prediction
//#define OUTPUT_FOR_LS_WEIGHTS
#ifdef OUTPUT_FOR_LS_WEIGHTS
FILE* lsoutfCorner;
FILE* lsoutfBridge;
FILE* lsoutfGap;
FILE* lsoutfPit;
#endif

//output the residuals in an ASCII FILE
//#define OUTPUT_RESIDUALS
#ifdef OUTPUT_RESIDUALS
#include <map>
std::map<I32, int> residualHistogram;
int nResiduals = 0;
#ifdef SVH_CELL_DATA_SECONDORDER_PREDICTION
  std::map<I32, int> residual2Histogram;
#endif
#endif
//output the quantized cell data in an ASCII FILE
//#define OUTPUT_QUANTIZED_CELL_DATA
#ifdef OUTPUT_QUANTIZED_CELL_DATA
#include <map>
std::map<I32, int> quantizedHistogram;
int nQuantized = 0;
#endif

// rangedecoder

#include "rangedecoder.h"

static RangeDecoder* rd_conn;
static RangeDecoder* rd_conn_op;
static RangeDecoder* rd_conn_cache;
static RangeDecoder* rd_conn_index;
static RangeDecoder* rd_conn_candi;
static RangeDecoder* rd_conn_final;

static RangeDecoder* rd_geom;

#ifdef SVH_COMPRESS_CELL_DATA
static RangeDecoder* rd_celldata;
#endif

// statistics

#ifdef PRINT_CONTROL_OUTPUT
static int predict_lorenzo = 0;
static int predict_parallelogram = 0;
static int predict_shoot = 0;
static int predict_last = 0;
#endif

#if !defined(SVH_COMPRESS_GEOMETRY)
FILE *geofiled;
#endif

static void initDecoder(FILE* file)
{
  rd_conn = new RangeDecoder(file);
  rd_conn_op = rd_conn;
  rd_conn_cache = rd_conn;
  rd_conn_index = rd_conn;
  rd_conn_candi = rd_conn;
  rd_conn_final = rd_conn;
  rd_geom = rd_conn;
#ifdef SVH_COMPRESS_CELL_DATA
  rd_celldata = rd_conn;
#endif
}

static void finishDecoder()
{
  rd_conn->done();
  delete rd_conn;
}

bool SVreader_svh::open(FILE* file)
{
  if(file == 0)
  {
    return false;
  }

  this->file = file;

#ifdef _WIN32
  if(file == stdin)
  {
    if(_setmode(_fileno(stdin), _O_BINARY) == -1)
    {
      fprintf(stderr, "ERROR: cannot set stdin to binary (untranslated) mode\n");
    }
  }
#endif

  // read version
  I32 version = fgetc(file);

  // compiled version
  I32 compiled_version = 0;
#ifdef SVH_BETTER_SHOOT_PREDICTION
  compiled_version += 1;
#endif
#ifdef SVH_FACECACHE_DEPENDS_ON_FINALIZATION
  compiled_version += 2;
#endif
#ifdef SVH_RMOP_DEPENDS_ON_FINALIZATION
  compiled_version += 4;
#endif
#ifdef SVH_INITIALIZE_VISVERTICES
  compiled_version += 8;
#endif
#ifdef SVH_DUMB_JOIN_CODING
  compiled_version += 16;
#endif
#ifdef SVH_IMPLICIT_FINALIZATION
  compiled_version += 32;
#endif
#ifdef SVH_COMPRESS_CELL_DATA
  compiled_version += 64;
#endif
#ifdef SVH_CELL_DATA_SECONDORDER_PREDICTION
  compiled_version += 128;
#endif

  if(version != compiled_version)
  {
    fprintf(stderr,"ERROR: wrong SVreader (data is %d but compiled reader is %d)\n",version, compiled_version);
    if(version == EOF) fprintf(stderr,"       check ... maybe the file (or the stream) is empty\n");
    exit(0);
  }

  initDecoder(file);
  initModels(0);

  // read precision
  nbits = rd_conn->readRange(25);

  read_header();

  last_op = 0;
  last_fin = 0;

  initVertexBuffer(1024, v_pnum_f);
#ifdef SVH_COMPRESS_CELL_DATA
  initEdgeBuffer(1024, c_pnum_f);
#else
  initEdgeBuffer(1024);
#endif

  dv = new DynamicVector();
  lc = new LittleCache(8);
  lc_f = new LittleCache(6);

  v_count = 0;
  c_count = 0;

  if(v_pnum_f) v_prop_f = new float[v_pnum_f];
  if(c_pnum_f) c_prop_f = new float[c_pnum_f];

  // only HEX meshes are supported
  type = SV_HEX;
  c_nverts = 8;
  c_idx = new I32[8];
  c_final = new bool[8];
  new_vertices = new I32[8];
  finalized_vertices = new I32[8];

#if !defined(SVH_COMPRESS_GEOMETRY)
  geofiled = fopen("geofile","r");
#endif

#ifdef OUTPUT_FOR_LS_WEIGHTS
  lsoutfCorner = fopen("lsCorner", "w");
  lsoutfBridge = fopen("lsBridge", "w");
  lsoutfGap = fopen("lsGap", "w");
  lsoutfPit = fopen("lsPit", "w");
#endif

  return true;
}

#ifdef SVH_COMPRESS_CELL_DATA

#ifdef SVH_CELL_DATA_SECONDORDER_PREDICTION
static inline void decompressCellDataStart( float* cd, float* residual)
#else
static inline void decompressCellDataStart(float* cd)
#endif
{
  if(c_quantizer)
  {
    for(int i = 0; i < cell_pnum_f; i++)
    {
      ((I32*)cd)[i] = c_ic[i]->Decompress(0, SVH_START);
#ifdef SVH_CELL_DATA_SECONDORDER_PREDICTION
      ((I32*)residual)[i] = 0;
#endif
    }
  }
  else
  {
    for(int i=0; i<cell_pnum_f; ++i)
    {
#ifdef SVH_CELL_DATA_SECONDORDER_PREDICTION
      residual[i] = 0;
#endif
      cd[i] = c_fc[i]->Decompress(0.0f, SVH_START);
#ifdef PRINT_CELLDATA_DEBUG_OUTPUT
      fprintf(stderr, "cell data actual      : %f\n", cd[i]);
#endif
    }
  }
}

#ifdef SVH_CELL_DATA_SECONDORDER_PREDICTION
static inline void decompressCellDataHut( const float *d0, float* cd, const float *r0, float* residual)
#else
static inline void decompressCellDataHut( const float *d0, float* cd)
#endif
{
  if(c_quantizer)
  {
    for(int i = 0; i < cell_pnum_f; i++)
    {
#ifdef SVH_CELL_DATA_SECONDORDER_PREDICTION
      ((I32*)residual)[i] = c_ic[i]->Decompress(((I32*)r0)[i], SVH_HUT);
      if(((I32*)residual)[i] > c_quantizer->m_aiQuantRange[i])
        ((I32*)residual)[i] -= 2*c_quantizer->m_aiQuantRange[i];
      ((I32*)cd)[i] = ((I32*) residual)[i] + ((I32*)d0)[i];
#else
      ((I32*)cd)[i] = c_ic[i]->Decompress(((I32*)d0)[i], SVH_HUT);
#endif
#ifdef PRINT_CELLDATA_DEBUG_OUTPUT
      fprintf(stderr, "cell data predicted/actual      : %i / %i\n", ((I32*)d0)[i], ((I32*)cd)[i]);
  #ifdef SVH_CELL_DATA_SECONDORDER_PREDICTION
      fprintf(stderr, "cell data predicted/actual residual: %i / %i\n", ((I32*)r0)[i], ((I32*)residual)[i]);
  #endif
#endif
    }
  }
  else
  {
    for(int i=0; i<cell_pnum_f; ++i)
    {
#ifdef SVH_CELL_DATA_SECONDORDER_PREDICTION
  #ifdef PRINT_CELLDATA_DEBUG_OUTPUT
      fprintf(stderr, "cell data predicted residual: %f\n", r0[i]);
  #endif
      residual[i] = c_fc[i]->Decompress(r0[i], SVH_HUT);
      cd[i] = residual[i] + d0[i];
#else
      cd[i] = c_fc[i]->Decompress(d0[i], SVH_HUT);
#endif
  #ifdef PRINT_CELLDATA_DEBUG_OUTPUT
      fprintf(stderr, "cell data predicted/actual      : %f / %f\n", d0[i], cd[i]);
  #endif
    }
  }
}

#ifdef SVH_CELL_DATA_SECONDORDER_PREDICTION
static inline void decompressCellDataRoof( const float *d0, const float *d5, float* cd, const float *r0, const float *r5, float* residual)
#else
static inline void decompressCellDataRoof( const float *d0, const float *d5, float* cd)
#endif
{
  if(c_quantizer)
  {
    for(int i = 0; i < cell_pnum_f; i++)
    {
      I32 pred = (((I32*)d0)[i] + ((I32*)d5)[i])/2;
#ifdef SVH_CELL_DATA_SECONDORDER_PREDICTION
      I32 residualPred = (((I32*)r0)[i]+ ((I32*)r5)[i])/2;
      ((I32*)residual)[i] = c_ic[i]->Decompress(residualPred, SVH_ROOF);
      if(((I32*)residual)[i] > c_quantizer->m_aiQuantRange[i])
        ((I32*)residual)[i] -= 2*c_quantizer->m_aiQuantRange[i];
      ((I32*)cd)[i] = ((I32*) residual)[i] + pred;
#else
      ((I32*)cd)[i] = c_ic[i]->Decompress(pred, SVH_ROOF);
#endif
#ifdef PRINT_CELLDATA_DEBUG_OUTPUT
      fprintf(stderr, "cell data predicted/actual      : %i / %i\n", pred, ((I32*)cd)[i]);
  #ifdef SVH_CELL_DATA_SECONDORDER_PREDICTION
      fprintf(stderr, "cell data predicted/actual residual: %i / %i\n", residualPred, ((I32*)residual)[i]);
  #endif
#endif
    }
  }
  else
  {
    for(int i=0; i<cell_pnum_f; ++i)
    {
      float pred;
      pred = (d0[i] + d5[i])/2;
#ifdef SVH_CELL_DATA_SECONDORDER_PREDICTION
      float residualPred = (r0[i] + r5[i])/2;
  #ifdef PRINT_CELLDATA_DEBUG_OUTPUT
      fprintf(stderr, "cell data predicted residual: %f\n", residualPred);
  #endif
      residual[i] = c_fc[i]->Decompress(residualPred, SVH_ROOF);
      cd[i] = residual[i] + pred;
#else
      cd[i] = c_fc[i]->Decompress(pred, SVH_ROOF);
#endif
  #ifdef PRINT_CELLDATA_DEBUG_OUTPUT
      fprintf(stderr, "cell data predicted/actual      : %f / %f\n", pred, cd[i]);
  #endif
    }
  }
}

#ifdef SVH_CELL_DATA_SECONDORDER_PREDICTION
static inline void decompressCellDataStep( const float *d0, const float *d1, float* cd, const float *r0, const float *r1, float* residual)
#else
static inline void decompressCellDataStep( const float *d0, const float *d1, float* cd)
#endif
{
  if(c_quantizer)
  {
    for(int i = 0; i < cell_pnum_f; i++)
    {
      I32 pred = (((I32*)d0)[i] + ((I32*)d1)[i])/2;
#ifdef SVH_CELL_DATA_SECONDORDER_PREDICTION
      I32 residualPred = (((I32*)r0)[i] + ((I32*)r1)[i])/2;
      ((I32*)residual)[i] = c_ic[i]->Decompress(residualPred, SVH_STEP);
      if(((I32*)residual)[i] > c_quantizer->m_aiQuantRange[i])
        ((I32*)residual)[i] -= 2*c_quantizer->m_aiQuantRange[i];
      ((I32*)cd)[i] = ((I32*) residual)[i] + pred;
#else
      ((I32*)cd)[i] = c_ic[i]->Decompress(pred, SVH_STEP);
#endif
#ifdef PRINT_CELLDATA_DEBUG_OUTPUT
      fprintf(stderr, "cell data predicted/actual      : %i / %i\n", pred, ((I32*)cd)[i]);
  #ifdef SVH_CELL_DATA_SECONDORDER_PREDICTION
      fprintf(stderr, "cell data predicted/actual residual: %i / %i\n", residualPred, ((I32*)residual)[i]);
  #endif
#endif
    }
  }
  else
  {
    for(int i=0; i<cell_pnum_f; ++i)
    {
      float pred;
      pred = (d0[i] + d1[i])/2;
#ifdef SVH_CELL_DATA_SECONDORDER_PREDICTION
      float residualPred = (r0[i] + r1[i])/2;
  #ifdef PRINT_CELLDATA_DEBUG_OUTPUT
      fprintf(stderr, "cell data predicted residual: %f\n", residualPred);
  #endif
      residual[i] = c_fc[i]->Decompress(residualPred, SVH_STEP);
      cd[i] = residual[i] + pred;
#else
      cd[i] = c_fc[i]->Decompress(pred, SVH_STEP);
#endif
  #ifdef PRINT_CELLDATA_DEBUG_OUTPUT
      fprintf(stderr, "cell data predicted/actual      : %f / %f\n", pred, cd[i]);
  #endif
    }
  }
}

#ifdef SVH_CELL_DATA_SECONDORDER_PREDICTION
static inline void decompressCellDataCorner( const float *d0, const float *d1, const float *d2, float* cd, const float *r0, const float *r1, const float *r2, float* residual)
#else
static inline void decompressCellDataCorner( const float *d0, const float *d1, const float *d2, float* cd)
#endif
{
  if(c_quantizer)
  {
    for(int i = 0; i < cell_pnum_f; i++)
    {
      I32 pred = (((I32*)d0)[i] + ((I32*)d1)[i] + ((I32*)d2)[i])/3;
#ifdef SVH_CELL_DATA_SECONDORDER_PREDICTION
      I32 residualPred = (((I32*)r0)[i] + ((I32*)r1)[i] + ((I32*)r2)[i])/3;
      ((I32*)residual)[i] = c_ic[i]->Decompress(residualPred, SVH_CORNER);
      if(((I32*)residual)[i] > c_quantizer->m_aiQuantRange[i])
        ((I32*)residual)[i] -= 2*c_quantizer->m_aiQuantRange[i];
      ((I32*)cd)[i] = ((I32*) residual)[i] + pred;
#else
      ((I32*)cd)[i] = c_ic[i]->Decompress(pred, SVH_CORNER);
#endif
#ifdef PRINT_CELLDATA_DEBUG_OUTPUT
      fprintf(stderr, "cell data predicted/actual      : %i / %i\n", pred, ((I32*)cd)[i]);
  #ifdef SVH_CELL_DATA_SECONDORDER_PREDICTION
      fprintf(stderr, "cell data predicted/actual residual: %i / %i\n", residualPred, ((I32*)residual)[i]);
  #endif
#endif
#ifdef OUTPUT_RESIDUALS
      {
      I32 residual = ((I32*)cd)[i]-pred;
      std::map<I32, int>::iterator mit = residualHistogram.find(residual);
      if(mit==residualHistogram.end())
        residualHistogram.insert(std::pair<I32, int>(residual, 1));
      else
        mit->second ++;
      nResiduals++;
      }
  #ifdef SVH_CELL_DATA_SECONDORDER_PREDICTION
      {
      I32 residual2 = ((I32*)residual)[i]-residualPred;
      std::map<I32, int>::iterator mit = residual2Histogram.find(residual2);
      if(mit==residual2Histogram.end())
        residual2Histogram.insert(std::pair<I32, int>(residual2, 1));
      else
        mit->second ++;
      }
  #endif
#endif
#ifdef OUTPUT_QUANTIZED_CELL_DATA
      {
      std::map<I32, int>::iterator mit = quantizedHistogram.find(((I32*)cd)[i]);
      if(mit==quantizedHistogram.end())
        quantizedHistogram.insert(std::pair<I32, int>(((I32*)cd)[i], 1));
      else
        mit->second ++;
      nQuantized++;
      }
#endif
    }
  }
  else
  {
    for(int i=0; i<cell_pnum_f; ++i)
    {
      float pred;
      pred = (d0[i] + d1[i] + d2[i])/3; 
#ifdef SVH_CELL_DATA_SECONDORDER_PREDICTION
      float residualPred = (r0[i] + r1[i] + r2[i])/3;
  #ifdef PRINT_CELLDATA_DEBUG_OUTPUT
      fprintf(stderr, "cell data predicted residual: %f\n", residualPred);
  #endif
      residual[i] = c_fc[i]->Decompress(residualPred, SVH_CORNER);
      cd[i] = residual[i] + pred;
#else
      cd[i] = c_fc[i]->Decompress(pred, SVH_CORNER);
#endif
  #ifdef PRINT_CELLDATA_DEBUG_OUTPUT
      fprintf(stderr, "cell data predicted/actual      : %f / %f\n", pred, cd[i]);
  #endif
    }
  }
#ifdef OUTPUT_FOR_LS_WEIGHTS
  fwrite(d0, sizeof(float), 1, lsoutfCorner);
  fwrite(d1, sizeof(float), 1, lsoutfCorner);
  fwrite(d2, sizeof(float), 1, lsoutfCorner);
  fwrite(cd, sizeof(float), 1, lsoutfCorner);
#endif
}

#ifdef SVH_CELL_DATA_SECONDORDER_PREDICTION
static inline void decompressCellDataBridge( const float *d0, const float *d1, const float *d3, float* cd, const float *r0, const float *r1, const float *r3, float* residual)
#else
static inline void decompressCellDataBridge( const float *d0, const float *d1, const float *d3, float* cd)
#endif
{
  if(c_quantizer)
  {
    for(int i = 0; i < cell_pnum_f; i++)
    {
      I32 pred = (((I32*)d1)[i] + ((I32*)d3)[i])/2; 
#ifdef SVH_CELL_DATA_SECONDORDER_PREDICTION
      I32 residualPred = (((I32*)r1)[i] + ((I32*)r3)[i])/2;
      ((I32*)residual)[i] = c_ic[i]->Decompress(residualPred, SVH_BRIDGE);
      if(((I32*)residual)[i] > c_quantizer->m_aiQuantRange[i])
        ((I32*)residual)[i] -= 2*c_quantizer->m_aiQuantRange[i];
      ((I32*)cd)[i] = ((I32*) residual)[i] + pred;
#else
      ((I32*)cd)[i] = c_ic[i]->Decompress(pred, SVH_BRIDGE);
#endif
#ifdef PRINT_CELLDATA_DEBUG_OUTPUT
      fprintf(stderr, "cell data predicted/actual      : %i / %i\n", pred, ((I32*)cd)[i]);
  #ifdef SVH_CELL_DATA_SECONDORDER_PREDICTION
      fprintf(stderr, "cell data predicted/actual residual: %i / %i\n", residualPred, ((I32*)residual)[i]);
  #endif
#endif
    }
  }
  else
  {
    for(int i=0; i<cell_pnum_f; ++i)
    {
      float pred;
      pred = (d1[i] + d3[i])/2;
#ifdef SVH_CELL_DATA_SECONDORDER_PREDICTION
      float residualPred = (r1[i] + r3[i])/2;
  #ifdef PRINT_CELLDATA_DEBUG_OUTPUT
      fprintf(stderr, "cell data predicted residual: %f\n", residualPred);
  #endif
      residual[i] = c_fc[i]->Decompress(residualPred, SVH_BRIDGE);
      cd[i] = residual[i] + pred;
#else
      cd[i] = c_fc[i]->Decompress(pred, SVH_BRIDGE);
#endif
  #ifdef PRINT_CELLDATA_DEBUG_OUTPUT
      fprintf(stderr, "cell data predicted/actual      : %f / %f\n", pred, cd[i]);
  #endif
    }
  }
#ifdef OUTPUT_FOR_LS_WEIGHTS
  fwrite(d0, sizeof(float), 1, lsoutfBridge);
  fwrite(d1, sizeof(float), 1, lsoutfBridge);
  fwrite(d3, sizeof(float), 1, lsoutfBridge);
  fwrite(cd, sizeof(float), 1, lsoutfBridge);
#endif
}

#ifdef SVH_CELL_DATA_SECONDORDER_PREDICTION
static inline void decompressCellDataTunnel( const float *d0, const float *d1, const float *d3, const float *d5, float* cd, const float *r0, const float *r1, const float *r3, const float *r5, float* residual)
#else
static inline void decompressCellDataTunnel( const float *d0, const float *d1, const float *d3, const float *d5, float* cd)
#endif
{
  if(c_quantizer)
  {
    for(int i = 0; i < cell_pnum_f; i++)
    {
      I32 pred = (((I32*)d0)[i] + ((I32*)d1)[i] + ((I32*)d3)[i] + ((I32*)d5)[i])/4;
#ifdef SVH_CELL_DATA_SECONDORDER_PREDICTION
      I32 residualPred = (((I32*)r0)[i] + ((I32*)r1)[i] + ((I32*)r3)[i] + ((I32*)r5)[i])/4;
      ((I32*)residual)[i] = c_ic[i]->Decompress(residualPred, SVH_TUNNEL);
      if(((I32*)residual)[i] > c_quantizer->m_aiQuantRange[i])
        ((I32*)residual)[i] -= 2*c_quantizer->m_aiQuantRange[i];
      ((I32*)cd)[i] = ((I32*) residual)[i] + pred;
#else
      ((I32*)cd)[i] = c_ic[i]->Decompress(pred, SVH_TUNNEL);
#endif
#ifdef PRINT_CELLDATA_DEBUG_OUTPUT
      fprintf(stderr, "cell data predicted/actual      : %i / %i\n", pred, ((I32*)cd)[i]);
  #ifdef SVH_CELL_DATA_SECONDORDER_PREDICTION
      fprintf(stderr, "cell data predicted/actual residual: %i / %i\n", residualPred, ((I32*)residual)[i]);
  #endif
#endif
    }
  }
  else
  {
    for(int i=0; i<cell_pnum_f; ++i)
    {
      float pred;
      pred = (d0[i] + d1[i] + d3[i] + d5[i])/4;
#ifdef SVH_CELL_DATA_SECONDORDER_PREDICTION
      float residualPred = (r0[i] + r1[i] + r3[i] + r5[i])/4;
  #ifdef PRINT_CELLDATA_DEBUG_OUTPUT
      fprintf(stderr, "cell data predicted residual: %f\n", residualPred);
  #endif
      residual[i] = c_fc[i]->Decompress(residualPred, SVH_TUNNEL);
      cd[i] = residual[i] + pred;
#else
      cd[i] = c_fc[i]->Decompress(pred, SVH_TUNNEL);
#endif
  #ifdef PRINT_CELLDATA_DEBUG_OUTPUT
      fprintf(stderr, "cell data predicted/actual      : %f / %f\n", pred, cd[i]);
  #endif
    }
  }
}

#ifdef SVH_CELL_DATA_SECONDORDER_PREDICTION
static inline void decompressCellDataGap( const float *d0, const float *d1, const float *d2, const float *d3, float* cd, const float *r0, const float *r1, const float *r2, const float *r3, float* residual)
#else
static inline void decompressCellDataGap( const float *d0, const float *d1, const float *d2, const float *d3, float* cd)
#endif
{
  if(c_quantizer)
  {
    for(int i = 0; i < cell_pnum_f; i++)
    {
      I32 pred = (((I32*)d1)[i] + ((I32*)d3)[i])/2; 
#ifdef SVH_CELL_DATA_SECONDORDER_PREDICTION
      I32 residualPred =(((I32*)r1)[i] + ((I32*)r3)[i])/2;
      ((I32*)residual)[i] = c_ic[i]->Decompress(residualPred, SVH_GAP);
      if(((I32*)residual)[i] > c_quantizer->m_aiQuantRange[i])
        ((I32*)residual)[i] -= 2*c_quantizer->m_aiQuantRange[i];
      ((I32*)cd)[i] = ((I32*) residual)[i] + pred;
#else
      ((I32*)cd)[i] = c_ic[i]->Decompress(pred, SVH_GAP);
#endif
#ifdef PRINT_CELLDATA_DEBUG_OUTPUT
      fprintf(stderr, "cell data predicted/actual      : %i / %i\n", pred, ((I32*)cd)[i]);
  #ifdef SVH_CELL_DATA_SECONDORDER_PREDICTION
      fprintf(stderr, "cell data predicted/actual residual: %i / %i\n", residualPred, ((I32*)residual)[i]);
  #endif
#endif
    }
  }
  else
  {
    for(int i=0; i<cell_pnum_f; ++i)
    {
      float pred;
      pred = (d1[i] + d3[i])/2; 
#ifdef SVH_CELL_DATA_SECONDORDER_PREDICTION
      float residualPred =(r1[i] + r3[i])/2;
  #ifdef PRINT_CELLDATA_DEBUG_OUTPUT
      fprintf(stderr, "cell data predicted residual: %f\n", residualPred);
  #endif
      residual[i] = c_fc[i]->Decompress(residualPred, SVH_PIT);
      cd[i] = residual[i] + pred;
#else
      cd[i] = c_fc[i]->Decompress(pred, SVH_PIT);
#endif
  #ifdef PRINT_CELLDATA_DEBUG_OUTPUT
      fprintf(stderr, "cell data predicted/actual      : %f / %f\n", pred, cd[i]);
  #endif
    }
  }
#ifdef OUTPUT_FOR_LS_WEIGHTS
  fwrite(d0, sizeof(float), 1, lsoutfGap);
  fwrite(d1, sizeof(float), 1, lsoutfGap);
  fwrite(d2, sizeof(float), 1, lsoutfGap);
  fwrite(d3, sizeof(float), 1, lsoutfGap);
  fwrite(cd, sizeof(float), 1, lsoutfGap);
#endif
}

#ifdef SVH_CELL_DATA_SECONDORDER_PREDICTION
static inline void decompressCellDataPit( const float *d0, const float *d1, const float *d2, const float *d3, const float *d4, float* cd, const float *r0, const float *r1, const float *r2, const float *r3, const float *r4, float* residual)
#else
static inline void decompressCellDataPit( const float *d0, const float *d1, const float *d2, const float *d3, const float *d4, float* cd)
#endif
{
  if(c_quantizer)
  {
    for(int i = 0; i < cell_pnum_f; i++)
    {
      I32 pred = (((I32*)d1)[i] + ((I32*)d2)[i] + ((I32*)d3)[i] + ((I32*)d4)[i])/4; //TODO: compute the right weights
#ifdef SVH_CELL_DATA_SECONDORDER_PREDICTION
      I32 residualPred = (((I32*)r1)[i] + ((I32*)r2)[i] + ((I32*)r3)[i] + ((I32*)r4)[i])/4;
      ((I32*)residual)[i] = c_ic[i]->Decompress(residualPred, SVH_PIT);
      if(((I32*)residual)[i] > c_quantizer->m_aiQuantRange[i])
        ((I32*)residual)[i] -= 2*c_quantizer->m_aiQuantRange[i];
      ((I32*)cd)[i] = ((I32*) residual)[i] + pred;
#else
      ((I32*)cd)[i] = c_ic[i]->Decompress(pred, SVH_PIT);
#endif
#ifdef PRINT_CELLDATA_DEBUG_OUTPUT
      fprintf(stderr, "cell data predicted/actual      : %i / %i\n", pred, ((I32*)cd)[i]);
  #ifdef SVH_CELL_DATA_SECONDORDER_PREDICTION
      fprintf(stderr, "cell data predicted/actual residual: %i / %i\n", residualPred, ((I32*)residual)[i]);
  #endif
#endif
    }
  }
  else
  {
    for(int i=0; i<cell_pnum_f; ++i)
    {
      float pred;
      pred = (d1[i] + d2[i] + d3[i] + d4[i])/4; //TODO: compute the right weights
#ifdef SVH_CELL_DATA_SECONDORDER_PREDICTION
      float residualPred = (r1[i] + r2[i] + r3[i] + r4[i])/4;
  #ifdef PRINT_CELLDATA_DEBUG_OUTPUT
      fprintf(stderr, "cell data predicted residual: %f\n", residualPred);
  #endif
      residual[i] = c_fc[i]->Decompress(residualPred, SVH_PIT);
      cd[i] = residual[i] + pred;
#else
      cd[i] = c_fc[i]->Decompress(pred, SVH_PIT);
#endif
#ifdef PRINT_CELLDATA_DEBUG_OUTPUT
      fprintf(stderr, "cell data predicted/actual      : %f / %f\n", pred, cd[i]);
#endif
    }
  }
#ifdef OUTPUT_FOR_LS_WEIGHTS
  fwrite(d0, sizeof(float), 1, lsoutfPit);
  fwrite(d1, sizeof(float), 1, lsoutfPit);
  fwrite(d2, sizeof(float), 1, lsoutfPit);
  fwrite(d3, sizeof(float), 1, lsoutfPit);
  fwrite(d4, sizeof(float), 1, lsoutfPit);
  fwrite(cd, sizeof(float), 1, lsoutfPit);
#endif
}

#ifdef SVH_CELL_DATA_SECONDORDER_PREDICTION
static inline void decompressCellDataDen( const float *d0, const float *d1, const float *d2, const float *d3, const float *d4, const float *d5, float* cd, const float *r0, const float *r1, const float *r2, const float *r3, const float *r4, const float *r5, float* residual)
#else
static inline void decompressCellDataDen( const float *d0, const float *d1, const float *d2, const float *d3, const float *d4, const float *d5, float* cd)
#endif
{
  if(c_quantizer)
  {
    for(int i = 0; i < cell_pnum_f; i++)
    {
      I32 pred = (((I32*)d0)[i] + ((I32*)d1)[i] + ((I32*)d2)[i] + ((I32*)d3)[i] + ((I32*)d4)[i] + ((I32*)d5)[i])/6;
#ifdef SVH_CELL_DATA_SECONDORDER_PREDICTION
      I32 residualPred = (((I32*)r0)[i] + ((I32*)r1)[i] + ((I32*)r2)[i] + ((I32*)r3)[i] + ((I32*)r4)[i] + ((I32*)r5)[i])/6;
      ((I32*)residual)[i] = c_ic[i]->Decompress(residualPred, SVH_DEN);
      if(((I32*)residual)[i] > c_quantizer->m_aiQuantRange[i])
        ((I32*)residual)[i] -= 2*c_quantizer->m_aiQuantRange[i];
      ((I32*)cd)[i] = ((I32*) residual)[i] + pred;
#else
      ((I32*)cd)[i] = c_ic[i]->Decompress(pred, SVH_DEN);
#endif
#ifdef PRINT_CELLDATA_DEBUG_OUTPUT
      fprintf(stderr, "cell data predicted/actual      : %i / %i\n", pred, ((I32*)cd)[i]);
  #ifdef SVH_CELL_DATA_SECONDORDER_PREDICTION
      fprintf(stderr, "cell data predicted/actual residual: %i / %i\n", residualPred, ((I32*)residual)[i]);
  #endif
#endif
    }
  }
  else
  {
    for(int i=0; i<cell_pnum_f; ++i)
    {
      float pred = (d0[i] + d1[i] + d2[i] + d3[i] + d4[i]+ d5[i])/6; 
#ifdef SVH_CELL_DATA_SECONDORDER_PREDICTION
      float residualPred = (r0[i] + r1[i] + r2[i] + r3[i] + r4[i]+ r5[i])/6;
  #ifdef PRINT_CELLDATA_DEBUG_OUTPUT
      fprintf(stderr, "cell data predicted residual: %f\n", residualPred);
  #endif
      residual[i] = c_fc[i]->Decompress(residualPred, SVH_DEN);
      cd[i] = residual[i] + pred;
#else
      cd[i] = c_fc[i]->Decompress(pred, SVH_DEN);
#endif
#ifdef PRINT_CELLDATA_DEBUG_OUTPUT
    fprintf(stderr, "cell data predicted/actual      : %f / %f\n", pred, cd[i]);
#endif
    }
  }
}
#endif //SVH_COMPRESS_CELL_DATA

#ifdef SVH_COMPRESS_GEOMETRY
/**
 * decompress the vertex properties of v6 using the lorenzo predictor
 * @param p0,p1,p2,p3,p4,p5,p7: vertex properties known to the decoder
 * @param p6: vertex properties unknown to the decoder
 */
static void decompressVertexGeometryLorenzo(const float* p0, const float* p1, const float* p2, const float* p3, const float* p4, const float* p5, float* p6, const float* p7)
{
#ifdef PRINT_CONTROL_OUTPUT
  predict_lorenzo++;
#endif
  if(v_pq)
  {
    //\bar{p6} = p0 - p1 + p2 - p3 - p4 + p5 + p7
    static I32 pred[3];
    pred[0] = ((I32*)p0)[0] - ((I32*)p1)[0] + ((I32*)p2)[0] - ((I32*)p3)[0] - ((I32*)p4)[0] + ((I32*)p5)[0] + ((I32*)p7)[0];
    pred[1] = ((I32*)p0)[1] - ((I32*)p1)[1] + ((I32*)p2)[1] - ((I32*)p3)[1] - ((I32*)p4)[1] + ((I32*)p5)[1] + ((I32*)p7)[1];
    pred[2] = ((I32*)p0)[2] - ((I32*)p1)[2] + ((I32*)p2)[2] - ((I32*)p3)[2] - ((I32*)p4)[2] + ((I32*)p5)[2] + ((I32*)p7)[2];

#ifdef PRINT_GEOM_DEBUG_OUTPUT
    fprintf(stderr, "predicted(L): %i\t%i\t%i\n", pred[0], pred[1], pred[2]);
#endif
    for(int i = 0; i < 3; i++)
    {
     (( I32*) p6) [i] = v_ic[i]->Decompress(((I32*) pred) [i], SVH_PREDICT_LORENZO);
    }
#ifdef PRINT_GEOM_DEBUG_OUTPUT
    fprintf(stderr, "output(L): %i\t%i\t%i\n",(( I32*) p6) [0],(( I32*) p6) [1],(( I32*) p6) [2]);
#endif
  }
  else
  {
    static float pred[3];
    pred[0] = p0[0] - p1[0] + p2[0] - p3[0] - p4[0] + p5[0] + p7[0];
    pred[1] = p0[1] - p1[1] + p2[1] - p3[1] - p4[1] + p5[1] + p7[1];
    pred[2] = p0[2] - p1[2] + p2[2] - p3[2] - p4[2] + p5[2] + p7[2];

#ifdef PRINT_GEOM_DEBUG_OUTPUT
    fprintf(stderr, "predicted(L): %f\t%f\t%f\n", pred[0], pred[1], pred[2]);
#endif
    for(int i = 0; i < 3; i++)
    {
      p6[i] = v_fc[i]->Decompress(pred[i], SVH_PREDICT_LORENZO);
    }
#ifdef PRINT_GEOM_DEBUG_OUTPUT
    fprintf(stderr, "output(L): %f\t%f\t%f\n", p6[0], p6[1], p6[2]);
#endif
  }
}

/**
 * decompress the vertex properties of vb using the 'last decoded vertex' predictor
 * @param pa: vertex to predict
 */
static void decompressVertexGeometryLast(float* pa)
{
#ifdef PRINT_CONTROL_OUTPUT
  predict_last++;
#endif
  if(v_pq)
  {
    for(int i = 0; i < 3; i++)
    {
     (( I32*) pa) [i] = v_ic[i]->Decompress(0, SVH_PREDICT_NONE);
    }
#ifdef PRINT_GEOM_DEBUG_OUTPUT
    fprintf(stderr, "output(N): %i\t%i\t%i\n",(( I32*) pa) [0],(( I32*) pa) [1],(( I32*) pa) [2]);
#endif
  }
  else
  {
    for(int i = 0; i < 3; i++)
    {
      pa[i] = v_fc[i]->Decompress(0.0f, SVH_PREDICT_NONE);
    }
#ifdef PRINT_GEOM_DEBUG_OUTPUT
    fprintf(stderr, "output(N): %f\t%f\t%f\n", pa[0], pa[1], pa[2]);
#endif
  }
}

/**
 * decompress the vertex properties of vb using the 'shoot' predictor
 * @param pb: vertex to predict
 * @param pa: vertex below it
 */
static void decompressVertexGeometryShoot(float* pb, const float* pa)
{
#ifdef PRINT_CONTROL_OUTPUT
  predict_shoot++;
#endif
  if(v_pq)
  {
#ifdef PRINT_GEOM_DEBUG_OUTPUT
    fprintf(stderr, "prediction(S): %i\t%i\t%i\n",(( I32*) pa) [0],(( I32*) pa) [1],(( I32*) pa) [2]);
#endif
    for(int i = 0; i < 3; i++)
    {
     (( I32*) pb) [i] = v_ic[i]->Decompress(((I32*) pa) [i], SVH_PREDICT_SHOOT);
    }
#ifdef PRINT_GEOM_DEBUG_OUTPUT
    fprintf(stderr, "output(S): %i\t%i\t%i\n",(( I32*) pb) [0],(( I32*) pb) [1],(( I32*) pb) [2]);
#endif
  }
  else
  {
#ifdef PRINT_GEOM_DEBUG_OUTPUT
    fprintf(stderr, "predicted(S): %f\t%f\t%f\n", pa[0], pa[1], pa[2]);
#endif
    for(int i = 0; i < 3; i++)
    {
      pb[i] = v_fc[i]->Decompress(pa[i], SVH_PREDICT_SHOOT);
    }
#ifdef PRINT_GEOM_DEBUG_OUTPUT
    fprintf(stderr, "output(S): %f\t%f\t%f\n", pb[0], pb[1], pb[2]);
#endif
  }
}

/**
 * decompress the vertex properties of vd using the parallelogram predictor
 * @param pb: vertex before pa when enumerating the vertices of the face
 * @param pa: opposite vertex
 * @param pc: vertex after pa when enumerating the vertices of the face
 * @param pd: vertex properties unknown to the decoder
 */
static void decompressVertexGeometryParallelogram(const float* pb, const float* pa, const float* pc, float* pd)
{
#ifdef PRINT_CONTROL_OUTPUT
  predict_parallelogram++;
#endif
  if(v_pq)
  {
    //\bar{pd} = pb + pc - pa
    static I32 pred[3];
    pred[0] = ((I32*)pb)[0] - ((I32*)pa)[0] + ((I32*)pc)[0] ;
    pred[1] = ((I32*)pb)[1] - ((I32*)pa)[1] + ((I32*)pc)[1] ;
    pred[2] = ((I32*)pb)[2] - ((I32*)pa)[2] + ((I32*)pc)[2] ;

#ifdef PRINT_GEOM_DEBUG_OUTPUT
    fprintf(stderr, "predicted(P): %i\t%i\t%i\n", pred[0], pred[1], pred[2]);
#endif
    for(int i = 0; i < 3; i++)
    {
     (( I32*) pd) [i] = v_ic[i]->Decompress(pred[i], SVH_PREDICT_PARALLELOGRAM);
    }
#ifdef PRINT_GEOM_DEBUG_OUTPUT
    fprintf(stderr, "output(P): %i\t%i\t%i\n",(( I32*) pd) [0],(( I32*) pd) [1],(( I32*) pd) [2]);
#endif
  }
  else
  {
    static float pred[3];

#ifdef PRINT_GEOM_DEBUG_OUTPUT
    fprintf(stderr, "pb: %f\t%f\t%f\npa: %f\t%f\t%f\npc: %f\t%f\t%f\n", pb[0], pb[1], pb[2], pa[0], pa[1], pa[2], pc[0], pc[1], pc[2]);
#endif

    pred[0] = pb[0] - pa[0] + pc[0];
    pred[1] = pb[1] - pa[1] + pc[1];
    pred[2] = pb[2] - pa[2] + pc[2];

#ifdef PRINT_GEOM_DEBUG_OUTPUT
    fprintf(stderr, "predicted(P): %f\t%f\t%f\n", pred[0], pred[1], pred[2]);
#endif

    for(int i = 0; i < 3; i++)
    {
      pd[i] = v_fc[i]->Decompress(pred[i], SVH_PREDICT_PARALLELOGRAM);
    }
#ifdef PRINT_GEOM_DEBUG_OUTPUT
    fprintf(stderr, "output(P): %f\t%f\t%f\n", pd[0], pd[1], pd[2]);
#endif
  }

}
#endif //SVH_COMPRESS_GEOMETRY

/**
 * helper functions to create and decompress each vertex
 * @param e: the edge number
 * @param i0,i1: vertex indices
 * @return
 */

void SVreader_svh::createAddAndDecompressV0(void *p_vertices)
{
  SVvertex **cvertices = (SVvertex **)p_vertices;
  // create a new vertex
  cvertices[0] = allocVertex();
  // give it its index
  cvertices[0]->index = v_count + have_new;
#ifdef SVH_COMPRESS_GEOMETRY
  decompressVertexGeometryLast(cvertices[0]->v_prop_f);
#else
  fread(cvertices[0]->v_prop_f, 1, sizeof(float) *3, geofiled);
#endif
  // insert it into the indexable data structure
  dv->addElement(cvertices[0]);
  // enlist it for being output as a new vertex
  new_vertices[have_new] = 0;
  have_new++;
}
void SVreader_svh::createAddAndDecompressV1(void *p_vertices)
{
  SVvertex **cvertices = (SVvertex **)p_vertices;
  // create a new vertex
  cvertices[1] = allocVertex();
  // give it its index
  cvertices[1]->index = v_count + have_new;
#ifdef SVH_COMPRESS_GEOMETRY
  decompressVertexGeometryShoot(cvertices[1]->v_prop_f, cvertices[0]->v_prop_f);
#else
  fread(cvertices[1]->v_prop_f, 1, sizeof(float) *3, geofiled);
#endif
  // insert it into the indexable data structure
  dv->addElement(cvertices[1]);
  // enlist it for being output as a new vertex
  new_vertices[have_new] = 1;
  have_new++;
}
void SVreader_svh::createAddAndDecompressV2(void *p_vertices)
{
  SVvertex **cvertices = (SVvertex **)p_vertices;
// create a new vertex
  cvertices[2] = allocVertex();
  // give it its index
  cvertices[2]->index = v_count + have_new;
#ifdef SVH_COMPRESS_GEOMETRY
  decompressVertexGeometryShoot(cvertices[2]->v_prop_f, cvertices[1]->v_prop_f);
#else
  fread(cvertices[2]->v_prop_f, 1, sizeof(float) *3, geofiled);
#endif
  // insert it into the indexable data structure
  dv->addElement(cvertices[2]);
  // enlist it for being output as a new vertex
  new_vertices[have_new] = 2;
  have_new++;
}
void SVreader_svh::createAddAndDecompressV3(void *p_vertices)
{
  SVvertex **cvertices = (SVvertex **)p_vertices;
  // create a new vertex
  cvertices[3] = allocVertex();
  // give it its index
  cvertices[3]->index = v_count + have_new;
#ifdef SVH_COMPRESS_GEOMETRY
  decompressVertexGeometryParallelogram(cvertices[2]->v_prop_f, cvertices[1]->v_prop_f, cvertices[0]->v_prop_f, cvertices[3]->v_prop_f);
#else
  fread(cvertices[3]->v_prop_f, 1, sizeof(float) *3, geofiled);
#endif
  // insert it into the indexable data structure
  dv->addElement(cvertices[3]);
  // enlist it for being output as a new vertex
  new_vertices[have_new] = 3;
  have_new++;
}
void SVreader_svh::createAddAndDecompressV4(void *p_vertices)
{
  SVvertex **cvertices = (SVvertex **)p_vertices;
  // create a new vertex
  cvertices[4] = allocVertex();
  // give it its index
  cvertices[4]->index = v_count + have_new;
#ifdef SVH_COMPRESS_GEOMETRY
  decompressVertexGeometryShoot(cvertices[4]->v_prop_f, cvertices[0]->v_prop_f);
#else
  fread(cvertices[4]->v_prop_f, 1, sizeof(float) *3, geofiled);
#endif
  // insert it into the indexable data structure
  dv->addElement(cvertices[4]);
  // enlist it for being output as a new vertex
  new_vertices[have_new] = 4;
  have_new++;
}
void SVreader_svh::createAddAndDecompressV7(void *p_vertices)
{
  SVvertex **cvertices = (SVvertex **)p_vertices;
  // create a new vertex
  cvertices[7] = allocVertex();
  // give it its index
  cvertices[7]->index = v_count + have_new;
#ifdef SVH_COMPRESS_GEOMETRY
  decompressVertexGeometryParallelogram(cvertices[3]->v_prop_f, cvertices[0]->v_prop_f, cvertices[4]->v_prop_f, cvertices[7]->v_prop_f);
#else
  fread(cvertices[7]->v_prop_f, 1, sizeof(float) *3, geofiled);
#endif
  // insert it into the indexable data structure
  dv->addElement(cvertices[7]);
  // enlist it for being output as a new vertex
  new_vertices[have_new] = 7;
  have_new++;
}
void SVreader_svh::createAddAndDecompressV5(void *p_vertices)
{
  SVvertex **cvertices = (SVvertex **)p_vertices;
  // create a new vertex
  cvertices[5] = allocVertex();
  // give it its index
  cvertices[5]->index = v_count + have_new;
#ifdef SVH_COMPRESS_GEOMETRY
  decompressVertexGeometryParallelogram(cvertices[4]->v_prop_f, cvertices[0]->v_prop_f, cvertices[1]->v_prop_f, cvertices[5]->v_prop_f);
#else
  fread(cvertices[5]->v_prop_f, 1, sizeof(float) *3, geofiled);
#endif
  // insert it into the indexable data structure
  dv->addElement(cvertices[5]);
  // enlist it for being output as a new vertex
  new_vertices[have_new] = 5;
  have_new++;
}
void SVreader_svh::createAddAndDecompressV6(void *p_vertices)
{
  SVvertex **cvertices = (SVvertex **)p_vertices;
  // create a new vertex
  cvertices[6] = allocVertex();
  // give it its index
  cvertices[6]->index = v_count + have_new;
#ifdef SVH_COMPRESS_GEOMETRY
  decompressVertexGeometryLorenzo(cvertices[0]->v_prop_f, cvertices[1]->v_prop_f, cvertices[2]->v_prop_f, cvertices[3]->v_prop_f, cvertices[4]->v_prop_f, cvertices[5]->v_prop_f, cvertices[6]->v_prop_f, cvertices[7]->v_prop_f);
#else
  fread(cvertices[6]->v_prop_f, 1, sizeof(float) *3, geofiled);
#endif
  // insert it into the indexable data structure
  dv->addElement(cvertices[6]);
  // enlist it for being output as a new vertex
  new_vertices[have_new] = 6;
  have_new++;
}


/**
 * helper function to get the edge number from two vertex local indices
 * @param e: the edge number
 * @param i1,i0: vertex indices (edge goes from i1 to i0)
 * @return
 */
static void getLocalIndicesFromEdgeNumber(int e, int &i1, int &i0)
{
  switch(e)
  {
    case 0:
      i1 = 1; i0 = 0;
      return;
    case 1:
      i1 = 0; i0 = 3;
      return;
    case 2:
      i1 = 3; i0 = 2;
      return;
    case 3:
      i1 = 2; i0 = 1;
      return;
    case 4:
      i1 = 3; i0 = 0;
      return;
    case 5:
      i1 = 0; i0 = 4;
      return;
    case 6:
      i1 = 4; i0 = 7;
      return;
    case 7:
      i1 = 7; i0 = 3;
      return;
    case 8:
      i1 = 4; i0 = 0;
      return;
    case 9:
      i1 = 0; i0 = 1;
      return;
    case 10:
      i1 = 1; i0 = 5;
      return;
    case 11:
      i1 = 5; i0 = 4;
      return;
    case 12:
      i1 = 5; i0 = 1;
      return;
    case 13:
      i1 = 1; i0 = 2;
      return;
    case 14:
      i1 = 2; i0 = 6;
      return;
    case 15:
      i1 = 6; i0 = 5;
      return;
    case 16:
      i1 = 6; i0 = 2;
      return;
    case 17:
      i1 = 2; i0 = 3;
      return;
    case 18:
      i1 = 3; i0 = 7;
      return;
    case 19:
      i1 = 7; i0 = 6;
      return;
    case 20:
      i1 = 7; i0 = 4;
      return;
    case 21:
      i1 = 4; i0 = 5;
      return;
    case 22:
      i1 = 5; i0 = 6;
      return;
    case 23:
      i1 = 6; i0 = 7;
      return;
    default:
      assert(false);
      return;
  }
}

static SVvertex* vertices[8];
int SVreader_svh::decompress_hexahedron()
{
#ifdef PRINT_DEBUG_OUTPUT
  fprintf(stderr, "\nBEGIN (con/de)compress_hexahedron()\n");
#endif

  int i;
  int op;
  int lc_pos;

  SVedge* f_visited[]={NULL,NULL,NULL,NULL,NULL,NULL};

  if(edge_buffer_size)
  {
#ifdef SVH_RMOP_DEPENDS_ON_FINALIZATION
    op = rd_conn_op->decode(rmOpX[last_fin][last_op]);
#else
    op = rd_conn_op->decode(rmOp[last_op]);
#endif
  }
  else
  {
    if(rd_conn_op->decode(rmDone))
    {
      return 0;
    }
    op = SVH_START;
  }

  if(op == SVH_START)
  {
#ifdef PRINT_DEBUG_OUTPUT
    fprintf(stderr,  "SVH_START\n");
#endif
    unsigned int joinedVertices = rd_conn_cache->decode(rmStart_VisitedVertices);

#ifdef PRINT_DEBUG_OUTPUT
    fprintf(stderr,  "joinedVertices is %i\n", joinedVertices);
#endif

    //see if this is the end of the stream
    if(joinedVertices==256)
      return 0;

    // decompress vertices
    if(joinedVertices&1)
    {
      lc_pos = rd_conn_cache->decode(rmS_Cache[0]);
      if(lc_pos == 8) //not in the chache, get it from the dynamic vector
      {
        int dv_index;
        dv_index = rd_conn_index->readRange(dv->size());
        vertices[0] =(SVvertex*) dv->getElementWithRelativeIndex(dv_index);
      }
      else //get it from the little cache
      {
        vertices[0] =(SVvertex*) lc->get(lc_pos);
      }
    }
    else
    {
      createAddAndDecompressV0(vertices);
    }
    if(joinedVertices&2)
    {
      lc_pos = rd_conn_cache->decode(rmS_Cache[1]);
      if(lc_pos == 8) //not in the chache, get it from the dynamic vector
      {
        int dv_index;
        dv_index = rd_conn_index->readRange(dv->size());
        vertices[1] =(SVvertex*) dv->getElementWithRelativeIndex(dv_index);
      }
      else //get it from the little cache
      {
        vertices[1] =(SVvertex*) lc->get(lc_pos);
      }
    }
    else
    {
      createAddAndDecompressV1(vertices);
    }
    if(joinedVertices&4)
    {
      lc_pos = rd_conn_cache->decode(rmS_Cache[2]);
      if(lc_pos == 8) //not in the chache, get it from the dynamic vector
      {
        int dv_index;
        dv_index = rd_conn_index->readRange(dv->size());
        vertices[2] =(SVvertex*) dv->getElementWithRelativeIndex(dv_index);
      }
      else //get it from the little cache
      {
        vertices[2] =(SVvertex*) lc->get(lc_pos);
      }
    }
    else
    {
      createAddAndDecompressV2(vertices);
    }
    if(joinedVertices&8)
    {
      lc_pos = rd_conn_cache->decode(rmS_Cache[3]);
      if(lc_pos == 8) //not in the chache, get it from the dynamic vector
      {
        int dv_index;
        dv_index = rd_conn_index->readRange(dv->size());
        vertices[3] =(SVvertex*) dv->getElementWithRelativeIndex(dv_index);
      }
      else //get it from the little cache
      {
        vertices[3] =(SVvertex*) lc->get(lc_pos);
      }
    }
    else
    {
      createAddAndDecompressV3(vertices);
    }
    if(joinedVertices&16)
    {
      lc_pos = rd_conn_cache->decode(rmS_Cache[4]);
      if(lc_pos == 8) //not in the chache, get it from the dynamic vector
      {
        int dv_index;
        dv_index = rd_conn_index->readRange(dv->size());
        vertices[4] =(SVvertex*) dv->getElementWithRelativeIndex(dv_index);
      }
      else //get it from the little cache
      {
        vertices[4] =(SVvertex*) lc->get(lc_pos);
      }
    }
    else
    {
      createAddAndDecompressV4(vertices);
    }
    if(joinedVertices&128)
    {
      lc_pos = rd_conn_cache->decode(rmS_Cache[7]);
      if(lc_pos == 8) //not in the chache, get it from the dynamic vector
      {
        int dv_index;
        dv_index = rd_conn_index->readRange(dv->size());
        vertices[7] =(SVvertex*) dv->getElementWithRelativeIndex(dv_index);
      }
      else //get it from the little cache
      {
        vertices[7] =(SVvertex*) lc->get(lc_pos);
      }
    }
    else
    {
      createAddAndDecompressV7(vertices);
    }
    if(joinedVertices&32)
    {
      lc_pos = rd_conn_cache->decode(rmS_Cache[5]);
      if(lc_pos == 8) //not in the chache, get it from the dynamic vector
      {
        int dv_index;
        dv_index = rd_conn_index->readRange(dv->size());
        vertices[5] =(SVvertex*) dv->getElementWithRelativeIndex(dv_index);
      }
      else //get it from the little cache
      {
        vertices[5] =(SVvertex*) lc->get(lc_pos);
      }
    }
    else
    {
      createAddAndDecompressV5(vertices);
    }
    // compress vertices
    if(joinedVertices&64)
    {
      lc_pos = rd_conn_cache->decode(rmS_Cache[6]);
      if(lc_pos == 8) //not in the chache, get it from the dynamic vector
      {
        int dv_index;
        dv_index = rd_conn_index->readRange(dv->size());
        vertices[6] =(SVvertex*) dv->getElementWithRelativeIndex(dv_index);
      }
      else //get it from the little cache
      {
        vertices[6] =(SVvertex*) lc->get(lc_pos);
      }
    }
    else
    {
      createAddAndDecompressV6(vertices);
    }

#ifdef SVH_COMPRESS_CELL_DATA
    if(c_pnum_f)
    {
#ifdef SVH_CELL_DATA_SECONDORDER_PREDICTION
      decompressCellDataStart(c_prop_f, tmpCellDataResidual);
#else
      decompressCellDataStart(c_prop_f);
#endif
    }
#endif  
  }
  else if(op == SVH_HUT)
  {
#ifdef PRINT_DEBUG_OUTPUT
    fprintf(stderr, "SVH_HUT\n");
#endif
#ifdef SVH_FACECACHE_DEPENDS_ON_FINALIZATION
    lc_pos = rd_conn_cache->decode(rmHut_FaceCacheX[last_fin][last_op]);
#else
    lc_pos = rd_conn_cache->decode(rmHut_FaceCache[last_op]);
#endif
#ifdef PRINT_DEBUG_OUTPUT
    fprintf(stderr, "lc_pos_f is %i\n", lc_pos);
#endif
    if(lc_pos < 6) //f0 is is in the face cache
    {
      f_visited[0] =(SVedge*) lc_f->get(lc_pos);
      vertices[0] = f_visited[0]->target;
      vertices[3] = f_visited[0]->next->target;
      vertices[2] = f_visited[0]->next->next->target;
      vertices[1] = f_visited[0]->next->next->next->target;
    }
    else //f0 is is not in the face cache
    {
      lc_pos = rd_conn_cache->decode(rmHut_EdgeCache[last_op]);
#ifdef PRINT_DEBUG_OUTPUT
      fprintf(stderr, "lc_pos_e is %i\n", lc_pos);
#endif
      if(lc_pos < 24) //f0 has edge 1->0 in the edge cache
      {
        int i0,i1;
        //get which edge in the cache this is
        getLocalIndicesFromEdgeNumber(lc_pos, i1, i0);
        vertices[1] =(SVvertex*) lc->get(i1);
        vertices[0] =(SVvertex*) lc->get(i0);
        // encode which of the faces incident to v1 and v0 is f0
        int candidate=-1, candidate_count=0;
        for(i = 0; i < vertices[1]->list_size; i++)
        {
          if(vertices[1]->list[i]->target == vertices[0])
          {
            f_visited[0] = vertices[1]->list[i];
            candidate_count++;
          }
        }
        assert(candidate_count != 0);
        if(candidate_count > 1)
        {
          if(candidate_count == 2)
          {
            candidate = rd_conn_candi->decode(rm_FaceCandi);
          }
          else
          {
            candidate = rd_conn_candi->readRange(candidate_count);
          }
          if(candidate != candidate_count - 1)
          {
            for(i = 0; i < vertices[1]->list_size; i++)
            {
              if(vertices[1]->list[i]->target == vertices[0])
              {
                if(candidate == 0)
                {
                  f_visited[0] = vertices[1]->list[i];
                  break;
                }
                else
                  candidate--;
              }
            }
          }
        }
        vertices[3] = f_visited[0]->next->target;
        vertices[2] = f_visited[0]->next->next->target;
        assert(vertices[1] == f_visited[0]->next->next->next->target);
      }
      else //f0 does not have an edge in the edge cache
      {
        lc_pos = rd_conn_cache->decode(rmHut_VertexCache[last_op]);
#ifdef PRINT_DEBUG_OUTPUT
        fprintf(stderr, "lc_pos_v is %i\n", lc_pos);
#endif
        if(lc_pos < 8) //f0 has a v1 in the vertex cache
        {
          vertices[1] =(SVvertex*) lc->get(lc_pos);
        }
        else
        {
          int dv_index;
          dv_index = rd_conn_index->readRange(dv->size());
          vertices[1] =(SVvertex*) dv->getElementWithRelativeIndex(dv_index);
        }

        // decode which of the faces incident to v1 is f0
        if(vertices[1]->list_size > 1)
        {
          if(vertices[1]->list_size < SVH_MAGIC_NUMBER_F0_AROUND_V1_RM)
            f_visited[0] = vertices[1]->list[ rd_conn->decode(rm_VertexCandi[vertices[1]->list_size]) ];
          else
            f_visited[0] = vertices[1]->list[ rd_conn->readRange(vertices[1]->list_size) ];
        }
        else
          f_visited[0] = vertices[1]->list[0];

        vertices[0] = f_visited[0]->target;
        vertices[3] = f_visited[0]->next->target;
        vertices[2] = f_visited[0]->next->next->target;
        assert(vertices[1] == f_visited[0]->next->next->next->target);
      }
    }

    // decode v4,v5,v6,v7
    //we first need to read whether there is a vertex or edge join
    if(rd_conn_op->decode(rmHut_HasJoin) == 1) //if there are joins
    {
#ifdef SVH_DUMB_JOIN_CODING
      unsigned int joinedVertices = 1 + rd_conn_op->decode(rmHut_JoinedVertices);

#ifdef PRINT_DEBUG_OUTPUT
      fprintf(stderr, "joinedString is %i\n", joinedVertices);
#endif

      if(joinedVertices&1)
      {
        int dv_index;
        dv_index = rd_conn_index->readRange(dv->size());
        vertices[4] =(SVvertex*) dv->getElementWithRelativeIndex(dv_index);
#ifdef PRINT_DEBUG_OUTPUT
        fprintf(stderr, "vertex 4 is joined to index %i (dv->size() %i)\n", dv_index, dv->size());
#endif
      }
      else
      {
        createAddAndDecompressV4(vertices);
      }
      if(joinedVertices&8)
      {
        int dv_index;
        dv_index = rd_conn_index->readRange(dv->size());
        vertices[7] =(SVvertex*) dv->getElementWithRelativeIndex(dv_index);
#ifdef PRINT_DEBUG_OUTPUT
        fprintf(stderr, "vertex 7 is joined to index %i (dv->size() %i)\n", dv_index, dv->size());
#endif
      }
      else
      {
        createAddAndDecompressV7(vertices);
      }
      if(joinedVertices&2)
      {
        int dv_index;
        dv_index = rd_conn_index->readRange(dv->size());
        vertices[5] =(SVvertex*) dv->getElementWithRelativeIndex(dv_index);
#ifdef PRINT_DEBUG_OUTPUT
        fprintf(stderr, "vertex 5 is joined to index %i (dv->size() %i)\n", dv_index, dv->size());
#endif
      }
      else
      {
        createAddAndDecompressV5(vertices);
      }
      // compress vertices
      if(joinedVertices&4)
      {
        int dv_index;
        dv_index = rd_conn_index->readRange(dv->size());
        vertices[6] =(SVvertex*) dv->getElementWithRelativeIndex(dv_index);
#ifdef PRINT_DEBUG_OUTPUT
        fprintf(stderr, "vertex 6 is joined to index %i (dv->size() %i)\n", dv_index, dv->size());
#endif
      }
      else
      {
        createAddAndDecompressV6(vertices);
      }
#else
      fprintf(stderr, "NOT YET IMPLEMENTED: intelligent join for SVH_HUT\n");
      exit(1);
#endif
    }
    else
    {
      createAddAndDecompressV4(vertices);
      createAddAndDecompressV7(vertices);
      createAddAndDecompressV5(vertices);
      createAddAndDecompressV6(vertices);
    }
#ifdef SVH_COMPRESS_CELL_DATA
    //now decompress the cell data.
    if(c_pnum_f)
    {
#ifdef SVH_CELL_DATA_SECONDORDER_PREDICTION
      decompressCellDataHut(f_visited[0]->c_prop_f, c_prop_f, f_visited[0]->c_prop_f+c_pnum_f, tmpCellDataResidual);
#else
      decompressCellDataHut(f_visited[0]->c_prop_f, c_prop_f);
#endif
    }
#endif  
  }
  else if(op == SVH_ROOF)
  {
#ifdef PRINT_DEBUG_OUTPUT
    fprintf(stderr, "SVH_ROOF\n");
#endif
#ifdef SVH_FACECACHE_DEPENDS_ON_FINALIZATION
    lc_pos = rd_conn_cache->decode(rmRoof_FaceCacheX[last_fin][last_op]);
#else
    lc_pos = rd_conn_cache->decode(rmRoof_FaceCache[last_op]);
#endif
#ifdef PRINT_DEBUG_OUTPUT
    fprintf(stderr, "lc_pos_f is %i\n", lc_pos);
#endif
    if(lc_pos < 6) //f0 is is in the face cache
    {
      f_visited[0] =(SVedge*) lc_f->get(lc_pos);
      vertices[0] = f_visited[0]->target;
      vertices[3] = f_visited[0]->next->target;
      vertices[2] = f_visited[0]->next->next->target;
      vertices[1] = f_visited[0]->next->next->next->target;
    }
    else //f0 is is not in the face cache
    {
      lc_pos = rd_conn_cache->decode(rmRoof_EdgeCache[last_op]);
#ifdef PRINT_DEBUG_OUTPUT
      fprintf(stderr, "lc_pos_e is %i\n", lc_pos);
#endif
      if(lc_pos < 24) //f0 has edge 1->0 in the edge cache
      {
        int i0,i1;
        //get which edge in the cache this is
        getLocalIndicesFromEdgeNumber(lc_pos, i1, i0);
        vertices[1] =(SVvertex*) lc->get(i1);
        vertices[0] =(SVvertex*) lc->get(i0);
        // encode which of the faces incident to v1 and v0 is f0
        int candidate=-1, candidate_count=0;
        for(i = 0; i < vertices[1]->list_size; i++)
        {
          if(vertices[1]->list[i]->target == vertices[0])
          {
            f_visited[0] = vertices[1]->list[i];
            candidate_count++;
          }
        }
        assert(candidate_count != 0);
        if(candidate_count > 1)
        {
          if(candidate_count == 2)
          {
            candidate = rd_conn_candi->decode(rm_FaceCandi);
          }
          else
          {
            candidate = rd_conn_candi->readRange(candidate_count);
          }
          if(candidate != candidate_count - 1)
          {
            for(i = 0; i < vertices[1]->list_size; i++)
            {
              if(vertices[1]->list[i]->target == vertices[0])
              {
                if(candidate == 0)
                {
                  f_visited[0] = vertices[1]->list[i];
                  break;
                }
                else
                  candidate--;
              }
            }
          }
        }
        vertices[3] = f_visited[0]->next->target;
        vertices[2] = f_visited[0]->next->next->target;
        assert(vertices[1] == f_visited[0]->next->next->next->target);
      }
      else //f0 does not have an edge in the edge cache
      {
        lc_pos = rd_conn_cache->decode(rmRoof_VertexCache[last_op]);
#ifdef PRINT_DEBUG_OUTPUT
        fprintf(stderr, "lc_pos_v is %i\n", lc_pos);
#endif
        if(lc_pos < 8) //f0 has a v1 in the vertex cache
        {
          vertices[1] =(SVvertex*) lc->get(lc_pos);
        }
        else
        {
          int dv_index;
          dv_index = rd_conn_index->readRange(dv->size());
          //printf("dv_index: %i/%i\n", dv_index, dv->size());
          vertices[1] =(SVvertex*) dv->getElementWithRelativeIndex(dv_index);
        }

        // decode which of the faces incident to v1 is f0
        //printf("%i\n", vertices[1]->list_size);
        if(vertices[1]->list_size > 1)
        {
          if(vertices[1]->list_size < SVH_MAGIC_NUMBER_F0_AROUND_V1_RM)
            f_visited[0] = vertices[1]->list[ rd_conn->decode(rm_VertexCandi[vertices[1]->list_size]) ];
          else
            f_visited[0] = vertices[1]->list[ rd_conn->readRange(vertices[1]->list_size) ];
        }
        else
          f_visited[0] = vertices[1]->list[0];

        vertices[0] = f_visited[0]->target;
        vertices[3] = f_visited[0]->next->target;
        vertices[2] = f_visited[0]->next->next->target;
        assert(vertices[1] == f_visited[0]->next->next->next->target);
      }
    }

    // decode f5: first decode v7
    int dv_index;
    dv_index = rd_conn_index->readRange(dv->size());
    vertices[7] =(SVvertex*) dv->getElementWithRelativeIndex(dv_index);

    // then decode which of the faces incident to v7 is f5
    if(vertices[7]->list_size > 1)
    {
      if(vertices[7]->list_size < SVH_MAGIC_NUMBER_F0_AROUND_V1_RM)
        f_visited[5] = vertices[7]->list[ rd_conn->decode(rm_VertexCandi[vertices[7]->list_size]) ];
      else
        f_visited[5] = vertices[7]->list[ rd_conn->readRange(vertices[7]->list_size) ];
    }
    else
      f_visited[5] = vertices[7]->list[0];

    vertices[4] = f_visited[5]->target;
    vertices[5] = f_visited[5]->next->target;
    vertices[6] = f_visited[5]->next->next->target;
    assert(vertices[7] == f_visited[5]->next->next->next->target);

#ifdef SVH_COMPRESS_CELL_DATA
    //now decompress the cell data.
    if(c_pnum_f)
    {
#ifdef SVH_CELL_DATA_SECONDORDER_PREDICTION
      decompressCellDataRoof(f_visited[0]->c_prop_f, f_visited[5]->c_prop_f, c_prop_f, f_visited[0]->c_prop_f+c_pnum_f, f_visited[5]->c_prop_f+c_pnum_f, tmpCellDataResidual);
#else
      decompressCellDataRoof(f_visited[0]->c_prop_f, f_visited[5]->c_prop_f, c_prop_f);
#endif
    }
#endif  

  }
  else if(op == SVH_STEP) // SVH_STEP, SVH_CORNER, SVH_GAP, SVH_PIT, SVH_DEN
  {
#ifdef PRINT_DEBUG_OUTPUT
    fprintf(stderr, "SVH_STEP\n");
#endif
#ifdef SVH_FACECACHE_DEPENDS_ON_FINALIZATION
    lc_pos = rd_conn_cache->decode(rmSCGPD_FaceCacheX[last_fin][last_op]);
#else
    lc_pos = rd_conn_cache->decode(rmSCGPD_FaceCache[last_op]);
#endif
#ifdef PRINT_DEBUG_OUTPUT
    fprintf(stderr, "lc_pos_f is %i\n", lc_pos);
#endif
    if(lc_pos < 24) //f0 is in the face cache
    {
      if(lc_pos < 6)
        f_visited[0] =(( SVedge*) lc_f->get(lc_pos));
      else if(lc_pos < 12)
        f_visited[0] =(( SVedge*) lc_f->get(lc_pos-6))->next->next->next;
      else if(lc_pos < 18)
        f_visited[0] =(( SVedge*) lc_f->get(lc_pos-12))->next->next;
      else
        f_visited[0] =(( SVedge*) lc_f->get(lc_pos-18))->next;

      vertices[0] = f_visited[0]->target;
      vertices[3] = f_visited[0]->next->target;
      vertices[2] = f_visited[0]->next->next->target;
      vertices[1] = f_visited[0]->next->next->next->target;
    }
    else //f0 is is not in the face cache
    {
      lc_pos = rd_conn_cache->decode(rmSCGPD_EdgeCache[last_op]);
#ifdef PRINT_DEBUG_OUTPUT
      fprintf(stderr, "lc_pos_e is %i\n", lc_pos);
#endif
      if(lc_pos < 96) //f0 has an edge in the edge cache
      {
        if(lc_pos < 24)
        {
          //get which edge in the cache this is
          int i0,i1;
          getLocalIndicesFromEdgeNumber(lc_pos, i1, i0);
          vertices[1] =(SVvertex*) lc->get(i1);
          vertices[0] =(SVvertex*) lc->get(i0);
          // encode which of the faces incident to v1 and v0 is f0
          int candidate=-1, candidate_count=0;
          for(i = 0; i < vertices[1]->list_size; i++)
          {
            if(vertices[1]->list[i]->target == vertices[0])
            {
              f_visited[0] = vertices[1]->list[i];
              candidate_count++;
            }
          }
          assert(candidate_count != 0);
          if(candidate_count > 1)
          {
            if(candidate_count == 2)
            {
              candidate = rd_conn_candi->decode(rm_FaceCandi);
            }
            else
            {
              candidate = rd_conn_candi->readRange(candidate_count);
            }
            if(candidate != candidate_count - 1)
            {
              for(i = 0; i < vertices[1]->list_size; i++)
              {
                if(vertices[1]->list[i]->target == vertices[0])
                {
                  if(candidate == 0)
                  {
                    f_visited[0] = vertices[1]->list[i];
                    break;
                  }
                  else
                    candidate--;
                }
              }
            }
          }
          vertices[3] = f_visited[0]->next->target;
          vertices[2] = f_visited[0]->next->next->target;
        }
        else if(lc_pos < 48)
        {
          //get which edge in the cache this is
          int i0,i1;
          getLocalIndicesFromEdgeNumber(lc_pos-24, i1, i0);
          vertices[0] =(SVvertex*) lc->get(i1);
          vertices[3] =(SVvertex*) lc->get(i0);
          // decode which of the faces incident to v0 and v3 is f0
          int candidate=-1, candidate_count=0;
          for(i = 0; i < vertices[0]->list_size; i++)
          {
            if(vertices[0]->list[i]->target == vertices[3])
            {
              f_visited[0] = vertices[0]->list[i]->next->next->next;
              candidate_count++;
            }
          }
          assert(candidate_count != 0);
          if(candidate_count > 1)
          {
            if(candidate_count == 2)
            {
              candidate = rd_conn_candi->decode(rm_FaceCandi);
            }
            else
            {
              candidate = rd_conn_candi->readRange(candidate_count);
            }
            if(candidate != candidate_count - 1)
            {
              for(i = 0; i < vertices[0]->list_size; i++)
              {
                if(vertices[0]->list[i]->target == vertices[3])
                {
                  if(candidate == 0)
                  {
                    f_visited[0] = vertices[0]->list[i]->next->next->next;
                    break;
                  }
                  else
                    candidate--;
                }
              }
            }
          }
          vertices[1] = f_visited[0]->next->next->next->target;
          vertices[2] = f_visited[0]->next->next->target;
        }
        else if(lc_pos < 72)
        {
          //get which edge in the cache this is
          int i0,i1;
          getLocalIndicesFromEdgeNumber(lc_pos-48, i1, i0);
          vertices[3] =(SVvertex*) lc->get(i1);
          vertices[2] =(SVvertex*) lc->get(i0);
          // decode which of the faces incident to v3 and v2 is f0
          int candidate=-1, candidate_count=0;
          for(i = 0; i < vertices[3]->list_size; i++)
          {
            if(vertices[3]->list[i]->target == vertices[2])
            {
              f_visited[0] = vertices[3]->list[i]->next->next;
              candidate_count++;
            }
          }
          assert(candidate_count != 0);
          if(candidate_count > 1)
          {
            if(candidate_count == 2)
            {
              candidate = rd_conn_candi->decode(rm_FaceCandi);
            }
            else
            {
              candidate = rd_conn_candi->readRange(candidate_count);
            }
            if(candidate != candidate_count - 1)
            {
              for(i = 0; i < vertices[3]->list_size; i++)
              {
                if(vertices[3]->list[i]->target == vertices[2])
                {
                  if(candidate == 0)
                  {
                    f_visited[0] = vertices[3]->list[i]->next->next;
                    break;
                  }
                  else
                    candidate--;
                }
              }
            }
          }
          vertices[1] = f_visited[0]->next->next->next->target;
          vertices[0] = f_visited[0]->target;
        }
        else //lc_pos<96
        {
          //get which edge in the cache this is
          int i0,i1;
          getLocalIndicesFromEdgeNumber(lc_pos-72, i1, i0);
          vertices[2] =(SVvertex*) lc->get(i1);
          vertices[1] =(SVvertex*) lc->get(i0);
          // decode which of the faces incident to v2 and v1 is f0
          int candidate=-1, candidate_count=0;
          for(i = 0; i < vertices[2]->list_size; i++)
          {
            if(vertices[2]->list[i]->target == vertices[1])
            {
              f_visited[0] = vertices[2]->list[i]->next;
              candidate_count++;
            }
          }
          assert(candidate_count != 0);
          if(candidate_count > 1)
          {
            if(candidate_count == 2)
            {
              candidate = rd_conn_candi->decode(rm_FaceCandi);
            }
            else
            {
              candidate = rd_conn_candi->readRange(candidate_count);
            }
            if(candidate != candidate_count - 1)
            {
              for(i = 0; i < vertices[2]->list_size; i++)
              {
                if(vertices[2]->list[i]->target == vertices[1])
                {
                  if(candidate == 0)
                  {
                    f_visited[0] = vertices[2]->list[i]->next;
                    break;
                  }
                  else
                    candidate--;
                }
              }
            }
          }
          vertices[0] = f_visited[0]->target;
          vertices[3] = f_visited[0]->next->target;
        }
        assert(vertices[1] == f_visited[0]->next->next->next->target);
      }
      else //f0 does not have an edge in the edge cache
      {
        lc_pos = rd_conn_cache->decode(rmSCGPD_VertexCache[last_op]);
#ifdef PRINT_DEBUG_OUTPUT
        fprintf(stderr, "lc_pos_v is %i\n", lc_pos);
#endif
        if(lc_pos < 32) //f0 has a vertex in the vertex cache
        {
          if(lc_pos < 8)
          {
            vertices[1] =(SVvertex*) lc->get(lc_pos);
            // decode which of the faces incident to this vertex is f0
            if(vertices[1]->list_size > 1)
            {
              if(vertices[1]->list_size < SVH_MAGIC_NUMBER_F0_AROUND_V1_RM)
                f_visited[0] = vertices[1]->list[ rd_conn->decode(rm_VertexCandi[vertices[1]->list_size]) ];
              else
                f_visited[0] = vertices[1]->list[ rd_conn->readRange(vertices[1]->list_size) ];
            }
            else
              f_visited[0] = vertices[1]->list[0];
            vertices[0] = f_visited[0]->target;
            vertices[3] = f_visited[0]->next->target;
            vertices[2] = f_visited[0]->next->next->target;
          }
          else if(lc_pos < 16)
          {
            vertices[0] =(SVvertex*) lc->get(lc_pos - 8);
            // decode which of the faces incident to this vertex is f0
            if(vertices[0]->list_size > 1)
            {
              if(vertices[0]->list_size < SVH_MAGIC_NUMBER_F0_AROUND_V1_RM)
                f_visited[0] = vertices[0]->list[ rd_conn->decode(rm_VertexCandi[vertices[0]->list_size]) ]->next->next->next;
              else
                f_visited[0] = vertices[0]->list[ rd_conn->readRange(vertices[0]->list_size) ]->next->next->next;
            }
            else
              f_visited[0] = vertices[0]->list[0]->next->next->next;
            vertices[3] = f_visited[0]->next->target;
            vertices[2] = f_visited[0]->next->next->target;
            vertices[1] = f_visited[0]->next->next->next->target;
          }
          else if(lc_pos < 24)
          {
            vertices[3] =(SVvertex*) lc->get(lc_pos - 16);
            // decode which of the faces incident to this vertex is f0
            if(vertices[3]->list_size > 1)
            {
              if(vertices[3]->list_size < SVH_MAGIC_NUMBER_F0_AROUND_V1_RM)
                f_visited[0] = vertices[3]->list[ rd_conn->decode(rm_VertexCandi[vertices[3]->list_size]) ]->next->next;
              else
                f_visited[0] = vertices[3]->list[ rd_conn->readRange(vertices[3]->list_size) ]->next->next;
            }
            else
              f_visited[0] = vertices[0]->list[0]->next->next->next;
            vertices[0] = f_visited[0]->target;
            vertices[2] = f_visited[0]->next->next->target;
            vertices[1] = f_visited[0]->next->next->next->target;
          }
          else
          {
            vertices[2] =(SVvertex*) lc->get(lc_pos - 24);
            // decode which of the faces incident to this vertex is f0
            if(vertices[2]->list_size > 1)
            {
              if(vertices[2]->list_size < SVH_MAGIC_NUMBER_F0_AROUND_V1_RM)
                f_visited[0] = vertices[2]->list[ rd_conn->decode(rm_VertexCandi[vertices[2]->list_size]) ]->next;
              else
                f_visited[0] = vertices[2]->list[ rd_conn->readRange(vertices[2]->list_size) ]->next;
            }
            else
              f_visited[0] = vertices[2]->list[0]->next;
            vertices[0] = f_visited[0]->target;
            vertices[3] = f_visited[0]->next->target;
            vertices[1] = f_visited[0]->next->next->next->target;
          }
        }
        else
        {
          int dv_index;
          dv_index = rd_conn_index->readRange(dv->size());
          vertices[1] =(SVvertex*) dv->getElementWithRelativeIndex(dv_index);
          // decode which of the faces incident to this vertex is f0
          if(vertices[1]->list_size > 1)
          {
            if(vertices[1]->list_size < SVH_MAGIC_NUMBER_F0_AROUND_V1_RM)
              f_visited[0] = vertices[1]->list[ rd_conn->decode(rm_VertexCandi[vertices[1]->list_size]) ];
            else
              f_visited[0] = vertices[1]->list[ rd_conn->readRange(vertices[1]->list_size) ];
          }
          else
            f_visited[0] = vertices[1]->list[0];
          vertices[0] = f_visited[0]->target;
          vertices[3] = f_visited[0]->next->target;
          vertices[2] = f_visited[0]->next->next->target;
        }
        assert(vertices[1] == f_visited[0]->next->next->next->target);
      }
    }

    // decode which of the faces incident to v3 and v0 is f1
    int candidate=-1, candidate_count=0;
    for(i = 0; i < vertices[3]->list_size; i++)
    {
      if(vertices[3]->list[i]->target == vertices[0])
      {
        f_visited[1] = vertices[3]->list[i];
        candidate_count++;
      }
    }
    assert(candidate_count != 0);
    if(candidate_count > 1)
    {
      if(candidate_count == 2)
      {
        candidate = rd_conn_candi->decode(rm_FaceCandi);
      }
      else
      {
        candidate = rd_conn_candi->readRange(candidate_count);
      }
      if(candidate != candidate_count - 1)
      {
        for(i = 0; i < vertices[3]->list_size; i++)
        {
          if(vertices[3]->list[i]->target == vertices[0])
          {
            if(candidate == 0)
            {
              f_visited[1] = vertices[3]->list[i];
              break;
            }
            else
              candidate--;
          }
        }
      }
    }
    vertices[4] = f_visited[1]->next->target;
    vertices[7] = f_visited[1]->next->next->target;

    //now try to see if we are more than a SVH_STEP operation...
    //look if f2 (v4->v0->v1) or f4 (v2->v3->v7) exist

    //note: be careful! there can be a very weird phenomenon where the current hex shares f0 and f1 with the SAME hexahedron. this is the case in the shaft model at hexa 1966 (c_count==1965). See the file backtobackhexas.svg for details. If it is the case, v0->list_size is 3. In this case, we must explicitely say whether the faces join or not.
    for(i = 0; i < vertices[4]->list_size; i++)
    {
      if(( vertices[4]->list[i]->target == vertices[0]) &&(vertices[4]->list[i]->next->target == vertices[1]))
      {
        if(rd_conn->decode(rm_ConfirmFaceMerge) ==1)
        {
          f_visited[2] = vertices[4]->list[i];
          vertices[5] = f_visited[2]->next->next->target;
        }
        break;
      }
    }

    //we cannot have f3 without f2, else this would be a bridge
    if(f_visited[2])
    {
      //look if f3 (v5->v1->v2) exists
      for(i = 0; i < vertices[5]->list_size; i++)
      {
        if(( vertices[5]->list[i]->target == vertices[1]) &&(vertices[5]->list[i]->next->target == vertices[2]))
        {
          if(rd_conn->decode(rm_ConfirmFaceMerge) ==1)
          {
            f_visited[3] = vertices[5]->list[i];
            vertices[6] = f_visited[3]->next->next->target;
          }
          break;
        }
      }
      //we cannot have f4 without f3, else this would be an incorrectly oriented GAP
      if(f_visited[3])
      {
        //look if f4 exists (no need for confirmation since we now have 4 verts)
        for(i = 0; i < vertices[6]->list_size; i++)
        {
          if(( vertices[6]->list[i]->target == vertices[2]) &&(vertices[6]->list[i]->next->target == vertices[3]) &&(vertices[6]->list[i]->next->next->target == vertices[7]))
          {
            f_visited[4] = vertices[6]->list[i];
            break;
          }
        }
        //we cannot have f5 without f4, else incorrectly oriented PIT
        if(f_visited[4])
        {
          //look if f5 (v7->v4->v5->v6) exists (same note)
          for(i = 0; i < vertices[7]->list_size; i++)
          {
            if(( vertices[7]->list[i]->target == vertices[4]) &&(vertices[7]->list[i]->next->target == vertices[5]) &&(vertices[7]->list[i]->next->next->target == vertices[6]))
            {
              f_visited[5] = vertices[7]->list[i];
              break;
            }
          }

          if(f_visited[5])
          {
            op = SVH_DEN;
          }
          else
          {
            op = SVH_PIT;
          }
        }
        else
        {
          op = SVH_GAP;
        }
      }
      else
      {
        op = SVH_CORNER;
      }
    }
    else
    {
      op = SVH_STEP;
    }

#ifdef PRINT_DEBUG_OUTPUT
    if(op==SVH_DEN)
      fprintf(stderr, "actual operation is SVH_DEN\n");
    else if(op==SVH_PIT)
      fprintf(stderr, "actual operation is SVH_PIT\n");
    else if(op==SVH_GAP)
      fprintf(stderr, "actual operation is SVH_GAP\n");
    else if(op==SVH_CORNER)
      fprintf(stderr, "actual operation is SVH_CORNER\n");
    else if(op==SVH_STEP)
      fprintf(stderr, "actual operation is SVH_STEP\n");
#endif

    // decode the unknown vertices (there can be unknown vertices is op==SVH_STEP or SVH_CORNER)
    //we first need to read whether there is a vertex or edge join
    if(op==SVH_STEP)
    {
      if(( rd_conn_op->decode(rmStep_HasJoin) == 1)) //if there are joins
      {
#ifdef SVH_DUMB_JOIN_CODING
        unsigned int joinedVertices = 1 + rd_conn_op->decode(rmStep_JoinedVertices);

#ifdef PRINT_DEBUG_OUTPUT
        fprintf(stderr, "joinedString is %i\n", joinedVertices);
#endif

        if(joinedVertices&1)
        {
          int dv_index;
          dv_index = rd_conn_index->readRange(dv->size());
          vertices[5] =(SVvertex*) dv->getElementWithRelativeIndex(dv_index);
#ifdef PRINT_DEBUG_OUTPUT
          fprintf(stderr, "vertex 5 is joined to index %i\n", dv_index);
#endif
        }
        else
        {
          createAddAndDecompressV5(vertices);
        }
        if(joinedVertices&2)
        {
          int dv_index;
          dv_index = rd_conn_index->readRange(dv->size());
          vertices[6] =(SVvertex*) dv->getElementWithRelativeIndex(dv_index);
#ifdef PRINT_DEBUG_OUTPUT
          fprintf(stderr, "vertex 6 is joined to index %i\n", dv_index);
#endif
        }
        else
        {
          createAddAndDecompressV6(vertices);
        }
#else
        fprintf(stderr, "NOT YET IMPLEMENTED: intelligent join for SVH_STEP\n");
        exit(1);
#endif
      }
      else
      {
        createAddAndDecompressV5(vertices);
        createAddAndDecompressV6(vertices);
      }
    }
    else if(op==SVH_CORNER)
    {
      if(( rd_conn_op->decode(rmCorner_HasJoin) == 1)) //if there are joins
      {
#ifdef SVH_DUMB_JOIN_CODING
        int dv_index;
        dv_index = rd_conn_index->readRange(dv->size());
        vertices[6] =(SVvertex*) dv->getElementWithRelativeIndex(dv_index);
#ifdef PRINT_DEBUG_OUTPUT
        fprintf(stderr, "vertex 6 is joined to index %i\n", dv_index);
#endif
#else
        fprintf(stderr, "NOT YET IMPLEMENTED: intelligent join for SVH_CORNER\n");
        exit(1);
#endif
      }
      else
      {
        createAddAndDecompressV6(vertices);
      }
    }

    //now decompress cell data
    
    if(op==SVH_DEN)
    {
#ifdef SVH_COMPRESS_CELL_DATA
      //now decompress the cell data.
      if(c_pnum_f)
      {
#ifdef SVH_CELL_DATA_SECONDORDER_PREDICTION
        decompressCellDataDen(f_visited[0]->c_prop_f, f_visited[1]->c_prop_f, f_visited[2]->c_prop_f, f_visited[3]->c_prop_f, f_visited[4]->c_prop_f, f_visited[5]->c_prop_f, c_prop_f, f_visited[0]->c_prop_f+c_pnum_f, f_visited[1]->c_prop_f+c_pnum_f, f_visited[2]->c_prop_f+c_pnum_f, f_visited[3]->c_prop_f+c_pnum_f, f_visited[4]->c_prop_f+c_pnum_f, f_visited[5]->c_prop_f+c_pnum_f, tmpCellDataResidual);
#else
        decompressCellDataDen(f_visited[0]->c_prop_f, f_visited[1]->c_prop_f, f_visited[2]->c_prop_f, f_visited[3]->c_prop_f, f_visited[4]->c_prop_f, f_visited[5]->c_prop_f, c_prop_f);
#endif
      }
#endif  
    }
    else if(op==SVH_PIT)
    {
#ifdef SVH_COMPRESS_CELL_DATA
      //now decompress the cell data.
      if(c_pnum_f)
      {
#ifdef SVH_CELL_DATA_SECONDORDER_PREDICTION
        decompressCellDataPit(f_visited[0]->c_prop_f, f_visited[1]->c_prop_f, f_visited[2]->c_prop_f, f_visited[3]->c_prop_f, f_visited[4]->c_prop_f, c_prop_f, f_visited[0]->c_prop_f+c_pnum_f, f_visited[1]->c_prop_f+c_pnum_f, f_visited[2]->c_prop_f+c_pnum_f, f_visited[3]->c_prop_f+c_pnum_f, f_visited[4]->c_prop_f+c_pnum_f, tmpCellDataResidual);
#else
        decompressCellDataPit(f_visited[0]->c_prop_f, f_visited[1]->c_prop_f, f_visited[2]->c_prop_f, f_visited[3]->c_prop_f, f_visited[4]->c_prop_f, c_prop_f);
#endif
      }
#endif  
    }
    else if(op==SVH_GAP)
    {
#ifdef SVH_COMPRESS_CELL_DATA
      //now decompress the cell data.
      if(c_pnum_f)
      {
#ifdef SVH_CELL_DATA_SECONDORDER_PREDICTION
        decompressCellDataGap(f_visited[0]->c_prop_f, f_visited[1]->c_prop_f, f_visited[2]->c_prop_f, f_visited[3]->c_prop_f, c_prop_f, f_visited[0]->c_prop_f+c_pnum_f, f_visited[1]->c_prop_f+c_pnum_f, f_visited[2]->c_prop_f+c_pnum_f, f_visited[3]->c_prop_f+c_pnum_f, tmpCellDataResidual);
#else
        decompressCellDataGap(f_visited[0]->c_prop_f, f_visited[1]->c_prop_f, f_visited[2]->c_prop_f, f_visited[3]->c_prop_f, c_prop_f);
#endif
      }
#endif
    }
    else if(op==SVH_CORNER)
    {
#ifdef SVH_COMPRESS_CELL_DATA
      //now decompress the cell data.
      if(c_pnum_f)
      {
#ifdef SVH_CELL_DATA_SECONDORDER_PREDICTION
        decompressCellDataCorner(f_visited[0]->c_prop_f, f_visited[1]->c_prop_f, f_visited[2]->c_prop_f, c_prop_f, f_visited[0]->c_prop_f+c_pnum_f, f_visited[1]->c_prop_f+c_pnum_f, f_visited[2]->c_prop_f+c_pnum_f, tmpCellDataResidual);
#else
        decompressCellDataCorner(f_visited[0]->c_prop_f, f_visited[1]->c_prop_f, f_visited[2]->c_prop_f, c_prop_f);
#endif
      }
#endif
    }
    else if(op==SVH_STEP)
    {
#ifdef SVH_COMPRESS_CELL_DATA
      //now decompress the cell data.
      if(c_pnum_f)
      {
#ifdef SVH_CELL_DATA_SECONDORDER_PREDICTION
        decompressCellDataStep(f_visited[0]->c_prop_f, f_visited[1]->c_prop_f, c_prop_f, f_visited[0]->c_prop_f+c_pnum_f, f_visited[1]->c_prop_f+c_pnum_f, tmpCellDataResidual);
#else
        decompressCellDataStep(f_visited[0]->c_prop_f, f_visited[1]->c_prop_f, c_prop_f);
#endif
      }
#endif
    }

  }
  else if(op == SVH_BRIDGE) // SVH_BRIDGE, SVH_TUNNEL
  {
#ifdef PRINT_DEBUG_OUTPUT
    fprintf(stderr, "SVH_BRIDGE\n");
#endif
#ifdef SVH_FACECACHE_DEPENDS_ON_FINALIZATION
    lc_pos = rd_conn_cache->decode(rmBT_FaceCacheX[last_fin][last_op]);
#else
    lc_pos = rd_conn_cache->decode(rmBT_FaceCache[last_op]);
#endif
#ifdef PRINT_DEBUG_OUTPUT
    fprintf(stderr, "lc_pos_f is %i\n", lc_pos);
#endif
    if(lc_pos < 12) //f0 has an edge in the face cache
    {
      if(lc_pos < 6)
        f_visited[0] =(( SVedge*) lc_f->get(lc_pos));
      else if(lc_pos < 12)
        f_visited[0] =(( SVedge*) lc_f->get(lc_pos-6))->next->next->next;

      vertices[0] = f_visited[0]->target;
      vertices[3] = f_visited[0]->next->target;
      vertices[2] = f_visited[0]->next->next->target;
      vertices[1] = f_visited[0]->next->next->next->target;
    }
    else //f0 is is not in the face cache
    {
      lc_pos = rd_conn_cache->decode(rmBT_EdgeCache[last_op]);
#ifdef PRINT_DEBUG_OUTPUT
      fprintf(stderr, "lc_pos_e is %i\n", lc_pos);
#endif
      if(lc_pos < 48) //f0 has an edge in the edge cache
      {
        if(lc_pos < 24)
        {
          //get which edge in the cache this is
          int i0,i1;
          getLocalIndicesFromEdgeNumber(lc_pos, i1, i0);
          vertices[1] =(SVvertex*) lc->get(i1);
          vertices[0] =(SVvertex*) lc->get(i0);
          // encode which of the faces incident to v1 and v0 is f0
          int candidate=-1, candidate_count=0;
          for(i = 0; i < vertices[1]->list_size; i++)
          {
            if(vertices[1]->list[i]->target == vertices[0])
            {
              f_visited[0] = vertices[1]->list[i];
              candidate_count++;
            }
          }
          assert(candidate_count != 0);
          if(candidate_count > 1)
          {
            if(candidate_count == 2)
            {
              candidate = rd_conn_candi->decode(rm_FaceCandi);
            }
            else
            {
              candidate = rd_conn_candi->readRange(candidate_count);
            }
            if(candidate != candidate_count - 1)
            {
              for(i = 0; i < vertices[1]->list_size; i++)
              {
                if(vertices[1]->list[i]->target == vertices[0])
                {
                  if(candidate == 0)
                  {
                    f_visited[0] = vertices[1]->list[i];
                    break;
                  }
                  else
                    candidate--;
                }
              }
            }
          }
          vertices[3] = f_visited[0]->next->target;
          vertices[2] = f_visited[0]->next->next->target;
        }
        else if(lc_pos < 48)
        {
          //get which edge in the cache this is
          int i0,i1;
          getLocalIndicesFromEdgeNumber(lc_pos-24, i1, i0);
          vertices[0] =(SVvertex*) lc->get(i1);
          vertices[3] =(SVvertex*) lc->get(i0);
          // decode which of the faces incident to v0 and v3 is f0
          int candidate=-1, candidate_count=0;
          for(i = 0; i < vertices[0]->list_size; i++)
          {
            if(vertices[0]->list[i]->target == vertices[3])
            {
              f_visited[0] = vertices[0]->list[i]->next->next->next;
              candidate_count++;
            }
          }
          assert(candidate_count != 0);
          if(candidate_count > 1)
          {
            if(candidate_count == 2)
            {
              candidate = rd_conn_candi->decode(rm_FaceCandi);
            }
            else
            {
              candidate = rd_conn_candi->readRange(candidate_count);
            }
            if(candidate != candidate_count - 1)
            {
              for(i = 0; i < vertices[0]->list_size; i++)
              {
                if(vertices[0]->list[i]->target == vertices[3])
                {
                  if(candidate == 0)
                  {
                    f_visited[0] = vertices[0]->list[i]->next->next->next;
                    break;
                  }
                  else
                    candidate--;
                }
              }
            }
          }
          vertices[1] = f_visited[0]->next->next->next->target;
          vertices[2] = f_visited[0]->next->next->target;
        }
        assert(vertices[1] == f_visited[0]->next->next->next->target);
      }
      else //f0 does not have an edge in the edge cache
      {
        lc_pos = rd_conn_cache->decode(rmBT_VertexCache[last_op]);
#ifdef PRINT_DEBUG_OUTPUT
        fprintf(stderr, "lc_pos_v is %i\n", lc_pos);
#endif
        if(lc_pos < 16) //f0 has a vertex in the vertex cache
        {
          if(lc_pos < 8)
          {
            vertices[1] =(SVvertex*) lc->get(lc_pos);
            // decode which of the faces incident to this vertex is f0
            if(vertices[1]->list_size > 1)
            {
              if(vertices[1]->list_size < SVH_MAGIC_NUMBER_F0_AROUND_V1_RM)
                f_visited[0] = vertices[1]->list[ rd_conn->decode(rm_VertexCandi[vertices[1]->list_size]) ];
              else
                f_visited[0] = vertices[1]->list[ rd_conn->readRange(vertices[1]->list_size) ];
            }
            else
              f_visited[0] = vertices[1]->list[0];
            vertices[0] = f_visited[0]->target;
            vertices[3] = f_visited[0]->next->target;
            vertices[2] = f_visited[0]->next->next->target;
          }
          else if(lc_pos < 16)
          {
            vertices[0] =(SVvertex*) lc->get(lc_pos - 8);
            // decode which of the faces incident to this vertex is f0
            if(vertices[0]->list_size > 1)
            {
              if(vertices[0]->list_size < SVH_MAGIC_NUMBER_F0_AROUND_V1_RM)
                f_visited[0] = vertices[0]->list[ rd_conn->decode(rm_VertexCandi[vertices[0]->list_size]) ]->next->next->next;
              else
                f_visited[0] = vertices[0]->list[ rd_conn->readRange(vertices[0]->list_size) ]->next->next->next;
            }
            else
              f_visited[0] = vertices[0]->list[0]->next->next->next;
            vertices[3] = f_visited[0]->next->target;
            vertices[2] = f_visited[0]->next->next->target;
            vertices[1] = f_visited[0]->next->next->next->target;
          }
        }
        else
        {
          int dv_index;
          dv_index = rd_conn_index->readRange(dv->size());
          vertices[1] =(SVvertex*) dv->getElementWithRelativeIndex(dv_index);
          // decode which of the faces incident to this vertex is f0
          if(vertices[1]->list_size > 1)
          {
            if(vertices[1]->list_size < SVH_MAGIC_NUMBER_F0_AROUND_V1_RM)
              f_visited[0] = vertices[1]->list[ rd_conn->decode(rm_VertexCandi[vertices[1]->list_size]) ];
            else
              f_visited[0] = vertices[1]->list[ rd_conn->readRange(vertices[1]->list_size) ];
          }
          else
            f_visited[0] = vertices[1]->list[0];
          vertices[0] = f_visited[0]->target;
          vertices[3] = f_visited[0]->next->target;
          vertices[2] = f_visited[0]->next->next->target;
        }
        assert(vertices[1] == f_visited[0]->next->next->next->target);
      }
    }

    // decode which of the faces incident to v3 and v0 is f1
    int candidate=-1, candidate_count=0;
    for(i = 0; i < vertices[3]->list_size; i++)
    {
      if(vertices[3]->list[i]->target == vertices[0])
      {
        f_visited[1] = vertices[3]->list[i];
        candidate_count++;
      }
    }
    assert(candidate_count != 0);
    if(candidate_count > 1)
    {
      if(candidate_count == 2)
      {
        candidate = rd_conn_candi->decode(rm_FaceCandi);
      }
      else
      {
        candidate = rd_conn_candi->readRange(candidate_count);
      }
      if(candidate != candidate_count - 1)
      {
        for(i = 0; i < vertices[3]->list_size; i++)
        {
          if(vertices[3]->list[i]->target == vertices[0])
          {
            if(candidate == 0)
            {
              f_visited[1] = vertices[3]->list[i];
              break;
            }
            else
              candidate--;
          }
        }
      }
    }
    vertices[4] = f_visited[1]->next->target;
    vertices[7] = f_visited[1]->next->next->target;

    // decode which of the faces incident to v1 and v2 is f3
    candidate=-1; candidate_count=0;
    for(i = 0; i < vertices[1]->list_size; i++)
    {
      if(vertices[1]->list[i]->target == vertices[2])
      {
        f_visited[3] = vertices[1]->list[i]->next->next->next;
        candidate_count++;
      }
    }
    assert(candidate_count != 0);
    if(candidate_count > 1)
    {
      if(candidate_count == 2)
      {
        candidate = rd_conn_candi->decode(rm_FaceCandi);
      }
      else
      {
        candidate = rd_conn_candi->readRange(candidate_count);
      }
      if(candidate != candidate_count - 1)
      {
        for(i = 0; i < vertices[1]->list_size; i++)
        {
          if(vertices[1]->list[i]->target == vertices[2])
          {
            if(candidate == 0)
            {
              f_visited[3] = vertices[1]->list[i]->next->next->next;
              break;
            }
            else
              candidate--;
          }
        }
      }
    }
    vertices[5] = f_visited[3]->next->next->next->target;
    vertices[6] = f_visited[3]->next->next->target;


    //now try to see if we are more than a SVH_BRIDGE operation...
    //look if f5 exists
    for(i = 0; i < vertices[7]->list_size; i++)
    {
      if(( vertices[7]->list[i]->target == vertices[4]) &&(vertices[7]->list[i]->next->target == vertices[5]))
      {
        f_visited[5] = vertices[7]->list[i];
        break;
      }
    }

    //if we have added f5, then the operation is SVH_TUNNEL
    if(f_visited[5])
    {
      op=SVH_TUNNEL;
#ifdef SVH_COMPRESS_CELL_DATA
      //now decompress the cell data.
      if(c_pnum_f)
      {
#ifdef SVH_CELL_DATA_SECONDORDER_PREDICTION
        decompressCellDataTunnel(f_visited[0]->c_prop_f, f_visited[1]->c_prop_f, f_visited[3]->c_prop_f, f_visited[5]->c_prop_f, c_prop_f, f_visited[0]->c_prop_f+c_pnum_f, f_visited[1]->c_prop_f+c_pnum_f, f_visited[3]->c_prop_f+c_pnum_f, f_visited[5]->c_prop_f+c_pnum_f, tmpCellDataResidual);
#else
        decompressCellDataTunnel(f_visited[0]->c_prop_f, f_visited[1]->c_prop_f, f_visited[3]->c_prop_f, f_visited[5]->c_prop_f, c_prop_f);
#endif
      }
#endif  
    }
#ifdef SVH_COMPRESS_CELL_DATA
    else
    {
      //now decompress the cell data.
      if(c_pnum_f)
      {
#ifdef SVH_CELL_DATA_SECONDORDER_PREDICTION
        decompressCellDataBridge(f_visited[0]->c_prop_f, f_visited[1]->c_prop_f, f_visited[3]->c_prop_f, c_prop_f, f_visited[0]->c_prop_f+c_pnum_f, f_visited[1]->c_prop_f+c_pnum_f, f_visited[3]->c_prop_f+c_pnum_f, tmpCellDataResidual);
#else
        decompressCellDataBridge(f_visited[0]->c_prop_f, f_visited[1]->c_prop_f, f_visited[3]->c_prop_f, c_prop_f);
#endif
      }
    }
#endif  

#ifdef PRINT_DEBUG_OUTPUT
    if(op==SVH_TUNNEL)
      fprintf(stderr, "actual operation is SVH_TUNNEL\n");
    else
      fprintf(stderr, "actual operation is SVH_BRIDGE\n");
#endif
  }
  else //OTHER NOT IMPLEMENTED YET OPERATIONS
  {
    fprintf(stderr,  "OPERATION %i NOT IMPLEMENTED YET.\n", op);
    exit(1);
  }



  // statistic
#ifdef PRINT_DEBUG_OUTPUT
  fprintf(stderr,"op %d %d\n",op,c_count);
#endif

//  fprintf(stderr,"%d %d %d %d\n",vertices[0]->index, vertices[1]->index, vertices[2]->index, vertices[3]->index);

  // update data structures
  if(f_visited[0])
  {
    removeEdgeFromVertex(f_visited[0]->next->next->next, vertices[2]);
    removeEdgeFromVertex(f_visited[0]->next->next, vertices[3]);
    removeEdgeFromVertex(f_visited[0]->next, vertices[0]);
    removeEdgeFromVertex(f_visited[0], vertices[1]);

    deallocEdge(f_visited[0]->next->next->next);
    deallocEdge(f_visited[0]->next->next);
    deallocEdge(f_visited[0]->next);
    deallocEdge(f_visited[0]);

    f_visited[0] = 0;
  }
  else
  {
    SVedge* e0 = allocEdge();
    SVedge* e1 = allocEdge();
    SVedge* e2 = allocEdge();
    SVedge* e3 = allocEdge();

    e0->next = e1;
    e1->next = e2;
    e2->next = e3;
    e3->next = e0;

#ifdef SVH_COMPRESS_CELL_DATA
    if(c_pnum_f)
    {
#ifdef SVH_CELL_DATA_SECONDORDER_PREDICTION
      memcpy(e0->c_prop_f, c_prop_f, c_pnum_f*sizeof(float));
      memcpy(e1->c_prop_f, c_prop_f, c_pnum_f*sizeof(float));
      memcpy(e2->c_prop_f, c_prop_f, c_pnum_f*sizeof(float));
      memcpy(e3->c_prop_f, c_prop_f, c_pnum_f*sizeof(float));
      memcpy(e0->c_prop_f+c_pnum_f, tmpCellDataResidual, c_pnum_f*sizeof(float));
      memcpy(e1->c_prop_f+c_pnum_f, tmpCellDataResidual, c_pnum_f*sizeof(float));
      memcpy(e2->c_prop_f+c_pnum_f, tmpCellDataResidual, c_pnum_f*sizeof(float));
      memcpy(e3->c_prop_f+c_pnum_f, tmpCellDataResidual, c_pnum_f*sizeof(float));
#else
      memcpy(e0->c_prop_f, c_prop_f, c_pnum_f*sizeof(float));
      memcpy(e1->c_prop_f, c_prop_f, c_pnum_f*sizeof(float));
      memcpy(e2->c_prop_f, c_prop_f, c_pnum_f*sizeof(float));
      memcpy(e3->c_prop_f, c_prop_f, c_pnum_f*sizeof(float));
#endif
    }
#endif

    e0->target = vertices[1];
    e1->target = vertices[2];
    e2->target = vertices[3];
    e3->target = vertices[0];

    addEdgeToVertex(e0, vertices[0]);
    addEdgeToVertex(e1, vertices[1]);
    addEdgeToVertex(e2, vertices[2]);
    addEdgeToVertex(e3, vertices[3]);

    f_visited[0] = e0;
  }

  if(f_visited[1])
  {
    removeEdgeFromVertex(f_visited[1]->next->next->next, vertices[7]);
    removeEdgeFromVertex(f_visited[1]->next->next, vertices[4]);
    removeEdgeFromVertex(f_visited[1]->next, vertices[0]);
    removeEdgeFromVertex(f_visited[1], vertices[3]);

    deallocEdge(f_visited[1]->next->next->next);
    deallocEdge(f_visited[1]->next->next);
    deallocEdge(f_visited[1]->next);
    deallocEdge(f_visited[1]);

    f_visited[1] = 0;
  }
  else
  {
    SVedge* e0 = allocEdge();
    SVedge* e1 = allocEdge();
    SVedge* e2 = allocEdge();
    SVedge* e3 = allocEdge();

    e0->next = e1;
    e1->next = e2;
    e2->next = e3;
    e3->next = e0;

#ifdef SVH_COMPRESS_CELL_DATA
    if(c_pnum_f)
    {
#ifdef SVH_CELL_DATA_SECONDORDER_PREDICTION
      memcpy(e0->c_prop_f, c_prop_f, c_pnum_f*sizeof(float));
      memcpy(e1->c_prop_f, c_prop_f, c_pnum_f*sizeof(float));
      memcpy(e2->c_prop_f, c_prop_f, c_pnum_f*sizeof(float));
      memcpy(e3->c_prop_f, c_prop_f, c_pnum_f*sizeof(float));
      memcpy(e0->c_prop_f+c_pnum_f, tmpCellDataResidual, c_pnum_f*sizeof(float));
      memcpy(e1->c_prop_f+c_pnum_f, tmpCellDataResidual, c_pnum_f*sizeof(float));
      memcpy(e2->c_prop_f+c_pnum_f, tmpCellDataResidual, c_pnum_f*sizeof(float));
      memcpy(e3->c_prop_f+c_pnum_f, tmpCellDataResidual, c_pnum_f*sizeof(float));
#else
      memcpy(e0->c_prop_f, c_prop_f, c_pnum_f*sizeof(float));
      memcpy(e1->c_prop_f, c_prop_f, c_pnum_f*sizeof(float));
      memcpy(e2->c_prop_f, c_prop_f, c_pnum_f*sizeof(float));
      memcpy(e3->c_prop_f, c_prop_f, c_pnum_f*sizeof(float));
#endif
    }
#endif
    e0->target = vertices[3];
    e1->target = vertices[7];
    e2->target = vertices[4];
    e3->target = vertices[0];

    addEdgeToVertex(e0, vertices[0]);
    addEdgeToVertex(e1, vertices[3]);
    addEdgeToVertex(e2, vertices[7]);
    addEdgeToVertex(e3, vertices[4]);

    f_visited[1] = e0;
  }

  if(f_visited[2])
  {
    removeEdgeFromVertex(f_visited[2]->next->next->next, vertices[5]);
    removeEdgeFromVertex(f_visited[2]->next->next, vertices[1]);
    removeEdgeFromVertex(f_visited[2]->next, vertices[0]);
    removeEdgeFromVertex(f_visited[2], vertices[4]);

    deallocEdge(f_visited[2]->next->next->next);
    deallocEdge(f_visited[2]->next->next);
    deallocEdge(f_visited[2]->next);
    deallocEdge(f_visited[2]);

    f_visited[2] = 0;
  }
  else
  {
    SVedge* e0 = allocEdge();
    SVedge* e1 = allocEdge();
    SVedge* e2 = allocEdge();
    SVedge* e3 = allocEdge();

    e0->next = e1;
    e1->next = e2;
    e2->next = e3;
    e3->next = e0;

#ifdef SVH_COMPRESS_CELL_DATA
    if(c_pnum_f)
    {
#ifdef SVH_CELL_DATA_SECONDORDER_PREDICTION
      memcpy(e0->c_prop_f, c_prop_f, c_pnum_f*sizeof(float));
      memcpy(e1->c_prop_f, c_prop_f, c_pnum_f*sizeof(float));
      memcpy(e2->c_prop_f, c_prop_f, c_pnum_f*sizeof(float));
      memcpy(e3->c_prop_f, c_prop_f, c_pnum_f*sizeof(float));
      memcpy(e0->c_prop_f+c_pnum_f, tmpCellDataResidual, c_pnum_f*sizeof(float));
      memcpy(e1->c_prop_f+c_pnum_f, tmpCellDataResidual, c_pnum_f*sizeof(float));
      memcpy(e2->c_prop_f+c_pnum_f, tmpCellDataResidual, c_pnum_f*sizeof(float));
      memcpy(e3->c_prop_f+c_pnum_f, tmpCellDataResidual, c_pnum_f*sizeof(float));
#else
      memcpy(e0->c_prop_f, c_prop_f, c_pnum_f*sizeof(float));
      memcpy(e1->c_prop_f, c_prop_f, c_pnum_f*sizeof(float));
      memcpy(e2->c_prop_f, c_prop_f, c_pnum_f*sizeof(float));
      memcpy(e3->c_prop_f, c_prop_f, c_pnum_f*sizeof(float));
#endif
    }
#endif
    e0->target = vertices[4];
    e1->target = vertices[5];
    e2->target = vertices[1];
    e3->target = vertices[0];

    addEdgeToVertex(e0, vertices[0]);
    addEdgeToVertex(e1, vertices[4]);
    addEdgeToVertex(e2, vertices[5]);
    addEdgeToVertex(e3, vertices[1]);

    f_visited[2] = e0;
  }

  if(f_visited[3])
  {
    removeEdgeFromVertex(f_visited[3]->next->next->next, vertices[6]);
    removeEdgeFromVertex(f_visited[3]->next->next, vertices[2]);
    removeEdgeFromVertex(f_visited[3]->next, vertices[1]);
    removeEdgeFromVertex(f_visited[3], vertices[5]);

    deallocEdge(f_visited[3]->next->next->next);
    deallocEdge(f_visited[3]->next->next);
    deallocEdge(f_visited[3]->next);
    deallocEdge(f_visited[3]);

    f_visited[3] = 0;
  }
  else
  {
    SVedge* e0 = allocEdge();
    SVedge* e1 = allocEdge();
    SVedge* e2 = allocEdge();
    SVedge* e3 = allocEdge();

    e0->next = e1;
    e1->next = e2;
    e2->next = e3;
    e3->next = e0;

#ifdef SVH_COMPRESS_CELL_DATA
    if(c_pnum_f)
    {
#ifdef SVH_CELL_DATA_SECONDORDER_PREDICTION
      memcpy(e0->c_prop_f, c_prop_f, c_pnum_f*sizeof(float));
      memcpy(e1->c_prop_f, c_prop_f, c_pnum_f*sizeof(float));
      memcpy(e2->c_prop_f, c_prop_f, c_pnum_f*sizeof(float));
      memcpy(e3->c_prop_f, c_prop_f, c_pnum_f*sizeof(float));
      memcpy(e0->c_prop_f+c_pnum_f, tmpCellDataResidual, c_pnum_f*sizeof(float));
      memcpy(e1->c_prop_f+c_pnum_f, tmpCellDataResidual, c_pnum_f*sizeof(float));
      memcpy(e2->c_prop_f+c_pnum_f, tmpCellDataResidual, c_pnum_f*sizeof(float));
      memcpy(e3->c_prop_f+c_pnum_f, tmpCellDataResidual, c_pnum_f*sizeof(float));
#else
      memcpy(e0->c_prop_f, c_prop_f, c_pnum_f*sizeof(float));
      memcpy(e1->c_prop_f, c_prop_f, c_pnum_f*sizeof(float));
      memcpy(e2->c_prop_f, c_prop_f, c_pnum_f*sizeof(float));
      memcpy(e3->c_prop_f, c_prop_f, c_pnum_f*sizeof(float));
#endif
    }
#endif
    e0->target = vertices[5];
    e1->target = vertices[6];
    e2->target = vertices[2];
    e3->target = vertices[1];

    addEdgeToVertex(e0, vertices[1]);
    addEdgeToVertex(e1, vertices[5]);
    addEdgeToVertex(e2, vertices[6]);
    addEdgeToVertex(e3, vertices[2]);

    f_visited[3] = e0;
  }

  if(f_visited[4])
  {
    removeEdgeFromVertex(f_visited[4]->next->next->next, vertices[7]);
    removeEdgeFromVertex(f_visited[4]->next->next, vertices[3]);
    removeEdgeFromVertex(f_visited[4]->next, vertices[2]);
    removeEdgeFromVertex(f_visited[4], vertices[6]);

    deallocEdge(f_visited[4]->next->next->next);
    deallocEdge(f_visited[4]->next->next);
    deallocEdge(f_visited[4]->next);
    deallocEdge(f_visited[4]);

    f_visited[4] = 0;
  }
  else
  {
    SVedge* e0 = allocEdge();
    SVedge* e1 = allocEdge();
    SVedge* e2 = allocEdge();
    SVedge* e3 = allocEdge();

    e0->next = e1;
    e1->next = e2;
    e2->next = e3;
    e3->next = e0;

#ifdef SVH_COMPRESS_CELL_DATA
    if(c_pnum_f)
    {
#ifdef SVH_CELL_DATA_SECONDORDER_PREDICTION
      memcpy(e0->c_prop_f, c_prop_f, c_pnum_f*sizeof(float));
      memcpy(e1->c_prop_f, c_prop_f, c_pnum_f*sizeof(float));
      memcpy(e2->c_prop_f, c_prop_f, c_pnum_f*sizeof(float));
      memcpy(e3->c_prop_f, c_prop_f, c_pnum_f*sizeof(float));
      memcpy(e0->c_prop_f+c_pnum_f, tmpCellDataResidual, c_pnum_f*sizeof(float));
      memcpy(e1->c_prop_f+c_pnum_f, tmpCellDataResidual, c_pnum_f*sizeof(float));
      memcpy(e2->c_prop_f+c_pnum_f, tmpCellDataResidual, c_pnum_f*sizeof(float));
      memcpy(e3->c_prop_f+c_pnum_f, tmpCellDataResidual, c_pnum_f*sizeof(float));
#else
      memcpy(e0->c_prop_f, c_prop_f, c_pnum_f*sizeof(float));
      memcpy(e1->c_prop_f, c_prop_f, c_pnum_f*sizeof(float));
      memcpy(e2->c_prop_f, c_prop_f, c_pnum_f*sizeof(float));
      memcpy(e3->c_prop_f, c_prop_f, c_pnum_f*sizeof(float));
#endif
    }
#endif
    e0->target = vertices[6];
    e1->target = vertices[7];
    e2->target = vertices[3];
    e3->target = vertices[2];

    addEdgeToVertex(e0, vertices[2]);
    addEdgeToVertex(e1, vertices[6]);
    addEdgeToVertex(e2, vertices[7]);
    addEdgeToVertex(e3, vertices[3]);

    f_visited[4] = e0;
  }

  if(f_visited[5])
  {
    removeEdgeFromVertex(f_visited[5]->next->next->next, vertices[6]);
    removeEdgeFromVertex(f_visited[5]->next->next, vertices[5]);
    removeEdgeFromVertex(f_visited[5]->next, vertices[4]);
    removeEdgeFromVertex(f_visited[5], vertices[7]);

    deallocEdge(f_visited[5]->next->next->next);
    deallocEdge(f_visited[5]->next->next);
    deallocEdge(f_visited[5]->next);
    deallocEdge(f_visited[5]);

    f_visited[5] = 0;
  }
  else
  {
    SVedge* e0 = allocEdge();
    SVedge* e1 = allocEdge();
    SVedge* e2 = allocEdge();
    SVedge* e3 = allocEdge();

    e0->next = e1;
    e1->next = e2;
    e2->next = e3;
    e3->next = e0;

#ifdef SVH_COMPRESS_CELL_DATA
    if(c_pnum_f)
    {
#ifdef SVH_CELL_DATA_SECONDORDER_PREDICTION
      memcpy(e0->c_prop_f, c_prop_f, c_pnum_f*sizeof(float));
      memcpy(e1->c_prop_f, c_prop_f, c_pnum_f*sizeof(float));
      memcpy(e2->c_prop_f, c_prop_f, c_pnum_f*sizeof(float));
      memcpy(e3->c_prop_f, c_prop_f, c_pnum_f*sizeof(float));
      memcpy(e0->c_prop_f+c_pnum_f, tmpCellDataResidual, c_pnum_f*sizeof(float));
      memcpy(e1->c_prop_f+c_pnum_f, tmpCellDataResidual, c_pnum_f*sizeof(float));
      memcpy(e2->c_prop_f+c_pnum_f, tmpCellDataResidual, c_pnum_f*sizeof(float));
      memcpy(e3->c_prop_f+c_pnum_f, tmpCellDataResidual, c_pnum_f*sizeof(float));
#else
      memcpy(e0->c_prop_f, c_prop_f, c_pnum_f*sizeof(float));
      memcpy(e1->c_prop_f, c_prop_f, c_pnum_f*sizeof(float));
      memcpy(e2->c_prop_f, c_prop_f, c_pnum_f*sizeof(float));
      memcpy(e3->c_prop_f, c_prop_f, c_pnum_f*sizeof(float));
#endif
    }
#endif
    e0->target = vertices[7];
    e1->target = vertices[6];
    e2->target = vertices[5];
    e3->target = vertices[4];

    addEdgeToVertex(e0, vertices[4]);
    addEdgeToVertex(e1, vertices[7]);
    addEdgeToVertex(e2, vertices[6]);
    addEdgeToVertex(e3, vertices[5]);

    f_visited[5] = e0;
  }

  // update last_op
  last_op = op;

  // increment vertex use_counts and copy hexahedron indices into API
  for(i = 0; i < 8; i++)
  {
    vertices[i]->use_count++;
    c_idx[i] = vertices[i]->index;
  }

  // read finalization info and finalize vertices
#if defined(SVH_FACECACHE_DEPENDS_ON_FINALIZATION) || defined(SVH_RMOP_DEPENDS_ON_FINALIZATION)
  last_fin = 0;
#endif
  for(i = 0; i < 8; i++)
  {

#ifdef SVH_IMPLICIT_FINALIZATION
    if(vertices[i]->list_size == 0)
      c_final[i] = 1;
    else
      c_final[i] = 0;
#else
    int degree_one = vertices[i]->list_size;
    if(degree_one >= SVH_MAX_DEGREE_ONE)
    {
      degree_one = SVH_MAX_DEGREE_ONE-1;
    }
    int use_count = vertices[i]->use_count;
    if(use_count >= SVH_MAX_USE_COUNT)
    {
      use_count = SVH_MAX_USE_COUNT-1;
    }
    c_final[i] =(rd_conn_final->decode(rmFinalized[degree_one][use_count]) == 1);
#endif

    if(c_final[i])
    {
      SVvertex* vertex = vertices[i];
#if defined(SVH_FACECACHE_DEPENDS_ON_FINALIZATION) || defined(SVH_RMOP_DEPENDS_ON_FINALIZATION)
      last_fin = last_fin |(1<<i);
#endif
      dv->removeElement(vertex);
      for(int j = 0; j < vertex->list_size; j++)
      {
        SVedge* edge = vertex->list[j];
        removeEdgeFromVertex(edge->next->next->next, edge->next->next->target);
        removeEdgeFromVertex(edge->next->next, edge->next->target);
        removeEdgeFromVertex(edge->next, edge->target);
        //removeEdgeFromVertex(edge, vertex);
        deallocEdge(edge->next->next->next);
        deallocEdge(edge->next->next);
        deallocEdge(edge->next);
        deallocEdge(edge);
      }
      deallocVertex(vertex);
      finalized_vertices[have_finalized] = i;
      have_finalized++;
    }
  }

  // clean cache flag of current elements in cache
  for(i = 0; i < 8; i++)
  {
    SVvertex* vertex =(SVvertex*) lc->get(i);
    if(vertex) vertex->in_cache = -1;
  }

  // update cache elements
  lc->put(vertices[0], vertices[1], vertices[2], vertices[3], vertices[4], vertices[5], vertices[6], vertices[7]);
  lc_f->put(f_visited[0], f_visited[1], f_visited[2], f_visited[3], f_visited[4], f_visited[5]);

  // set cache flag of current elements in cache
  for(i = 0; i < 8; i++)
  {
    if(vertices[i]) vertices[i]->in_cache = i;
  }

#ifdef PRINT_DEBUG_OUTPUT
  fprintf(stderr, "END (con/de)compress_hexahedron()\n");
#endif


  have_hexahedron = 1;
  return 1;
}

void SVreader_svh::close(bool close_file)
{

#ifdef SVH_CELL_DATA_SECONDORDER_PREDICTION
  if(tmpCellDataResidual) 
  {
    delete[] tmpCellDataResidual;
    tmpCellDataResidual = 0;
  }
#endif

  // close of SVreader_svh interface
  finishDecoder();
  if(v_pq)
  {
#ifdef PRINT_CONTROL_OUTPUT
    fprintf(stderr,"predict_last %d, predict_shoot %d, predict_parallelogram %d, predict_lorenzo %d\n", predict_last, predict_shoot, predict_parallelogram, predict_lorenzo);
#endif
    v_ic[0]->FinishDecompressor();
    v_ic[1]->FinishDecompressor();
    v_ic[2]->FinishDecompressor();
    delete v_ic[0];
    delete v_ic[1];
    delete v_ic[2];
    delete v_pq;
  }
  else
  {
    v_fc[0]->FinishDecompressor();
    v_fc[1]->FinishDecompressor();
    v_fc[2]->FinishDecompressor();
    delete v_fc[0];
    delete v_fc[1];
    delete v_fc[2];
  }
#ifdef SVH_COMPRESS_CELL_DATA
  if(c_pnum_f)
  {
    if (c_quantizer)
    {
      for(int i=0; i<c_pnum_f; ++i)
      {
	c_ic[i]->FinishCompressor();
	delete c_ic[i];
      }
      free(c_ic);
      delete c_quantizer;
    }
    else
    {
      for(int i=0; i<c_pnum_f; ++i)
      {
	c_fc[i]->FinishCompressor();
	delete c_fc[i];
      }
      free(c_fc);
    }
  }
#endif // SVH_COMPRESS_CELL_DATA
  finishModels();

  destroyVertexBuffer();
  destroyEdgeBuffer();

  if(dv->size()) fprintf(stderr,"WARNING: there are %d unfinalized vertices\n",dv->size());
  delete dv;
  delete lc;
  delete lc_f;
  nbits = -1;

#ifdef OUTPUT_FOR_LS_WEIGHTS
  fclose(lsoutfCorner);
  fclose(lsoutfBridge);
  fclose(lsoutfGap);
  fclose(lsoutfPit);
#endif

#ifdef OUTPUT_RESIDUALS
  {
  FILE * outResiduals = fopen("residuals.txt", "w");
  double entropy=0;
  for(std::map<I32, int>::const_iterator mit = residualHistogram.begin(); mit != residualHistogram.end(); ++mit)
  {
    fprintf(outResiduals, "%i\t%i\n", mit->first, mit->second);
    entropy -= log2((double)mit->second/nResiduals) * (double)mit->second/nResiduals;
  }
  fclose(outResiduals);
  fprintf(stderr, "Entropy of cell data residuals: %f (warning: hurts performance)\n", entropy);
  }
  #ifdef SVH_CELL_DATA_SECONDORDER_PREDICTION
  {
  FILE * outResiduals = fopen("residuals2.txt", "w");
  double entropy=0;
  for(std::map<I32, int>::const_iterator mit = residual2Histogram.begin(); mit != residual2Histogram.end(); ++mit)
  {
    fprintf(outResiduals, "%i\t%i\n", mit->first, mit->second);
    entropy -= log2((double)mit->second/nResiduals) * (double)mit->second/nResiduals;
  }
  fclose(outResiduals);
  fprintf(stderr, "Entropy of 2nd order cell data residuals: %f (warning: hurts performance)\n", entropy);
  }
  #endif
#endif

#ifdef OUTPUT_QUANTIZED_CELL_DATA
  {
  FILE * outQuantized = fopen("quantized.txt", "w");
  double entropy=0;
  for(std::map<I32, int>::const_iterator mit = quantizedHistogram.begin(); mit != quantizedHistogram.end(); ++mit)
  {
    fprintf(outQuantized, "%i\t%i\n", mit->first, mit->second);
    entropy -= log2((double)mit->second/nQuantized) * (double)mit->second/nQuantized;
  }
  fclose(outQuantized);
  fprintf(stderr, "Entropy of quantized cell data: %f (warning: hurts performance)\n", entropy);
  }
#endif
  if (new_vertices) delete [] new_vertices;
  new_vertices = 0;
  if (finalized_vertices) delete [] finalized_vertices;
  finalized_vertices = 0;

  // close of SVreader interface
  v_count = -1;
  c_count = -1;
  if(v_prop_f) delete [] v_prop_f;
  v_prop_f = 0;
  if(c_idx) delete [] c_idx;
  c_idx = 0;
  if(c_final) delete [] c_final;
  c_final = 0;
  if (close_file && file && file != stdin) fclose(file);
}

void SVreader_svh::read_header()
{
  int i;

  // read nverts
  if(rd_conn->readRange(2))
  {
    nverts = rd_conn->readInt();
#ifdef PRINT_CONTROL_OUTPUT
    fprintf(stderr,"nverts: %d\n",nverts);
#endif
  }

  // read ncells
  if(rd_conn->readRange(2))
  {
    ncells = rd_conn->readInt();
#ifdef PRINT_CONTROL_OUTPUT
    fprintf(stderr,"ncells: %d\n",ncells);
#endif
  }

  // read which properties are there
  int prop = rd_geom->readByte();
#ifdef PRINT_CONTROL_OUTPUT
  printf("property mapping is %d\n",prop);
#endif

  if(prop & 1)
  {
    v_pnum_f = rd_geom->readByte();
#ifdef PRINT_CONTROL_OUTPUT
    printf("have %d float vertex attributes\n",v_pnum_f);
#endif
    if(rd_geom->readRange(2))
    {
      if(v_pmin_f) delete [] v_pmin_f;
      v_pmin_f = new float[v_pnum_f];
      if(v_pmax_f) delete [] v_pmax_f;
      v_pmax_f = new float[v_pnum_f];
      for(i=0; i < v_pnum_f; i++)
      {
        v_pmin_f[i] = rd_geom->readFloat();
        v_pmax_f[i] = rd_geom->readFloat();
      }
#ifdef PRINT_CONTROL_OUTPUT
      fprintf(stderr,"bbox is %f %f %f, %f %f %f\n",v_pmin_f[0], v_pmin_f[1], v_pmin_f[2], v_pmax_f[0], v_pmax_f[1], v_pmax_f[2]);
#endif
    }
  }
  else
  {
#ifdef PRINT_CONTROL_OUTPUT
    printf("have no float vertex attributes\n");
#endif
  }

  if(prop & 2)
  {
#ifdef PRINT_CONTROL_OUTPUT
    printf("have integer vertex attributes\n");
#endif
  }
  else
  {
#ifdef PRINT_CONTROL_OUTPUT
    printf("have no integer vertex attributes\n");
#endif
  }

  if(prop & 4)
  {
    c_pnum_f = rd_geom->readByte();
    if(rd_geom->readRange(2))
    {
      if(c_pmin_f) delete [] c_pmin_f;
      c_pmin_f = new float[c_pnum_f];
      if(c_pmax_f) delete [] c_pmax_f;
      c_pmax_f = new float[c_pnum_f];
      for(i=0; i < c_pnum_f; i++)
      {
        c_pmin_f[i] = rd_geom->readFloat();
        c_pmax_f[i] = rd_geom->readFloat();
      }
    }
  }
  else
  {
#ifdef PRINT_CONTROL_OUTPUT
    printf("have no float cell attributes\n");
#endif
  }

  if(prop & 8)
  {
#ifdef PRINT_CONTROL_OUTPUT
    printf("have integer cell attributes\n");
#endif
  }
  else
  {
#ifdef PRINT_CONTROL_OUTPUT
    printf("have no integer cell attributes\n");
#endif
  }

  if(rd_conn->readRange(2)) // if we have comments
  {
    /* yet to be implemented */
  }
  else
  {
#ifdef PRINT_CONTROL_OUTPUT
    printf("have no comments\n");
#endif
  }

  // setup float quantizers and integer compressors

  int high_bits = rd_conn->readRange(17);

#ifdef PRINT_CONTROL_OUTPUT
  printf("high_bits for v_prop_f %d\n", high_bits);
#endif

  if(v_pmin_f && v_pmax_f && nbits)
  {
#ifdef PRINT_CONTROL_OUTPUT
    fprintf(stderr,"dequantizing v_prop_f with %d bits\n",nbits);
#endif
    v_pq = new PositionQuantizerNew();
    v_pq->SetMinMax(v_pmin_f, v_pmax_f);
    v_pq->SetPrecision(nbits);
    v_pq->SetupQuantizer();

#ifdef PRINT_CONTROL_OUTPUT
    printf("aiQuantRange %d %d %d across\n", v_pq->m_aiQuantRange[0], v_pq->m_aiQuantRange[1], v_pq->m_aiQuantRange[2]);
#endif

    v_ic[0] = new RangeIntegerCompressor();
    v_ic[0]->SetRange(v_pq->m_aiQuantRange[0]);
    v_ic[0]->SetupDecompressor(rd_geom, SVH_NUMBER_OF_PREDICTIONS, high_bits);

    v_ic[1] = new RangeIntegerCompressor();
    v_ic[1]->SetRange(v_pq->m_aiQuantRange[1]);
    v_ic[1]->SetupDecompressor(rd_geom, SVH_NUMBER_OF_PREDICTIONS, high_bits);

    v_ic[2] = new RangeIntegerCompressor();
    v_ic[2]->SetRange(v_pq->m_aiQuantRange[2]);
    v_ic[2]->SetupDecompressor(rd_geom, SVH_NUMBER_OF_PREDICTIONS, high_bits);
  }
  else
  {
#ifdef PRINT_CONTROL_OUTPUT
    if(nbits)
    {
      fprintf(stderr,"no bounding box. dequantizing v_prop_f adaptive to %d bits\n",nbits);
    }
    else
    {
      fprintf(stderr,"no quantization. lossless compression of floating-point numbers.\n");
    }
#endif
    v_pq = 0;

    v_fc[0] = new RangeFloatCompressor();
    v_fc[0]->SetPrecision(nbits);
    v_fc[0]->SetupDecompressor(rd_geom, SVH_NUMBER_OF_PREDICTIONS);

    v_fc[1] = new RangeFloatCompressor();
    v_fc[1]->SetPrecision(nbits);
    v_fc[1]->SetupDecompressor(rd_geom, SVH_NUMBER_OF_PREDICTIONS);

    v_fc[2] = new RangeFloatCompressor();
    v_fc[2]->SetPrecision(nbits);
    v_fc[2]->SetupDecompressor(rd_geom, SVH_NUMBER_OF_PREDICTIONS);
  }
  
#ifdef SVH_COMPRESS_CELL_DATA
  if(c_pnum_f)
  {
    cellDataBits = rd_conn->readRange(25);
    if(cellDataBits)
    {
      int cellDataHighBits = rd_conn->readRange(17);
#ifdef PRINT_CONTROL_OUTPUT
      printf("high_bits for c_prop_f %d\n", cellDataHighBits);
#endif
      if(!(c_pmin_f && c_pmax_f))
      {
        fprintf(stderr,"FATAL ERROR: quantizeCellData set to TRUE but no BBox.\n");
        exit(1);
      }
#ifdef PRINT_CONTROL_OUTPUT
      fprintf(stderr,"dequantizing c_prop_f with %d bits\n", cellDataBits);
#endif
      c_quantizer = new PositionQuantizerNew();
      c_quantizer->SetMinMax(c_pmin_f, c_pmax_f);
      c_quantizer->SetPrecision(cellDataBits);
      c_quantizer->SetupQuantizer();
      c_ic = (RangeIntegerCompressor**)malloc(sizeof(RangeIntegerCompressor*)*c_pnum_f);
      for(int i=0; i<c_pnum_f; ++i)
      {
        c_ic[i] = new RangeIntegerCompressor();
#ifdef SVH_CELL_DATA_SECONDORDER_PREDICTION
        c_ic[i]->SetRange(2*c_quantizer->m_aiQuantRange[i]);
        c_ic[i]->SetupDecompressor(rd_celldata, SVH_NUMBER_OF_OPERATIONS, cellDataHighBits + 1);
#else
        c_ic[i]->SetRange(c_quantizer->m_aiQuantRange[i]);
        c_ic[i]->SetupDecompressor(rd_celldata, SVH_NUMBER_OF_OPERATIONS, cellDataHighBits);
#endif
        
      }
    }
    else
    {
#ifdef PRINT_CONTROL_OUTPUT
      fprintf(stderr,"lossless c_prop_f floating-point compression\n");
#endif
      c_quantizer= NULL;
      c_fc = (RangeFloatCompressor**)malloc(sizeof(RangeFloatCompressor*)*c_pnum_f);
      for(int i=0; i<c_pnum_f; ++i)
      {
        c_fc[i] = new RangeFloatCompressor();
        c_fc[i]->SetPrecision(0);
        c_fc[i]->SetupDecompressor(rd_celldata, SVH_NUMBER_OF_OPERATIONS);
      }
    }
  }
#endif
}

SVevent SVreader_svh::read_element()
{
  if(have_hexahedron == 0)
  {
    assert(have_new == 0);
    next_new = 0;
    have_finalized = 0;
    if(decompress_hexahedron() == 0)
    {
      if(nverts != -1 && v_count != nverts)
      {
        fprintf(stderr,"WARNING: wrong vertex count: v_count (%d) != nverts (%d)\n", v_count, nverts);
      }
      nverts = v_count;
      if(ncells != -1 && c_count != ncells)
      {
        fprintf(stderr,"WARNING: wrong cell count: c_count (%d) != ncells (%d)\n", c_count, ncells);
      }
      ncells = c_count;
      have_hexahedron = -1;
      return SV_EOF;
    }
  }
  if(have_new)
  {
    v_idx = c_idx[new_vertices[next_new]];
    if(v_pq)
    {
      v_pq->DeQuantize(( I32*)(vertices[new_vertices[next_new]]->v_prop_f), v_prop_f);
    }
    else
    {
      VecCopy3fv(v_prop_f, vertices[new_vertices[next_new]]->v_prop_f);
    }
    have_new--; next_new++;
    v_count++;
    return SV_VERTEX;
  }
  else if(have_hexahedron > 0)
  {
    have_hexahedron = 0;
    c_count++;
#ifdef SVH_COMPRESS_CELL_DATA
    if(c_quantizer)
    {
      c_quantizer->DeQuantize(( I32*)(c_prop_f), c_prop_f);
    }
#endif
    return SV_HEXAHEDRON;
  }
  else if(have_hexahedron < 0)
  {
    return SV_EOF;
  }
  return SV_ERROR;
}

SVevent SVreader_svh::read_event()
{
  if(have_hexahedron == 0 && have_finalized == 0)
  {
    assert(have_new == 0);
    next_new = 0;
    next_finalized = 0;
    if(decompress_hexahedron() == 0)
    {
      if(nverts != -1 && v_count != nverts)
      {
        fprintf(stderr,"WARNING: wrong vertex count: v_count (%d) != nverts (%d)\n", v_count, nverts);
      }
      nverts = v_count;
      if(ncells != -1 && c_count != ncells)
      {
        fprintf(stderr,"WARNING: wrong cell count: c_count (%d) != ncells (%d)\n", c_count, ncells);
      }
      ncells = c_count;
      have_hexahedron = -1;
      return SV_EOF;
    }
  }
  if(have_new)
  {
//    v_idx = c_idx[new_vertices[next_new]];
    if(v_pq)
    {
      v_pq->DeQuantize(( I32*)(vertices[new_vertices[next_new]]->v_prop_f), v_prop_f);
    }
    else
    {
      VecCopy3fv(v_prop_f, vertices[new_vertices[next_new]]->v_prop_f);
    }
    have_new--; next_new++;
    v_count++;
    return SV_VERTEX;
  }
  else if(have_hexahedron > 0)
  {
    have_hexahedron = 0;
    c_count++;
    return SV_HEXAHEDRON;
  }
  else if(have_finalized)
  {
    final_idx = c_idx[finalized_vertices[next_finalized]];
    have_finalized--; next_finalized++;
    return SV_FINALIZED;
  }
  else if(have_hexahedron < 0)
  {
    return SV_EOF;
  }
  return SV_ERROR;
}

SVreader_svh::SVreader_svh()
{
  // init of SVreader_svh
  have_new = 0; next_new = 0; new_vertices = 0;
  have_hexahedron = 0;
  have_finalized = 0; next_finalized = 0; finalized_vertices = 0;

  nbits = -1;
}

SVreader_svh::~SVreader_svh()
{
  // clean-up for SVreader_svh
}
