/*
===============================================================================

  FILE:  SVreader_tet.cpp
  
  CONTENTS:
  
    see corresponding header file
  
  PROGRAMMERS:
  
    martin isenburg@cs.unc.edu
  
  COPYRIGHT:
  
    copyright (C) 2005  martin isenburg@cs.unc.edu
    
    This software is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
  
  CHANGE HISTORY:
  
    see corresponding header file
  
===============================================================================
*/
#include "svreader_tet.h"

#ifdef _WIN32
#include <fcntl.h>
#include <io.h>
#endif

#include <stdlib.h>
#include <string.h>
#include <ctype.h>

#include "vecnfv.h"
#include "vecniv.h"

bool SVreader_tet::open(FILE* file, bool compute_boundingbox)
{
  if (file == 0)
  {
    return false;
  }

#ifdef _WIN32
  if (file == stdin)
  {
    if(_setmode( _fileno( stdin ), _O_TEXT ) == -1 )
    {
      fprintf(stderr, "ERROR: cannot set stdin to text (translated) mode\n");
    }
  }
#endif

  this->file = file;

  // only TET meshes are supported
  type = SV_TET;
  post_order = false;

  minus = 0;
  skipped_lines = 0;
  line = (char*)malloc(sizeof(char)*256);
  if (fgets(line, sizeof(char) * 256, file) == 0)
  {
    free(line);
    line = 0;
    return false;
  }

  // look for header information
  char dummy[256];

  while (line && line[0] != 'v')
  {
    if (strstr(line, "nNodes"))
    {
      sscanf(&(line[1]), "%s %d", dummy, &nverts);
    }
    else if (strstr(line, "nTetrahedra"))
    {
      sscanf(&(line[1]), "%s %d", dummy, &ncells);
    }
    else if (strstr(line, "vertices"))
    {
      sscanf(&(line[1]), "%d %s", &nverts, dummy);
    }
    else if (strstr(line, "tetrahedra"))
    {
      sscanf(&(line[1]), "%d %s", &ncells, dummy);
    }
    else if (line[0] == 'p')
    {
      fprintf(stderr,"WARNING: input seems to be a point cloud\n");
    }
    else if (line[0] == 'f')
    {
      fprintf(stderr,"WARNING: input seems to be a surface mesh\n");
    }
    else if (line[0] == '#')
    {
// silently skipping comments
    }
    else if (line[0] == 'm' && line[1] == 'i' && line[2] == 'n' && line[3] == 'u')
    {
      sscanf(&(line[6]), "%d", &minus);
    }
    else
    {
      int j = (int)strlen(line)-1;
      if (j > 0)
      {
        for (int i = 0; i < j; i++)
        {
          if (!isprint(line[i]))
          {
            fprintf(stderr,"FATAL ERROR: input file seems to be binary\n");
            break;
          }
        }
        line[(10<j?10:j)] = '\0';
        if (skipped_lines < 5)
        {
          fprintf(stderr,"WARNING: skipping line '%s...'.\n",line);
        }
        else if (skipped_lines == 5)
        {
          fprintf(stderr,"WARNING: skipping more than 5 lines. being quiet.\n");
        }
        skipped_lines++;
      }
    }
    if (fgets(line, sizeof(char) * 256, file) == 0)
    {
      free(line);
      line = 0;
      return false;
    }
  }

  if (line[0] == 'v')
  {
    // how many vertex properties (all float)
    v_pnum_f = sscanf(&(line[1]), "%s %s %s %s %s %s %s", dummy, dummy, dummy, dummy, dummy, dummy, dummy);
    v_prop_f = new float[v_pnum_f];
    // computing bounding box
    if (compute_boundingbox)
    {
      // create bounding box
      v_pmin_f = new float[v_pnum_f];
      v_pmax_f = new float[v_pnum_f];
      // get properties of first vertex
      if (v_pnum_f == 3)
      {
        sscanf(&(line[1]), "%f %f %f", &(v_prop_f[0]), &(v_prop_f[1]), &(v_prop_f[2]));
      }
      else if (v_pnum_f == 4)
      {
        sscanf(&(line[1]), "%f %f %f %f", &(v_prop_f[0]), &(v_prop_f[1]), &(v_prop_f[2]), &(v_prop_f[3]));
      }
      else if (v_pnum_f == 5)
      {
        sscanf(&(line[1]), "%f %f %f %f %f", &(v_prop_f[0]), &(v_prop_f[1]), &(v_prop_f[2]), &(v_prop_f[3]), &(v_prop_f[4]));
      }
      // init bounding box
      VecCopyNfv(v_pmin_f, v_prop_f, v_pnum_f);
      VecCopyNfv(v_pmax_f, v_prop_f, v_pnum_f);
      fgets(line, sizeof(char) * 256, file);
      while (line[0] != 't')
      {
        if (line[0] == 'v')
        {
          if (v_pnum_f == 3)
          {
            sscanf(&(line[1]), "%f %f %f", &(v_prop_f[0]), &(v_prop_f[1]), &(v_prop_f[2]));
          }
          else if (v_pnum_f == 4)
          {
            sscanf(&(line[1]), "%f %f %f %f", &(v_prop_f[0]), &(v_prop_f[1]), &(v_prop_f[2]), &(v_prop_f[3]));
          }
          else if (v_pnum_f == 5)
          {
            sscanf(&(line[1]), "%f %f %f %f %f", &(v_prop_f[0]), &(v_prop_f[1]), &(v_prop_f[2]), &(v_prop_f[3]), &(v_prop_f[4]));
          }
          VecUpdateMinMaxNfv(v_pmin_f, v_pmax_f, v_prop_f, v_pnum_f);
        }
        fgets(line, sizeof(char) * 256, file);
      }
      // move file pointer back to beginning
      fseek(file, 0, SEEK_SET);
      // scan forward to the first vertex
      fgets(line, sizeof(char) * 256, file);
      while (line[0] != 'v')
      {
        fgets(line, sizeof(char) * 256, file);
      }
    }
  }

  v_count = 0;
  c_count = 0;

  // only TET meshes are supported
  c_nverts = 4;
  c_idx = new I32[4];
  c_final = new bool[4];

  c_final[0] = false; 
  c_final[1] = false;
  c_final[2] = false;
  c_final[3] = false;

  return true;
}

void SVreader_tet::close()
{
  // close of SMreader interface
  v_count = -1;
  c_count = -1;

  if (v_pnum_f) {v_pnum_f = 0; delete [] v_prop_f; v_prop_f = 0;};
  delete [] c_idx; c_idx = 0;
  delete [] c_final; c_final = 0;

  // close of SVreader_tet
  if (skipped_lines) fprintf(stderr,"WARNING: skipped %d lines.\n",skipped_lines);
  file = 0;
  skipped_lines = 0;
  if (line)
  {
    free(line);
    line = 0;
  }
}

SVevent SVreader_tet::read_element()
{
  while (line)
  {
    if ((line[0] == 'v') && (line[1] == ' '))
    {
      v_idx = v_count;
      if (v_pnum_f == 3)
      {
        sscanf(&(line[1]), "%f %f %f", &(v_prop_f[0]), &(v_prop_f[1]), &(v_prop_f[2]));
      }
      else if (v_pnum_f == 4)
      {
        sscanf(&(line[1]), "%f %f %f %f", &(v_prop_f[0]), &(v_prop_f[1]), &(v_prop_f[2]), &(v_prop_f[3]));
      }
      else if (v_pnum_f == 5)
      {
        sscanf(&(line[1]), "%f %f %f %f %f", &(v_prop_f[0]), &(v_prop_f[1]), &(v_prop_f[2]), &(v_prop_f[3]), &(v_prop_f[4]));
      }
      v_count++;
      if (fgets(line, sizeof(char) * 256, file) == 0)
      {
        free(line);
        line = 0;
      }
      return SV_VERTEX;
    }
    else if ((line[0] == 't') && (line[1] == ' '))
    {
      sscanf(&(line[1]), "%d %d %d %d", &(c_idx[0]), &(c_idx[1]), &(c_idx[2]), &(c_idx[3]));
      if (minus)
      {
        for (int i = 0; i < 4; i++) c_idx[i] -= minus;
      }
      c_count++;
      if (fgets(line, sizeof(char) * 256, file) == 0)
      {
        free(line);
        line = 0;
      }
      return SV_TETRAHEDRON;
    }
    else if (line[0] == 'm' && line[1] == 'i' && line[2] == 'n' && line[3] == 'u')
    {
      sscanf(&(line[6]), "%d", &minus);
      if (fgets(line, sizeof(char) * 256, file) == 0)
      {
        free(line);
        line = 0;
      }
    }
    else if (line[0] == '#')
    {
      // comments in the body are silently ignored
      if (fgets(line, sizeof(char) * 256, file) == 0)
      {
        free(line);
        line = 0;
      }
    }
    else
    {
      int j = (int)strlen(line)-1;
      if (j)
      {
        for (int i = 0; i < j; i++)
        {
          if (!isprint(line[i]))
          {
            fprintf(stderr,"WARNING: input file seems to be binary\n");
            break;
          }
        }
        line[(10<j?10:j)] = '\0';
        if (skipped_lines < 5)
        {
          fprintf(stderr,"WARNING: skipping line '%s...'.\n",line);
        }
        else if (skipped_lines == 5)
        {
          fprintf(stderr,"WARNING: skipping more than 5 lines. being quiet.\n");
        }
        skipped_lines++;
      }
      if (fgets(line, sizeof(char) * 256, file) == 0)
      {
        free(line);
        line = 0;
      }
    }
  }

  if (nverts == -1)
  {
    nverts = v_count;
  }
  else
  {
    if (v_count != nverts)
    {
      fprintf(stderr,"WARNING: wrong vertex count: v_count (%d) != nverts (%d)\n", v_count, nverts);
    }
  }
  if (ncells == -1)
  {
    ncells = c_count;
  }
  else
  {
    if (c_count != ncells)
    {
      fprintf(stderr,"WARNING: wrong cell count: c_count (%d) != ncells (%d)\n", c_count, ncells);
    }
  }
  return SV_EOF;
}

SVevent SVreader_tet::read_event()
{
  return read_element();
}

SVreader_tet::SVreader_tet()
{
  // init of SVreader_tet
  file = 0;
  line = 0;
}

SVreader_tet::~SVreader_tet()
{
  // clean-up for SVreader_tet
}
