/*
===============================================================================

  FILE:  svreadopener.cpp
  
  CONTENTS:
  
    see corresponding header file
  
  PROGRAMMERS:
  
    martin isenburg@cs.unc.edu
  
  COPYRIGHT:
  
    copyright (C) 2007  martin isenburg@cs.unc.edu
    
    This software is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
  
  CHANGE HISTORY:
  
    see corresponding header file
  
===============================================================================
*/
#include "svreadopener.h"

// streaming mesh readers
#include "svreader_sva.h"
#include "svreader_svb.h"
#include "svreader_svc.h"
#include "svreader_svh.h"
#include "svreader_svg.h"
#include "svreader_silo.h"

#ifdef BUILD_HZIP
#include "svreader_hzip.h"
#endif
#ifdef BUILD_GEO
#include "svreader_geo.h"
#endif

// non-streaming mesh readers
#include "svreader_tet.h"
#include "svreader_off.h"
#include "svreader_vtk.h"
#include "svreader_jrs.h"
#include "svreader_mesh3d.h"
#include "svreader_sms.h"

#include <stdlib.h>
#include <string.h>

#ifdef _WIN32
#include <fcntl.h>
#include <io.h>
extern "C" FILE* fopenGzipped(const char* filename, const char* mode);
#endif

void SVreadOpener::set_file_name(const char* file_name)
{
  if (this->file_name) free(this->file_name);
  this->file_name = strdup(file_name);
}

void SVreadOpener::set_file_format(const char* file_format)
{
  if (this->file_format) free(this->file_format);
  this->file_format = strdup(file_format);
}

SVreader* SVreadOpener::open()
{
  // if the file format is not given explicitely we try to guess it
  if (file_format == 0)
  {
    if (file_name == 0)
    {
      fprintf(stderr, "ERROR: neither a file name nor a file format were specified\n");
      return 0;
    }
    else if (strstr(file_name, ".sva") || strstr(file_name, ".obj") || strstr(file_name, ".hex"))
      file_format = strdup("sva");
    else if (strstr(file_name, ".svb"))
      file_format = strdup("svb");
    else if (strstr(file_name, ".svc"))
      file_format = strdup("svc");
    else if (strstr(file_name, ".svh"))
      file_format = strdup("svh");
    else if (strstr(file_name, ".svg"))
      file_format = strdup("svg");
#ifdef BUILD_HZIP
    else if (strstr(file_name, ".hzip"))
      file_format = strdup("hzip");
#endif
    else if (strstr(file_name, ".geo"))
      file_format = strdup("geo");
    else if (strstr(file_name, ".silo"))
      file_format = strdup("silo");
    else if (strstr(file_name, ".tet"))
      file_format = strdup("tet");
    else if (strstr(file_name, ".hex"))
      file_format = strdup("hex");
    else if (strstr(file_name, ".off"))
      file_format = strdup("off");
    else if (strstr(file_name, ".node"))
      file_format = strdup("jrs");
    else if (strstr(file_name, ".vtk"))
      file_format = strdup("vtk");
    else if (strstr(file_name, ".mesh3d"))
      file_format = strdup("mesh3d");
    else if (strstr(file_name, ".sms"))
      file_format = strdup("sms");
    else
    {
      fprintf(stderr, "ERROR: could not guess file format from file name '%s'\n", file_name);
      return 0;
    }
  }

  // depending on the file format we open the file in binary or text mode
  bool is_binary = true;
  if (strcmp(file_format, "sva") == 0 || strcmp(file_format, "tet") == 0 || strcmp(file_format, "off") == 0 || strcmp(file_format, "vtk") == 0 || strcmp(file_format, "jrs") == 0 || strcmp(file_format, "mesh3d") == 0 || strcmp(file_format, "sms") == 0)
    is_binary = false;

  FILE* file = 0;

  // if we have no file name we read from stdin, otherwise we open the file
  if (file_name == 0)
  {
    file = stdin;
#ifdef _WIN32
    if (is_binary)
      if(_setmode( _fileno( stdin ), _O_BINARY ) == -1 )
      {
        fprintf(stderr, "ERROR: cannot set stdin to binary (untranslated) mode\n");
      }
    else
      if(_setmode( _fileno( stdin ), _O_TEXT ) == -1 )
      {
        fprintf(stderr, "ERROR: cannot set stdin to text (translated) mode\n");
      }
#endif
  }
  else
  {
    if (strstr(file_name, ".gz"))
    {
#ifdef _WIN32
      file = fopenGzipped(file_name, (is_binary ? "rb" : "r"));
#else
      fprintf(stderr,"ERROR: cannot open gzipped file '%s'\n",file_name);
      return 0;
#endif
    }
    else
    {
      file = fopen(file_name, (is_binary ? "rb" : "r"));
    }
    if (file == 0)
    {
      fprintf(stderr,"ERROR: cannot open file '%s'\n",file_name);
      return 0;
    }
  }

  // now we open the respective reader

  if (strcmp(file_format, "sva") == 0)
  {
    SVreader_sva* svreader_sva = new SVreader_sva();
    if (svreader_sva->open(file)) return svreader_sva;
    delete svreader_sva;
  }
  else if (strcmp(file_format, "svb") == 0)
  {
    SVreader_svb* svreader_svb = new SVreader_svb();
    if (svreader_svb->open(file)) return svreader_svb;
    delete svreader_svb;
  }
  else if (strcmp(file_format, "svc") == 0)
  {
    SVreader_svc* svreader_svc = new SVreader_svc();
    if (svreader_svc->open(file)) return svreader_svc;
    delete svreader_svc;
  }
  else if (strcmp(file_format, "svh") == 0)
  {
    SVreader_svh* svreader_svh = new SVreader_svh();
    if (svreader_svh->open(file)) return svreader_svh;
    delete svreader_svh;
  }
  else if (strcmp(file_format, "svg") == 0)
  {
    SVreader_svg* svreader_svg = new SVreader_svg();
    if (svreader_svg->open(file)) return svreader_svg;
    delete svreader_svg;
  }
#ifdef BUILD_HZIP
  else if (strcmp(file_format, "hzip") == 0)
  {
    SVreader_hzip* svreader_hzip = new SVreader_hzip();
    if (svreader_hzip->open(file)) return svreader_hzip;
    delete svreader_hzip;
  }
#endif
#ifdef BUILD_GEO
  else if (strcmp(file_format, "geo") == 0)
  {
    SVreader_geo* svreader_geo = new SVreader_geo();
    if (svreader_geo->open(file)) return svreader_geo;
    delete svreader_geo;
  }
#endif
  else if (strcmp(file_format, "silo") == 0)
  {
    SVreader_silo* svreader_silo = new SVreader_silo();
    if (svreader_silo->open(file)) return svreader_silo;
    delete svreader_silo;
  }
  else if (strcmp(file_format, "tet") == 0)
  {
    SVreader_tet* svreader_tet = new SVreader_tet();
    if (svreader_tet->open(file)) return svreader_tet;
    delete svreader_tet;
  }
  else if (strcmp(file_format, "off") == 0)
  {
    SVreader_off* svreader_off = new SVreader_off();
    if (svreader_off->open(file)) return svreader_off;
    delete svreader_off;
  }
  else if (strcmp(file_format, "vtk") == 0)
  {
    SVreader_vtk* svreader_vtk = new SVreader_vtk();
    if (svreader_vtk->open(file)) return svreader_vtk;
    delete svreader_vtk;
  }
  else if (strcmp(file_format, "jrs") == 0)
  {
    FILE* file_other;
    char* file_name_other = strdup(file_name);
    int len = strlen(file_name);
    do len--; while (len > 0 && file_name_other[len] != '.');
    if (strstr(file_name, ".gz"))
    {
      do len--; while (len > 0 && file_name_other[len] != '.');
      file_name_other[len+1] = 'e';
      file_name_other[len+2] = 'l';
      file_name_other[len+3] = 'e';
      file_name_other[len+4] = '.';
      file_name_other[len+5] = 'g';
      file_name_other[len+6] = 'z';
      file_name_other[len+7] = '\0';
#ifdef _WIN32
      file_other = fopenGzipped(file_name_other, "r");
#else      
      file_other = 0;
#endif
    }
    else
    {
      file_name_other[len+1] = 'e';
      file_name_other[len+2] = 'l';
      file_name_other[len+3] = 'e';
      file_name_other[len+4] = '\0';
      file_other = fopen(file_name_other, "r");
    }
    free (file_name_other);
    if (file_other == 0)
    {
      fprintf(stderr,"ERROR: cannot open other file '%s'\n",file_name);
      return 0;
    }
    SVreader_jrs* svreader_jrs = new SVreader_jrs();
    if (svreader_jrs->open(file,file_other)) return svreader_jrs;
    delete svreader_jrs;
  }
  else if (strcmp(file_format, "mesh3d") == 0)
  {
    SVreader_mesh3d* svreader_mesh3d = new SVreader_mesh3d();
    if (svreader_mesh3d->open(file)) return svreader_mesh3d;
    delete svreader_mesh3d;
  }
  else if (strcmp(file_format, "sms") == 0)
  {
    SVreader_sms* svreader_sms = new SVreader_sms();
    if (svreader_sms->open(file)) return svreader_sms;
    delete svreader_sms;
  }
  else
  {
    fprintf(stderr,"ERROR: cannot determine the reader for format '%s'\n", file_format);
  }
  return 0;
}

bool SVreadOpener::reopen(SVreader* svreader)
{
  if (file_name == 0)
  {
    fprintf(stderr, "ERROR: cannot reopen() without a file name\n");
    return false;
  }
  else if (file_format == 0)
  {
    fprintf(stderr, "ERROR: cannot reopen() without a successful open()\n");
    return false;
  }
  else if (svreader == 0)
  {
    fprintf(stderr, "ERROR: cannot reopen() when svreader is a null pointer\n");
    return false;
  }

  // close the reader
  svreader->close();

  // depending on the file format we open the file in binary or text mode
  bool is_binary = true;
  if (strcmp(file_format, "sva") == 0 || strcmp(file_format, "tet") == 0 || strcmp(file_format, "off") == 0 || strcmp(file_format, "vtk") == 0 || strcmp(file_format, "jrs") == 0 || strcmp(file_format, "mesh3d") == 0 || strcmp(file_format, "sms") == 0)
    is_binary = false;

  FILE* file;

  // reopen the file
  if (strstr(file_name, ".gz"))
  {
#ifdef _WIN32
    file = fopenGzipped(file_name, (is_binary ? "rb" : "r"));
#else
    fprintf(stderr,"ERROR: cannot reopen gzipped file '%s'\n",file_name);
    return false;
#endif
  }
  else
  {
    file = fopen(file_name, (is_binary ? "rb" : "r"));
  }
  if (file == 0)
  {
    fprintf(stderr,"ERROR: cannot reopen file '%s'\n",file_name);
    return false;
  }

  // reopen the respective reader
  if (strcmp(file_format, "sva") == 0)
  {
    SVreader_sva* svreader_sva = (SVreader_sva*)svreader;
    return svreader_sva->open(file);
  }
  else if (strcmp(file_format, "svb") == 0)
  {
    SVreader_svb* svreader_svb = (SVreader_svb*)svreader;
    return svreader_svb->open(file);
  }
  else if (strcmp(file_format, "svc") == 0)
  {
    SVreader_svc* svreader_svc = (SVreader_svc*)svreader;
    return svreader_svc->open(file);
  }
  else if (strcmp(file_format, "svh") == 0)
  {
    SVreader_svh* svreader_svh = (SVreader_svh*)svreader;
    return svreader_svh->open(file);
  }
  else if (strcmp(file_format, "svg") == 0)
  {
    SVreader_svg* svreader_svg = (SVreader_svg*)svreader;
    return svreader_svg->open(file);
  }
#ifdef BUILD_HZIP
  else if (strcmp(file_format, "hzip") == 0)
  {
    SVreader_hzip* svreader_hzip = (SVreader_hzip*)svreader;
    return svreader_hzip->open(file);
  }
#endif
#ifdef BUILD_GEO
  else if (strcmp(file_format, "geo") == 0)
  {
    SVreader_geo* svreader_geo = (SVreader_geo*)svreader;
    return svreader_geo->open(file);
  }
#endif
  else if (strcmp(file_format, "silo") == 0)
  {
    SVreader_silo* svreader_silo = (SVreader_silo*)svreader;
    return svreader_silo->open(file);
  }
  else if (strcmp(file_format, "tet") == 0)
  {
    SVreader_tet* svreader_tet = (SVreader_tet*)svreader;
    return svreader_tet->open(file);
  }
  else if (strcmp(file_format, "off") == 0)
  {
    SVreader_off* svreader_off = (SVreader_off*)svreader;
    return svreader_off->open(file);
  }
  else if (strcmp(file_format, "vtk") == 0)
  {
    SVreader_vtk* svreader_vtk = (SVreader_vtk*)svreader;
    return svreader_vtk->open(file);
  }
  else if (strcmp(file_format, "jrs") == 0)
  {
    FILE* file_other;
    char* file_name_other = strdup(file_name);
    int len = strlen(file_name);
    do len--; while (len > 0 && file_name_other[len] != '.');
    if (strstr(file_name, ".gz"))
    {
      do len--; while (len > 0 && file_name_other[len] != '.');
      file_name_other[len+1] = 'e';
      file_name_other[len+2] = 'l';
      file_name_other[len+3] = 'e';
      file_name_other[len+4] = '.';
      file_name_other[len+5] = 'g';
      file_name_other[len+6] = 'z';
      file_name_other[len+7] = '\0';
#ifdef _WIN32
      file_other = fopenGzipped(file_name_other, "r");
#else      
      fprintf(stderr,"ERROR: cannot reopen other gzipped file '%s'\n",file_name_other);
      return false;
#endif
    }
    else
    {
      file_name_other[len+1] = 'e';
      file_name_other[len+2] = 'l';
      file_name_other[len+3] = 'e';
      file_name_other[len+4] = '\0';
      file_other = fopen(file_name_other, "r");
    }
    if (file_other == 0)
    {
      fprintf(stderr,"ERROR: cannot reopen other file '%s'\n",file_name);
      return false;
    }
    SVreader_jrs* svreader_jrs = (SVreader_jrs*)svreader;
    return svreader_jrs->open(file,file_other);
  }
  else if (strcmp(file_format, "mesh3d") == 0)
  {
    SVreader_mesh3d* svreader_mesh3d = (SVreader_mesh3d*)svreader;
    return svreader_mesh3d->open(file);
  }
  else if (strcmp(file_format, "sms") == 0)
  {
    SVreader_sms* svreader_sms = (SVreader_sms*)svreader;
    return svreader_sms->open(file);
  }
  else
  {
    fprintf(stderr,"ERROR: cannot determine the reader for format '%s'\n", file_format);
  }
  return false;
}

SVreadOpener::SVreadOpener()
{
  file_name = 0;
  file_format = 0;
}

SVreadOpener::~SVreadOpener()
{
  if (file_format) free(file_format);
  if (file_name) free(file_name);
}
