/*
===============================================================================

  FILE:  SVreadPreAsCompactPre.cpp
  
  CONTENTS:
  
    see corresponding header file

  PROGRAMMERS:
  
    martin isenburg@cs.unc.edu
  
  COPYRIGHT:
  
    copyright (C) 2005  martin isenburg@cs.unc.edu
    
    This software is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
  
  CHANGE HISTORY:
  
    see corresponding header file
  
===============================================================================
*/
#include "svreadpreascompactpre.h"

#include <stdlib.h>
#include <string.h>
#include <stdio.h>
#include <assert.h>

#include "vecnfv.h"
#include "vecniv.h"

//this is for compatibility with the new g++
#ifdef __GNUC__
	#include <ext/hash_map>
	namespace std { using namespace __gnu_cxx; }
	using namespace std;
#else
	#include <hash_map.h>
#endif

#include "dynamicvector.h"

struct SVcell;

typedef struct SVvertex
{
  SVvertex* buffer_next;    // used for efficient memory management
  float* prop_f;
  int index;
  SVcell* last_cell;
} SVvertex;

typedef struct SVcell
{
  int dynamicvector;        // used by the dynamicvector
  SVcell* buffer_next;      // used for efficient memory management
  float* prop_f;
  SVvertex* vertices[8];
  char ready_vertices;
  char number_vertices;
} SVcell;

typedef hash_map<int, SVvertex*> my_hash;

static my_hash* vertex_hash;

static DynamicVector* waiting_area;
static DynamicVector* output_cells;

static SVcell* new_cell;

// efficient memory allocation

static int vertex_pnum_f = 0;
static SVvertex* vertex_buffer_next;
static int vertex_buffer_increment;
static int vertex_buffer_number_chunks;
static SVvertex* vertex_buffer_chunks[16];
#ifdef PRINT_CONTROL_OUTPUT
static int vertex_buffer_size;
static int vertex_buffer_maxsize;
static int vertex_buffer_alloc;
#endif // PRINT_CONTROL_OUTPUT

static void initVertexBuffer(int increment, int pnum_f)
{
  vertex_pnum_f = pnum_f;
  vertex_buffer_next = 0;
  vertex_buffer_increment = increment;
  vertex_buffer_number_chunks = 0;
#ifdef PRINT_CONTROL_OUTPUT
  vertex_buffer_size = 0;
  vertex_buffer_maxsize = 0;
  vertex_buffer_alloc = 0;
#endif // PRINT_CONTROL_OUTPUT
}

static SVvertex* allocVertex()
{
  if (vertex_buffer_next == 0)
  {
    int increment = vertex_buffer_increment;
    vertex_buffer_increment *= 2;
    vertex_buffer_next = (SVvertex*)malloc(sizeof(SVvertex)*increment);
    if (vertex_buffer_next == 0)
    {
      fprintf(stderr,"ERROR (SVreadPreAsCompactPre): malloc for vertex buffer %d failed\n", increment);
      return 0;
    }
    else
    {
      vertex_buffer_chunks[vertex_buffer_number_chunks] = vertex_buffer_next;
      vertex_buffer_number_chunks++;
    }
#ifdef PRINT_CONTROL_OUTPUT
    vertex_buffer_alloc += increment;
#endif // PRINT_CONTROL_OUTPUT
    if (vertex_pnum_f)
    {
      vertex_buffer_next[0].prop_f = (float*) malloc (sizeof (float) * increment * vertex_pnum_f);
      if (vertex_buffer_next[0].prop_f == 0)
      {
        fprintf (stderr,"ERROR (SVreadPreAsCompactPre): malloc for vertex_buffer_next->prop_f %d failed\n", increment);
        return 0;
      }
    }
    for (int i = 0; i < increment; i++)
    {
      vertex_buffer_next[i].buffer_next = &(vertex_buffer_next[i+1]);
      if ( vertex_pnum_f )
      {
        vertex_buffer_next[i].prop_f = & ( vertex_buffer_next[0].prop_f[i*vertex_pnum_f] );
      }
    }
    vertex_buffer_next[increment-1].buffer_next = 0;
  }
  // get pointer to next available vertex
  SVvertex* vertex = vertex_buffer_next;
  vertex_buffer_next = vertex->buffer_next;

  // initialize data fields
  vertex->index = -1;
  vertex->last_cell = 0;

#ifdef PRINT_CONTROL_OUTPUT
  vertex_buffer_size++;
  if (vertex_buffer_size > vertex_buffer_maxsize) vertex_buffer_maxsize = vertex_buffer_size;
#endif // PRINT_CONTROL_OUTPUT

  return vertex;
}

static void deallocVertex(SVvertex* vertex)
{
  vertex->buffer_next = vertex_buffer_next;
  vertex_buffer_next = vertex;
#ifdef PRINT_CONTROL_OUTPUT
  vertex_buffer_size--;
#endif // PRINT_CONTROL_OUTPUT
}

static void destroyVertexBuffer()
{
  for (int i = 0; i < vertex_buffer_number_chunks; i++)
  {
    if ( vertex_pnum_f )
    {
      free(vertex_buffer_chunks[i][0].prop_f);
    }
    free(vertex_buffer_chunks[i]);
  }
}

static int cell_pnum_f;
static SVcell* cell_buffer_next;
static int cell_buffer_increment;
static int cell_buffer_number_chunks;
static SVcell* cell_buffer_chunks[16];
#ifdef PRINT_CONTROL_OUTPUT
static int cell_buffer_size;
static int cell_buffer_maxsize;
static int cell_buffer_alloc;
#endif // PRINT_CONTROL_OUTPUT

static void initCellBuffer(int increment, int pnum_f)
{
  cell_pnum_f = pnum_f;
  cell_buffer_next = 0;
  cell_buffer_increment = increment;
  cell_buffer_number_chunks = 0;
#ifdef PRINT_CONTROL_OUTPUT
  cell_buffer_size = 0;
  cell_buffer_maxsize = 0;
  cell_buffer_alloc = 0;
#endif // PRINT_CONTROL_OUTPUT
}

static SVcell* allocCell()
{
  if (cell_buffer_next == 0)
  {
    int increment = cell_buffer_increment;
    cell_buffer_increment*=2;
    cell_buffer_next = (SVcell*)malloc(sizeof(SVcell)*increment);
    if (cell_buffer_next == 0)
    {
      fprintf(stderr,"ERROR (SVreadPreAsCompactPre): malloc for cell buffer %d failed\n", increment);
      return 0;
    }
    else
    {
      cell_buffer_chunks[cell_buffer_number_chunks] = cell_buffer_next;
      cell_buffer_number_chunks++;
    }
#ifdef PRINT_CONTROL_OUTPUT
    cell_buffer_alloc += increment;
#endif // PRINT_CONTROL_OUTPUT
    if ( cell_pnum_f )
    {
      cell_buffer_next[0].prop_f = ( float* ) malloc ( sizeof ( float ) * increment * cell_pnum_f );
      if ( cell_buffer_next[0].prop_f == 0 )
      {
        fprintf ( stderr,"ERROR (SVreadPreAsCompactPre): malloc for cell_buffer_next->prop_f %d failed\n", increment);
        return 0;
      }
    }
    for (int i = 0; i < increment; i++)
    {
      cell_buffer_next[i].buffer_next = &(cell_buffer_next[i+1]);
      cell_buffer_next[i].prop_f = 0;
      if ( cell_pnum_f )
      {
        cell_buffer_next[i].prop_f = & ( cell_buffer_next[0].prop_f[i*cell_pnum_f] );
      }
    }
    cell_buffer_next[increment-1].buffer_next = 0;
  }
  // get pointer to next available cell
  SVcell* cell = cell_buffer_next;
  cell_buffer_next = cell->buffer_next;
 
  // initialize data fields
  cell->ready_vertices = 0;

#ifdef PRINT_CONTROL_OUTPUT
  cell_buffer_size++;
  if (cell_buffer_size > cell_buffer_maxsize) cell_buffer_maxsize = cell_buffer_size;
#endif // PRINT_CONTROL_OUTPUT

  return cell;
}

static void deallocCell(SVcell* cell)
{
  cell->buffer_next = cell_buffer_next;
  cell_buffer_next = cell;
#ifdef PRINT_CONTROL_OUTPUT
  cell_buffer_size--;
#endif // PRINT_CONTROL_OUTPUT
}

static void destroyCellBuffer()
{
  for (int i = 0; i < cell_buffer_number_chunks; i++)
  {
    free(cell_buffer_chunks[i]);
  }
}

bool SVreadPreAsCompactPre::open(SVreader* svreader, bool preserve_order)
{
  if (svreader == 0 || svreader->post_order)
  {
    return false;
  }
  this->svreader = svreader;
  this->preserve_order = preserve_order;

  if (!preserve_order)
  {
    fprintf(stderr, "ERROR (SVreadPreAsCompactPre): preserve_order == false not supported right now.\n");
    return false;
  }

  nverts = svreader->nverts;
  ncells = svreader->ncells;

  v_pnum_f = svreader->v_pnum_f;
  v_pmin_f = svreader->v_pmin_f;
  v_pmax_f = svreader->v_pmax_f;

  c_pnum_f = svreader->c_pnum_f;
  c_pmin_f = svreader->c_pmin_f;
  c_pmax_f = svreader->c_pmax_f;

  v_count = 0;
  c_count = 0;

  c_idx = new int[8];
  c_final = new bool[8];
  c_v_prop_f = new float*[8];

  vertex_hash = new my_hash;

  waiting_area = new DynamicVector();
  output_cells = new DynamicVector();

  initVertexBuffer(1024, v_pnum_f);
  initCellBuffer(2048, c_pnum_f);

  return true;
}

void SVreadPreAsCompactPre::close(bool close_file)
{
  // close of SVreadPreAsCompactPre
  svreader->close(close_file);
  delete vertex_hash;
  if (waiting_area) delete waiting_area;
  delete output_cells;
  destroyVertexBuffer();
  destroyCellBuffer();
  // close of SVreadDereferenced
  if (c_v_prop_f) delete [] c_v_prop_f; c_v_prop_f = 0;
  // close of SVreader interface
  v_count = -1;
  c_count = -1;
  if (c_idx) delete [] c_idx; c_idx = 0;
  if (c_final) delete [] c_final; c_final = 0;
}

int SVreadPreAsCompactPre::read_preorder_input()
{
  int i;
  SVvertex* vertex;
  SVcell* cell;
  my_hash::iterator hash_element;

  if (waiting_area == 0)
  {
    return 0; // svreader had already returned SV_EOF
  }

  SVevent event = svreader->read_element();
    
  if (event == SV_TETRAHEDRON || event == SV_HEXAHEDRON)
  {
    assert(svreader->c_nverts == 4 || svreader->c_nverts == 8);
    // create cell
    cell = allocCell();
    cell->number_vertices = svreader->c_nverts;
    // insert cell into waiting area
    waiting_area->addElement(cell);
    // copy cell's float properties
    if (c_pnum_f)
    {
      if (cell->prop_f == 0) cell->prop_f = new float[c_pnum_f];
      VecCopyNfv(cell->prop_f, svreader->c_prop_f, c_pnum_f);
    }
    // process the cell's vertices
    for (i = 0; i < svreader->c_nverts; i++)
    {
      // look for vertex in the hash
      hash_element = vertex_hash->find(svreader->c_idx[i]);
      // vertices must preceed cells in a pre-order mesh  
      if (hash_element == vertex_hash->end())
      {
        // fatal error
        fprintf(stderr, "FATAL ERROR: vertex of cell not in hash. corrupt pre-order mesh.\n");
        return -1;
      }
      else
      {
        // use vertex found in hash
        vertex = (*hash_element).second;
      }
      // add vertex to cell
      cell->vertices[i] = vertex;
      // does this vertex have a last cell
      if (vertex->last_cell)
      {
        // this advances the last cell's readiness for output
        vertex->last_cell->ready_vertices++;
        // maybe the last cell is now ready for output?
        if (vertex->last_cell->ready_vertices == vertex->last_cell->number_vertices)
        {
          if (preserve_order) // we have to preserve the cell order
          {
            if (vertex->last_cell == waiting_area->getFirstElement())
            {
              waiting_area->removeFirstElement();
              output_cells->addElement(vertex->last_cell);

              // behind this last cell additional ready cells may be waiting

              while (waiting_area->size())
              {
                SVcell* next_cell = (SVcell*)waiting_area->getFirstElement();
                if (next_cell->ready_vertices == next_cell->number_vertices)
                {
                  waiting_area->removeFirstElement();
                  output_cells->addElement(next_cell);
                }
                else
                {
                  break;
                }
              }
            }
          }
          else // we do not have to preserve the cell order
          {
            waiting_area->removeElement(vertex->last_cell);
            output_cells->addElement(vertex->last_cell);
          }
        }
      }

      // is this cell finalizes this vertex
      if (svreader->c_final && svreader->c_final[i])
      {
        // this advances this cell's readiness for output
        cell->ready_vertices++;
        // maybe this cell is now ready for output?
        if (cell->ready_vertices == cell->number_vertices)
        {
          if (preserve_order) // we have to preserve the cell order
          {
            if (cell == waiting_area->getFirstElement())
            {
              waiting_area->removeFirstElement();
              output_cells->addElement(cell);

              // behind this cell additional ready cells may be waiting

              while (waiting_area->size())
              {
                cell = (SVcell*)waiting_area->getFirstElement();
                if (cell->ready_vertices == 3)
                {
                  waiting_area->removeFirstElement();
                  output_cells->addElement(cell);
                }
                else
                {
                  break;
                }
              }
            }
          }
          else // we do not have to preserve the cell order
          {
            waiting_area->removeElement(cell);
            output_cells->addElement(cell);
          }
        }
        // store with vertex the cell that finalizes it
        vertex->last_cell = cell;
        // remove vertex from hash
        vertex_hash->erase(hash_element);
      }
      else
      {
        // this cell becomes the last cell for this vertex
        vertex->last_cell = cell;
      }
    }
  }
  else if (event == SV_VERTEX)
  {
    // create vertex
    vertex = allocVertex();
    // insert vertex into hash
    vertex_hash->insert(my_hash::value_type(svreader->v_idx, vertex));
    // copy vertex's float properties
    if (v_pnum_f)
    {
      VecCopyNfv(vertex->prop_f, svreader->v_prop_f, v_pnum_f);
    }
  }
  else if (event == SV_FINALIZED)
  {
    // look for finalized vertex in the vertex hash
    hash_element = vertex_hash->find(svreader->final_idx);
    // vertices must preceed their finalization in a pre-order mesh  
    if (hash_element == vertex_hash->end())
    {
      // fatal error
      fprintf(stderr, "ERROR (SVreadPreAsCompactPre): finalized vertex %d not in hash. corrupt pre-order mesh.\n", svreader->final_idx);
      return -1;
    }
    // use vertex found in hash
    vertex = (*hash_element).second;
    // does this vertex have a last cell
    if (vertex->last_cell)
    {
      // this advances the last cell's readiness for output
      vertex->last_cell->ready_vertices++;
      // maybe the last cell is now ready for output?
      if (vertex->last_cell->ready_vertices == 3)
      {
        if (preserve_order) // we have to preserve the cell order
        {
          if (vertex->last_cell == waiting_area->getFirstElement())
          {
            waiting_area->removeFirstElement();
            output_cells->addElement(vertex->last_cell);

            // behind this last cell additional ready cells may be waiting

            while (waiting_area->size())
            {
              cell = (SVcell*)waiting_area->getFirstElement();
              if (cell->ready_vertices == 3)
              {
                waiting_area->removeFirstElement();
                output_cells->addElement(cell);
              }
              else
              {
                break;
              }
            }
          }
        }
        else // we do not have to preserve the cell order
        {
          waiting_area->removeElement(vertex->last_cell);
          output_cells->addElement(vertex->last_cell);
        }
      }
    }
    // remove vertex from hash
    vertex_hash->erase(hash_element);
  }
  else if (event == SV_EOF)
  {
    // all remaining vertices are implicitely finalized
    vertex_hash->clear();
    // are there any cells remaining in the waiting area
    if (!waiting_area->size())
    {
      delete waiting_area;
      waiting_area = 0;
      // apparently not
      return 0;
    }
    else
    {
      if (!output_cells->size())
      {
        delete output_cells;
        output_cells = waiting_area;
      }
      else
      {
        while (waiting_area->size())
        {
          cell = (SVcell*)waiting_area->getFirstElement();
          waiting_area->removeFirstElement();
          output_cells->addElement(cell);
        }
        delete waiting_area;
      }
      waiting_area = 0;
    }
  }
  return 1;
}

int SVreadPreAsCompactPre::read_cell()
{
  int i;
  SVvertex* vertex;
  SVcell* cell;

  // read from preorder input until we have a compact preorder cell for output (or fail)

  while (!output_cells->size())
  {
    int ok = read_preorder_input();
    if (ok <= 0)
    {
      return ok;
    }
  }

  // get next output cell

  cell = (SVcell*)output_cells->getAndRemoveFirstElement();

  // now we have a cell

  have_cell = cell->number_vertices;

  // look for new or finalized vertices and fill in indices, finalization, and positions

  for (i = 0; i < cell->number_vertices; i++)
  {
    vertex = cell->vertices[i];
    // is this vertex a new vertex?
    if (vertex->index == -1)
    {
      vertex->index = v_count + have_new;
      new_vertices[have_new++] = i;
    }
    // is this vertex to be finalized?
    if (vertex->last_cell == cell)
    {
      c_final[i] = true;
      finalized_vertices[have_finalized++] = i;
      deallocVertex(vertex); // it can still be used until the next vertex is alloced.
    }
    else
    {
      c_final[i] = false;
    }
    // fill index array
    c_idx[i] = vertex->index;
    // set additional position array
    c_v_prop_f[i] = vertex->prop_f;
  }

  new_cell = cell;

  // deallocate this cell

  deallocCell(cell); // it can still be used until the next cell is alloced.

  return 1;
}

SVevent SVreadPreAsCompactPre::read_element()
{
  if ((have_new + have_cell) == 0)
  {
    next_new = 0;
    have_finalized = next_finalized = 0;
    
    int ok = read_cell();
    if (ok <= 0)
    {
      return (SVevent)ok;
    }
  }
  if (have_new)
  {
    // reference vertex properties in API interface
    v_idx = c_idx[new_vertices[next_new]];
    v_prop_f = c_v_prop_f[new_vertices[next_new]];
    have_new--; next_new++;
    v_count++;
    return SV_VERTEX;
  }
  else
  {
    // reference cell properties in API interface
    if (c_pnum_f) c_prop_f = new_cell->prop_f;
    c_nverts = have_cell;
    have_cell = 0;
    c_count++;
    if (c_nverts == 4)
      return SV_TETRAHEDRON;
    else
      return SV_HEXAHEDRON;
  }
}

SVevent SVreadPreAsCompactPre::read_event()
{
  if ((have_new + have_cell + have_finalized) == 0)
  {
    return read_element();
  }
  if (have_new)
  {
    // reference vertex properties in API interface
    v_idx = c_idx[new_vertices[next_new]];
    v_prop_f = c_v_prop_f[new_vertices[next_new]];
    have_new--; next_new++;
    v_count++;
    return SV_VERTEX;
  }
  else if (have_cell)
  {
    // reference cell properties in API interface
    if (c_pnum_f) c_prop_f = new_cell->prop_f;
    c_nverts = have_cell;
    have_cell = 0;
    c_count++;
    if (c_nverts == 4)
      return SV_TETRAHEDRON;
    else
      return SV_HEXAHEDRON;
  }
  else if (have_finalized)
  {
    final_idx = new_cell->vertices[finalized_vertices[next_finalized]]->index;
    have_finalized--; next_finalized++;
    return SV_FINALIZED;
  }
  return SV_ERROR;
}

SVreadPreAsCompactPre::SVreadPreAsCompactPre()
{
  // init of SVreadPreAsCompactPre
  c_v_prop_f = 0;

  have_new = 0; next_new = 0;
  have_cell = 0;
  have_finalized = 0; next_finalized = 0;

  svreader = 0;
  preserve_order = true;
}

SVreadPreAsCompactPre::~SVreadPreAsCompactPre()
{
  // clean-up for SVreadPreAsCompactPre
  v_prop_f = 0;
  v_pnum_f = 0;
  v_pmin_f = 0;
  v_pmax_f = 0;
  c_prop_f = 0;
  c_pnum_f = 0;
  c_pmin_f = 0;
  c_pmax_f = 0;
  if (svreader) delete svreader;
}
