/*
===============================================================================

  FILE:  SVwriteBuffered.cpp
  
  CONTENTS:
  
    see corresponding header file

  PROGRAMMERS:
  
    martin isenburg@cs.unc.edu
  
  COPYRIGHT:
  
    copyright (C) 2003-07 martin isenburg@cs.unc.edu
    
    This software is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
  
  CHANGE HISTORY:
  
    see corresponding header file
  
===============================================================================
*/
#include "svwritebuffered.h"

#include <stdlib.h>
#include <string.h>
#include <stdio.h>

#include "vecnfv.h"

//this is for compatibility with the new g++
#ifdef __GNUC__
  #include <ext/hash_map>
  namespace std { using namespace __gnu_cxx; }
  using namespace std;
#else
  #include <hash_map.h>
#endif

#include "dynamicqueue.h"

#define PRINT_CONTROL_OUTPUT
#undef PRINT_CONTROL_OUTPUT

// data structures used for streaming re-ordering 

struct SVtetrahedron;

typedef struct SVvertex
{
  SVvertex* buffer_next; // used for efficient memory management
  I32 index;
  bool finalized;
  float* prop_f;
  // tetrahedra
  int incoming_size;
  int incoming_alloc;
  SVtetrahedron** incoming;
} SVvertex;

typedef struct SVtetrahedron
{
  SVtetrahedron* buffer_next; // used for efficient memory management (and by the waiting queue)
  int dirty;                  // to dirty tetrahedra in the buffer (so they are only checked once)
  SVvertex* vertices[4];
} SVtetrahedron;

typedef hash_map<I32, SVvertex*> my_hash;

static my_hash* vertex_hash; // for matching vertices

static DynamicQueue* waiting_queue; // for tetrahedra ready for output

static SVvertex* cache[4]; // for simulating the cache 

// statistics

#ifdef PRINT_CONTROL_OUTPUT
static int max_in_width = 0;
static int max_in_span = 0;
static int cur_out_width = 0;
static int max_out_width = 0;
static int max_out_span = 0;
#endif

// efficient memory allocation for vertices

static int vertex_pnum_f = 0;
static SVvertex* vertex_buffer_next;
static int vertex_buffer_alloc;
static int vertex_buffer_increment;
static int vertex_buffer_number_chunks;
static SVvertex* vertex_buffer_chunks[16];
#ifdef PRINT_CONTROL_OUTPUT
static int vertex_buffer_size;
static int vertex_buffer_maxsize;
#endif

static void initVertexBuffer(int increment, int pnum_f)
{
  vertex_pnum_f = pnum_f;
  vertex_buffer_next = 0;
  vertex_buffer_alloc = 0;
  vertex_buffer_increment = increment;
  vertex_buffer_number_chunks = 0;
#ifdef PRINT_CONTROL_OUTPUT
  vertex_buffer_size = 0;
  vertex_buffer_maxsize = 0;
#endif
}

static SVvertex* allocVertex()
{
  if ( vertex_buffer_next == 0 )
  {
    int increment = vertex_buffer_increment;
    vertex_buffer_increment *= 2;
    vertex_buffer_next = ( SVvertex* ) malloc ( sizeof ( SVvertex ) * increment );
    if ( vertex_buffer_next == 0 )
    {
      fprintf ( stderr,"ERROR (SVreader_svh): malloc for vertex buffer %d failed\n", increment );
      return 0;
    }
    else
    {
      vertex_buffer_chunks[vertex_buffer_number_chunks] = vertex_buffer_next;
      vertex_buffer_number_chunks++;      
    }
    vertex_buffer_alloc += increment;
    if ( vertex_pnum_f )
    {
      vertex_buffer_next[0].prop_f = ( float* ) malloc ( sizeof ( float ) * increment * vertex_pnum_f );
      if ( vertex_buffer_next[0].prop_f == 0 )
      {
        fprintf ( stderr,"ERROR (SVreader_svh): malloc for vertex_buffer_next->prop_f %d failed\n", increment );
        return 0;
      }
    }
    for ( int i = 0; i < increment; i++ )
    {
      vertex_buffer_next[i].buffer_next = & ( vertex_buffer_next[i+1] );
      vertex_buffer_next[i].incoming = 0;
      if ( vertex_pnum_f )
      {
        vertex_buffer_next[i].prop_f = & ( vertex_buffer_next[0].prop_f[i*vertex_pnum_f] );
      }
    }
    vertex_buffer_next[increment-1].buffer_next = 0;
  }
  // get pointer to next available vertex
  SVvertex* vertex = vertex_buffer_next;
  vertex_buffer_next = vertex->buffer_next;

  vertex->finalized = false;
  vertex->index = -1;
  if (vertex->incoming == 0)
  {
    vertex->incoming = (SVtetrahedron**)malloc(sizeof(SVtetrahedron*)*20);
    vertex->incoming_alloc = 20;
  }
  vertex->incoming_size = 0;

#ifdef PRINT_CONTROL_OUTPUT
  vertex_buffer_size++;
  if ( vertex_buffer_size > vertex_buffer_maxsize ) vertex_buffer_maxsize = vertex_buffer_size;
#endif

  return vertex;
}

static void deallocVertex ( SVvertex* vertex )
{
  vertex->buffer_next = vertex_buffer_next;
  vertex_buffer_next = vertex;
#ifdef PRINT_CONTROL_OUTPUT
  vertex_buffer_size--;
#endif
}

static void destroyVertexBuffer()
{
  for (int i = vertex_buffer_number_chunks-1; i >= 0; i--)
  {
    vertex_buffer_increment/=2;
    for (int j = 0; j < vertex_buffer_increment; j++)
    {
      if (vertex_buffer_chunks[i][j].incoming) free(vertex_buffer_chunks[i][j].incoming);
    }
    if ( vertex_pnum_f )
    {
      free(vertex_buffer_chunks[i][0].prop_f);
    }
    free(vertex_buffer_chunks[i]);
  }
  vertex_buffer_number_chunks = 0;
  vertex_buffer_next = 0;
  vertex_buffer_alloc = 0;
}

// efficient memory allocation for tetrahedra

static SVtetrahedron* tetrahedron_buffer_next;
static int tetrahedron_buffer_alloc;
static int tetrahedron_buffer_increment;
static int tetrahedron_buffer_number_chunks;
static SVtetrahedron* tetrahedron_buffer_chunks[16];
#ifdef PRINT_CONTROL_OUTPUT
static int tetrahedron_buffer_size;
static int tetrahedron_buffer_maxsize;
#endif

static void initTetrahedronBuffer ( int alloc )
{
  tetrahedron_buffer_next = 0;
  tetrahedron_buffer_alloc = 0;
  tetrahedron_buffer_increment = alloc;
  tetrahedron_buffer_number_chunks = 0;
#ifdef PRINT_CONTROL_OUTPUT
  tetrahedron_buffer_size = 0;
  tetrahedron_buffer_maxsize = 0;
#endif
}

static SVtetrahedron* allocTetrahedron()
{
  if ( tetrahedron_buffer_next == 0 )
  {
    int increment = tetrahedron_buffer_increment;
    tetrahedron_buffer_increment *= 2;
    tetrahedron_buffer_next = ( SVtetrahedron* ) malloc ( sizeof ( SVtetrahedron ) * increment );
    if ( tetrahedron_buffer_next == 0 )
    {
      fprintf ( stderr,"malloc for tetrahedron buffer failed\n" );
      return 0;
    }
    else
    {
      tetrahedron_buffer_chunks[tetrahedron_buffer_number_chunks] = tetrahedron_buffer_next;
      tetrahedron_buffer_number_chunks++;      
    }
    tetrahedron_buffer_alloc += increment;
    for ( int i = 0; i < increment; i++ )
    {
      tetrahedron_buffer_next[i].buffer_next = & ( tetrahedron_buffer_next[i+1] );
    }
    tetrahedron_buffer_next[increment-1].buffer_next = 0;
  }
  // get index of next available tetrahedron
  SVtetrahedron* tetrahedron = tetrahedron_buffer_next;
  tetrahedron_buffer_next = tetrahedron->buffer_next;

  tetrahedron->dirty = -1;

#ifdef PRINT_CONTROL_OUTPUT
  tetrahedron_buffer_size++;
  if ( tetrahedron_buffer_size > tetrahedron_buffer_maxsize ) tetrahedron_buffer_maxsize = tetrahedron_buffer_size;
#endif

  return tetrahedron;
}

static void deallocTetrahedron ( SVtetrahedron* tetrahedron )
{
  tetrahedron->buffer_next = tetrahedron_buffer_next;
  tetrahedron_buffer_next = tetrahedron;
#ifdef PRINT_CONTROL_OUTPUT
  tetrahedron_buffer_size--;
#endif
}

static void destroyTetrahedronBuffer()
{
  for (int i = tetrahedron_buffer_number_chunks-1; i >= 0; i--)
  {
    tetrahedron_buffer_increment/=2;
    free(tetrahedron_buffer_chunks[i]);
  }
  tetrahedron_buffer_number_chunks = 0;
  tetrahedron_buffer_next = 0;
  tetrahedron_buffer_alloc = 0;
}

static void addToVertex(SVtetrahedron* tetrahedron, SVvertex* vertex)
{
  if (vertex->incoming_size == vertex->incoming_alloc)
  {
    vertex->incoming_alloc += 2;
    vertex->incoming = (SVtetrahedron**)realloc(vertex->incoming, sizeof(SVtetrahedron*)*vertex->incoming_alloc);
  }
  vertex->incoming[vertex->incoming_size] = tetrahedron;
  vertex->incoming_size++;
}

static void addToVertices(SVtetrahedron* tetrahedron)
{
  addToVertex(tetrahedron, tetrahedron->vertices[0]);
  addToVertex(tetrahedron, tetrahedron->vertices[1]);
  addToVertex(tetrahedron, tetrahedron->vertices[2]);
  addToVertex(tetrahedron, tetrahedron->vertices[3]);
}

static void removeFromVertex(SVtetrahedron* tetrahedron, SVvertex* vertex)
{
  for (int i = 0; i < vertex->incoming_size; i++)
  {
    if (vertex->incoming[i] == tetrahedron)
    {
      vertex->incoming_size--;
      vertex->incoming[i] = vertex->incoming[vertex->incoming_size];
      return;
    }
  }
  fprintf(stderr,"ERROR: did not find tetrahedron in incoming\n");
  exit(0);
}

static void removeFromVertices(SVtetrahedron* tetrahedron)
{
  removeFromVertex(tetrahedron, tetrahedron->vertices[0]);
  removeFromVertex(tetrahedron, tetrahedron->vertices[1]);
  removeFromVertex(tetrahedron, tetrahedron->vertices[2]);
  removeFromVertex(tetrahedron, tetrahedron->vertices[3]);
}

void SVwriteBuffered::add_comment(const char* comment)
{
  fprintf(stderr,"WARNING: add_comments not yet implemented\n");
}

void SVwriteBuffered::set_nverts(I32 nverts)
{
  svwriter->set_nverts(nverts);
  this->nverts = svwriter->nverts;
}

void SVwriteBuffered::set_ncells(I32 ncells)
{
  svwriter->set_ncells(ncells);
  this->ncells = svwriter->ncells;
}

void SVwriteBuffered::set_v_pnum_f(I32 v_pnum_f)
{
  svwriter->set_v_pnum_f(v_pnum_f);
  this->v_pnum_f = svwriter->v_pnum_f;
}

void SVwriteBuffered::set_v_pbox_f(const float* v_pmin_f, const float* v_pmax_f)
{
  svwriter->set_v_pbox_f(v_pmin_f, v_pmax_f);
  this->v_pmin_f = svwriter->v_pmin_f;
  this->v_pmax_f = svwriter->v_pmax_f;
}

void SVwriteBuffered::set_c_pnum_f(I32 c_pnum_f)
{
  svwriter->set_c_pnum_f(c_pnum_f);
  this->c_pnum_f = svwriter->c_pnum_f;
}

void SVwriteBuffered::set_c_pbox_f(const float* c_pmin_f, const float* c_pmax_f)
{
  svwriter->set_c_pbox_f(c_pmin_f, c_pmax_f);
  this->c_pmin_f = svwriter->c_pmin_f;
  this->c_pmax_f = svwriter->c_pmax_f;
}

bool SVwriteBuffered::open(SVwriter* svwriter, int max_delay)
{
  if (svwriter == 0)
  {
    return false;
  }
  this->svwriter = svwriter;
  this->max_delay = max_delay;

#ifdef PRINT_CONTROL_OUTPUT
  fprintf(stderr,"maximal tetrahedron delay is %d\n",max_delay);
#endif
  
  nverts = svwriter->nverts;
  ncells = svwriter->ncells;

  v_count = 0;
  c_count = 0;

  v_pnum_f = svwriter->v_pnum_f;
  v_pmin_f = svwriter->v_pmin_f;
  v_pmax_f = svwriter->v_pmax_f;

  c_pnum_f = svwriter->c_pnum_f;
  c_pmin_f = svwriter->c_pmin_f;
  c_pmax_f = svwriter->c_pmax_f;

  vertex_hash = new my_hash;

  waiting_queue = new DynamicQueue();

  cache[0] = 0;
  cache[1] = 0;
  cache[2] = 0;
  cache[3] = 0;

  return true;
}

void SVwriteBuffered::close(bool close_file, bool update_header)
{
  while (waiting_queue->elements() != 0)
  {
    write_tetrahedron_delayed();
  }

  if (nverts != -1) if (nverts != v_count)  fprintf(stderr,"WARNING (SVwriteBuffered): set nverts %d but v_count %d\n",nverts,v_count);
  if (ncells != -1) if (ncells != c_count)  fprintf(stderr,"WARNING (SVwriteBuffered): set ncells %d but c_count %d\n",ncells,c_count);

  v_count = -1;
  c_count = -1;

  svwriter->close(close_file, update_header);

#ifdef PRINT_CONTROL_OUTPUT
  fprintf(stderr,"%d %d max_in_width %d max_out_width %d diff %6.3f\n",vertex_hash->size(),cur_out_width,max_in_width,max_out_width,100.0f*(max_out_width-max_in_width)/max_in_width);
  fprintf(stderr,"max_in_span %d max_out_span %d diff  %6.3f \n",max_in_span,max_out_span,100.0f*(max_out_span-max_in_span)/max_in_span);
#endif

  delete vertex_hash;
  delete waiting_queue;
  destroyVertexBuffer();
  destroyTetrahedronBuffer();
}

static SVtetrahedron* find_tetrahedron(int dirty)
{
  int i,j,k, in_cache, un_used, idx_1st_oldest, idx_2nd_oldest;
  SVtetrahedron* tetrahedron;
  SVvertex** verts;
  int ccca_idx_1st_oldest = 0x7FFFFFFF;
  int ccca_idx_2nd_oldest = 0x7FFFFFFF;
  SVtetrahedron* ccca = 0; // tetrahedron with three cache vertices and one active vertex
  int cccx_idx_1st_oldest = 0x7FFFFFFF;
  int cccx_idx_2nd_oldest = 0x7FFFFFFF;
  SVtetrahedron* cccx = 0; // tetrahedron with three cache vertices
  int ccaa_idx_1st_oldest = 0x7FFFFFFF;
  int ccaa_idx_2nd_oldest = 0x7FFFFFFF;
  SVtetrahedron* ccaa = 0; // tetrahedron with two cache vertices and two active vertices
  int ccax_idx_1st_oldest = 0x7FFFFFFF;
  int ccax_idx_2nd_oldest = 0x7FFFFFFF;
  SVtetrahedron* ccax = 0; // tetrahedron with two cache vertices and one active vertex
  int caaa_idx_1st_oldest = 0x7FFFFFFF;
  SVtetrahedron* caaa = 0; // tetrahedron with one cache vertex and three active vertices
  int caax_idx_1st_oldest = 0x7FFFFFFF;
  SVtetrahedron* caax = 0; // tetrahedron with one cache vertex and two active vertices
  SVtetrahedron* ccxx = 0; // tetrahedron with two cache vertices 
  SVtetrahedron* caxx = 0; // tetrahedron with one cache vertex and one active vertices
  SVtetrahedron* cxxx = 0; // tetrahedron with one cache vertex

  // check if there are good tetrahedra around cache entry i

  for (i = 0; i < 4; i++)
  {
    if (cache[i] && cache[i]->incoming_size) // does this cache vertex has tetrahedra
    {
      if (cache[i]->incoming_size == 1 && cache[i]->finalized) // does the next tetrahedron finalize this vertex
      {
        return cache[i]->incoming[0]; // do it. finalized vertices can not cause future cache misses
      }
      for (j = 0; j < cache[i]->incoming_size; j++) // iterate over the tetrahedra of cache vertex i
      {
        tetrahedron = cache[i]->incoming[j];
        if (tetrahedron->dirty != dirty) // if not checked yet
        {
          tetrahedron->dirty = dirty; // mark as checked
          verts = tetrahedron->vertices; // get its vertices
          in_cache = 0;
          un_used = 0;
          idx_1st_oldest = 0x7FFFFFFF;
          idx_2nd_oldest = 0x7FFFFFFF;
          for (k = 0; k < 4; k++) // iterate over vertices and count number of cache hits
          {
            if (verts[k]->index == -1)
            {
              un_used++;
            }
            else
            {
              if (cache[0] == verts[k] || cache[1] == verts[k] || cache[2] == verts[k] || cache[3] == verts[k])
              {
                in_cache++;
                if (verts[k]->index < idx_1st_oldest) {idx_2nd_oldest = idx_1st_oldest; idx_1st_oldest = verts[k]->index;}
                else if (verts[k]->index < idx_2nd_oldest) {idx_2nd_oldest = verts[k]->index;}
              }
            }
          }
          if (in_cache == 3)
          {
            if (un_used == 0)
            {
              if (idx_1st_oldest <= ccca_idx_1st_oldest && idx_2nd_oldest <= ccca_idx_2nd_oldest)
              {
                ccca_idx_1st_oldest = idx_1st_oldest;
                ccca_idx_2nd_oldest = idx_2nd_oldest;
                ccca = tetrahedron;
              }
            }
            else 
            {
              if (idx_1st_oldest <= cccx_idx_1st_oldest && idx_2nd_oldest <= cccx_idx_2nd_oldest)
              {
                cccx_idx_1st_oldest = idx_1st_oldest;
                cccx_idx_2nd_oldest = idx_2nd_oldest;
                cccx = tetrahedron;
              }
            }
          }
          else if (in_cache == 2)
          {
            if (un_used == 0)
            {
              if (idx_1st_oldest <= ccaa_idx_1st_oldest && idx_2nd_oldest <= ccaa_idx_2nd_oldest)
              {
                ccaa_idx_1st_oldest = idx_1st_oldest;
                ccaa_idx_2nd_oldest = idx_2nd_oldest;
                ccaa = tetrahedron;
              }
            }
            else if (un_used == 1)
            {
              if (idx_1st_oldest <= ccax_idx_1st_oldest && idx_2nd_oldest <= ccax_idx_2nd_oldest)
              {
                ccax_idx_1st_oldest = idx_1st_oldest;
                ccax_idx_2nd_oldest = idx_2nd_oldest;
                ccax = tetrahedron;
              }
            }
            else
            {
              ccxx = tetrahedron;
            }
          }
          else
          {
            if (un_used == 0)
            {
              if (idx_1st_oldest <= caaa_idx_1st_oldest)
              {
                caaa_idx_1st_oldest = idx_1st_oldest;
                caaa = tetrahedron;
              }
            }
            else if (un_used == 1)
            {
              if (idx_1st_oldest <= caax_idx_1st_oldest)
              {
                caax_idx_1st_oldest = idx_1st_oldest;
                caax = tetrahedron;
              }
            }
            else if (un_used == 2)
            {
              caxx = tetrahedron;
            }
            else
            {
              cxxx = tetrahedron;
            }
          }
        }
      }
    }
  }

  if (ccca) return ccca; // first triple cache hits without new vertices
  else if (cccx) return cccx; // then triple cache hits that introduce one new vertex
  else if (ccaa) return ccaa; // then double cache hits without new vertices
  else if (ccax) return ccax; // then double cache hits that introduce one new vertex
  else if (caaa) return caaa; // then single cache hits without new vertices
  else if (caax) return caax; // then single cache hits that introduce one new vertex
  else if (ccxx) return ccxx; // then double cache hits that introduce two new vertices
  else if (caxx) return caxx; // then single cache hits that introduce two new vertices
  else return cxxx;           // then single cache hits that introduce three new vertices (or nothing)
}

void SVwriteBuffered::write_tetrahedron_delayed()
{
  I32 c_idx[4];
  bool c_final[4];
  SVtetrahedron* tetrahedron = find_tetrahedron(svwriter->c_count);

  if (tetrahedron)
  {
    waiting_queue->removeElement(tetrahedron);
  }
  else
  {
    tetrahedron = (SVtetrahedron*)waiting_queue->getAndRemoveFirstElement();
  }

  removeFromVertices(tetrahedron);

  SVvertex**vertices = tetrahedron->vertices;

  for (int i = 0; i < 4; i++)
  {
    if (vertices[i]->index == -1)
    {
      vertices[i]->index = svwriter->v_count;
      svwriter->write_vertex(vertices[i]->prop_f);
#ifdef PRINT_CONTROL_OUTPUT
      cur_out_width++;
      if (cur_out_width > max_out_width) max_out_width = cur_out_width;
#endif
    }
    c_idx[i] = vertices[i]->index;
    if (vertices[i]->finalized && vertices[i]->incoming_size == 0)
    {
      c_final[i] = true;
      deallocVertex(vertices[i]);
      cache[i] = 0;
#ifdef PRINT_CONTROL_OUTPUT
      cur_out_width--;
      if ((svwriter->v_count-c_idx[i]+1) > max_out_span) max_out_span = (svwriter->v_count-c_idx[i]+1);
#endif
    }
    else
    {
      c_final[i] = false;
      cache[i] = vertices[i];
    }
  }
  deallocTetrahedron(tetrahedron);

  svwriter->write_tetrahedron(c_idx, c_final);
}

void SVwriteBuffered::write_vertex(const float* v_data_f)
{
  if (v_count == 0) initVertexBuffer(1024, v_pnum_f);
  SVvertex* vertex = allocVertex();
  VecCopyNfv(vertex->prop_f, v_data_f, v_pnum_f);
  vertex_hash->insert(my_hash::value_type(v_count, vertex));
  v_count++;
#ifdef PRINT_CONTROL_OUTPUT
  if (vertex_hash->size() > max_in_width) max_in_width = vertex_hash->size();
#endif
}

void SVwriteBuffered::write_tetrahedron(const I32* c_idx, const bool* c_final)
{
  int i;
  my_hash::iterator hash_elements[4];

  if (c_final == 0)
  {
    fprintf(stderr,"FATAL ERROR: need immediate finalization but c_final is a zero pointer\n");
    exit(0);
  }

  if (c_count == 0) initTetrahedronBuffer(2048);
  SVtetrahedron* tetrahedron = allocTetrahedron();

  // get vertices from hash
  for (i = 0; i < 4; i++)
  {
    hash_elements[i] = vertex_hash->find(c_idx[i]);
    if (hash_elements[i] == vertex_hash->end())
    {
      fprintf(stderr,"FATAL ERROR: vertex not in hash. need pre-order mesh\n");
      exit(0);
    }
    else
    {
      tetrahedron->vertices[i] = (*hash_elements[i]).second;
    }
  }
  
  addToVertices(tetrahedron);

  // check for finalization
  for (i = 0; i < 4; i++)
  {
    if (c_final[i])
    {
      tetrahedron->vertices[i]->finalized = true;
      vertex_hash->erase(hash_elements[i]);
#ifdef PRINT_CONTROL_OUTPUT
      if ((v_count - c_idx[i] + 1) > max_in_span) max_in_span = (v_count - c_idx[i] + 1);
#endif
    }
  }

  // add tetrahedron to buffer
  waiting_queue->addElement(tetrahedron);

  // compress a tetrahedron if the buffer is full
  while (waiting_queue->size() == max_delay)
  {
    write_tetrahedron_delayed();
  }

  c_count++;
}

void SVwriteBuffered::write_hexahedron(const I32* c_idx, const bool* c_final)
{
  fprintf(stderr,"FATAL ERROR: write_hexahedron(const int* c_idx, const bool* c_final) not implemented\n");
  exit(0);
}

void SVwriteBuffered::write_tetrahedron(const I32* c_idx, const float* c_data_f, const bool* c_final)
{
  write_tetrahedron(c_idx, c_final);
}

void SVwriteBuffered::write_hexahedron(const I32* c_idx, const float* c_data_f, const bool* c_final)
{
  fprintf(stderr,"FATAL ERROR: write_hexahedron(const int* c_idx, const float* c_data_f, const bool* c_final) not implemented\n");
  exit(0);
}

void SVwriteBuffered::write_finalized(I32 final_idx)
{
  fprintf(stderr,"FATAL ERROR: need immediate finalization (e.g. a tail-compact pre-order mesh)\n");
  exit(0);
}

SVwriteBuffered::SVwriteBuffered()
{
  // init of SVwriteBuffered
  svwriter = 0;
  max_delay = -1;
}

SVwriteBuffered::~SVwriteBuffered()
{
  // clean-up for SVwriteBuffered
  v_pmin_f = 0;
  v_pmax_f = 0;
  c_pmin_f = 0;
  c_pmax_f = 0;
  if (svwriter) delete svwriter;
}
