/*
===============================================================================

  FILE:  SVwriteBufferedHex.cpp

  CONTENTS:

    see corresponding header file

  PROGRAMMERS:

    martin isenburg@cs.unc.edu

  COPYRIGHT:

    copyright (C) 2003-07 martin isenburg@cs.unc.edu

    This software is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.

  CHANGE HISTORY:

    see corresponding header file

===============================================================================
*/
#include "svwritebufferedhex.h"

#include <stdlib.h>
#include <string.h>
#include <stdio.h>

#include "vecnfv.h"

//this is for compatibility with the new g++
#ifdef __GNUC__
#include <ext/hash_map>
namespace std { using namespace __gnu_cxx; }
using namespace std;
#else
#include <hash_map.h>
#endif

#include "dynamicqueue.h"

#include <assert.h>

#define PRINT_CONTROL_OUTPUT
#undef PRINT_CONTROL_OUTPUT

#define PRINT_DEBUG_OUTPUT
#undef PRINT_DEBUG_OUTPUT

#define SVH_START 	0
#define SVH_HUT 	1
#define SVH_ROOF 	2
#define SVH_STEP 	3
#define SVH_CORNER 	4
#define SVH_BRIDGE 	5
#define SVH_TUNNEL 	6
#define SVH_GAP 	7
#define SVH_PIT 	8
#define SVH_DEN 	9

#ifdef PICK_BEST_OP_WITH_FACE_IN_CACHE
//SVH_DEN is the best because it finalizes 8 vertices, does not need joining and does not impact other operations
//SVH_PIT is second because it finalizes 4 vertices, does not need joining and does not impact other operations
//SVH_GAP is third because it finalizes 2 vertices, does not need joining and does not impact other operations
//SVH_TUNNEL is fourth because it does not need joining and does not impact other operations
//SVH_BRIDGE is fifth because it does not need joining and does not impact other operations
//SVH_ROOF has high value because is joins 4 vertices by specifying only one index and a face, and will avoid later corner/step joins
//SVH_CORNER then because it is better than step
//SVH_STEP then because it is better than hut
//SVH_HUT then because it is better than start
//SVH_START is the worst
static int opValue[] = {0, 1, 4, 2, 3, 5, 6, 7, 8, 9};

#ifdef PRINT_DEBUG_OUTPUT
static char opName[][11] = {"SVH_START ", "SVH_HUT   ", "SVH_ROOF  ", "SVH_STEP  ", "SVH_BRIDGE", "SVH_CORNER", "SVH_TUNNEL", "SVH_GAP   ", "SVH_PIT   ", "SVH_DEN   " };
#endif
#endif

// data structures used for streaming re-ordering

struct SVhexahedron;

typedef struct SVvertex
{
  SVvertex* buffer_next; // used for efficient memory management
  I32 index;
  bool finalized;
  float *prop_f;
  // tetrahedra
  int incoming_size;
  int incoming_alloc;
  SVhexahedron** incoming;
#ifdef PICK_MOST_SPIRALLING
  int in_cache; //position of the vertex in the cache (-1 if not in the cache)
#else
  bool in_cache; //is the vertex in the cache ?
#endif
#ifdef PICK_OLDEST_WITH_MOST_FACES_IN_EXTENDED_CACHE
  bool in_tb_cache; //is the vertex in the tb_cache ?
#endif
} SVvertex;

typedef struct SVhexahedron
{
  SVhexahedron* buffer_next; // used for efficient memory management (and by the waiting queue)
  I32 dirty;                  // to dirty tetrahedra in the buffer (so they are only checked once)
  SVvertex* vertices[8];
  I32 order_index; //this is the index of the hexahedron in the stream
  float *prop_f;
} SVhexahedron;

#ifdef PICK_MOST_SPIRALLING
#define SPIRALLING_UP 8
#define SPIRALLING_DOWN 4
#define SPIRALLING_LEFT 2
#define SPIRALLING_RIGHT 1
#define SPIRALLING_STRAIGHT 16
static int currentSpirallingDirection = 0; 
static int previousSpirallingDirection = 0; 
//these are the vertices of the face that was in common between cached hex and previous cached hex (-1 if no face in common). the order is topRight, topLeft, bottomLeft, bottomRight
static SVvertex* lastFaceEntry[4] = {NULL, NULL, NULL, NULL};
//this is the vertex shot from lastFaceEntry[0] toward the opposite face
static SVvertex* lastFaceZeroShoot = NULL;
#endif

//helper: indices of faces
static const int faceIndices[6][4] = {{1,0,3,2},{3,0,4,7},{4,0,1,5},{5,1,2,6},{6,2,3,7},{7,4,5,6}};
//helper: the index of the opposite face vertex for the given vertex 
static const int oppositeFaceVertices[6][4] = {{5,4,7,6},{2,1,5,6},{7,3,2,6},{4,0,3,7},{5,1,0,4},{3,0,1,2}};

typedef hash_map<I32, SVvertex*> my_hash;

static my_hash* vertex_hash; // for matching vertices

static DynamicQueue* waiting_queue; // for hexahedron ready for output

static SVvertex* cache[8] = {NULL, NULL, NULL, NULL, NULL, NULL, NULL, NULL}; // for simulating the cache

#ifdef PICK_OLDEST_WITH_MOST_FACES_IN_EXTENDED_CACHE
  static SVvertex* b_cache[8] = {NULL, NULL, NULL, NULL, NULL, NULL, NULL, NULL};; // this caches the vertices of the hex before the one in cache
  static SVvertex* tb_cache[8] = {NULL, NULL, NULL, NULL, NULL, NULL, NULL, NULL};; // this caches the vertices of the hex 2-before the one in cache
#endif

// statistics

#ifdef PRINT_CONTROL_OUTPUT
static int max_in_width = 0;
static int max_in_span = 0;
static int cur_out_width = 0;
static int max_out_width = 0;
static int max_out_span = 0;
#endif

// efficient memory allocation for vertices

static int vertex_pnum_f = 0;
static SVvertex* vertex_buffer_next;
static int vertex_buffer_alloc;
static int vertex_buffer_increment;
static int vertex_buffer_number_chunks;
static SVvertex* vertex_buffer_chunks[16];
#ifdef PRINT_CONTROL_OUTPUT
static int vertex_buffer_size;
static int vertex_buffer_maxsize;
#endif

static void initVertexBuffer(int increment, int v_pnum_f)
{
  vertex_pnum_f = v_pnum_f;
  vertex_buffer_next = 0;
  vertex_buffer_alloc = 0;
  vertex_buffer_increment = increment;
  vertex_buffer_number_chunks = 0;
#ifdef PRINT_CONTROL_OUTPUT
  vertex_buffer_size = 0;
  vertex_buffer_maxsize = 0;
#endif
}

static SVvertex* allocVertex()
{
  if(vertex_buffer_next == 0)
  {
    int increment = vertex_buffer_increment;
    vertex_buffer_increment *= 2;
    vertex_buffer_next =(SVvertex*) malloc(sizeof(SVvertex) * increment);
    if(vertex_buffer_next == 0)
    {
      fprintf(stderr,"ERROR (SVreader_svh): malloc for vertex buffer %d failed\n", increment);
      return 0;
    }
    else
    {
      vertex_buffer_chunks[vertex_buffer_number_chunks] = vertex_buffer_next;
      vertex_buffer_number_chunks++;      
    }
    vertex_buffer_alloc += increment;
    if(vertex_pnum_f)
    {
      vertex_buffer_next[0].prop_f =(float*) malloc(sizeof(float) * increment * vertex_pnum_f);
      if(vertex_buffer_next[0].prop_f == 0)
      {
        fprintf(stderr,"ERROR (SVreader_svh): malloc for vertex_buffer_next->prop_f %d failed\n", increment);
        return 0;
      }
    }
    for(int i = 0; i < increment; i++)
    {
      vertex_buffer_next[i].buffer_next = &(vertex_buffer_next[i+1]);
      vertex_buffer_next[i].incoming = 0;
      if(vertex_pnum_f)
      {
        vertex_buffer_next[i].prop_f = &(vertex_buffer_next[0].prop_f[i*vertex_pnum_f]);
      }
    }
    vertex_buffer_next[increment-1].buffer_next = 0;
  }
  // get pointer to next available vertex
  SVvertex* vertex = vertex_buffer_next;
  vertex_buffer_next = vertex->buffer_next;

  vertex->finalized = false;
  vertex->index = -1;
#ifdef PICK_MOST_SPIRALLING
  vertex-> in_cache = -1;
#else
  vertex-> in_cache = false;
#endif
#ifdef PICK_OLDEST_WITH_MOST_FACES_IN_EXTENDED_CACHE
  vertex-> in_tb_cache = false;
#endif
  if (vertex->incoming == 0)
  {
    vertex->incoming = (SVhexahedron**)malloc(sizeof(SVhexahedron*)*10);
    vertex->incoming_alloc = 10;
  }
  vertex->incoming_size = 0;

#ifdef PRINT_CONTROL_OUTPUT
  vertex_buffer_size++;
  if(vertex_buffer_size > vertex_buffer_maxsize) vertex_buffer_maxsize = vertex_buffer_size;
#endif

  return vertex;
}

static void deallocVertex(SVvertex* vertex)
{
  vertex->buffer_next = vertex_buffer_next;
  vertex_buffer_next = vertex;
#ifdef PRINT_CONTROL_OUTPUT
  vertex_buffer_size--;
#endif
}

static void destroyVertexBuffer()
{
  for (int i = vertex_buffer_number_chunks-1; i >= 0; i--)
  {
    vertex_buffer_increment/=2;
    for (int j = 0; j < vertex_buffer_increment; j++)
    {
      if (vertex_buffer_chunks[i][j].incoming) free(vertex_buffer_chunks[i][j].incoming);
    }
    if(vertex_pnum_f)
    {
      free(vertex_buffer_chunks[i][0].prop_f);
    }
    free(vertex_buffer_chunks[i]);
  }
  vertex_buffer_number_chunks = 0;
  vertex_buffer_next = 0;
  vertex_buffer_alloc = 0;
}

// efficient memory allocation for hexahedra

static int hexahedron_pnum_f = 0;
static SVhexahedron* hexahedron_buffer_next;
static int hexahedron_buffer_alloc;
static int hexahedron_buffer_increment;
static int hexahedron_buffer_number_chunks;
static SVhexahedron* hexahedron_buffer_chunks[16];
#ifdef PRINT_CONTROL_OUTPUT
static int hexahedron_buffer_size;
static int hexahedron_buffer_maxsize;
#endif

static void initHexahedronBuffer(int increment, int c_pnum_f)
{
  hexahedron_pnum_f = c_pnum_f;
  hexahedron_buffer_next = 0;
  hexahedron_buffer_alloc = 0;
  hexahedron_buffer_increment = increment;
  hexahedron_buffer_number_chunks = 0;
#ifdef PRINT_CONTROL_OUTPUT
  hexahedron_buffer_size = 0;
  hexahedron_buffer_maxsize = 0;
#endif
}

static SVhexahedron* allocHexahedron()
{
  if(hexahedron_buffer_next == 0)
  {
    int increment = hexahedron_buffer_increment;
    hexahedron_buffer_increment *= 2;
    hexahedron_buffer_next =(SVhexahedron*) malloc(sizeof(SVhexahedron) * increment);
    if(hexahedron_buffer_next == 0)
    {
      fprintf(stderr,"ERROR (SVreader_svh): malloc for hexahedron buffer %d failed\n", increment);
      return 0;
    }
    else
    {
      hexahedron_buffer_chunks[hexahedron_buffer_number_chunks] = hexahedron_buffer_next;
      hexahedron_buffer_number_chunks++;      
    }
    hexahedron_buffer_alloc += increment;
    if(hexahedron_pnum_f)
    {
      hexahedron_buffer_next[0].prop_f =(float*) malloc(sizeof(float) * increment * hexahedron_pnum_f);
      if(hexahedron_buffer_next[0].prop_f == 0)
      {
        fprintf(stderr,"ERROR (SVreader_svh): malloc for hexahedron_buffer_next->prop_f %d failed\n", increment);
        return 0;
      }
    }
    for(int i = 0; i < increment; i++)
    {
      hexahedron_buffer_next[i].buffer_next = &(hexahedron_buffer_next[i+1]);
      if(hexahedron_pnum_f)
      {
        hexahedron_buffer_next[i].prop_f = &(hexahedron_buffer_next[0].prop_f[i*hexahedron_pnum_f]);
      }
    }
    hexahedron_buffer_next[increment-1].buffer_next = 0;
  }
  // get pointer to next available hexahedron
  SVhexahedron* hexahedron = hexahedron_buffer_next;
  hexahedron_buffer_next = hexahedron->buffer_next;

  hexahedron->dirty = -1;

#ifdef PRINT_CONTROL_OUTPUT
  hexahedron_buffer_size++;
  if(hexahedron_buffer_size > hexahedron_buffer_maxsize) hexahedron_buffer_maxsize = hexahedron_buffer_size;
#endif

  return hexahedron;
}

static void deallocHexahedron(SVhexahedron* hexahedron)
{
  hexahedron->buffer_next = hexahedron_buffer_next;
  hexahedron_buffer_next = hexahedron;
#ifdef PRINT_CONTROL_OUTPUT
  hexahedron_buffer_size--;
#endif
}

static void destroyHexahedronBuffer()
{
  for (int i = hexahedron_buffer_number_chunks-1; i >= 0; i--)
  {
    hexahedron_buffer_increment/=2;
    if(hexahedron_pnum_f)
    {
      free(hexahedron_buffer_chunks[i][0].prop_f);
    }
    free(hexahedron_buffer_chunks[i]);
  }
  hexahedron_buffer_number_chunks = 0;
  hexahedron_buffer_next = 0;
  hexahedron_buffer_alloc = 0;
}

static void addToVertex(SVhexahedron* hexahedron, SVvertex* vertex)
{
  if(vertex->incoming_size == vertex->incoming_alloc)
  {
    vertex->incoming_alloc += 2;
    vertex->incoming =(SVhexahedron**) realloc(vertex->incoming, sizeof(SVhexahedron*) *vertex->incoming_alloc);
  }
  vertex->incoming[vertex->incoming_size] = hexahedron;
  vertex->incoming_size++;
}

static void addToVertices(SVhexahedron* hexahedron)
{
  addToVertex(hexahedron, hexahedron->vertices[0]);
  addToVertex(hexahedron, hexahedron->vertices[1]);
  addToVertex(hexahedron, hexahedron->vertices[2]);
  addToVertex(hexahedron, hexahedron->vertices[3]);
  addToVertex(hexahedron, hexahedron->vertices[4]);
  addToVertex(hexahedron, hexahedron->vertices[5]);
  addToVertex(hexahedron, hexahedron->vertices[6]);
  addToVertex(hexahedron, hexahedron->vertices[7]);
}

static void removeFromVertex(SVhexahedron* hexahedron, SVvertex* vertex)
{
  for(int i = 0; i < vertex->incoming_size; i++)
  {
    if(vertex->incoming[i] == hexahedron)
    {
      vertex->incoming_size--;
      vertex->incoming[i] = vertex->incoming[vertex->incoming_size];
      return;
    }
  }
  fprintf(stderr,"ERROR: did not find hexahedron in incoming\n");
  exit(0);
}

static void removeFromVertices(SVhexahedron* hexahedron)
{
  removeFromVertex(hexahedron, hexahedron->vertices[0]);
  removeFromVertex(hexahedron, hexahedron->vertices[1]);
  removeFromVertex(hexahedron, hexahedron->vertices[2]);
  removeFromVertex(hexahedron, hexahedron->vertices[3]);
  removeFromVertex(hexahedron, hexahedron->vertices[4]);
  removeFromVertex(hexahedron, hexahedron->vertices[5]);
  removeFromVertex(hexahedron, hexahedron->vertices[6]);
  removeFromVertex(hexahedron, hexahedron->vertices[7]);
}

void SVwriteBufferedHex::add_comment(const char* comment)
{
  fprintf(stderr,"WARNING: add_comments not yet implemented\n");
}

void SVwriteBufferedHex::set_nverts(I32 nverts)
{
  svwriter->set_nverts(nverts);
  this->nverts = svwriter->nverts;
}

void SVwriteBufferedHex::set_ncells(I32 ncells)
{
  svwriter->set_ncells(ncells);
  this->ncells = svwriter->ncells;
}

void SVwriteBufferedHex::set_v_pnum_f(I32 v_pnum_f)
{
  if(v_pnum_f != 3)
  {
    fprintf(stderr,"WARNING: v_pnum_f = %d not implemented. changed to v_pnum_f = 3\n", v_pnum_f);
    v_pnum_f = 3;
  }
  svwriter->set_v_pnum_f(v_pnum_f);
  this->v_pnum_f = svwriter->v_pnum_f;
}

void SVwriteBufferedHex::set_v_pbox_f(const float* v_pmin_f, const float* v_pmax_f)
{
  svwriter->set_v_pbox_f(v_pmin_f, v_pmax_f);
  this->v_pmin_f = svwriter->v_pmin_f;
  this->v_pmax_f = svwriter->v_pmax_f;
}


void SVwriteBufferedHex::set_c_pnum_f(I32 c_pnum_f)
{
  svwriter->set_c_pnum_f(c_pnum_f);
  if(this->c_pnum_f != c_pnum_f)
  {
    this->c_pnum_f = svwriter->c_pnum_f;
  }
}

void SVwriteBufferedHex::set_c_pbox_f(const float* c_pmin_f, const float* c_pmax_f)
{
  svwriter->set_c_pbox_f(c_pmin_f, c_pmax_f);
  this->c_pmin_f = svwriter->c_pmin_f;
  this->c_pmax_f = svwriter->c_pmax_f;
}

#ifdef PICK_BEST_OP_WITH_FACE_IN_CACHE
bool SVwriteBufferedHex::open(SVwriter_svh* svwriter, int max_delay)
#else
bool SVwriteBufferedHex::open(SVwriter* svwriter, int max_delay)
#endif
{
  if(svwriter == 0)
  {
    return false;
  }
  this->svwriter = svwriter;
  this->max_delay = max_delay;

#ifdef PRINT_CONTROL_OUTPUT
  fprintf(stderr,"maximal hexahedron delay is %d\n",max_delay);
#endif

  nverts = svwriter->nverts;
  ncells = svwriter->ncells;

  v_count = 0;
  c_count = 0;

  v_pnum_f = svwriter->v_pnum_f;
  v_pmin_f = svwriter->v_pmin_f;
  v_pmax_f = svwriter->v_pmax_f;

  c_pnum_f = svwriter->c_pnum_f;
  c_pmin_f = svwriter->c_pmin_f;
  c_pmax_f = svwriter->c_pmax_f;

  vertex_hash = new my_hash;

  waiting_queue = new DynamicQueue();

  cache[0] = NULL;
  cache[1] = NULL;
  cache[2] = NULL;
  cache[3] = NULL;
  cache[4] = NULL;
  cache[5] = NULL;
  cache[6] = NULL;
  cache[7] = NULL;

#ifdef PICK_OLDEST_WITH_MOST_FACES_IN_EXTENDED_CACHE
  tb_cache[0] = NULL;
  tb_cache[1] = NULL;
  tb_cache[2] = NULL;
  tb_cache[3] = NULL;
  tb_cache[4] = NULL;
  tb_cache[5] = NULL;
  tb_cache[6] = NULL;
  tb_cache[7] = NULL;
#endif

  return true;
}

void SVwriteBufferedHex::close(bool close_file, bool update_header)
{
  while(waiting_queue->elements() != 0)
  {
    write_hexahedron_delayed();
  }

  if(nverts != -1) if(nverts != v_count)  fprintf(stderr,"WARNING (SVwriteBufferedHex): set nverts %d but v_count %d\n",nverts,v_count);
  if(ncells != -1) if(ncells != c_count)  fprintf(stderr,"WARNING (SVwriteBufferedHex): set ncells %d but c_count %d\n",ncells,c_count);

  v_count = -1;
  c_count = -1;

  svwriter->close(close_file, update_header);

#ifdef PRINT_CONTROL_OUTPUT
  fprintf(stderr,"%d %d max_in_width %d max_out_width %d diff %6.3f\n",vertex_hash->size(),cur_out_width,max_in_width,max_out_width,100.0f*(max_out_width-max_in_width) /max_in_width);
  fprintf(stderr,"max_in_span %d max_out_span %d diff  %6.3f \n",max_in_span,max_out_span,100.0f*(max_out_span-max_in_span) /max_in_span);
#endif

  delete vertex_hash;
  delete waiting_queue;
  destroyVertexBuffer();
  destroyHexahedronBuffer();
}

#ifdef PICK_MOST_SPIRALLING
//helper function, looks whether the vertices a,b,c,d are a face of the cache, 
//if it is, it returns whether the face is up/down/left/right with respect to the current hexahedron
static int faceInCache(SVvertex *a, SVvertex *b, SVvertex *c, SVvertex *d)
{
  if(a->in_cache!=-1 && b->in_cache!=-1 && c->in_cache!=-1 && d->in_cache!=-1)
  {
#ifdef PRINT_DEBUG_OUTPUT
    fprintf(stderr, "\t\ta:%x, b:%x, c:%x, d:%x\n", (int)a, (int)b, (int)c, (int)d);
    fprintf(stderr, "\t\tLFE0:%x LFE1:%x LFE2:%x LFE3:%x\n", (int)lastFaceEntry[0], (int)lastFaceEntry[1], (int)lastFaceEntry[2], (int)lastFaceEntry[3]);
#endif
    //try to find two verts in common with left edge (verts 1-2)
    if( (a==lastFaceEntry[1]&&b==lastFaceEntry[2]) || (b==lastFaceEntry[1]&&c==lastFaceEntry[2]) || (c==lastFaceEntry[1]&&d==lastFaceEntry[2]) || (d==lastFaceEntry[1]&&a==lastFaceEntry[2]))
    {
      return SPIRALLING_LEFT;
    }
    //try to find two verts in common with down edge (verts 2-3)
    else if( (a==lastFaceEntry[2]&&b==lastFaceEntry[3]) || (b==lastFaceEntry[2]&&c==lastFaceEntry[3]) || (c==lastFaceEntry[2]&&d==lastFaceEntry[3]) || (d==lastFaceEntry[2]&&a==lastFaceEntry[3]))
    {
      return SPIRALLING_DOWN;
    }
    //try to find two verts in common with right edge (verts 3-0)
    else if((a==lastFaceEntry[3]&&b==lastFaceEntry[0]) || (b==lastFaceEntry[3]&&c==lastFaceEntry[0]) || (c==lastFaceEntry[3]&&d==lastFaceEntry[0]) || (d==lastFaceEntry[3]&&a==lastFaceEntry[0]))
    {
      return SPIRALLING_RIGHT;
    }
    //try to find two verts in common with up edge (verts 0-1)
    else if((a==lastFaceEntry[0]&&b==lastFaceEntry[1]) || (b==lastFaceEntry[0]&&c==lastFaceEntry[1]) || (c==lastFaceEntry[0]&&d==lastFaceEntry[1]) || (d==lastFaceEntry[0]&&a==lastFaceEntry[1]))
    {
      return SPIRALLING_UP;
    }
    else
      return SPIRALLING_STRAIGHT;
  }
  else
    return 0;
}

//helper function, finds the vertices that the given hexahedron shares with the cache, and builds the corresponding lastFaceEntry
static void updateSpirallingState(SVvertex **vertices, int bestFace, int bestDirection)
{
#ifdef PRINT_DEBUG_OUTPUT
  fprintf(stderr, "updateSpirallingState(..., bestFace=%i, bestDirection=%i)\n", bestFace, bestDirection);
#endif
  if(bestFace==-1)
  {
    lastFaceEntry[0] = NULL;
    lastFaceEntry[1] = NULL;
    lastFaceEntry[2] = NULL;
    lastFaceEntry[3] = NULL;
    previousSpirallingDirection = 0;
    currentSpirallingDirection = 0;
#ifdef PRINT_DEBUG_OUTPUT
    fprintf(stderr, "update: no face in chache, starting anew.\n");
#endif
  }
  else
  {
    if((bestDirection == currentSpirallingDirection) || (bestDirection == SPIRALLING_STRAIGHT))
      ;//nothing to do
    else if (bestDirection == previousSpirallingDirection)
    {
      previousSpirallingDirection = currentSpirallingDirection;
      //previousSpirallingDirection = 0;
      currentSpirallingDirection = bestDirection;
    }
    else
    {
      currentSpirallingDirection = bestDirection;
      previousSpirallingDirection = 0;
    }
    if(bestDirection & SPIRALLING_LEFT)
    {
      if(vertices[faceIndices[bestFace][1]]==lastFaceEntry[1])
      {
        lastFaceEntry[0] = vertices[faceIndices[bestFace][0]];
        lastFaceEntry[1] = vertices[faceIndices[bestFace][1]];
        lastFaceEntry[2] = vertices[faceIndices[bestFace][2]];
        lastFaceEntry[3] = vertices[faceIndices[bestFace][3]];
        lastFaceZeroShoot = vertices[oppositeFaceVertices[bestFace][0]];
        return;
      }
      else if(vertices[faceIndices[bestFace][2]]==lastFaceEntry[1])
      {
        lastFaceEntry[0] = vertices[faceIndices[bestFace][1]];
        lastFaceEntry[1] = vertices[faceIndices[bestFace][2]];
        lastFaceEntry[2] = vertices[faceIndices[bestFace][3]];
        lastFaceEntry[3] = vertices[faceIndices[bestFace][0]];
        lastFaceZeroShoot = vertices[oppositeFaceVertices[bestFace][1]];
        return;
      }
      else if(vertices[faceIndices[bestFace][3]]==lastFaceEntry[1])
      {
        lastFaceEntry[0] = vertices[faceIndices[bestFace][2]];
        lastFaceEntry[1] = vertices[faceIndices[bestFace][3]];
        lastFaceEntry[2] = vertices[faceIndices[bestFace][0]];
        lastFaceEntry[3] = vertices[faceIndices[bestFace][1]];
        lastFaceZeroShoot = vertices[oppositeFaceVertices[bestFace][2]];
        return;
      }
      else if(vertices[faceIndices[bestFace][0]]==lastFaceEntry[1])
      {
        lastFaceEntry[0] = vertices[faceIndices[bestFace][3]];
        lastFaceEntry[1] = vertices[faceIndices[bestFace][0]];
        lastFaceEntry[2] = vertices[faceIndices[bestFace][1]];
        lastFaceEntry[3] = vertices[faceIndices[bestFace][2]];
        lastFaceZeroShoot = vertices[oppositeFaceVertices[bestFace][3]];
        return;
      }
      else
        assert(false);
    }
    else if(bestDirection & SPIRALLING_UP)
    {
      if(vertices[faceIndices[bestFace][0]]==lastFaceEntry[0])
      {
        lastFaceEntry[0] = vertices[faceIndices[bestFace][0]];
        lastFaceEntry[1] = vertices[faceIndices[bestFace][1]];
        lastFaceEntry[2] = vertices[faceIndices[bestFace][2]];
        lastFaceEntry[3] = vertices[faceIndices[bestFace][3]];
        lastFaceZeroShoot = vertices[oppositeFaceVertices[bestFace][0]];
        return;
      }
      else if(vertices[faceIndices[bestFace][1]]==lastFaceEntry[0])
      {
        lastFaceEntry[0] = vertices[faceIndices[bestFace][1]];
        lastFaceEntry[1] = vertices[faceIndices[bestFace][2]];
        lastFaceEntry[2] = vertices[faceIndices[bestFace][3]];
        lastFaceEntry[3] = vertices[faceIndices[bestFace][0]];
        lastFaceZeroShoot = vertices[oppositeFaceVertices[bestFace][1]];
        return;
      }
      else if(vertices[faceIndices[bestFace][2]]==lastFaceEntry[0])
      {
        lastFaceEntry[0] = vertices[faceIndices[bestFace][2]];
        lastFaceEntry[1] = vertices[faceIndices[bestFace][3]];
        lastFaceEntry[2] = vertices[faceIndices[bestFace][0]];
        lastFaceEntry[3] = vertices[faceIndices[bestFace][1]];
        lastFaceZeroShoot = vertices[oppositeFaceVertices[bestFace][2]];
        return;
      }
      else if(vertices[faceIndices[bestFace][3]]==lastFaceEntry[0])
      {
        lastFaceEntry[0] = vertices[faceIndices[bestFace][3]];
        lastFaceEntry[1] = vertices[faceIndices[bestFace][0]];
        lastFaceEntry[2] = vertices[faceIndices[bestFace][1]];
        lastFaceEntry[3] = vertices[faceIndices[bestFace][2]];
        lastFaceZeroShoot = vertices[oppositeFaceVertices[bestFace][3]];
        return;
      }
      else
        assert(false);
    }
    else if(bestDirection & SPIRALLING_RIGHT)
    {
      if(vertices[faceIndices[bestFace][3]]==lastFaceEntry[3])
      {
        lastFaceEntry[0] = vertices[faceIndices[bestFace][0]];
        lastFaceEntry[1] = vertices[faceIndices[bestFace][1]];
        lastFaceEntry[2] = vertices[faceIndices[bestFace][2]];
        lastFaceEntry[3] = vertices[faceIndices[bestFace][3]];
        lastFaceZeroShoot = vertices[oppositeFaceVertices[bestFace][0]];
        return;
      }
      else if(vertices[faceIndices[bestFace][0]]==lastFaceEntry[3])
      {
        lastFaceEntry[0] = vertices[faceIndices[bestFace][1]];
        lastFaceEntry[1] = vertices[faceIndices[bestFace][2]];
        lastFaceEntry[2] = vertices[faceIndices[bestFace][3]];
        lastFaceEntry[3] = vertices[faceIndices[bestFace][0]];
        lastFaceZeroShoot = vertices[oppositeFaceVertices[bestFace][1]];
        return;
      }
      else if(vertices[faceIndices[bestFace][1]]==lastFaceEntry[3])
      {
        lastFaceEntry[0] = vertices[faceIndices[bestFace][2]];
        lastFaceEntry[1] = vertices[faceIndices[bestFace][3]];
        lastFaceEntry[2] = vertices[faceIndices[bestFace][0]];
        lastFaceEntry[3] = vertices[faceIndices[bestFace][1]];
        lastFaceZeroShoot = vertices[oppositeFaceVertices[bestFace][2]];
        return;
      }
      else if(vertices[faceIndices[bestFace][2]]==lastFaceEntry[3])
      {
        lastFaceEntry[0] = vertices[faceIndices[bestFace][3]];
        lastFaceEntry[1] = vertices[faceIndices[bestFace][0]];
        lastFaceEntry[2] = vertices[faceIndices[bestFace][1]];
        lastFaceEntry[3] = vertices[faceIndices[bestFace][2]];
        lastFaceZeroShoot = vertices[oppositeFaceVertices[bestFace][3]];
        return;
      }
      else
        assert(false);
    }
    else if(bestDirection & SPIRALLING_DOWN)
    {
      if(vertices[faceIndices[bestFace][2]]==lastFaceEntry[2])
      {
        lastFaceEntry[0] = vertices[faceIndices[bestFace][0]];
        lastFaceEntry[1] = vertices[faceIndices[bestFace][1]];
        lastFaceEntry[2] = vertices[faceIndices[bestFace][2]];
        lastFaceEntry[3] = vertices[faceIndices[bestFace][3]];
        lastFaceZeroShoot = vertices[oppositeFaceVertices[bestFace][02]];
        return;
      }
      else if(vertices[faceIndices[bestFace][3]]==lastFaceEntry[2])
      {
        lastFaceEntry[0] = vertices[faceIndices[bestFace][1]];
        lastFaceEntry[1] = vertices[faceIndices[bestFace][2]];
        lastFaceEntry[2] = vertices[faceIndices[bestFace][3]];
        lastFaceEntry[3] = vertices[faceIndices[bestFace][0]];
        lastFaceZeroShoot = vertices[oppositeFaceVertices[bestFace][1]];
        return;
      }
      else if(vertices[faceIndices[bestFace][0]]==lastFaceEntry[2])
      {
        lastFaceEntry[0] = vertices[faceIndices[bestFace][2]];
        lastFaceEntry[1] = vertices[faceIndices[bestFace][3]];
        lastFaceEntry[2] = vertices[faceIndices[bestFace][0]];
        lastFaceEntry[3] = vertices[faceIndices[bestFace][1]];
        lastFaceZeroShoot = vertices[oppositeFaceVertices[bestFace][2]];
        return;
      }
      else if(vertices[faceIndices[bestFace][1]]==lastFaceEntry[2])
      {
        lastFaceEntry[0] = vertices[faceIndices[bestFace][3]];
        lastFaceEntry[1] = vertices[faceIndices[bestFace][0]];
        lastFaceEntry[2] = vertices[faceIndices[bestFace][1]];
        lastFaceEntry[3] = vertices[faceIndices[bestFace][2]];
        lastFaceZeroShoot = vertices[oppositeFaceVertices[bestFace][3]];
        return;
      }
      else
        assert(false);
    }
    else if(bestDirection & SPIRALLING_STRAIGHT)
    {
#ifdef PRINT_DEBUG_OUTPUT
      fprintf(stderr, "vertices={%x, %x, %x, %x, %x, %x, %x, %x}\n", (int)vertices[0], (int)vertices[1], (int)vertices[2], (int)vertices[3], (int)vertices[4], (int)vertices[5], (int)vertices[6], (int)vertices[7]);
#endif
      if(lastFaceEntry[0]==NULL) //there was no previous direction
      {
        assert(lastFaceEntry[1]==NULL);
        assert(lastFaceEntry[2]==NULL);
        assert(lastFaceEntry[3]==NULL);
        //any order will do
        lastFaceEntry[0] = vertices[faceIndices[bestFace][0]];
        lastFaceEntry[1] = vertices[faceIndices[bestFace][1]];
        lastFaceEntry[2] = vertices[faceIndices[bestFace][2]];
        lastFaceEntry[3] = vertices[faceIndices[bestFace][3]];
        lastFaceZeroShoot = vertices[oppositeFaceVertices[bestFace][0]];
        return;
      }
      else if(lastFaceZeroShoot == vertices[faceIndices[bestFace][0]])
      {
        lastFaceEntry[0] = vertices[faceIndices[bestFace][0]];
        lastFaceEntry[1] = vertices[faceIndices[bestFace][1]];
        lastFaceEntry[2] = vertices[faceIndices[bestFace][2]];
        lastFaceEntry[3] = vertices[faceIndices[bestFace][3]];
        lastFaceZeroShoot = vertices[oppositeFaceVertices[bestFace][0]];
        return;
      }
      else if(lastFaceZeroShoot == vertices[faceIndices[bestFace][1]])
      {
        lastFaceEntry[0] = vertices[faceIndices[bestFace][1]];
        lastFaceEntry[1] = vertices[faceIndices[bestFace][2]];
        lastFaceEntry[2] = vertices[faceIndices[bestFace][3]];
        lastFaceEntry[3] = vertices[faceIndices[bestFace][0]];
        lastFaceZeroShoot = vertices[oppositeFaceVertices[bestFace][1]];
        return;
      }
      else if(lastFaceZeroShoot == vertices[faceIndices[bestFace][2]])
      {
        lastFaceEntry[0] = vertices[faceIndices[bestFace][2]];
        lastFaceEntry[1] = vertices[faceIndices[bestFace][3]];
        lastFaceEntry[2] = vertices[faceIndices[bestFace][0]];
        lastFaceEntry[3] = vertices[faceIndices[bestFace][1]];
        lastFaceZeroShoot = vertices[oppositeFaceVertices[bestFace][2]];
        return;
      }
      else if(lastFaceZeroShoot == vertices[faceIndices[bestFace][3]])
      {
        lastFaceEntry[0] = vertices[faceIndices[bestFace][3]];
        lastFaceEntry[1] = vertices[faceIndices[bestFace][0]];
        lastFaceEntry[2] = vertices[faceIndices[bestFace][1]];
        lastFaceEntry[3] = vertices[faceIndices[bestFace][2]];
        lastFaceZeroShoot = vertices[oppositeFaceVertices[bestFace][3]];
        return;
      }
      else
      {
        assert(false);
      }
    }
  }

}
#else
inline static bool isFaceInCache(SVvertex *a, SVvertex *b, SVvertex *c, SVvertex *d)
{
  return a->in_cache && b->in_cache && c->in_cache && d->in_cache;
}
#endif

#ifdef PICK_OLDEST_WITH_MOST_FACES_IN_EXTENDED_CACHE
inline int isFaceInTBCache(SVvertex *a, SVvertex *b, SVvertex *c, SVvertex *d)
{
  if(a->in_tb_cache && b->in_tb_cache && c->in_tb_cache && d->in_tb_cache)
    return 1;
  else
    return 0;
}
#endif

#ifdef PICK_BEST_OP_WITH_FACE_IN_CACHE
void* SVwriteBufferedHex::find_hexahedron(int dirty)
#elif defined(PICK_MOST_SPIRALLING)
SVhexahedron* find_hexahedron(int dirty, int &bestFace, int &bestFaceDirection)
#else
SVhexahedron* find_hexahedron(int dirty)
#endif
{

#ifdef PRINT_DEBUG_OUTPUT
#ifdef PICK_MOST_SPIRALLING
  fprintf(stderr, "\nbegin find_hexahedron() [currentSpirallingDirection=%i, previousSpirallingDirection=%i]\n", currentSpirallingDirection, previousSpirallingDirection);
#else
  fprintf(stderr, "\nbegin find_hexahedron()\n");
#endif
#endif
  SVhexahedron* best_hexahedron = NULL;

#ifdef PICK_OLDEST_WITH_FACE_IN_CACHE
  //for each vertex in the cache
  for(int i=0; i< 8; ++i)
  {
    SVvertex *cur_vertex = cache[i];
    if(cur_vertex)
    {
      //printf("trying vertex %i (%ith in cache)\n", cur_vertex->index, i);
      //for each incoming hexa of this vertex
      for(int j=0; j< cur_vertex->incoming_size; ++j)
      {
        SVhexahedron *cur_hexa = cur_vertex->incoming[j];
        if((cur_hexa->dirty != dirty) &&(best_hexahedron==NULL || cur_hexa->order_index < best_hexahedron->order_index))
        {
          cur_hexa->dirty = dirty;
          // check if a face has 4 vertices in cache, i.e. (except if I made a mistake) is in the face cache
          if(isFaceInCache(cur_hexa->vertices[0], cur_hexa->vertices[1], cur_hexa->vertices[2], cur_hexa->vertices[3]) || isFaceInCache(cur_hexa->vertices[0], cur_hexa->vertices[3], cur_hexa->vertices[7], cur_hexa->vertices[4]) || isFaceInCache(cur_hexa->vertices[0], cur_hexa->vertices[4], cur_hexa->vertices[5], cur_hexa->vertices[1]) || isFaceInCache(cur_hexa->vertices[1], cur_hexa->vertices[5], cur_hexa->vertices[6], cur_hexa->vertices[2]) || isFaceInCache(cur_hexa->vertices[2], cur_hexa->vertices[6], cur_hexa->vertices[7], cur_hexa->vertices[3]) || isFaceInCache(cur_hexa->vertices[4], cur_hexa->vertices[7], cur_hexa->vertices[6], cur_hexa->vertices[5]))
          {
            //printf("\t\tpicked.\n");
            best_hexahedron = cur_hexa;
          }
        }
      }
    }
  }
#elif defined(PICK_OLDEST_WITH_MOST_FACES_IN_EXTENDED_CACHE)
  int bestNbFacesInTB = 0;
  //for each vertex in the cache
  for(int i=0; i< 8; ++i)
  {
    SVvertex *cur_vertex = cache[i];
    if(cur_vertex)
    {
      //printf("trying vertex %i (%ith in cache)\n", cur_vertex->index, i);
      //for each incoming hexa of this vertex
      for(int j=0; j< cur_vertex->incoming_size; ++j)
      {
        SVhexahedron *cur_hexa = cur_vertex->incoming[j];
        if(cur_hexa->dirty != dirty)
        {
          cur_hexa->dirty = dirty;
          // check if a face has 4 vertices in cache, i.e. (except if I made a mistake) is in the face cache
          if(isFaceInCache(cur_hexa->vertices[0], cur_hexa->vertices[1], cur_hexa->vertices[2], cur_hexa->vertices[3]) || isFaceInCache(cur_hexa->vertices[0], cur_hexa->vertices[3], cur_hexa->vertices[7], cur_hexa->vertices[4]) || isFaceInCache(cur_hexa->vertices[0], cur_hexa->vertices[4], cur_hexa->vertices[5], cur_hexa->vertices[1]) || isFaceInCache(cur_hexa->vertices[1], cur_hexa->vertices[5], cur_hexa->vertices[6], cur_hexa->vertices[2]) || isFaceInCache(cur_hexa->vertices[2], cur_hexa->vertices[6], cur_hexa->vertices[7], cur_hexa->vertices[3]) || isFaceInCache(cur_hexa->vertices[4], cur_hexa->vertices[7], cur_hexa->vertices[6], cur_hexa->vertices[5]))
          {
            int nbFacesInTB = isFaceInTBCache(cur_hexa->vertices[0], cur_hexa->vertices[1], cur_hexa->vertices[2], cur_hexa->vertices[3]) + isFaceInTBCache(cur_hexa->vertices[0], cur_hexa->vertices[3], cur_hexa->vertices[7], cur_hexa->vertices[4]) + isFaceInTBCache(cur_hexa->vertices[0], cur_hexa->vertices[4], cur_hexa->vertices[5], cur_hexa->vertices[1]) + isFaceInTBCache(cur_hexa->vertices[1], cur_hexa->vertices[5], cur_hexa->vertices[6], cur_hexa->vertices[2]) + isFaceInTBCache(cur_hexa->vertices[2], cur_hexa->vertices[6], cur_hexa->vertices[7], cur_hexa->vertices[3]) + isFaceInTBCache(cur_hexa->vertices[4], cur_hexa->vertices[7], cur_hexa->vertices[6], cur_hexa->vertices[5]);
            if( (best_hexahedron == NULL) || (bestNbFacesInTB <= nbFacesInTB))
            {
              if(bestNbFacesInTB == nbFacesInTB)
              {
                //on equality, take the oldest
                if((best_hexahedron == NULL) || (cur_hexa->order_index < best_hexahedron->order_index))
                {
                  best_hexahedron = cur_hexa;
                  bestNbFacesInTB = nbFacesInTB;
                }
              }
              else
              {
                best_hexahedron = cur_hexa;
                bestNbFacesInTB = nbFacesInTB;
              }
            }
          }
        }
      }
    }
  }
#elif defined(PICK_MOST_VISITED_VERTICES_WITH_FACE_IN_CACHE)
  int best_visited = 0;
  //for each vertex in the cache
  for(int i=0; i< 8; ++i)
  {
    SVvertex *cur_vertex = cache[i];
    if(cur_vertex)
    {
      //printf("trying vertex %i (%ith in cache)\n", cur_vertex->index, i);
      //for each incoming hexa of this vertex
      for(int j=0; j< cur_vertex->incoming_size; ++j)
      {
        SVhexahedron *cur_hexa = cur_vertex->incoming[j];
        if((cur_hexa->dirty != dirty))
        {
          cur_hexa->dirty = dirty;
          // check if a face has 4 vertices in cache, i.e. (except if I made a mistake) is in the face cache
          if(isFaceInCache(cur_hexa->vertices[0], cur_hexa->vertices[1], cur_hexa->vertices[2], cur_hexa->vertices[3]) || isFaceInCache(cur_hexa->vertices[0], cur_hexa->vertices[3], cur_hexa->vertices[7], cur_hexa->vertices[4]) || isFaceInCache(cur_hexa->vertices[0], cur_hexa->vertices[4], cur_hexa->vertices[5], cur_hexa->vertices[1]) || isFaceInCache(cur_hexa->vertices[1], cur_hexa->vertices[5], cur_hexa->vertices[6], cur_hexa->vertices[2]) || isFaceInCache(cur_hexa->vertices[2], cur_hexa->vertices[6], cur_hexa->vertices[7], cur_hexa->vertices[3]) || isFaceInCache(cur_hexa->vertices[4], cur_hexa->vertices[7], cur_hexa->vertices[6], cur_hexa->vertices[5]))
          {
            //count the number of visited verts:
            int cur_visited = 0;
            for(int k=0; k<8; k++)
              if(cur_hexa->vertices[i]->index != -1)
                ++cur_visited;
            if(cur_visited > best_visited)
            {
              best_visited = cur_visited;
              best_hexahedron = cur_hexa;
            }
          }
        }
      }
    }
  }

#elif defined(PICK_BEST_OP_WITH_FACE_IN_CACHE)
  int bestOpValue = -1;
  //for each vertex in the cache
  for(int i=0; i< 8; ++i)
  {
    SVvertex *cur_vertex = cache[i];
    if(cur_vertex)
    {
#ifdef PRINT_DEBUG_OUTPUT
      fprintf(stderr, "vertex %i has %i incoming hexahedrons\n", i, cur_vertex->incoming_size);
#endif
      //for each incoming hexa of this vertex
      for(int j=0; j< cur_vertex->incoming_size; ++j)
      {
        SVhexahedron *cur_hexa = cur_vertex->incoming[j];
#ifdef PRINT_DEBUG_OUTPUT
        fprintf(stderr, "\thexa %i is %s\n", j, cur_hexa->dirty == dirty ? "dirty" : "not dirty");
#endif
        if(cur_hexa->dirty != dirty)
        {
          cur_hexa->dirty = dirty;
          // check if a face has 4 vertices in cache, i.e. (except if I made a mistake) is in the face cache
          if(isFaceInCache(cur_hexa->vertices[0], cur_hexa->vertices[1], cur_hexa->vertices[2], cur_hexa->vertices[3]) || isFaceInCache(cur_hexa->vertices[0], cur_hexa->vertices[3], cur_hexa->vertices[7], cur_hexa->vertices[4]) || isFaceInCache(cur_hexa->vertices[0], cur_hexa->vertices[4], cur_hexa->vertices[5], cur_hexa->vertices[1]) || isFaceInCache(cur_hexa->vertices[1], cur_hexa->vertices[5], cur_hexa->vertices[6], cur_hexa->vertices[2]) || isFaceInCache(cur_hexa->vertices[2], cur_hexa->vertices[6], cur_hexa->vertices[7], cur_hexa->vertices[3]) || isFaceInCache(cur_hexa->vertices[4], cur_hexa->vertices[7], cur_hexa->vertices[6], cur_hexa->vertices[5]))
          {
#ifdef PRINT_DEBUG_OUTPUT
            fprintf(stderr, "\t\thas face in cache\n");
#endif
            I32 c_idx[8];
            for(int k=0; k<8; ++k)
              c_idx[k] = cur_hexa->vertices[k]->index;
            int op = svwriter->test_write_hexahedron(c_idx);
#ifdef PRINT_DEBUG_OUTPUT
            fprintf(stderr, "\t\ttested and got a %s with value %i\n", opName[op], opValue[op]);
#endif
            if(best_hexahedron == NULL)
            {
#ifdef PRINT_DEBUG_OUTPUT
              fprintf(stderr, "\t\tno previous best: picked the %s with value %i\n", opName[op], opValue[op]);
#endif
              best_hexahedron = cur_hexa;
              bestOpValue = opValue[op];
            }
            else if(opValue[op] > bestOpValue)
            {
#ifdef PRINT_DEBUG_OUTPUT
              fprintf(stderr, "\t\tbetter opValue: picked the %s with value %i\n", opName[op], opValue[op]);
#endif
              best_hexahedron = cur_hexa;
              bestOpValue = opValue[op];
            }
            else if( (opValue[op] == bestOpValue) && (cur_hexa->order_index < best_hexahedron->order_index))
            {
#ifdef PRINT_DEBUG_OUTPUT
              fprintf(stderr, "\t\tsame opValue but older: picked the %s with value %i\n", opName[op], opValue[op]);
#endif
              best_hexahedron = cur_hexa;
            }
          }
#ifdef PRINT_DEBUG_OUTPUT
          else
          {
            fprintf(stderr, "\t\thas no face in cache\n");
          }
#endif
        }
      }
    }
  }
#elif defined(PICK_MOST_SPIRALLING)
  int bestScore = 0;
  bestFace = -1;
  bestFaceDirection = -1;
  //for each vertex in the cache
  for(int i=0; i< 8; ++i)
  {
    SVvertex *cur_vertex = cache[i];
    if(cur_vertex)
    {
#ifdef PRINT_DEBUG_OUTPUT
      //fprintf(stderr, "vertex %i has %i incoming hexahedrons\n", i, cur_vertex->incoming_size);
#endif
      //for each incoming hexa of this vertex
      for(int j=0; j< cur_vertex->incoming_size; ++j)
      {
        SVhexahedron *cur_hexa = cur_vertex->incoming[j];
#ifdef PRINT_DEBUG_OUTPUT
        //fprintf(stderr, "\thexa %i is %s\n", j, cur_hexa->dirty == dirty ? "dirty" : "not dirty");
        //fprintf(stderr, "\t\tvertices={%x, %x, %x, %x, %x, %x, %x, %x}\n", (int)cur_hexa->vertices[0], (int)cur_hexa->vertices[1], (int)cur_hexa->vertices[2], (int)cur_hexa->vertices[3], (int)cur_hexa->vertices[4], (int)cur_hexa->vertices[5], (int)cur_hexa->vertices[6], (int)cur_hexa->vertices[7]);
#endif
        if(cur_hexa->dirty != dirty)
        {
          cur_hexa->dirty = dirty;
          // check if a face has 4 vertices in cache, i.e. (except if I made a mistake) is in the face cache
          int cfDirection[6];
          for(int k=0; k<6; ++k)
            cfDirection[k] = faceInCache(cur_hexa->vertices[faceIndices[k][0]], cur_hexa->vertices[faceIndices[k][1]], cur_hexa->vertices[faceIndices[k][2]], cur_hexa->vertices[faceIndices[k][3]]);
          int direction = cfDirection[0] | cfDirection[1] | cfDirection[2] | cfDirection[3] | cfDirection[4] | cfDirection[5];

#ifdef PRINT_DEBUG_OUTPUT
          fprintf(stderr, "\t\tdirections: %i %i %i %i %i %i\n", cfDirection[0] , cfDirection[1] , cfDirection[2] , cfDirection[3] , cfDirection[4] , cfDirection[5]);
#endif
          if(direction==0)
          {
#ifdef PRINT_DEBUG_OUTPUT
            //fprintf(stderr, "\t\tno face in cache.\n");
#endif
            //no face in cache, worst case. Only pick if bestHex is NULL or if bestDirection is 10 and this hex is older.
            if((bestScore==0) && ((best_hexahedron==NULL) || (cur_hexa->order_index < best_hexahedron->order_index)))
            {
              best_hexahedron = cur_hexa;
            }
          }
          else if(direction & currentSpirallingDirection)
          {
            //there is a hex in the cache that spirals best, return it
#ifdef PRINT_DEBUG_OUTPUT
            fprintf(stderr, "\t\tcurrent spiralling direction\n");
#endif
            for(int k=0; k<6;++k)
            {
              if(cfDirection[k]==currentSpirallingDirection)
              {
                bestFaceDirection = currentSpirallingDirection;
                bestFace = k;
                return cur_hexa;
              }
            }
          }
          else if(direction & previousSpirallingDirection)
          {
            //there is a hex in the cache that spirals around the second best edge, pick it as curent best
#ifdef PRINT_DEBUG_OUTPUT
            fprintf(stderr, "\t\tprevious spiralling direction\n");
#endif
            for(int k=0; k<6;++k)
              if(cfDirection[k]==previousSpirallingDirection)
                bestFace = k;
            bestFaceDirection = previousSpirallingDirection;
            best_hexahedron = cur_hexa;
            bestScore = 100;
          }
          else if((bestScore<=99) && (direction & SPIRALLING_STRAIGHT))
          {
#ifdef PRINT_DEBUG_OUTPUT
            fprintf(stderr, "\t\tstraight\n");
#endif
            if((bestScore<99) || ((best_hexahedron==NULL) || (cur_hexa->order_index < best_hexahedron->order_index)))
            {
              bestFaceDirection = SPIRALLING_STRAIGHT;
              for(int k=0; k<6;++k)
                if(cfDirection[k]==SPIRALLING_STRAIGHT)
                  bestFace = k;
              best_hexahedron = cur_hexa;
              bestScore = 99;
            }
          }
          else if(bestScore<=98)//TODO: refine this case?
          {
#ifdef PRINT_DEBUG_OUTPUT
            fprintf(stderr, "\t\topposite direction\n");
#endif
            if((bestScore<98) || ((best_hexahedron==NULL) || (cur_hexa->order_index < best_hexahedron->order_index)))
            {
              bestFaceDirection = direction;
              for(int k=0; k<6;++k)
                if(cfDirection[k])
                  bestFace = k;
              best_hexahedron = cur_hexa;
              bestScore = 98;
            }
          }
#ifdef PRINT_DEBUG_OUTPUT
          else 
          {
            fprintf(stderr, "\t\tthere is a better one\n");
          }
#endif
        }
      }
    }
  }
#endif

  //printf("picked hexa %i\n", best_hexahedron ? best_hexahedron->order_index : -1);
#ifdef PRINT_DEBUG_OUTPUT
  fprintf(stderr, "end find_hexahedron()\n");
#endif
  return best_hexahedron;
}

void SVwriteBufferedHex::write_hexahedron_delayed()
{
  int i;
#ifdef PRINT_DEBUG_OUTPUT
  fprintf(stderr, "CELL %i\n", svwriter->c_count);
#endif
  I32 c_idx[8];
  bool c_final[8];
#ifdef PICK_MOST_SPIRALLING
  int bestFace, bestFaceDirection;
  SVhexahedron* hexahedron = (SVhexahedron*)find_hexahedron(svwriter->c_count, bestFace, bestFaceDirection);
#else
  SVhexahedron* hexahedron = (SVhexahedron*)find_hexahedron(svwriter->c_count);
#endif

  if (hexahedron)
  {
#ifdef PRINT_DEBUG_OUTPUT
    fprintf(stderr, "found a suitable hex.\n");
#endif
    waiting_queue->removeElement(hexahedron);
  }
  else
  {
#ifdef PRINT_DEBUG_OUTPUT
    fprintf(stderr, "did not find a suitable hex.\n");
#endif
    hexahedron =(SVhexahedron*) waiting_queue->getAndRemoveFirstElement();
  }

  removeFromVertices(hexahedron);

  SVvertex**vertices = hexahedron->vertices;

  for(i = 0; i < 8; i++)
  {
    //do not forget to put the state of the old cache vertices to not in cache:
#ifdef PICK_MOST_SPIRALLING
    if(cache[i]) cache[i]->in_cache = -1;
#else
    if(cache[i]) cache[i]->in_cache = false;
#endif
#ifdef PICK_OLDEST_WITH_MOST_FACES_IN_EXTENDED_CACHE
    //and update the before caches
    if(tb_cache[i]) tb_cache[i]->in_tb_cache = false;
    tb_cache[i] = b_cache[i];
    b_cache[i] = cache[i];
#endif

    if(vertices[i]->index == -1)
    {
      vertices[i]->index = svwriter->v_count;
      svwriter->write_vertex(vertices[i]->prop_f);
#ifdef PRINT_CONTROL_OUTPUT
      cur_out_width++;
      if(cur_out_width > max_out_width) max_out_width = cur_out_width;
#endif
    }
    c_idx[i] = vertices[i]->index;
    if(vertices[i]->finalized && vertices[i]->incoming_size == 0)
    {
      c_final[i] = true;
      deallocVertex(vertices[i]);
      cache[i] = NULL;
#ifdef PRINT_CONTROL_OUTPUT
      cur_out_width--;
      if(( svwriter->v_count-c_idx[i]+1) > max_out_span) max_out_span =(svwriter->v_count-c_idx[i]+1);
#endif
    }
    else
    {
      c_final[i] = false;
      cache[i] = vertices[i];
    }
  }

  //write hexahedron
  svwriter->write_hexahedron(c_idx, hexahedron->prop_f, c_final);

  //now we can deallocate it
  deallocHexahedron(hexahedron);

  for(i = 0; i < 8; i++)
  {
#ifdef PICK_MOST_SPIRALLING
    if(cache[i]) cache[i]->in_cache = i;
#else
    if(cache[i]) cache[i]->in_cache = true;
#endif
#ifdef PICK_OLDEST_WITH_MOST_FACES_IN_EXTENDED_CACHE
    if(tb_cache[i]) tb_cache[i]->in_tb_cache = true;
#endif
    //printf("setting vertices[%i]->in_cache = true\n", vertices[i]->index);
  }

#ifdef PICK_MOST_SPIRALLING
  updateSpirallingState(vertices, bestFace, bestFaceDirection);
#endif
}

void SVwriteBufferedHex::write_vertex(const float* v_data_f)
{
  if (v_count == 0) initVertexBuffer(1024, v_pnum_f);
  SVvertex* vertex = allocVertex();
  VecCopyNfv(vertex->prop_f, v_data_f, v_pnum_f);
  vertex_hash->insert(my_hash::value_type(v_count, vertex));
  v_count++;
#ifdef PRINT_CONTROL_OUTPUT
  if(vertex_hash->size() > max_in_width) max_in_width = vertex_hash->size();
#endif
}

void SVwriteBufferedHex::write_hexahedron(const I32* c_idx, const bool* c_final)
{
  fprintf(stderr,"FATAL ERROR: write_hexahedron(const int* c_idx, const bool* c_final) not implemented. Use write_hexahedron(const int* c_idx, BULL, const bool* c_final) instead.\n");
  exit(0);
}

void SVwriteBufferedHex::write_tetrahedron(const I32* c_idx, const bool* c_final)
{
  fprintf(stderr,"FATAL ERROR: write_tetrahedron(const int* c_idx, const bool* c_final) not implemented\n");
  exit(0);
}

void SVwriteBufferedHex::write_hexahedron(const I32* c_idx, const float* c_data_f, const bool* c_final)
{
  int i;
  my_hash::iterator hash_elements[8];

  if(c_final == 0)
  {
    fprintf(stderr,"FATAL ERROR: need immediate finalization but c_final is a zero pointer\n");
    exit(0);
  }

  if (c_count == 0) initHexahedronBuffer(1024, c_pnum_f);
  SVhexahedron* hexahedron = allocHexahedron();
  if(c_pnum_f)
  {
    if(c_data_f)
      VecCopyNfv(hexahedron->prop_f, c_data_f, c_pnum_f);
    else
    {
      fprintf(stderr, "WARNING (SVwriteBufferedHex): c_pnum_f==%i but write_hexahedron was called with c_prop_f==NULL\n", c_pnum_f);
    }
  }

  hexahedron->order_index = c_count;

  // get vertices from hash
  for(i = 0; i < 8; i++)
  {
    hash_elements[i] = vertex_hash->find(c_idx[i]);
    if(hash_elements[i] == vertex_hash->end())
    {
      fprintf(stderr,"ERROR (SVwriteBufferedHex): vertex not in hash. need pre-order mesh\n");
      exit(0);
    }
    else
    {
      hexahedron->vertices[i] =(*hash_elements[i]).second;
    }
  }

  addToVertices(hexahedron);

  // check for finalization
  for(i = 0; i < 8; i++)
  {
    if(c_final[i])
    {
      hexahedron->vertices[i]->finalized = true;
      vertex_hash->erase(hash_elements[i]);
#ifdef PRINT_CONTROL_OUTPUT
      if(( v_count - c_idx[i] + 1) > max_in_span) max_in_span =(v_count - c_idx[i] + 1);
#endif
    }
  }

  // add hexahedron to buffer
  waiting_queue->addElement(hexahedron);

  // compress a hexahedron if the buffer is full
  if(waiting_queue->elements() == max_delay)
  {

    write_hexahedron_delayed();
    assert(waiting_queue->elements()==max_delay-1);
  }

  c_count++;
}

void SVwriteBufferedHex::write_tetrahedron(const I32* c_idx, const float* c_data_f, const bool* c_final)
{
  fprintf(stderr,"FATAL ERROR: write_tetrahedron(const int* c_idx, const float* c_data_f, const bool* c_final) not implemented\n");
  exit(0);
}

void SVwriteBufferedHex::write_finalized(I32 final_idx)
{
  fprintf(stderr,"FATAL ERROR: need immediate finalization (e.g. a tail-compact pre-order mesh)\n");
  exit(0);
}

SVwriteBufferedHex::SVwriteBufferedHex()
{
  // init of SVwriteBuffered
  svwriter = 0;
  max_delay = -1;
}

SVwriteBufferedHex::~SVwriteBufferedHex()
{
  // clean-up for SVwriteBuffered
  v_pmin_f = 0;
  v_pmax_f = 0;
  c_pmin_f = 0;
  c_pmax_f = 0;
  if (svwriter) delete svwriter;
}
