/*
===============================================================================

  FILE:  svwriter_hzip.cpp

  CONTENTS:

    see corresponding header file

  PROGRAMMERS:

    clement.courbet@ecp.fr

  COPYRIGHT:

    copyright (C) 2009  clement.courbet@ecp.fr

    This software is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.

  CHANGE HISTORY:

    see corresponding header file

===============================================================================
*/
#include "svwriter_hzip.h"

#ifdef _WIN32
#include <fcntl.h>
#include <io.h>
#endif

#include <stdlib.h>
#include <string.h>
#include <assert.h>

#include "vecnfv.h"
#include "vecniv.h"

//this is for compatibility with the new g++
#ifdef __GNUC__
	#include <ext/hash_map>
	namespace std { using namespace __gnu_cxx; }
	using namespace std;
#else
	#include <hash_map.h>
#endif

#define HZIP_CUBIT_PERMUTATION 0x67542310

//#define PRINT_DEBUG_OUTPUT
//#define OUTPUT_COORDINATES

#ifdef OUTPUT_COORDINATES
FILE * coordFile=NULL;
#endif


//HZipNode struture
typedef struct HZipVertex
{
	HZipVertex* buffer_next;    // used for efficient memory management and by dynamicvector data structure
	float* prop_f;
	bool coded;
	bool finalize;
} HZipVertex;

typedef hash_map<int32_t, HZipVertex*> my_vertex_hash;
my_vertex_hash* vertex_hash;


// efficient memory allocation
static int vertex_pnum_f = 0;
static int vertex_buffer_size = 0;
static int vertex_buffer_alloc = 16;
static HZipVertex* vertex_buffer_next = 0;

static int initVertexBuffer ( int size, int32_t pnum_f )
{
	vertex_buffer_next = ( HZipVertex* ) malloc ( sizeof ( HZipVertex ) *size );
	vertex_pnum_f = pnum_f;

	if ( vertex_buffer_next == 0 )
	{
		fprintf ( stderr,"malloc for vertex buffer failed\n" );
		return 0;
	}
	if ( vertex_pnum_f )
	{
		vertex_buffer_next[0].prop_f = ( float* ) malloc ( sizeof ( float ) *size*vertex_pnum_f );
		if ( vertex_buffer_next[0].prop_f == 0 )
		{
			fprintf ( stderr,"malloc for vertex_buffer_next->prop_f failed\n" );
			return 0;
		}
	}

	for ( int i = 0; i < size; i++ )
	{
		vertex_buffer_next[i].buffer_next = & ( vertex_buffer_next[i+1] );
		if ( vertex_pnum_f )
		{
			vertex_buffer_next[i].prop_f = & ( vertex_buffer_next[0].prop_f[i*vertex_pnum_f] );
		}
	}
	vertex_buffer_next[size-1].buffer_next = 0;
	vertex_buffer_alloc = size;
	vertex_buffer_size = 0;
	return 1;
}

static HZipVertex* allocVertex()
{
	if ( vertex_buffer_next == 0 )
	{
		vertex_buffer_next = ( HZipVertex* ) malloc ( sizeof ( HZipVertex ) *vertex_buffer_alloc );
		if ( vertex_buffer_next == 0 )
		{
			fprintf ( stderr,"malloc for vertex buffer failed\n" );
			return 0;
		}
		if ( vertex_pnum_f )
		{
			vertex_buffer_next[0].prop_f = ( float* ) malloc ( sizeof ( float ) *vertex_buffer_alloc*vertex_pnum_f );
			if ( vertex_buffer_next[0].prop_f == 0 )
			{
				fprintf ( stderr,"malloc for vertex_buffer_next->prop_f failed\n" );
				return 0;
			}
		}

		for ( int i = 0; i < vertex_buffer_alloc; i++ )
		{
			vertex_buffer_next[i].buffer_next = & ( vertex_buffer_next[i+1] );
			if ( vertex_pnum_f )
			{
				vertex_buffer_next[i].prop_f = & ( vertex_buffer_next[0].prop_f[i*vertex_pnum_f] );
			}
		}
		vertex_buffer_next[vertex_buffer_alloc-1].buffer_next = 0;
		vertex_buffer_alloc = 2*vertex_buffer_alloc;
	}
	// get pointer to next available vertex
	HZipVertex* vertex = vertex_buffer_next;
	vertex_buffer_next = vertex->buffer_next;
	vertex_buffer_size++;

	vertex->coded = false;
	vertex->finalize = false;

#ifdef PRINT_CONTROL_OUTPUT
	if ( vertex_buffer_size > vertex_buffer_maxsize ) vertex_buffer_maxsize = vertex_buffer_size;
#endif

	return vertex;
}

static void deallocVertex ( HZipVertex* vertex )
{
	vertex->buffer_next = vertex_buffer_next;
	vertex_buffer_next = vertex;
	vertex_buffer_size--;
}



bool SVwriter_hzip::open ( FILE* file )
{
#ifndef MULTIPLEX
	fprintf ( stderr, "WARNING: SVwriter_hzip ignores the file argument, and outputs the connectivity and geometry in hzip_conn and hzip_geo respectively.\n" );
#endif

	if ( file == 0 )
	{
		fprintf ( stderr, "ERROR: zero file pointer not supported by SVwriter_hzip\n" );
		exit ( 0 );
	}

#ifdef _WIN32
	if ( file == stdout )
	{
		if ( _setmode ( _fileno ( stdout ), _O_TEXT ) == -1 )
		{
			fprintf ( stderr, "ERROR: cannot set stdout to text (translated) mode\n" );
			exit ( 0 );
		}
	}
#endif

// 	if(ncells<=0)
// 	{
// 		fprintf(stderr, "ERROR: SVwriter_hzip needs to know the number of cells in advance.\n");
// 		return false;
// 	}
// 	if(nverts<=0)
// 	{
// 		fprintf(stderr, "ERROR: SVwriter_hzip needs to know the number of vertices in advance.\n");
// 		return false;
// 	}

#ifndef MULTIPLEX
	conn_file = fopen ( "hzip_conn", "wb" );
	if ( !conn_file )
	{
		fprintf ( stderr, "ERROR: Cannot open hzip_conn for writing.\n" );
		return false;
	}
#endif
	
	this->file = file;
	v_count = 0;
	c_count = 0;

	if(vertex_hash) delete vertex_hash;
	vertex_hash = new my_vertex_hash();

	return true;
}

void SVwriter_hzip::write_vertex ( const float* v_prop_f )
{
	if(!v_count && !c_count)
		write_header();

	HZipVertex* vertex = allocVertex();
	vertex_hash->insert(my_vertex_hash::value_type(v_count, vertex));
	memcpy(vertex->prop_f, v_prop_f, v_pnum_f*sizeof(float));
	v_count++;
}

void SVwriter_hzip::write_tetrahedron ( const int32_t* c_idx, const bool* c_final )
{
	fprintf ( stderr, "ERROR: write_tetrahedron() not implemented.\n " );
	exit ( EXIT_FAILURE );
}

void SVwriter_hzip::write_tetrahedron ( const int* c_idx, const float* c_prop_f, const bool* c_final )
{
	fprintf ( stderr, "ERROR: write_tetrahedron() not implemented.\n " );
	exit ( EXIT_FAILURE );
}

void SVwriter_hzip::write_hexahedron ( const int32_t* c_idx, const bool* c_final )
{
	//get vertices from the hash
	my_vertex_hash::iterator hash_elements[8];
	HZipVertex* vertices[8];
	
#ifdef PRINT_DEBUG_OUTPUT
	fprintf(stderr, "writing hex ");
	for(int i=0; i<8; ++i)
	{
		fprintf(stderr, "%i ", c_idx[i]);
	}
	fprintf(stderr, "\n");
#endif

	// get vertices from hash
	for ( int i = 0; i < 8; i++ )
	{
		hash_elements[i] = vertex_hash->find ( c_idx[i] );
		if ( hash_elements[i] == vertex_hash->end() )
		{
			fprintf ( stderr,"FATAL ERROR: vertex %d with index %d not in hash\n", i, c_idx[i] );
			exit ( EXIT_FAILURE );
		}
		vertices[i] = ( *hash_elements[i] ).second;
		vertices[i]->finalize = c_final[i];
	}

	//check that the input is compact, pre-order
	for(int i=0; i<8; ++i)
	{
		if(c_idx[i]>=v_count) fprintf(stderr, "WARNING: the output will not be pre-order.\n");
	}

	//write indices
	hzip_mesh_write_cell(conn_stream, c_idx);

	//write properties of not yet compressed vertices
	static float s_v_prop[8];
	static bool coded[8];

	for(int k=0; k<v_pnum_f; ++k)
	{
		for(int i=0; i<8; ++i)
		{
			s_v_prop[i] = vertices[i]->prop_f[k];
			coded[i] = vertices[i]->coded;
		}
		hzip_node_write_unstreamed_cell_vertices(geo_stream[k], s_v_prop, coded);
		for(int i=0; i<8; ++i) assert(coded[i]);
	}

#ifdef OUTPUT_COORDINATES
	for(int i=0; i<8; ++i)
	{
		if(!vertices[i]->coded)
			fprintf(coordFile, "%i %f %f %f\n", c_idx[i], vertices[i]->prop_f[0], vertices[i]->prop_f[1], vertices[i]->prop_f[2]);
	}
#endif

	//delete finalized vertices, and mark vertices as coded:
	for(int i=0; i<8; ++i)
	{
		vertices[i]->coded = true;
		if(vertices[i]->finalize)
		{
			deallocVertex(vertices[i]);
			vertex_hash->erase(hash_elements[i]);
		}
	}

	c_count++;
}

void SVwriter_hzip::write_hexahedron ( const int32_t* c_idx, const float* c_prop_f, const bool* c_final )
{
	write_hexahedron ( c_idx, c_final );
}

void SVwriter_hzip::write_finalized ( int32_t final_idx )
{
	fprintf ( stderr, "WARNING: SVwriter_hzip does not support delayed finalization.\n " );
}

void SVwriter_hzip::write_header (  )
{
	param = hzm_codec_zlib;
	geo_stream = new HZNstream*[v_pnum_f];

#ifdef MULTIPLEX
	conn_stream = hzip_mesh_create_file ( file, HZIP_CUBIT_PERMUTATION, 3, ncells==-1 ? 0 : ncells, HZM_CODEC_ZLIB, &param );
	for(int i=0; i<v_pnum_f; ++i)
	{
	// 	if ( v_pmin_f && v_pmax_f )
	// 		geo_stream[i] = hzip_node_create_file ( geo_file[i], HZIP_CUBIT_PERMUTATION, 0, 3, nverts==-1 ? 0 : nverts, hzINT, HZM_CODEC_ZLIB, 0 );
	// 	else
			geo_stream[i] = hzip_node_create_file ( file, HZIP_CUBIT_PERMUTATION, 0, 3, nverts==-1 ? 0 : nverts, hzFLOAT, HZM_CODEC_ZLIB, 0 );
	}
#else
	conn_stream = hzip_mesh_create_file ( conn_file, HZIP_CUBIT_PERMUTATION, 3, ncells==-1 ? 0 : ncells, HZM_CODEC_ZLIB, &param );
	geo_file = new FILE*[v_pnum_f];
	char buf[1024];
	for(int i=0; i<v_pnum_f; ++i)
	{
		sprintf(buf, "hzip_geo_%i", i);
		geo_file[i] = fopen ( buf, "wb" );
		if ( !geo_file[i] )
		{
			fprintf ( stderr, "ERROR: Cannot open %s for writing.\n", buf );
			exit(EXIT_FAILURE);
		}

	// 	if ( v_pmin_f && v_pmax_f )
	// 		geo_stream[i] = hzip_node_create_file ( geo_file[i], HZIP_CUBIT_PERMUTATION, 0, 3, nverts==-1 ? 0 : nverts, hzINT, HZM_CODEC_ZLIB, 0 );
	// 	else
			geo_stream[i] = hzip_node_create_file ( geo_file[i], HZIP_CUBIT_PERMUTATION, 0, 3, nverts==-1 ? 0 : nverts, hzFLOAT, HZM_CODEC_ZLIB, 0 );
	}
#endif //MULTIPLEX

	initVertexBuffer(1024, v_pnum_f);

#ifdef OUTPUT_COORDINATES
	coordFile = fopen("coords", "wb");
#endif

}

void SVwriter_hzip::close ( bool close_file, bool update )
{
	int conn_outbytes, geo_outbytes;
	if ( (conn_outbytes = hzip_mesh_close ( conn_stream )) <0 )
	{
		fprintf ( stderr, "ERROR: hzip_mesh_close() failed.\n " );
		exit(EXIT_FAILURE);
	}
#ifndef MULTIPLEX
	fclose ( conn_file );
#endif

	geo_outbytes = 0;
	for(int i=0; i<v_pnum_f; ++i)
	{
		int t_geo_outbytes = hzip_node_close ( geo_stream[i] );
		if ( t_geo_outbytes <0 )
		{
			fprintf ( stderr, "ERROR: hzip_node_close() failed.\n " );
			exit(EXIT_FAILURE);
		}
		geo_outbytes += t_geo_outbytes;
#ifndef MULTIPLEX
		fclose ( geo_file[i] );
#endif
	}
#ifndef MULTIPLEX
	delete[] geo_file;
#endif
	delete[] geo_stream;
	
	if ( close_file )
		fclose ( file );

	if(vertex_hash)
	{
		delete vertex_hash;
		vertex_hash = NULL;
	}

	fprintf(stderr, "Connectivity: %.2f bph\n", (8.0*conn_outbytes)/c_count);
	fprintf(stderr, "Geometry: %.2f bpv\n", (8.0*geo_outbytes)/v_count);

	// close of SVwriter interface
	if ( nverts != -1 ) if ( nverts != v_count )  fprintf ( stderr,"WARNING: set nverts %d but v_count %d\n",nverts,v_count );
	if ( ncells != -1 ) if ( ncells != c_count )  fprintf ( stderr,"WARNING: set ncells %d but c_count %d\n",ncells,c_count );
	nverts = v_count;
	ncells = c_count;
	v_count = -1;
	c_count = -1;


#ifdef OUTPUT_COORDINATES
	fclose(coordFile);
#endif
}

SVwriter_hzip::SVwriter_hzip()
{
	// init of SVwriter_hzip
	file = 0;
}

SVwriter_hzip::~SVwriter_hzip()
{
	// clean-up for SVwriter_hzip
}
