/*
===============================================================================

  FILE:  SVwriter_off.cpp
  
  CONTENTS:

    see corresponding header file

  PROGRAMMERS:

    martin isenburg@cs.unc.edu

  COPYRIGHT:

    copyright (C) 2005  martin isenburg@cs.unc.edu

    This software is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.

  CHANGE HISTORY:

    see corresponding header file

===============================================================================
*/
#include "svwriter_off.h"

#ifdef _WIN32
#include <fcntl.h>
#include <io.h>
#endif

#include <stdlib.h>
#include <string.h>

#include "vecnfv.h"
#include "vecniv.h"

bool SVwriter_off::open(FILE* file)
{
  if (file == 0)
  {
    return false;
  }

#ifdef _WIN32
  if (file == stdout)
  {
    if(_setmode( _fileno( stdout ), _O_TEXT ) == -1 )
    {
      fprintf(stderr, "ERROR: cannot set stdout to text (translated) mode\n");
    }
  }
#endif

  this->file = file;
  v_count = 0;
  c_count = 0;
  vertex_buffer_alloc = 0;
  vertex_buffer = 0;
  cell_buffer_alloc = 0;
  cell_buffer = 0;
  return true;
}

void SVwriter_off::write_vertex(const float* v_prop_f)
{
  if (v_count == vertex_buffer_alloc)
  {
    if (vertex_buffer_alloc == 0)
    {
      if (nverts != -1)
      {
        vertex_buffer = (float*)realloc(vertex_buffer, sizeof(float)*v_pnum_f*nverts);
        vertex_buffer_alloc = nverts;
      }
      else
      {
        vertex_buffer = (float*)realloc(vertex_buffer, sizeof(float)*v_pnum_f*1024);
        vertex_buffer_alloc = 1024;
      }
    }
    else
    {
      vertex_buffer = (float*)realloc(vertex_buffer, sizeof(float)*v_pnum_f*vertex_buffer_alloc*2);
      vertex_buffer_alloc = vertex_buffer_alloc * 2;
    }
  }
  VecCopyNfv(&(vertex_buffer[v_count*v_pnum_f]),v_prop_f,v_pnum_f);

  v_count++;
}

void SVwriter_off::write_tetrahedron(const I32* c_idx, const bool* c_final)
{
  if (c_count == cell_buffer_alloc)
  {
    if (cell_buffer_alloc == 0)
    {
      if (ncells != -1)
      {
        cell_buffer = (int*)realloc(cell_buffer, sizeof(int)*4*ncells);
        cell_buffer_alloc = ncells;
      }
      else
      {
        cell_buffer = (int*)realloc(cell_buffer, sizeof(int)*4*2048);
        cell_buffer_alloc = 2048;
      }
    }
    else
    {
      cell_buffer = (int*)realloc(cell_buffer, sizeof(int)*4*cell_buffer_alloc*2);
      cell_buffer_alloc = cell_buffer_alloc * 2;
    }
  }
  VecCopyNiv(&(cell_buffer[c_count*4]),c_idx,4);

  c_count++;
}

void SVwriter_off::write_tetrahedron(const I32* c_idx, const float* c_prop_f, const bool* c_final)
{
  write_tetrahedron(c_idx, c_final);
}

void SVwriter_off::write_hexahedron(const I32* c_idx, const bool* c_final)
{
  fprintf(stderr, "WARNING: SVwriter_off does not implement write_hexahedron()\n");
}

void SVwriter_off::write_hexahedron(const I32* c_idx, const float* c_prop_f, const bool* c_final)
{
  fprintf(stderr, "WARNING: SVwriter_off does not implement write_hexahedron()\n");
}

void SVwriter_off::write_finalized(I32 final_idx)
{
}

void SVwriter_off::close(bool close_file, bool update_header)
{
  // close SVwriter_off
  int i;
  // write header
  fprintf(file, "OFF\012");
  fprintf(file, "%d %d 0\012",v_count,c_count);
  // write vertices
  for (i = 0; i < v_count; i++)
  {
    if (v_pnum_f == 3)
    {
      fprintf(file, "%.8g %.8g %.8g\012",vertex_buffer[3*i+0],vertex_buffer[3*i+1],vertex_buffer[3*i+2]);
    }
    else if (v_pnum_f == 4)
    {
      fprintf(file, "%.8g %.8g %.8g %.8g\012",vertex_buffer[4*i+0],vertex_buffer[4*i+1],vertex_buffer[4*i+2],vertex_buffer[4*i+3]);
    }
    else if (v_pnum_f == 5)
    {
      fprintf(file, "%.8g %.8g %.8g %.8g %.8g\012",vertex_buffer[5*i+0],vertex_buffer[5*i+1],vertex_buffer[5*i+2],vertex_buffer[5*i+3],vertex_buffer[5*i+4]);
    }
    else
    {
      fprintf(stderr, "ERROR: SVwriter_off::v_pnum_f == %d not implemented.\n",v_pnum_f);
      exit(0);
    }
  }
  // write tetrahedra
  for (i = 0; i < c_count; i++)
  {
    fprintf(file, "%d %d %d %d\012", cell_buffer[4*i+0],cell_buffer[4*i+1],cell_buffer[4*i+2],cell_buffer[4*i+3]);
  }

  vertex_buffer_alloc = 0;
  if (vertex_buffer) free(vertex_buffer);
  vertex_buffer = 0;

  cell_buffer_alloc = 0;
  if (cell_buffer) free(cell_buffer);
  cell_buffer = 0;

  if (close_file && file && file != stdout) fclose(file);
  file = 0;

  // close of SVwriter interface
  if (nverts != -1) if (nverts != v_count)  fprintf(stderr,"WARNING: set nverts %d but v_count %d\n",nverts,v_count);
  if (ncells != -1) if (ncells != c_count)  fprintf(stderr,"WARNING: set ncells %d but c_count %d\n",ncells,c_count);
  nverts = v_count;
  ncells = c_count;
  v_count = -1;
  c_count = -1;
}

SVwriter_off::SVwriter_off()
{
  // init of SVwriter_off
  file = 0;
  vertex_buffer_alloc = 0;
  vertex_buffer = 0;
  cell_buffer_alloc = 0;
  cell_buffer = 0;
}

SVwriter_off::~SVwriter_off()
{
  // clean-up for SVwriter_off
}
