/*
===============================================================================

  FILE:  SVwriter_svc.cpp
  
  CONTENTS:

    see corresponding header file

  PROGRAMMERS:

    martin isenburg@cs.unc.edu

  COPYRIGHT:

    copyright (C) 2005  martin isenburg@cs.unc.edu

    This software is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.

  CHANGE HISTORY:

    see corresponding header file

===============================================================================
*/
#include "svwriter_svc.h"

#ifdef _WIN32
#include <fcntl.h>
#include <io.h>
#endif

#include <stdlib.h>
#include <string.h>
#include <assert.h>

//this is for compatibility with the new g++
#ifdef __GNUC__
	#include <ext/hash_map>
	namespace std { using namespace __gnu_cxx; }
	using namespace std;
#else
	#include <hash_map.h>
#endif

#define PRINT_CONTROL_OUTPUT
#undef PRINT_CONTROL_OUTPUT

#define PRINT_DEBUG_OUTPUT
#undef PRINT_DEBUG_OUTPUT

#define SVC_CALCULATE_QUANTIZATION_ERROR
#undef SVC_CALCULATE_QUANTIZATION_ERROR

#define SVC_CALCULATE_PREDICTION_ERROR 
#undef SVC_CALCULATE_PREDICTION_ERROR

//code that is common to svreader_svh.cpp and svwriter_svh.cpp
#include "svcommon_svc.cpp"

// rangeencoder

#include "rangeencoder.h"

static RangeEncoder* re_conn;
static RangeEncoder* re_conn_op;
static RangeEncoder* re_conn_cache;
static RangeEncoder* re_conn_index;
static RangeEncoder* re_conn_candi;
static RangeEncoder* re_conn_final;

static RangeEncoder* re_geom;

// data structures used for streaming compression

typedef hash_map<I32, SVvertex*> my_vertex_hash;
static my_vertex_hash* vertex_hash;

// statistics

static int pos_vol = 0;
static int neg_vol = 0;

static int predict_none = 0;
static int predict_last = 0;
static int predict_inside = 0;
#if defined(SVC_USE_LOCAL_COORDINATE_SYSTEM)
static int predict_across = 0;
#endif

static int normal_not_valid_counter = 0;

static int op_start = 0;
static int op_start_x[] = {0,0,0,0,0};

#ifdef PRINT_CONTROL_OUTPUT

static int op_add = 0;
static int op_join = 0;
static int op_flip = 0;
static int op_fill = 0;
static int op_end = 0;

static int used_index = 0;
static int used_cache = 0;

static int start_hit_vertex = 0;
static int start_hit_edge_possible = 0;
static int start_miss = 0;

static int add_hit_face = 0;
static int add_hit_edge = 0;
static int add_hit_vertex = 0;
static int add_miss = 0;

static int join_hit_face = 0;
static int join_hit_edge = 0;
static int join_hit_vertex = 0;
static int join_hit_join_vertex = 0;
static int join_miss = 0;

static int join_edge_hist[] = {0,0,0,0,0,0,0,0,
                               0,0,0,0,0,0,0,0,
                               0,0,0,0,0,0,0,0,
                               0,0,0,0,0,0,0,0,
                               0,0,0,0,0,0,0,0,
                               0,0,0,0,0,0,0,0,
                               0,0,0,0,0,0,0,0,
                               0,0,0,0,0,0,0,0};

static int flip_hit_face = 0;
static int flip_hit_edge = 0;
static int flip_hit_vertex = 0;
static int flip_hit_edge_possible = 0;
static int flip_hit_edge_hist[] = {0,0,0,0,0};
static int flip_miss = 0;

static int fill_hit_face = 0;
static int fill_hit_edge = 0;
static int fill_hit_vertex = 0;
static int fill_hit_edge_possible = 0;
static int fill_hit_edge_hist[] = {0,0,0,0,0};
static int fill_miss = 0;

static int end_miss = 0;
static int end_hit_face = 0;
static int end_hit_vertex = 0;
#endif


static void initEncoder(FILE* file)
{
  if (file)
  {
    re_conn = new RangeEncoder(file);
    re_conn_op = re_conn;
    re_conn_cache = re_conn;
    re_conn_index = re_conn;
    re_conn_candi = re_conn;
    re_conn_final = re_conn;
    re_geom = re_conn;
  }
  else
  {
    re_conn = new RangeEncoder(0,false);
    re_conn_op = new RangeEncoder(0,false);
    re_conn_cache = new RangeEncoder(0,false);
    re_conn_index = new RangeEncoder(0,false);
    re_conn_candi = new RangeEncoder(0,false);
    re_conn_final = new RangeEncoder(0,false);
    re_geom = new RangeEncoder(0,false);
  }
}

static void finishEncoder(I32 nverts, I32 ncells)
{
  if (re_conn != re_conn_op)
  {
    re_conn_op->done();
    re_conn_cache->done();
    re_conn_index->done();
    re_conn_candi->done();
    re_conn_final->done();
    re_conn->done();
    re_geom->done();

    fprintf(stderr,"\n");
    fprintf(stderr,"*** compression rate details ***\n");
    fprintf(stderr,"\n");
    fprintf(stderr,"op:   \t%6.3f bpv \t%6.3f bpt\n", 8.0f/nverts*re_conn_op->getNumberBytes(),  8.0f/ncells*re_conn_op->getNumberBytes());
    fprintf(stderr,"cache:\t%6.3f bpv \t%6.3f bpt\n", 8.0f/nverts*re_conn_cache->getNumberBytes(), 8.0f/ncells*re_conn_cache->getNumberBytes());
    fprintf(stderr,"index:\t%6.3f bpv \t%6.3f bpt\n", 8.0f/nverts*re_conn_index->getNumberBytes(), 8.0f/ncells*re_conn_index->getNumberBytes());
    fprintf(stderr,"candi:\t%6.3f bpv \t%6.3f bpt\n", 8.0f/nverts*re_conn_candi->getNumberBytes(), 8.0f/ncells*re_conn_candi->getNumberBytes());
    fprintf(stderr,"final:\t%6.3f bpv \t%6.3f bpt\n", 8.0f/nverts*re_conn_final->getNumberBytes(), 8.0f/ncells*re_conn_final->getNumberBytes());
    fprintf(stderr,"rest: \t%6.3f bpv \t%6.3f bpt\n", 8.0f/nverts*re_conn->getNumberBytes(), 8.0f/ncells*re_conn->getNumberBytes());
    fprintf(stderr,"\n");
    fprintf(stderr,"conn: \t%6.3f bpv \t%6.3f bpt\n", 8.0f/nverts*(re_conn_op->getNumberBytes()+re_conn_cache->getNumberBytes()+re_conn_index->getNumberBytes()+re_conn_candi->getNumberBytes()+re_conn_final->getNumberBytes()+re_conn->getNumberBytes()), 8.0f/ncells*(re_conn_op->getNumberBytes()+re_conn_cache->getNumberBytes()+re_conn_index->getNumberBytes()+re_conn_candi->getNumberBytes()+re_conn_final->getNumberBytes()+re_conn->getNumberBytes()));
//    fprintf(stdout,"c: %5.2f bpt ", 8.0f/ncells*(re_conn_op->getNumberBytes()+re_conn_cache->getNumberBytes()+re_conn_index->getNumberBytes()+re_conn_candi->getNumberBytes()+re_conn_final->getNumberBytes()+re_conn->getNumberBytes()));
    fprintf(stderr,"geom: \t%6.3f bpv ", 8.0f/nverts*re_geom->getNumberBytes());
//   fprintf(stdout,"g: %5.2f bpv ", 8.0f/nverts*re_geom->getNumberBytes());
//    fprintf(stdout,"n %d l %d ", predict_none, predict_last);
//    fprintf(stdout,"%d [%d %d %d %d %d] ",op_start,op_start_x[0],op_start_x[1],op_start_x[2],op_start_x[3],op_start_x[4]);
    fprintf(stderr,"\n");
    fprintf(stderr,"total:\t%6.3f bpv\n", 8.0f/nverts*(re_geom->getNumberBytes()+re_conn_op->getNumberBytes()+re_conn_cache->getNumberBytes()+re_conn_index->getNumberBytes()+re_conn_candi->getNumberBytes()+re_conn_final->getNumberBytes()+re_conn->getNumberBytes()));
//    fprintf(stdout,"tot: %5.2f bpv ", 8.0f/nverts*(re_geom->getNumberBytes()+re_conn_op->getNumberBytes()+re_conn_cache->getNumberBytes()+re_conn_index->getNumberBytes()+re_conn_candi->getNumberBytes()+re_conn_final->getNumberBytes()+re_conn->getNumberBytes()));

//    fprintf(stderr,"predict_none %d predict_last %d predict_inside %d predict_across %d\n", predict_none, predict_last, predict_inside, predict_across);
    fprintf(stderr,"predict_none %d predict_last %d predict_midpoint %d\n", predict_none, predict_last, predict_inside);
#ifdef PRINT_CONTROL_OUTPUT
    fprintf(stderr,"ops:   start %d (%4.1f) add %d (%4.1f) join %d (%4.1f) flip %d (%4.1f) fill %d (%4.1f) end %d (%4.1f) TOTAL %d\n",
      op_start,100.0f*op_start/(op_start+op_add+op_join+op_flip+op_fill+op_end),
      op_add,100.0f*op_add/(op_start+op_add+op_join+op_flip+op_fill+op_end),
      op_join,100.0f*op_join/(op_start+op_add+op_join+op_flip+op_fill+op_end),
      op_flip,100.0f*op_flip/(op_start+op_add+op_join+op_flip+op_fill+op_end),
      op_fill,100.0f*op_fill/(op_start+op_add+op_join+op_flip+op_fill+op_end),
      op_end,100.0f*op_end/(op_start+op_add+op_join+op_flip+op_fill+op_end),
      op_start+op_add+op_join+op_flip+op_fill+op_end);
 
//      fprintf(stderr,"%4.1f & %4.1f & %4.1f & %4.1f          %5.1f\n",
      fprintf(stderr,"{%4.1f}{%4.1f}{%4.1f}          %4.1f          %5.1f\n",
        100.0f*(add_hit_face+join_hit_face+flip_hit_face+fill_hit_face+end_hit_face)/ncells,
        100.0f*(add_hit_edge+join_hit_edge+flip_hit_edge+fill_hit_edge)/ncells,
        100.0f*(start_hit_vertex+add_hit_vertex+join_hit_vertex+flip_hit_vertex+fill_hit_vertex+end_hit_vertex)/ncells,
        100.0f*(start_miss+add_miss+join_miss+flip_miss+fill_miss+end_miss)/ncells,
        100.0f*(add_hit_face+join_hit_face+flip_hit_face+fill_hit_face+end_hit_face)/ncells+
        100.0f*(add_hit_edge+join_hit_edge+flip_hit_edge+fill_hit_edge)/ncells+
        100.0f*(start_hit_vertex+add_hit_vertex+join_hit_vertex+flip_hit_vertex+fill_hit_vertex+end_hit_vertex)/ncells+
        100.0f*(start_miss+add_miss+join_miss+flip_miss+fill_miss+end_miss)/ncells);


      fprintf(stderr,"%d %4.1f %d %4.1f %d %4.1f %d %4.1f %d %d %d\n",
        add_hit_face+join_hit_face+flip_hit_face+fill_hit_face+end_hit_face,
        100.0f*(add_hit_face+join_hit_face+flip_hit_face+fill_hit_face+end_hit_face)/ncells,
        add_hit_edge+join_hit_edge+flip_hit_edge+fill_hit_edge,
        100.0f*(add_hit_edge+join_hit_edge+flip_hit_edge+fill_hit_edge)/ncells,
        start_hit_vertex+add_hit_vertex+join_hit_vertex+flip_hit_vertex+fill_hit_vertex+end_hit_vertex,
        100.0f*(start_hit_vertex+add_hit_vertex+join_hit_vertex+flip_hit_vertex+fill_hit_vertex+end_hit_vertex)/ncells,
        start_miss+add_miss+join_miss+flip_miss+fill_miss+end_miss,
        100.0f*(start_miss+add_miss+join_miss+flip_miss+fill_miss+end_miss)/ncells,
        add_hit_face+join_hit_face+flip_hit_face+fill_hit_face+end_hit_face+add_hit_edge+join_hit_edge+flip_hit_edge+fill_hit_edge+start_hit_vertex+add_hit_vertex+join_hit_vertex+flip_hit_vertex+fill_hit_vertex+end_hit_vertex+start_miss+add_miss+join_miss+flip_miss+fill_miss+end_miss,
        ncells,
        ncells-(add_hit_face+join_hit_face+flip_hit_face+fill_hit_face+end_hit_face+add_hit_edge+join_hit_edge+flip_hit_edge+fill_hit_edge+start_hit_vertex+add_hit_vertex+join_hit_vertex+flip_hit_vertex+fill_hit_vertex+end_hit_vertex+start_miss+add_miss+join_miss+flip_miss+fill_miss+end_miss));

      fprintf(stderr,"hit-f: s (%4.1f) a (%4.1f) j (%4.1f) fp (%4.1f) fl (%4.1f) e (%4.1f) TOTAL (%4.1f)\n",
      0.0f,
      100.0f*add_hit_face/op_add,
      100.0f*join_hit_face/op_join,
      100.0f*flip_hit_face/op_flip,
      100.0f*fill_hit_face/op_fill,
      100.0f*end_hit_face/op_end,
      100.0f*(add_hit_face+join_hit_face+flip_hit_face+fill_hit_face+end_hit_face)/(op_start+op_add+op_join+op_flip+op_fill+op_end));

      fprintf(stderr,"hit-e: s (%4.1f) a (%4.1f) j (%4.1f) fp (%4.1f) fl (%4.1f) e (%4.1f) TOTAL (%4.1f)\n",
      0.0f,
      100.0f*add_hit_edge/op_add,
      100.0f*join_hit_edge/op_join,
      100.0f*flip_hit_edge/op_flip,
      100.0f*fill_hit_edge/op_fill,
      0.0f,
      100.0f*(add_hit_edge+join_hit_edge+flip_hit_edge+fill_hit_edge)/(op_start+op_add+op_join+op_flip+op_fill+op_end));


      fprintf(stderr,"hit-v: s (%4.1f) a (%4.1f) j (%4.1f) fp (%4.1f) fl (%4.1f) e (%4.1f) TOTAL (%4.1f)\n",
      100.0f*start_hit_vertex/op_start,
      100.0f*add_hit_vertex/op_add,
      100.0f*join_hit_vertex/op_join,
      100.0f*flip_hit_vertex/op_flip,
      100.0f*fill_hit_vertex/op_fill,
      100.0f*end_hit_vertex/op_end,
      100.0f*(add_hit_vertex+join_hit_vertex+flip_hit_vertex+fill_hit_vertex+end_hit_vertex)/(op_start+op_add+op_join+op_flip+op_fill+op_end));

      fprintf(stderr,"miss : s (%4.1f) a (%4.1f) j (%4.1f) fp (%4.1f) fl (%4.1f) e (%4.1f) TOTAL (%4.1f)\n",
      100.0f*start_miss/op_start,
      100.0f*add_miss/op_add,
      100.0f*join_miss/op_join,
      100.0f*flip_miss/op_flip,
      100.0f*fill_miss/op_fill,
      100.0f*end_miss/op_end,
      100.0f*(add_miss+join_miss+flip_miss+fill_miss+end_miss)/(op_start+op_add+op_join+op_flip+op_fill+op_end));

      fprintf(stderr,"miss : s (%d) a (%d) j (%d) fp (%d) fl (%d) e (%d) TOTAL (%d) (%d)\n",
      start_miss,
      add_miss,
      join_miss,
      flip_miss,
      fill_miss,
      end_miss,
      (add_miss+join_miss+flip_miss+fill_miss+end_miss),op_join);

    int i;
    for (i = 0; i < 64; i++) if (join_edge_hist[i]) fprintf(stderr,"%d(%d) ",i,join_edge_hist[i]);
    fprintf(stderr,"\n");

    int k;
    int num;
    for (k = 0; k <=5 ;k++)
    {
      num = 0;
      for (i = 0; i < 64; i++) if (i & (1<<k)) num += join_edge_hist[i];
      fprintf(stderr,"%d(%d) ",k,num);
    }
    fprintf(stderr,"\n");

    int rank[] = {2,3,0,1,4,5};

    num = 0;
    for (i = 0; i < 64; i++) if (i & (1<<rank[0])) num += join_edge_hist[i];
    fprintf(stderr,"%d(%d) ",rank[0],num);
    num = 0;
    for (i = 0; i < 64; i++) if ( (i & (1<<rank[1])) && !(i & (1<<rank[0])) ) num += join_edge_hist[i];
    fprintf(stderr,"%d(%d) ",rank[1],num);
    num = 0;
    for (i = 0; i < 64; i++) if ( (i & (1<<rank[2])) && !(i & (1<<rank[0] | 1<<rank[1])) ) num += join_edge_hist[i];
    fprintf(stderr,"%d(%d) ",rank[2],num);
    num = 0;
    for (i = 0; i < 64; i++) if ( (i & (1<<rank[3])) && !(i & (1<<rank[0] | 1<<rank[1] | 1<<rank[2])) ) num += join_edge_hist[i];
    fprintf(stderr,"%d(%d) ",rank[3],num);
    num = 0;
    for (i = 0; i < 64; i++) if ( (i & (1<<rank[4])) && !(i & (1<<rank[0] | 1<<rank[1] | 1<<rank[2] | 1<<rank[3])) ) num += join_edge_hist[i];
    fprintf(stderr,"%d(%d) ",rank[4],num);
    num = 0;
    for (i = 0; i < 64; i++) if ( (i & (1<<rank[5])) && !(i & (1<<rank[0] | 1<<rank[1] | 1<<rank[2] | 1<<rank[3] | 1<<rank[4])) ) num += join_edge_hist[i];
    fprintf(stderr,"%d(%d) ",rank[5],num);

    fprintf(stderr,"\n");

    fprintf(stderr,"op_start [%d %d %d %d %d]\n",op_start_x[0],op_start_x[1],op_start_x[2],op_start_x[3],op_start_x[4]);
    fprintf(stderr,"start_hit_edge_possible %d (of start_hit_vertex %d)\n",start_hit_edge_possible,start_hit_vertex);
    fprintf(stderr,"join_hit_join_vertex %d (%4.1f)\n",join_hit_join_vertex,100.0f*join_hit_join_vertex/op_join);
    fprintf(stderr,"flip_hit_edge_possible %d (of flip_hit_vertex %d)\n",flip_hit_edge_possible,flip_hit_vertex);
    fprintf(stderr,"   hist %d %d %d %d %d\n",flip_hit_edge_hist[0],flip_hit_edge_hist[1],flip_hit_edge_hist[2],flip_hit_edge_hist[3],flip_hit_edge_hist[4]);

    fprintf(stderr,"fill_hit_edge_possible %d (of fill_hit_vertex %d)\n",fill_hit_edge_possible,fill_hit_vertex);
    fprintf(stderr,"   hist %d %d %d %d %d\n",fill_hit_edge_hist[0],fill_hit_edge_hist[1],fill_hit_edge_hist[2],fill_hit_edge_hist[3],fill_hit_edge_hist[4]);
//    fprintf(stderr,"used_index %d (%4.1f) used_cache %d  (%4.1f)\n",used_index,100.0f*used_index/(used_index+used_cache),used_cache,100.0f*used_cache/(used_index+used_cache));
#endif
    
    delete re_conn;
    delete re_conn_op;
    delete re_conn_cache;
    delete re_conn_index;
    delete re_conn_final;
    delete re_geom;
  }
  else
  {
    re_conn->done();

    fprintf(stderr,"predict_none %d predict_last %d predict_midpoint %d\n", predict_none, predict_last, predict_inside);
#ifdef PRINT_CONTROL_OUTPUT
    fprintf(stderr,"total: bytes %d bpv %f\n", re_conn->getNumberBytes(),(float)re_conn->getNumberBytes()*8/nverts);
#endif

        fprintf(stderr,"\n");
#ifdef PRINT_CONTROL_OUTPUT
    fprintf(stderr,"ops:   start %d (%4.1f) add %d (%4.1f) join %d (%4.1f) flip %d (%4.1f) fill %d (%4.1f) end %d (%4.1f) TOTAL %d\n",
      op_start,100.0f*op_start/(op_start+op_add+op_join+op_flip+op_fill+op_end),
      op_add,100.0f*op_add/(op_start+op_add+op_join+op_flip+op_fill+op_end),
      op_join,100.0f*op_join/(op_start+op_add+op_join+op_flip+op_fill+op_end),
      op_flip,100.0f*op_flip/(op_start+op_add+op_join+op_flip+op_fill+op_end),
      op_fill,100.0f*op_fill/(op_start+op_add+op_join+op_flip+op_fill+op_end),
      op_end,100.0f*op_end/(op_start+op_add+op_join+op_flip+op_fill+op_end),
      op_start+op_add+op_join+op_flip+op_fill+op_end);
 
//      fprintf(stderr,"%4.1f & %4.1f & %4.1f & %4.1f          %5.1f\n",
      fprintf(stderr,"{%4.1f}{%4.1f}{%4.1f}          %4.1f          %5.1f\n",
        100.0f*(add_hit_face+join_hit_face+flip_hit_face+fill_hit_face+end_hit_face)/ncells,
        100.0f*(add_hit_edge+join_hit_edge+flip_hit_edge+fill_hit_edge)/ncells,
        100.0f*(start_hit_vertex+add_hit_vertex+join_hit_vertex+flip_hit_vertex+fill_hit_vertex+end_hit_vertex)/ncells,
        100.0f*(start_miss+add_miss+join_miss+flip_miss+fill_miss+end_miss)/ncells,
        100.0f*(add_hit_face+join_hit_face+flip_hit_face+fill_hit_face+end_hit_face)/ncells+
        100.0f*(add_hit_edge+join_hit_edge+flip_hit_edge+fill_hit_edge)/ncells+
        100.0f*(start_hit_vertex+add_hit_vertex+join_hit_vertex+flip_hit_vertex+fill_hit_vertex+end_hit_vertex)/ncells+
        100.0f*(start_miss+add_miss+join_miss+flip_miss+fill_miss+end_miss)/ncells);


      fprintf(stderr,"%d %4.1f %d %4.1f %d %4.1f %d %4.1f %d %d %d\n",
        add_hit_face+join_hit_face+flip_hit_face+fill_hit_face+end_hit_face,
        100.0f*(add_hit_face+join_hit_face+flip_hit_face+fill_hit_face+end_hit_face)/ncells,
        add_hit_edge+join_hit_edge+flip_hit_edge+fill_hit_edge,
        100.0f*(add_hit_edge+join_hit_edge+flip_hit_edge+fill_hit_edge)/ncells,
        start_hit_vertex+add_hit_vertex+join_hit_vertex+flip_hit_vertex+fill_hit_vertex+end_hit_vertex,
        100.0f*(start_hit_vertex+add_hit_vertex+join_hit_vertex+flip_hit_vertex+fill_hit_vertex+end_hit_vertex)/ncells,
        start_miss+add_miss+join_miss+flip_miss+fill_miss+end_miss,
        100.0f*(start_miss+add_miss+join_miss+flip_miss+fill_miss+end_miss)/ncells,
        add_hit_face+join_hit_face+flip_hit_face+fill_hit_face+end_hit_face+add_hit_edge+join_hit_edge+flip_hit_edge+fill_hit_edge+start_hit_vertex+add_hit_vertex+join_hit_vertex+flip_hit_vertex+fill_hit_vertex+end_hit_vertex+start_miss+add_miss+join_miss+flip_miss+fill_miss+end_miss,
        ncells,
        ncells-(add_hit_face+join_hit_face+flip_hit_face+fill_hit_face+end_hit_face+add_hit_edge+join_hit_edge+flip_hit_edge+fill_hit_edge+start_hit_vertex+add_hit_vertex+join_hit_vertex+flip_hit_vertex+fill_hit_vertex+end_hit_vertex+start_miss+add_miss+join_miss+flip_miss+fill_miss+end_miss));

      fprintf(stderr,"hit-f: s (%4.1f) a (%4.1f) j (%4.1f) fp (%4.1f) fl (%4.1f) e (%4.1f) TOTAL (%4.1f)\n",
      0.0f,
      100.0f*add_hit_face/op_add,
      100.0f*join_hit_face/op_join,
      100.0f*flip_hit_face/op_flip,
      100.0f*fill_hit_face/op_fill,
      100.0f*end_hit_face/op_end,
      100.0f*(add_hit_face+join_hit_face+flip_hit_face+fill_hit_face+end_hit_face)/(op_start+op_add+op_join+op_flip+op_fill+op_end));

      fprintf(stderr,"hit-e: s (%4.1f) a (%4.1f) j (%4.1f) fp (%4.1f) fl (%4.1f) e (%4.1f) TOTAL (%4.1f)\n",
      0.0f,
      100.0f*add_hit_edge/op_add,
      100.0f*join_hit_edge/op_join,
      100.0f*flip_hit_edge/op_flip,
      100.0f*fill_hit_edge/op_fill,
      0.0f,
      100.0f*(add_hit_edge+join_hit_edge+flip_hit_edge+fill_hit_edge)/(op_start+op_add+op_join+op_flip+op_fill+op_end));


      fprintf(stderr,"hit-v: s (%4.1f) a (%4.1f) j (%4.1f) fp (%4.1f) fl (%4.1f) e (%4.1f) TOTAL (%4.1f)\n",
      100.0f*start_hit_vertex/op_start,
      100.0f*add_hit_vertex/op_add,
      100.0f*join_hit_vertex/op_join,
      100.0f*flip_hit_vertex/op_flip,
      100.0f*fill_hit_vertex/op_fill,
      100.0f*end_hit_vertex/op_end,
      100.0f*(add_hit_vertex+join_hit_vertex+flip_hit_vertex+fill_hit_vertex+end_hit_vertex)/(op_start+op_add+op_join+op_flip+op_fill+op_end));

      fprintf(stderr,"miss : s (%4.1f) a (%4.1f) j (%4.1f) fp (%4.1f) fl (%4.1f) e (%4.1f) TOTAL (%4.1f)\n",
      100.0f*start_miss/op_start,
      100.0f*add_miss/op_add,
      100.0f*join_miss/op_join,
      100.0f*flip_miss/op_flip,
      100.0f*fill_miss/op_fill,
      100.0f*end_miss/op_end,
      100.0f*(add_miss+join_miss+flip_miss+fill_miss+end_miss)/(op_start+op_add+op_join+op_flip+op_fill+op_end));

      fprintf(stderr,"miss : s (%d) a (%d) j (%d) fp (%d) fl (%d) e (%d) TOTAL (%d) (%d)\n",
      start_miss,
      add_miss,
      join_miss,
      flip_miss,
      fill_miss,
      end_miss,
      (add_miss+join_miss+flip_miss+fill_miss+end_miss),op_join);

    int i;
    for (i = 0; i < 64; i++) if (join_edge_hist[i]) fprintf(stderr,"%d(%d) ",i,join_edge_hist[i]);
    fprintf(stderr,"\n");

    int k;
    int num;
    for (k = 0; k <=5 ;k++)
    {
      num = 0;
      for (i = 0; i < 64; i++) if (i & (1<<k)) num += join_edge_hist[i];
      fprintf(stderr,"%d(%d) ",k,num);
    }
    fprintf(stderr,"\n");

    int rank[] = {2,3,0,1,4,5};

    num = 0;
    for (i = 0; i < 64; i++) if (i & (1<<rank[0])) num += join_edge_hist[i];
    fprintf(stderr,"%d(%d) ",rank[0],num);
    num = 0;
    for (i = 0; i < 64; i++) if ( (i & (1<<rank[1])) && !(i & (1<<rank[0])) ) num += join_edge_hist[i];
    fprintf(stderr,"%d(%d) ",rank[1],num);
    num = 0;
    for (i = 0; i < 64; i++) if ( (i & (1<<rank[2])) && !(i & (1<<rank[0] | 1<<rank[1])) ) num += join_edge_hist[i];
    fprintf(stderr,"%d(%d) ",rank[2],num);
    num = 0;
    for (i = 0; i < 64; i++) if ( (i & (1<<rank[3])) && !(i & (1<<rank[0] | 1<<rank[1] | 1<<rank[2])) ) num += join_edge_hist[i];
    fprintf(stderr,"%d(%d) ",rank[3],num);
    num = 0;
    for (i = 0; i < 64; i++) if ( (i & (1<<rank[4])) && !(i & (1<<rank[0] | 1<<rank[1] | 1<<rank[2] | 1<<rank[3])) ) num += join_edge_hist[i];
    fprintf(stderr,"%d(%d) ",rank[4],num);
    num = 0;
    for (i = 0; i < 64; i++) if ( (i & (1<<rank[5])) && !(i & (1<<rank[0] | 1<<rank[1] | 1<<rank[2] | 1<<rank[3] | 1<<rank[4])) ) num += join_edge_hist[i];
    fprintf(stderr,"%d(%d) ",rank[5],num);

    fprintf(stderr,"\n");

    fprintf(stderr,"op_start [%d %d %d %d %d]\n",op_start_x[0],op_start_x[1],op_start_x[2],op_start_x[3],op_start_x[4]);
    fprintf(stderr,"start_hit_edge_possible %d (of start_hit_vertex %d)\n",start_hit_edge_possible,start_hit_vertex);
    fprintf(stderr,"join_hit_join_vertex %d (%4.1f)\n",join_hit_join_vertex,100.0f*join_hit_join_vertex/op_join);
    fprintf(stderr,"flip_hit_edge_possible %d (of flip_hit_vertex %d)\n",flip_hit_edge_possible,flip_hit_vertex);
    fprintf(stderr,"   hist %d %d %d %d %d\n",flip_hit_edge_hist[0],flip_hit_edge_hist[1],flip_hit_edge_hist[2],flip_hit_edge_hist[3],flip_hit_edge_hist[4]);

    fprintf(stderr,"fill_hit_edge_possible %d (of fill_hit_vertex %d)\n",fill_hit_edge_possible,fill_hit_vertex);
    fprintf(stderr,"   hist %d %d %d %d %d\n",fill_hit_edge_hist[0],fill_hit_edge_hist[1],fill_hit_edge_hist[2],fill_hit_edge_hist[3],fill_hit_edge_hist[4]);
//    fprintf(stderr,"used_index %d (%4.1f) used_cache %d  (%4.1f)\n",used_index,100.0f*used_index/(used_index+used_cache),used_cache,100.0f*used_cache/(used_index+used_cache));
#endif

    delete re_conn;
  }
}

#ifdef SVC_CALCULATE_QUANTIZATION_ERROR 
static float quant_err;
static float max_quant_err = 0.0f;
#endif

#ifdef SVC_CALCULATE_PREDICTION_ERROR 
static double total_pred_err = 0.0;
static double total_pred_err_normal = 0.0;
static double total_pred_err_tangential = 0.0;
#endif

#if defined(SVC_USE_LOCAL_COORDINATE_SYSTEM)
static void compressVertexPropertiesLocal(float* p0, float* p1, float* p2, float* p3)
{
  predict_inside++;

  float P[3];
  float P1[3];
  float P2[3];
  float Pn[3];
  float Pp[3];

  VecSubtract3fv(P1, p2, p1);
  VecSubtract3fv(P2, p0, p1);

  VecCrossProd3fv(P, P1, P2);

  VecSelfNormalize3fv(P1);
  VecSelfNormalize3fv(P);

  VecCrossProd3fv(P2, P1, P);

  VecSelfNormalize3fv(P2);

#ifdef SVC_ADD_NORMAL_COMPONENT
  VecAdd3fv(Pp, p0, p1, p2);
  VecSelfScalarDiv3fv(Pp, 3.0f);
  float normal_f[3];
  float scale = sqrt(VecCcwNormNormalReturnLength3fv(normal_f, p0, p1, p2));
  VecSelfAddScalarMult3fv(Pp, normal_f, SVC_NORMAL_SCALE*scale);
  v_pq->Clamp(Pp);
#else // SVC_ADD_NORMAL_COMPONENT
  VecZero3fv(Pp);
  VecSelfAddScalarMult3fv(Pp, p0, 1.0f/3.0f);
  VecSelfAddScalarMult3fv(Pp, p1, 1.0f/3.0f);
  VecSelfAddScalarMult3fv(Pp, p2, 1.0f/3.0f);
#endif // SVC_ADD_NORMAL_COMPONENT

  VecSubtract3fv(Pn, p3, Pp);
  
  float diff[3];

  diff[0] = VecDotProd3fv(Pn, P1);
  diff[1] = VecDotProd3fv(Pn, P2);
  diff[2] = VecDotProd3fv(Pn, P);
 
#ifndef SVC_ADD_NORMAL_COMPONENT
  if (diff[2]<0) diff[2]=0;
#endif

  I32 a = quantize(diff[2]);
  I32 b = quantize(diff[0]);
  I32 c = quantize(diff[1]);

#ifdef SVC_ADD_NORMAL_COMPONENT
  local_ic[0]->Compress(maxAbso, a+maxAbso, 2);
  local_ic[1]->Compress(maxAbso, b+maxAbso, 2);
  local_ic[2]->Compress(maxAbso, c+maxAbso, 2);
#else
  local_ic[0]->Compress(0, a, 2);
  local_ic[1]->Compress(maxAbso, b+maxAbso, 2);
  local_ic[2]->Compress(maxAbso, c+maxAbso, 2);
#endif

  diff[2] = dequantize(a);
  diff[0] = dequantize(b);
  diff[1] = dequantize(c);

#ifdef SVC_CALCULATE_PREDICTION_ERROR 
  total_pred_err += VecLength3iv(a,b,c);
  total_pred_err_normal += abs(a);
  total_pred_err_tangential += sqrt((float)b*(float)b+(float)c*(float)c);
#endif

  float Ph[3];
  VecCopy3fv(Ph, Pp);

  VecSelfAddScalarMult3fv(Ph, P1, diff[0]);
  VecSelfAddScalarMult3fv(Ph, P2, diff[1]);
  VecSelfAddScalarMult3fv(Ph, P, diff[2]);

#ifdef SVC_CALCULATE_QUANTIZATION_ERROR 
  quant_err = VecDistance3fv(p3, Ph);
  if (quant_err > max_quant_err) max_quant_err = quant_err;
#endif

  VecCopy3fv(p3, Ph);
}

static void compressVertexPropertiesLocal(float* across_prop_f0, float* pred_prop_f0, float* pred_prop_f1, float* pred_prop_f2, float* real_prop_f)
{
  predict_across++;

  float P[3];
  float P1[3];
  float P2[3];
  float Pn[3];
  float Pp[3];
  float diff[3];

  // compute first tangential axis P1 (from p1 to p2)
  VecSubtract3fv(P1, pred_prop_f2, pred_prop_f1);
  // compute vector P2 (from p1 to p0) which is hopefully not parallel to P1
  VecSubtract3fv(P2, pred_prop_f0, pred_prop_f1);
  // compute normal axis P (orthogonal to P1 and P2)
  VecCrossProd3fv(P, P1, P2);
  // compute second tangential axis P2 (orthogonal to P1 and P)
  VecCrossProd3fv(P2, P1, P);

  // normalize coordinate system
  VecSelfNormalize3fv(P1);
  VecSelfNormalize3fv(P2);
  VecSelfNormalize3fv(P);

  // compute mid-point of triangle
  VecAdd3fv(Pp, pred_prop_f0, pred_prop_f1, pred_prop_f2);
  VecSelfScalarDiv3fv(Pp, 3.0f);

  // compute height of across point in respect to triangle
  VecSubtract3fv(diff, pred_prop_f0, across_prop_f0);
  float height = VecDotProd3fv(diff, P);
  // add normal scaled by height to mid-point
  VecSelfAddScalarMult3fv(Pp, P, height);

  // compute difference between prediction and real point
  VecSubtract3fv(Pn, real_prop_f, Pp);

  // rotate difference vector into local coordinate system
  diff[0] = VecDotProd3fv(Pn, P1);
  diff[1] = VecDotProd3fv(Pn, P2);
  diff[2] = VecDotProd3fv(Pn, P);

  // quantize the difference vector
  I32 a = quantize(diff[2]);
  I32 b = quantize(diff[0]);
  I32 c = quantize(diff[1]);

  // compress the coordinates of the difference vector
  local_ic[0]->Compress(maxAbso, a+maxAbso, 2);
  local_ic[1]->Compress(maxAbso, b+maxAbso, 2);
  local_ic[2]->Compress(maxAbso, c+maxAbso, 2);

  // dequantize the difference vector (replay of decoding process)
  diff[2] = dequantize(a);
  diff[0] = dequantize(b);
  diff[1] = dequantize(c);

#ifdef SVC_CALCULATE_PREDICTION_ERROR 
  total_pred_err += VecLength3iv(a,b,c);
  total_pred_err_normal += abs(a);
  total_pred_err_tangential += sqrt((float)b*(float)b+(float)c*(float)c);
#endif

  // copy the predictor (replay of decoding process)
  VecCopy3fv(Pn, Pp);

  // add the difference vector (replay of decoding process)
  VecSelfAddScalarMult3fv(Pn, P1, diff[0]);
  VecSelfAddScalarMult3fv(Pn, P2, diff[1]);
  VecSelfAddScalarMult3fv(Pn, P, diff[2]);

#ifdef SVC_CALCULATE_QUANTIZATION_ERROR 
  quant_err = VecDistance3fv(real_prop_f, Pn);
  if (quant_err > max_quant_err) max_quant_err = quant_err;
#endif

  // copy the quantized result back to the input
  VecCopy3fv(real_prop_f, Pn);
}

static void compressVertexPropertiesLocalFlip(float* across_prop_f0, float* pred_prop_f0, float* pred_prop_f1, float* pred_prop_f2, float* real_prop_f)
{
  predict_across++;

  float P[3];
  float P1[3];
  float P2[3];
  float Pn[3];
  float Pp[3];
  float diff[3];

  // compute first tangential axis P1 (from p1 to p2)
  VecSubtract3fv(P1, pred_prop_f2, pred_prop_f1);
  // compute vector P2 (from p1 to p0) which is hopefully not parallel to P1
  VecSubtract3fv(P2, pred_prop_f0, pred_prop_f1);
  // compute normal axis P (orthogonal to P1 and P2)
  VecCrossProd3fv(P, P1, P2);
  // compute second tangential axis P2 (orthogonal to P1 and P)
  VecCrossProd3fv(P2, P1, P);

  // normalize coordinate system
  VecSelfNormalize3fv(P1);
  VecSelfNormalize3fv(P2);
  VecSelfNormalize3fv(P);

  // compute mid-point of triangle
  VecAdd3fv(Pp, pred_prop_f0, pred_prop_f1, pred_prop_f2);
  VecSelfScalarDiv3fv(Pp, 3.0f);

  // mult the predictor by two
  VecSelfScalarMult3fv(Pp, 2.0f);

  // subtract the across position
  VecSelfSubtract3fv(Pp, across_prop_f0);

  // clamp the prediction into bounding box
  v_pq->Clamp(Pp);

  // compute difference between prediction and real point
  VecSubtract3fv(Pn, real_prop_f, Pp);

  // rotate difference vector into local coordinate system
  diff[0] = VecDotProd3fv(Pn, P1);
  diff[1] = VecDotProd3fv(Pn, P2);
  diff[2] = VecDotProd3fv(Pn, P);

  // quantize the difference vector
  I32 a = quantize(diff[2]);
  I32 b = quantize(diff[0]);
  I32 c = quantize(diff[1]);

  // compress the coordinates of the difference vector
  local_ic[0]->Compress(maxAbso, a+maxAbso, 2);
  local_ic[1]->Compress(maxAbso, b+maxAbso, 2);
  local_ic[2]->Compress(maxAbso, c+maxAbso, 2);

  // dequantize the difference vector (replay of decoding process)
  diff[2] = dequantize(a);
  diff[0] = dequantize(b);
  diff[1] = dequantize(c);

#ifdef SVC_CALCULATE_PREDICTION_ERROR 
  total_pred_err += VecLength3iv(a,b,c);
  total_pred_err_normal += abs(a);
  total_pred_err_tangential += sqrt((float)b*(float)b+(float)c*(float)c);
#endif

  // copy the predictor (replay of decoding process)
  VecCopy3fv(Pn, Pp);

  // add the difference vector (replay of decoding process)
  VecSelfAddScalarMult3fv(Pn, P1, diff[0]);
  VecSelfAddScalarMult3fv(Pn, P2, diff[1]);
  VecSelfAddScalarMult3fv(Pn, P, diff[2]);

#ifdef SVC_CALCULATE_QUANTIZATION_ERROR 
  quant_err = VecDistance3fv(real_prop_f, Pn);
  if (quant_err > max_quant_err) max_quant_err = quant_err;
#endif

  // copy the quantized result back to the input
  VecCopy3fv(real_prop_f, Pn);
}
#endif 

#ifdef SVC_DO_FLIP_PREDICTION
static void compressVertexPropertiesGlobalFlip(float* across_prop_f0, float* pred_prop_f0, float* pred_prop_f1, float* pred_prop_f2, float* real_prop_f)
{
  predict_across++;

  // compute quantized coordinates
  v_pq->EnQuantize(real_prop_f, (I32*)real_prop_f);

  // compute mid-point of triangle and multiply it times two
  float pred_f[3];
  VecAdd3fv(pred_f, (I32*)pred_prop_f0, (I32*)pred_prop_f1, (I32*)pred_prop_f2);
  // divide by three
  VecSelfScalarDiv3fv(pred_f, 3.0f);

  // quantize floating-point prediction onto integer grid
  I32 pred[3];
  v_pq->Clamp(pred_f, pred);

  // mult the predictor by two
  VecSelfScalarMult3iv(pred, 2);

  // subtract the across position
  VecSelfSubtract3iv(pred, (I32*)across_prop_f0);

  // clamp the prediction into bounding box
  v_pq->Clamp(pred);

#ifdef SVC_CALCULATE_PREDICTION_ERROR 
  total_pred_err += VecDistance3iv(pred, (I32*)real_prop_f);
#endif

#ifdef SVC_STORE_QUANTIZED_COORDINATES
  fwrite(pred_prop_f0, sizeof(I32), 3, quantized_coordinates_file);
  fwrite(pred_prop_f1, sizeof(I32), 3, quantized_coordinates_file);
  fwrite(pred_prop_f2, sizeof(I32), 3, quantized_coordinates_file);
  fwrite(pred, sizeof(I32), 3, quantized_coordinates_file);
  fwrite(real_prop_f, sizeof(I32), 3, quantized_coordinates_file);
#endif

  for (int i = 0; i < 3; i++)
  {
    global_ic[i]->Compress(((I32*)pred)[i],((I32*)real_prop_f)[i], 2);
  }

#ifdef PRINT_DEBUG_OUTPUT
  printf("v %d %d %d (%d,%d,%d) a\n", ((I32*)real_prop_f)[0], ((I32*)real_prop_f)[1], ((I32*)real_prop_f)[2],((I32*)pred)[0],((I32*)pred)[1], ((I32*)pred)[2]);
#endif
}
#endif //SVC_DO_FLIP_PREDICTION

#if defined(SVC_DO_TETRAHEDRAL_ACROSS_PREDICTION) && !defined(SVC_USE_LOCAL_COORDINATE_SYSTEM)
static void compressVertexPropertiesGlobal(float* across_prop_f0, float* pred_prop_f0, float* pred_prop_f1, float* pred_prop_f2, float* real_prop_f)
{
  predict_across++;

  // compute quantized coordinates
  v_pq->EnQuantize(real_prop_f, (I32*)real_prop_f);

  // compute mid-point of triangle
  float pred_f[3];
  VecAdd3fv(pred_f, (I32*)pred_prop_f0, (I32*)pred_prop_f1, (I32*)pred_prop_f2);
  VecSelfScalarDiv3fv(pred_f, 3.0f);
  // compute normal of triangle
  float normal_f[3];
  if (VecCcwNormNormal3fv(normal_f, (I32*)pred_prop_f0, (I32*)pred_prop_f1, (I32*)pred_prop_f2))
  {
    // compute height of across point in respect to triangle
    float temp_f[3];
    VecSubtract3fv(temp_f, (I32*)pred_prop_f0, (I32*)across_prop_f0);
    float height = VecDotProd3fv(temp_f, normal_f);
    // add normal scaled by height to mid-point
    VecSelfAddScalarMult3fv(pred_f, normal_f, height);
  }
  else
  {
    normal_not_valid_counter++; // normal is not valid ... simply use mid-point
  }
  // quantize floating-point prediction onto integer grid
  I32 pred[3];
  v_pq->Clamp(pred_f, pred);

#ifdef SVC_CALCULATE_PREDICTION_ERROR 
  total_pred_err += VecDistance3iv(pred, (I32*)real_prop_f);
#endif

#ifdef SVC_STORE_QUANTIZED_COORDINATES
  fwrite(pred_prop_f0, sizeof(I32), 3, quantized_coordinates_file);
  fwrite(pred_prop_f1, sizeof(I32), 3, quantized_coordinates_file);
  fwrite(pred_prop_f2, sizeof(I32), 3, quantized_coordinates_file);
  fwrite(pred, sizeof(I32), 3, quantized_coordinates_file);
  fwrite(real_prop_f, sizeof(I32), 3, quantized_coordinates_file);
#endif

  for (int i = 0; i < 3; i++)
  {
    global_ic[i]->Compress(((I32*)pred)[i],((I32*)real_prop_f)[i], 2);
  }

#ifdef PRINT_DEBUG_OUTPUT
  printf("v %d %d %d (%d,%d,%d) a\n", ((I32*)real_prop_f)[0], ((I32*)real_prop_f)[1], ((I32*)real_prop_f)[2],((I32*)pred)[0],((I32*)pred)[1], ((I32*)pred)[2]);
#endif
}
#endif //SVC_USE_LOCAL_COORDINATE_SYSTEM

static void compressVertexPropertiesGlobal(float* pred_prop_f0, float* pred_prop_f1, float* pred_prop_f2, float* real_prop_f)
{
  predict_inside++;

  if (v_pq)
  {
    I32 pred[3];
#ifdef SVC_ADD_NORMAL_COMPONENT
    if (SVC_NORMAL_SCALE == -8.0f)
    {
      VecAdd3iv(pred, (I32*)pred_prop_f0, (I32*)pred_prop_f1, (I32*)pred_prop_f2);
      VecSelfScalarDiv3iv(pred, 3);
    }
    else if (SVC_NORMAL_SCALE == -4.0f)
    {
      VecAdd3iv(pred, (I32*)pred_prop_f0, (I32*)pred_prop_f1, (I32*)pred_prop_f2);
      VecSelfAdd3iv(pred, 1);
      VecSelfScalarDiv3iv(pred, 3);
    }
    else if (SVC_NORMAL_SCALE == -2.0f)
    {
      float pred_f[3];
      VecAdd3fv(pred_f, (I32*)pred_prop_f0, (I32*)pred_prop_f1, (I32*)pred_prop_f2);
      VecSelfScalarDiv3fv(pred_f, 3.0f);
      v_pq->Clamp(pred_f, pred);
    }
    else
    {
      float pred_f[3];
      VecAdd3fv(pred_f, (int*)pred_prop_f0, (I32*)pred_prop_f1, (I32*)pred_prop_f2);
      VecSelfScalarDiv3fv(pred_f, 3.0f);
      float normal_f[3];
      float scale = sqrt(VecCcwNormNormalReturnLength3fv(normal_f, (I32*)pred_prop_f0, (I32*)pred_prop_f1, (I32*)pred_prop_f2));
      VecSelfAddScalarMult3fv(pred_f, normal_f, SVC_NORMAL_SCALE*scale);
      v_pq->Clamp(pred_f, pred);
    }
#else // SVC_ADD_NORMAL_COMPONENT
    VecAdd3iv(pred, (I32*)pred_prop_f0, (I32*)pred_prop_f1, (I32*)pred_prop_f2);
    VecSelfScalarDiv3iv(pred, 3);
#endif // SVC_ADD_NORMAL_COMPONENT

#ifdef SVC_CALCULATE_QUANTIZATION_ERROR 
    float before[3];
    VecCopy3fv(before, real_prop_f);
#endif

    v_pq->EnQuantize(real_prop_f, (I32*)real_prop_f);

#ifdef SVC_CALCULATE_PREDICTION_ERROR 
    total_pred_err += VecDistance3iv(pred, (I32*)real_prop_f);
#endif

#ifdef SVC_STORE_QUANTIZED_COORDINATES
    fwrite(pred_prop_f0, sizeof(I32), 3, quantized_coordinates_file);
    fwrite(pred_prop_f1, sizeof(I32), 3, quantized_coordinates_file);
    fwrite(pred_prop_f2, sizeof(I32), 3, quantized_coordinates_file);
    fwrite(pred, sizeof(I32), 3, quantized_coordinates_file);
    fwrite(real_prop_f, sizeof(I32), 3, quantized_coordinates_file);
#endif

#ifdef SVC_CALCULATE_QUANTIZATION_ERROR 
    float after[3];
    v_pq->DeQuantize((I32*)real_prop_f, after);
    quant_err = VecDistance3fv(before, after);
    if (quant_err > max_quant_err) max_quant_err = quant_err;
#endif

    for (int i = 0; i < 3; i++)
    {
      global_ic[i]->Compress(((I32*)pred)[i],((I32*)real_prop_f)[i], 2);
    }
#ifdef PRINT_DEBUG_OUTPUT
    printf("v %d %d %d (%d,%d,%d) a\n", ((I32*)real_prop_f)[0], ((I32*)real_prop_f)[1], ((I32*)real_prop_f)[2],((I32*)pred)[0],((I32*)pred)[1], ((I32*)pred)[2]);
#endif
  }
  else
  {
    float pred[3];
    VecAdd3fv(pred, pred_prop_f0, pred_prop_f1, pred_prop_f2);
    VecSelfScalarDiv3fv(pred, 3.0f);
    for (int i = 0; i < 3; i++)
    {
      real_prop_f[i] = v_fc[i]->Compress(pred[i],real_prop_f[i], 2);
    }
  }
}

static void compressVertexProperties(float* pred_prop_f, float* real_prop_f, I32 num_prop_f)
{
  predict_last++;
#ifdef PRINT_CONTROL_OUTPUT
#endif

  if (num_prop_f == 3)
  {
    if (v_pq)
    {
      v_pq->EnQuantize(real_prop_f, (I32*)real_prop_f);
#ifdef SVC_STORE_QUANTIZED_COORDINATES
      fwrite(real_prop_f, sizeof(I32), 3, quantized_coordinates_file);
#endif
      for (int i = 0; i < 3; i++)
      {
        global_ic[i]->Compress(((I32*)pred_prop_f)[i],((I32*)real_prop_f)[i], 1);
      }
#ifdef PRINT_DEBUG_OUTPUT
      printf("v %d %d %d (%d,%d,%d) l\n", ((I32*)real_prop_f)[0], ((I32*)real_prop_f)[1], ((I32*)real_prop_f)[2], ((I32*)pred_prop_f)[0], ((I32*)pred_prop_f)[1], ((I32*)pred_prop_f)[2]);
#endif
    }
    else
    {
      for (int i = 0; i < num_prop_f; i++)
      {
        real_prop_f[i] = v_fc[i]->Compress(pred_prop_f[i],real_prop_f[i], 1);
      }
    }
  }
}

static void compressVertexProperties(float* real_prop_f, I32 pnum_f)
{
  predict_none++;
#ifdef PRINT_CONTROL_OUTPUT
#endif

  if (pnum_f == 3)
  {
    if (v_pq)
    {
      v_pq->EnQuantize(real_prop_f, (I32*)real_prop_f);
#ifdef SVC_STORE_QUANTIZED_COORDINATES
      fwrite(real_prop_f, sizeof(I32), 3, quantized_coordinates_file);
#endif
      for (int i = 0; i < 3; i++)
      {
        global_ic[i]->Compress(0, ((I32*)real_prop_f)[i], 0);
      }
#ifdef PRINT_DEBUG_OUTPUT
      printf("vertex %d %d %d none\n", ((I32*)real_prop_f)[0], ((I32*)real_prop_f)[1], ((I32*)real_prop_f)[2]);
#endif
    }
    else
    {
      for (int i = 0; i < 3; i++)
      {
        real_prop_f[i] = v_fc[i]->Compress(0.0f, real_prop_f[i], 0);
      }
    }
  }
}

bool SVwriter_svc::open(FILE* file, int nbits, bool cbm_precision, int high_bits)
{
  // store precision 
  this->file = file;
  this->nbits = nbits;
  this->cbm_precision = cbm_precision;
  this->high_bits = high_bits;

  if (nbits < 0 || nbits > 24)
  {
    fprintf(stderr, "ERROR: nbits is %d. it should be in the range [0,24]\n", nbits);
    exit(0);
  }

  if (high_bits < 0 || high_bits > 16)
  {
    fprintf(stderr, "ERROR: high_bits is %d. it should be in the range [0,16]\n", high_bits);
    exit(0);
  }

#ifdef _WIN32
  if (file == stdout)
  {
    if(_setmode( _fileno( stdout ), _O_BINARY ) == -1 )
    {
      fprintf(stderr, "ERROR: cannot set stdout to binary (untranslated) mode\n");
      exit(0);
    }
  }
#endif

  // write version
  if (file)
  {
#ifdef SVC_DO_TETRAHEDRAL_ACROSS_PREDICTION
  #ifdef SVC_USE_LOCAL_COORDINATE_SYSTEM
    fputc(3, file);
  #else
    fputc(2, file);
  #endif
#elif SVC_DO_FLIP_PREDICTION
  #ifdef SVC_USE_LOCAL_COORDINATE_SYSTEM
    fputc(5, file);
  #else
    fputc(4, file);
  #endif
#elif SVC_ADD_NORMAL_COMPONENT
  #ifdef SVC_USE_LOCAL_COORDINATE_SYSTEM
    fputc(7, file);
  #else
    fputc(6, file);
  #endif
#else // simple mid-point
  #ifdef SVC_USE_LOCAL_COORDINATE_SYSTEM
    fputc(1, file);
  #else
    fputc(0, file);
  #endif
#endif
  }

  vertex_hash = new my_vertex_hash;
  dv = new DynamicVector();
  lc = new LittleCache(4);
  lc_f = new LittleCache(4);

  initEncoder(file);
  initModels(1);

  last_op = 0;
  last_fin = 0;

  // write precision
  re_conn->writeRange(25,nbits);

  v_count = 0;
  c_count = 0;

#ifdef SVC_STORE_QUANTIZED_COORDINATES
  quantized_coordinates_file = fopen("temp_file.bin","wb");
#endif

  return true;
}

void SVwriter_svc::set_v_pnum_f(I32 v_pnum_f)
{
  if (v_pnum_f != 3)
  {
    fprintf(stderr,"WARNING: v_pnum_f = %d not implemented. changed to v_pnum_f = 3\n", v_pnum_f);
    v_pnum_f = 3;
  }
  this->v_pnum_f = v_pnum_f;
}

void SVwriter_svc::write_vertex(const float* v_prop_f)
{
  if (v_count + c_count == 0) write_header();

  SVvertex* vertex = allocVertex();
  vertex_hash->insert(my_vertex_hash::value_type(v_count, vertex));
  vertex->index = v_count;
  VecCopyNfv(vertex->prop_f, v_prop_f, v_pnum_f);
  v_count++;
}

void SVwriter_svc::compress_tetrahedron(const I32* c_idx, const float* c_prop_f, const bool* c_final)
{
  int i,j;
  int op;
  I32 dv_index;
  int lc_pos_f = -1;
  int lc_pos_e = -1;
  int lc_pos_v = -1;
  int candidate;
  int candidate_count;
  int degree_one;
  int use_count;
  int rot[] = {0,1,2,3};

  my_vertex_hash::iterator hash_elements[4];
  SVvertex* vertices[4];

  // get vertices from hash 

  for (i = 0; i < 4; i++)
  {
    hash_elements[i] = vertex_hash->find(c_idx[i]);
    if (hash_elements[i] == vertex_hash->end())
    {
      fprintf(stderr,"FATAL ERROR: vertex %d with index %d not in hash\n", i, c_idx[i]);
      exit(0);
    }
    vertices[i] = (*hash_elements[i]).second;
    vertices[i]->finalize = c_final[i];
  }

  // check the orientation of tetrahedron

  float u1[3], u2[3], u3[3], u1u2[3];
  VecSubtract3fv(u1, vertices[1]->prop_f, vertices[0]->prop_f);
  VecSubtract3fv(u2, vertices[2]->prop_f, vertices[0]->prop_f);
  VecSubtract3fv(u3, vertices[3]->prop_f, vertices[0]->prop_f);
  VecCrossProd3fv(u1u2, u1, u2);
  if (VecDotProd3fv(u1u2, u3) > 0.0f)
  {
    pos_vol++;
  }
  else
  {
    neg_vol++;
    SVvertex* v = vertices[2];
    vertices[2] = vertices[3];
    vertices[3] = v;
    my_vertex_hash::iterator he = hash_elements[2];
    hash_elements[2] = hash_elements[3];
    hash_elements[3] = he;
  }
  
//  fprintf(stderr,"the input tetrahedron %d %d %d %d\n",vertices[0]->index, vertices[1]->index, vertices[2]->index, vertices[3]->index);
//  fprintf(stderr,"  face0 %d %d %d\n",vertices[0]->index, vertices[1]->index, vertices[2]->index);
//  fprintf(stderr,"  face1 %d %d %d\n",vertices[1]->index, vertices[0]->index, vertices[3]->index);
//  fprintf(stderr,"  face2 %d %d %d\n",vertices[2]->index, vertices[1]->index, vertices[3]->index);
//  fprintf(stderr,"  face3 %d %d %d\n",vertices[0]->index, vertices[2]->index, vertices[3]->index);

  // check which vertices are already visited

  SVvertex* v_visited[4];
  int num_v_visited = 0;
  int last_v_visited = 0;
  int last_v_unvisited = 0;
  int num_v_cached = 0;

  for (i = 0; i < 4; i++)
  {
    if (vertices[i]->use_count)
    {
      num_v_visited++;
      last_v_visited = i;
      v_visited[i] = vertices[i];
      if (vertices[i]->in_cache != -1)
      {
        num_v_cached++;
      }
    }
    else
    {
      last_v_unvisited = i;
      v_visited[i] = 0;
    }
  }

  if (num_v_cached >= 4)
  {
    fprintf(stderr,"WEIRD !!!!!!!!!!\n");
  }
//  assert(num_v_cached < 4);

  // check which faces are already visited

  SVedge* help;
  SVedge* f_visited[] = {0,0,0,0};
  int num_f_visited = 0;

  if (num_v_visited >= 3)
  {
    // check face 0 (0-1-2)
    if (v_visited[0] && v_visited[1] && v_visited[2])
    {
      for (i = 0; i < v_visited[1]->list_size; i++)
      {
        if (v_visited[1]->list[i]->target == v_visited[0])
        {
          if (v_visited[1]->list[i]->next->target == v_visited[2])
          {
            f_visited[0] = v_visited[1]->list[i];
            num_f_visited++;
            break;
          }
        }
      }
    }

    // check face 1 (1-0-3)
    if (v_visited[1] && v_visited[0] && v_visited[3])
    {
      for (i = 0; i < v_visited[0]->list_size; i++)
      {
        if (v_visited[0]->list[i]->target == v_visited[1])
        {
          if (v_visited[0]->list[i]->next->target == v_visited[3])
          {
            f_visited[1] = v_visited[0]->list[i];
            num_f_visited++;
            break;
          }
        }
      }
    }

    // check face 2 (2-1-3)
    if (v_visited[2] && v_visited[1] && v_visited[3])
    {
      for (i = 0; i < v_visited[1]->list_size; i++)
      {
        if (v_visited[1]->list[i]->target == v_visited[2])
        {
          if (v_visited[1]->list[i]->next->target == v_visited[3])
          {
            f_visited[2] = v_visited[1]->list[i];
            num_f_visited++;
            break;
          }
        }
      }
    }

    // check face 3 (2-3-0)
    if (v_visited[2] && v_visited[3] && v_visited[0])
    {
      for (i = 0; i < v_visited[2]->list_size; i++)
      {
        if (v_visited[2]->list[i]->target == v_visited[0])
        {
          if (v_visited[2]->list[i]->next->target == v_visited[3])
          {
            f_visited[3] = v_visited[2]->list[i];
            num_f_visited++;
            break;
          }
        }
      }
    }
  }

  if (num_f_visited == 0) // SVC_START or SVC_START (vertex-join of 1-4 vertices) or SVC_START (edge-join of 1-6 edges) 
  {
    op_start++;
    op_start_x[num_v_visited]++;
#ifdef PRINT_CONTROL_OUTPUT
    if (num_v_cached >= 2) start_hit_edge_possible++;
    if (num_v_cached) start_hit_vertex++;
    else  start_miss++;
#endif

    if (edge_buffer_size)
    {
//      re_conn_op->encode(rmOp[last_op], SVC_START);
      re_conn_op->encode(rmOpX[last_fin][last_op], SVC_START);
    }
    else
    {
      re_conn_op->encode(rmDone, 0); // more to come
    }
    op = SVC_START;

    // specify how many old (non-manifold) vertices there are

    re_conn_op->encode(rmS_Old, num_v_visited);

    // rotate such that (non-manifold) vertices are 0, 1, 2, 3 (in this order) and such that (if possible) 0 and 1 are in the cache
    if (num_v_visited)
    {
      if (num_v_visited == 1)
      {
        if (last_v_visited == 1)
        {
          rot[0] = 1;
          rot[1] = 2;
          rot[2] = 0;
        }
        else if (last_v_visited == 2)
        {
          rot[0] = 2;
          rot[1] = 0;
          rot[2] = 1;
        }
        else if (last_v_visited == 3)
        {
          rot[1] = 0;
          rot[0] = 3;
          rot[3] = 1;
        }
      }
      else if (num_v_visited == 2)
      {
        if (v_visited[2])
        {
          if (v_visited[3])      // move 2/3 to 0/1
          {
            rot[0] = 2;
            rot[1] = 3;
            rot[2] = 0;
            rot[3] = 1;
          }
          else if (v_visited[1]) // move 1/2 to 0/1
          {
            rot[0] = 1;
            rot[1] = 2;
            rot[2] = 0;
          }
          else if (v_visited[0]) // move 2/0 to 0/1
          {
            rot[0] = 2;
            rot[1] = 0;
            rot[2] = 1;
          }
        }
        else if (v_visited[3])
        {
          if (v_visited[1]) // move 3/1 to 1/0
          {
            rot[1] = 3;
            rot[0] = 1;
            rot[3] = 0;
          }
          else if (v_visited[0]) // move 0/3 to 1/0
          {
            rot[1] = 0;
            rot[0] = 3;
            rot[3] = 1;
          }
        }

        // put v0 and v1 into the right "cache" order
        if (num_v_cached == 1)
        {
          if (v_visited[rot[1]]->in_cache != -1)
          {
            i = rot[0];
            rot[0] = rot[1];
            rot[1] = i;
            i = rot[2];
            rot[2] = rot[3];
            rot[3] = i;
          }
        }
      }
      else if (num_v_visited == 3)
      {
        if (last_v_unvisited == 0)
        {
          rot[3] = 0;
          rot[1] = 3;
          rot[0] = 1;
        }
        else if (last_v_unvisited == 1)
        {
          rot[3] = 1;
          rot[1] = 0;
          rot[0] = 3;
        }
        else if (last_v_unvisited == 2)
        {
          rot[3] = 2;
          rot[2] = 1;
          rot[1] = 3;
        }

        // put v0, v1, and v2 into the right "cache" order
        if (num_v_cached == 1)
        {
          if (v_visited[rot[1]]->in_cache != -1)
          {
            i = rot[0];
            rot[0] = rot[1];
            rot[1] = rot[2];
            rot[2] = i;
          }
          else if (v_visited[rot[2]]->in_cache != -1)
          {
            i = rot[0];
            rot[0] = rot[2];
            rot[2] = rot[1];
            rot[1] = i;
          }
          assert(v_visited[rot[0]]->in_cache != -1);
        }
        else if (num_v_cached == 2)
        {
          if (v_visited[rot[0]]->in_cache == -1)
          {
            i = rot[0];
            rot[0] = rot[1];
            rot[1] = rot[2];
            rot[2] = i;
          }
          else if (v_visited[rot[1]]->in_cache == -1)
          {
            i = rot[0];
            rot[0] = rot[2];
            rot[2] = rot[1];
            rot[1] = i;
          }
          assert(v_visited[rot[0]]->in_cache != -1);
          assert(v_visited[rot[1]]->in_cache != -1);
        }
      }
    }

    // compress vertices

    SVvertex* vertex;

    for (int i = 0; i < 4; i++)
    {
      vertex = vertices[rot[i]];

      if (vertex->use_count)
      {
        num_v_visited--;
        // encode the index of v0 (-> improve with cache of with prediction)
        lc_pos_v = vertex->in_cache;
        if (lc_pos_v == -1)
        {
          re_conn_cache->encode(rmS_Cache[i],4);        
          // encode its index among all active vertices
          dv_index = dv->getRelativeIndex(vertex);
          re_conn_index->writeRange(dv->size(),dv_index);        
  #ifdef PRINT_CONTROL_OUTPUT
          used_index++;
  #endif
        }
        else
        {
          re_conn_cache->encode(rmS_Cache[i],lc_pos_v);        
  #ifdef PRINT_CONTROL_OUTPUT
          used_cache++;
  #endif
        }
      }
      else
      {
        assert(num_v_visited == 0);
        // encode its position
        if (i == 3)
        {
#if defined(SVC_USE_LOCAL_COORDINATE_SYSTEM)
          compressVertexPropertiesLocal(vertices[rot[0]]->prop_f, vertices[rot[1]]->prop_f, vertices[rot[2]]->prop_f, vertices[rot[3]]->prop_f);
          if (v_pq)
          {
            v_pq->EnQuantize(vertices[rot[3]]->prop_f, (I32*)(vertices[rot[3]]->prop_f));
          }
          else
          {
            VecCopy3fv(vertices[rot[3]]->prop_f, vertices[rot[3]]->prop_f);
          }
#else
          compressVertexPropertiesGlobal(vertices[rot[0]]->prop_f, vertices[rot[1]]->prop_f, vertices[rot[2]]->prop_f, vertices[rot[3]]->prop_f);
#endif
        }
        else
        {
          if (i > 0)
          {
            compressVertexProperties(vertices[rot[i-1]]->prop_f, vertex->prop_f, v_pnum_f);
          }
          else
          {
            compressVertexProperties(vertex->prop_f, v_pnum_f);
          }
#if defined(SVC_USE_LOCAL_COORDINATE_SYSTEM)
          if (v_pq)
          {
            v_pq->DeQuantize((I32*)(vertex->prop_f), vertex->prop_f);
          }
          else
          {
            VecCopy3fv(vertex->prop_f, vertex->prop_f);
          }
#endif
        }
        // insert it into the indexable data structure
        dv->addElement(vertex);
      }
    }
  }
  else if (num_f_visited == 1) // SVC_ADD or SVC_ADD (vertex-join) or SVC_ADD (edge-join of 1, 2, or 3 edges) 
  {
    if (num_v_visited == 3)
    {
//      re_conn_op->encode(rmOp[last_op], SVC_ADD);
      re_conn_op->encode(rmOpX[last_fin][last_op], SVC_ADD);
      op = SVC_ADD;
#ifdef PRINT_CONTROL_OUTPUT
      op_add++;
#endif
    }
    else
    {
//      re_conn_op->encode(rmOp[last_op], SVC_JOIN);
      re_conn_op->encode(rmOpX[last_fin][last_op], SVC_JOIN);
      op = SVC_JOIN;
#ifdef PRINT_CONTROL_OUTPUT
      op_join++;
#endif
    }

    // rotate the tetrahedron such that the known face is face 0 and
    // put the vertices in the appropriately rotated order.

    if (f_visited[1])
    {
      f_visited[0] = f_visited[1];
      f_visited[1] = 0;
      rot[0] = 1;
      rot[1] = 0;
      rot[2] = 3;
      rot[3] = 2;
    }
    else if (f_visited[2])
    {
      f_visited[0] = f_visited[2];
      f_visited[2] = 0;
      rot[0] = 2;
      rot[1] = 1;
      rot[2] = 3;
      rot[3] = 0;
    }
    else if (f_visited[3])
    {
      f_visited[0] = f_visited[3];
      f_visited[3] = 0;
      rot[0] = 0;
      rot[1] = 2;
      rot[2] = 3;
      rot[3] = 1;
    }
    
    assert(f_visited[0]);
    
    // check whether an edge of the known face is in the face cache
    // if yes, then rotate the face appropriately.

    if (num_v_cached == 3)
    {
      lc_pos_f = lc_f->pos(f_visited[0]);
      if (lc_pos_f != -1) // check the first edge
      {
        // great!!!
      }
      else
      {
        lc_pos_f = lc_f->pos(f_visited[0]->next);
        if (lc_pos_f != -1) // check the second edge
        {
          // great!!!
          f_visited[0] = f_visited[0]->next;
          i = rot[0];
          rot[0] = rot[2];
          rot[2] = rot[1];
          rot[1] = i;
        }
        else
        {
          lc_pos_f = lc_f->pos(f_visited[0]->next->next);
          if (lc_pos_f != -1) // check the third edge
          {
            // great!!!
            f_visited[0] = f_visited[0]->next->next;
            i = rot[0];
            rot[0] = rot[1];
            rot[1] = rot[2];
            rot[2] = i;
          }
        }
      }
    }
    
    assert(f_visited[0]->target == vertices[rot[0]]);
    assert(f_visited[0]->next->target == vertices[rot[2]]);
    assert(f_visited[0]->next->next->target == vertices[rot[1]]);

    if (lc_pos_f != -1)
    {
      // yes ... we can use a face from the cache
//      re_conn_cache->encode(rmAJ_FaceCache[last_op],lc_pos_f);
      re_conn_cache->encode(rmAJ_FaceCacheX[last_fin][last_op],lc_pos_f);
#ifdef PRINT_CONTROL_OUTPUT
      if (op == SVC_ADD)
      {
        add_hit_face++;
      }
      else
      {
        join_hit_face++;
      }
      assert(lc_f->pos(f_visited[0]) == lc_pos_f);
#endif
    }
    else
    {
      // no ... we cannot use a face from the cache
//      re_conn_cache->encode(rmAJ_FaceCache[last_op],4);
      re_conn_cache->encode(rmAJ_FaceCacheX[last_fin][last_op],4);

      // does that mean that the number of cached vertices must be
      // smaller than three? no ... there may be duplicate or wrongly
      // oriented tetrahedra

      if (num_v_cached >= 2 && (op == SVC_ADD || vertices[rot[3]]->in_cache == -1))
      {
#ifdef PRINT_CONTROL_OUTPUT
        if (op == SVC_ADD)
        {
          add_hit_edge++;
        }
        else
        {
          join_hit_edge++;
        }
#endif

        if (vertices[rot[1]]->in_cache == -1)
        {
          // need to rotate triangle
          f_visited[0] = f_visited[0]->next;
          i = rot[0];
          rot[0] = rot[2];
          rot[2] = rot[1];
          rot[1] = i;
        }
        else if (vertices[rot[0]]->in_cache == -1)
        {
          // need to rotate triangle
          f_visited[0] = f_visited[0]->next->next;
          i = rot[0];
          rot[0] = rot[1];
          rot[1] = rot[2];
          rot[2] = i;
        }
        i = vertices[rot[1]]->in_cache * 4 + vertices[rot[0]]->in_cache - 1;

        re_conn_cache->encode(rmAJ_EdgeCache[last_op],i);

        // encode which of the faces incident to v1 and v0 is f0
        // how many potential candidates are around that vertex?
        // encode which of the faces incident to v0 and v1 is f1
        candidate_count = 0;
        candidate = -1;
        // how many potential candidates are around the vertex with the shorter face list
        if (vertices[rot[0]]->list_size < vertices[rot[1]]->list_size)
        {
          for (i = 0; i < vertices[rot[0]]->list_size; i++)
          {
            if (vertices[rot[0]]->list[i]->next->target == vertices[rot[1]])
            {
              if (vertices[rot[0]]->list[i]->next->next == f_visited[0])
              {
                assert(candidate == -1);
                candidate = candidate_count;
              }
              candidate_count++;
            }
          }
          assert(candidate != -1);
          if (candidate_count > 1)
          {
            if (candidate_count == 2)
            {
              re_conn_candi->encode(rm_FaceCandi,candidate);
            }
            else
            {
              re_conn_candi->writeRange(candidate_count,candidate);
            }
          }
        }
        else
        {
          for (i = 0; i < vertices[rot[1]]->list_size; i++)
          {
            if (vertices[rot[1]]->list[i]->target == vertices[rot[0]])
            {
              if (vertices[rot[1]]->list[i] == f_visited[0])
              {
                assert(candidate == -1);
                candidate = candidate_count;
              }
              candidate_count++;
            }
          }
          assert(candidate != -1);
          if (candidate_count > 1)
          {
            if (candidate_count == 2)
            {
              re_conn_candi->encode(rm_FaceCandi,candidate);
            }
            else
            {
              re_conn_candi->writeRange(candidate_count,candidate);
            }
          }
        }
      }
      else 
      {
        // no .... we cannot use an edge from the cache
        re_conn_cache->encode(rmAJ_EdgeCache[last_op],4);

        // check whether at least one vertex of the known face is in the vertex
        // cache ... if yes, then rotate the face appropriately
        if (num_v_cached)
        {
          if (vertices[rot[1]]->in_cache != -1) 
          {
            // great
            lc_pos_v = vertices[rot[1]]->in_cache;
          }
          else if (vertices[rot[0]]->in_cache != -1) // this is vertex f_visited[0]->target
          {
            // great
            lc_pos_v = vertices[rot[0]]->in_cache;
            // but need to rotate triangle
            f_visited[0] = f_visited[0]->next;
            i = rot[0];
            rot[0] = rot[2];
            rot[2] = rot[1];
            rot[1] = i;
          }
          else if (vertices[rot[2]]->in_cache != -1) // this is vertex f_visited[0]->next->target
          {
            // great
            lc_pos_v = vertices[rot[2]]->in_cache;
            // but need to rotate triangle
            f_visited[0] = f_visited[0]->next->next;
            i = rot[0];
            rot[0] = rot[1];
            rot[1] = rot[2];
            rot[2] = i;
          }
        }

        assert(f_visited[0]->target == vertices[rot[0]]);
        assert(f_visited[0]->next->target == vertices[rot[2]]);
        assert(f_visited[0]->next->next->target == vertices[rot[1]]);

        if (lc_pos_v != -1)
        {
          // yes ... we can use a vertex from the cache
          re_conn_cache->encode(rmAJ_VertexCache[last_op],lc_pos_v);
  #ifdef PRINT_CONTROL_OUTPUT
          if (op == SVC_JOIN)
          {
            join_hit_vertex++;
          }
          else
          {
            add_hit_vertex++;
          }
          assert(vertices[rot[1]]->in_cache == lc_pos_v);
  #endif
        }
        else
        {
          // no ... we cannot use a vertex from the cache
          re_conn_cache->encode(rmAJ_VertexCache[last_op],4);
          dv_index = dv->getRelativeIndex(vertices[rot[1]]);
          re_conn_index->writeRange(dv->size(),dv_index);
  #ifdef PRINT_CONTROL_OUTPUT
          if (op == SVC_JOIN)
          {
            join_miss++;
          }
          else
          {
            add_miss++;
          }
  #endif
        }

        // encode which of the faces incident to v1 is f0
        // how many potential candidates are around that vertex?
        for (i = 0; i < vertices[rot[1]]->list_size; i++)
        {
          if (vertices[rot[1]]->list[i] == f_visited[0])
          {
            if (vertices[rot[1]]->list_size > 1)
            {
              if (vertices[rot[1]]->list_size < 14)
              {
                re_conn_candi->encode(rm_VertexCandi[vertices[rot[1]]->list_size],i);
              }
              else
              {
                re_conn_candi->writeRange(vertices[rot[1]]->list_size,i);
              }
            }
            break;
          }
        }
        assert(i < vertices[rot[1]]->list_size);
      }
    }

    // encode v3
    if (op == SVC_JOIN)
    {
#ifdef PRINT_CONTROL_OUTPUT
      if (vertices[rot[3]]->in_cache != -1)
      {
        join_hit_join_vertex++;
      }
#endif
      // try to find the join vertex


      bool encoded = false;

      int flag = 0;

      int order[] = {1,0,2};

      for (int j = 0; j<3; j++)
      {
        int k = order[j];
        for (i = 0; i < vertices[rot[k]]->list_size; i++)
        {
          if (vertices[rot[k]]->list[i]->target == vertices[rot[3]])
          {
//            fprintf(stderr,"join is target of %d edge of %d\n",i,k); 
            flag = flag | (1 << (2*k));
            if (!encoded)
            {
              re_conn->encode(rmAJ_LocalJoin[last_op],2*k);
              re_conn->writeRange(vertices[rot[k]]->list_size,i);
              encoded = true;
            }
          }
          if (vertices[rot[k]]->list[i]->next->target == vertices[rot[3]])
          {
//            fprintf(stderr,"join is next->target of %d edge of %d\n",i,k); 
            flag = flag | (1 << (2*k+1));
            if (!encoded)
            {
              re_conn->encode(rmAJ_LocalJoin[last_op],2*k+1);
              re_conn->writeRange(vertices[rot[k]]->list_size,i);
              encoded = true;
            }
          }
          assert(vertices[rot[k]]->list[i]->next->next->target == vertices[rot[k]]);
        }
      }
#ifdef PRINT_CONTROL_OUTPUT
      join_edge_hist[flag]++; 
#endif
      if (!encoded)
      {
        re_conn->encode(rmAJ_LocalJoin[last_op],6);
        // old position
        dv_index = dv->getRelativeIndex(vertices[rot[3]]);
        re_conn_index->writeRange(dv->size(),dv_index);
      }
    }
    else
    {
      // new position
#if defined(SVC_USE_LOCAL_COORDINATE_SYSTEM)

  #if defined(SVC_DO_TETRAHEDRAL_ACROSS_PREDICTION)
      assert(VecEqual3fv(f_visited[0]->across,f_visited[0]->next->across));
      compressVertexPropertiesLocal(f_visited[0]->across, vertices[rot[0]]->pos, vertices[rot[1]]->pos, vertices[rot[2]]->pos, vertices[rot[3]]->pos);
  #elif defined(SVC_DO_FLIP_PREDICTION)
      assert(VecEqual3fv(f_visited[0]->across,f_visited[0]->next->across));
      compressVertexPropertiesLocalFlip(f_visited[0]->across, vertices[rot[0]]->pos, vertices[rot[1]]->pos, vertices[rot[2]]->pos, vertices[rot[3]]->pos);
  #else // mid-point (maybe with base normal prediction)
      compressVertexPropertiesLocal(vertices[rot[0]]->pos, vertices[rot[1]]->pos, vertices[rot[2]]->pos, vertices[rot[3]]->pos);
  #endif

#else

  #if defined(SVC_DO_TETRAHEDRAL_ACROSS_PREDICTION)
      assert(VecEqual3fv(f_visited[0]->across,f_visited[0]->next->across));
      compressVertexPropertiesGlobal(f_visited[0]->across, vertices[rot[0]]->prop_f, vertices[rot[1]]->prop_f, vertices[rot[2]]->prop_f, vertices[rot[3]]->prop_f);
  #elif defined(SVC_DO_FLIP_PREDICTION)
      assert(VecEqual3fv(f_visited[0]->across,f_visited[0]->next->across));
      compressVertexPropertiesGlobalFlip(f_visited[0]->across, vertices[rot[0]]->prop_f, vertices[rot[1]]->prop_f, vertices[rot[2]]->prop_f, vertices[rot[3]]->prop_f);
  #else // mid-point (maybe with base normal prediction)
      compressVertexPropertiesGlobal(vertices[rot[0]]->prop_f, vertices[rot[1]]->prop_f, vertices[rot[2]]->prop_f, vertices[rot[3]]->prop_f);
  #endif

#endif

#if defined(SVC_USE_LOCAL_COORDINATE_SYSTEM)
      if (v_pq)
      {
        v_pq->EnQuantize(vertices[rot[3]]->pos, (I32*)(vertices[rot[3]]->prop_f));
      }
      else
      {
        VecCopy3fv(vertices[rot[3]]->pos, vertices[rot[3]]->prop_f);
      }
#endif

      assert(vertices[rot[3]]->use_count == 0);
      // insert it into the indexable data structure
      dv->addElement(vertices[rot[3]]);
    }
  }
  else
  {
#ifdef PRINT_CONTROL_OUTPUT
    if (num_f_visited == 2)
      op_flip++;
    else if (num_f_visited == 3)
      op_fill++;
    else
      op_end++;
#endif

//    re_conn_op->encode(rmOp[last_op], SVC_FLIP);
    re_conn_op->encode(rmOpX[last_fin][last_op], SVC_FLIP);

    op = SVC_FLIP;

    // rotate such that the known faces are face 0 and face 1 with the vertices in the appropriate order 
    if (num_f_visited == 2)
    {
      if (f_visited[0] && f_visited[1])
      {
      }
      else if (f_visited[0] && f_visited[2])
      {
        f_visited[0] = f_visited[0]->next->next;
        f_visited[1] = f_visited[2];
        f_visited[2] = 0;
        rot[0] = 1;
        rot[1] = 2;
        rot[2] = 0;
        rot[3] = 3;
      }
      else if (f_visited[0] && f_visited[3])
      {
        f_visited[0] = f_visited[0]->next;
        f_visited[1] = f_visited[3];
        f_visited[3] = 0;
        rot[0] = 2;
        rot[1] = 0;
        rot[2] = 1;
        rot[3] = 3;
      }
      else if (f_visited[1] && f_visited[2])
      {
        f_visited[0] = f_visited[1]->next;
        f_visited[1] = f_visited[2]->next->next;
        f_visited[2] = 0;
        rot[0] = 3;
        rot[1] = 1;
        rot[2] = 0;
        rot[3] = 2;
      }
      else if (f_visited[1] && f_visited[3])
      {
        f_visited[0] = f_visited[1]->next->next;
        f_visited[1] = f_visited[3]->next;
        f_visited[3] = 0;
        rot[0] = 0;
        rot[1] = 3;
        rot[2] = 1;
        rot[3] = 2;
      }
      else
      {
        assert(f_visited[2] && f_visited[3]);

        f_visited[0] = f_visited[2]->next;
        f_visited[1] = f_visited[3]->next->next;
        f_visited[2] = 0;
        f_visited[3] = 0;
        rot[0] = 3;
        rot[1] = 2;
        rot[2] = 1;
        rot[3] = 0;
      }

      assert(f_visited[0]->target == vertices[rot[0]]);
      assert(f_visited[0]->next->target == vertices[rot[2]]);
      assert(f_visited[0]->next->next->target == vertices[rot[1]]);

      assert(f_visited[1]->target == vertices[rot[1]]);
      assert(f_visited[1]->next->target == vertices[rot[3]]);
      assert(f_visited[1]->next->next->target == vertices[rot[0]]);

      // check whether an edge of one of the known faces is in the
      // face cache if yes, then rotate the face appropriately.

      if (num_v_cached == 3)
      {
        lc_pos_f = lc_f->pos(f_visited[0]);
        if (lc_pos_f != -1) // check the first edge
        {
          // great!!!
        }
        else
        {
          lc_pos_f = lc_f->pos(f_visited[1]);
          if (lc_pos_f != -1) // check the symmetrically other first edge
          {
            // great!!!
            help = f_visited[0];
            f_visited[0] = f_visited[1];
            f_visited[1] = help;
            i = rot[0];
            rot[0] = rot[1];
            rot[1] = i;
            i = rot[2];
            rot[2] = rot[3];
            rot[3] = i;
          }
          else
          {
            lc_pos_f = lc_f->pos(f_visited[0]->next);
            if (lc_pos_f != -1) // check the second edge
            {
              // great!!!
              lc_pos_f += 4;
            }
            else
            {
              lc_pos_f = lc_f->pos(f_visited[1]->next);
              if (lc_pos_f != -1) // check the symmetrically other second edge
              {
                // great!!!
                lc_pos_f += 4;
                help = f_visited[0];
                f_visited[0] = f_visited[1];
                f_visited[1] = help;
                i = rot[0];
                rot[0] = rot[1];
                rot[1] = i;
                i = rot[2];
                rot[2] = rot[3];
                rot[3] = i;
              }
              else
              {
                lc_pos_f = lc_f->pos(f_visited[0]->next->next);
                if (lc_pos_f != -1) // check the third edge
                {
                  // great!!!
                  lc_pos_f += 8;
                }
                else
                {
                  lc_pos_f = lc_f->pos(f_visited[1]->next->next);
                  if (lc_pos_f != -1) // check the symmetrically other third edge
                  {
                    // great!!!
                    lc_pos_f += 8;
                    help = f_visited[0];
                    f_visited[0] = f_visited[1];
                    f_visited[1] = help;
                    i = rot[0];
                    rot[0] = rot[1];
                    rot[1] = i;
                    i = rot[2];
                    rot[2] = rot[3];
                    rot[3] = i;
                  }
                }                
              }
            }
          }
        }
      }

      if (lc_pos_f != -1)
      {
#ifdef PRINT_CONTROL_OUTPUT
        flip_hit_face++;
#endif
        if (lc_pos_f < 4)
          assert(lc_f->pos(f_visited[0]) == lc_pos_f);
        else if (lc_pos_f < 8)
          assert(lc_f->pos(f_visited[0]->next) == lc_pos_f-4);
        else 
          assert(lc_f->pos(f_visited[0]->next->next) == lc_pos_f-8);
      }
      else
      {
        // check whether two vertices or one vertex of one of the known
        // faces are in vertex cache. if yes, then rotate the face appropriately

        if (num_v_cached)
        {
          if (vertices[rot[1]]->in_cache != -1)
          {
            if (vertices[rot[2]]->in_cache != -1)
            {
              // edge v1->v2 is in cache
              lc_pos_e = 1;
            }
            else if (vertices[rot[3]]->in_cache != -1)
            {
              // edge v0->v2 (rot) is in cache
              lc_pos_e = 2;
              // we have to rotate the tetrahedron 
              help = f_visited[0];
              f_visited[0] = f_visited[1];
              f_visited[1] = help;
              i = rot[0];
              rot[0] = rot[1];
              rot[1] = i;
              i = rot[2];
              rot[2] = rot[3];
              rot[3] = i;
            }
            else
            {
              // vertex v1 is in cache
              lc_pos_v = vertices[rot[1]]->in_cache;
            }
          }
          else if (vertices[rot[0]]->in_cache != -1)
          {
            if (vertices[rot[3]]->in_cache != -1)
            {
              // edge v1->v2 (rot) is in cache
              lc_pos_e = 1;
              // we have to rotate the tetrahedron 
              help = f_visited[0];
              f_visited[0] = f_visited[1];
              f_visited[1] = help;
              i = rot[0];
              rot[0] = rot[1];
              rot[1] = i;
              i = rot[2];
              rot[2] = rot[3];
              rot[3] = i;
            }
            else if (vertices[rot[2]]->in_cache != -1)
            {
              // edge v0->v2 is in cache
              lc_pos_e = 2; 
            }
            else
            {
              // vertex v1 (rot) is in cache
              lc_pos_v = vertices[rot[0]]->in_cache;
              // we have to rotate the tetrahedron 
              help = f_visited[0];
              f_visited[0] = f_visited[1];
              f_visited[1] = help;
              i = rot[0];
              rot[0] = rot[1];
              rot[1] = i;
              i = rot[2];
              rot[2] = rot[3];
              rot[3] = i;
            }
          }
          else if (vertices[rot[2]]->in_cache != -1) // this is vertex f_visited[0]->next->target
          {
            // vertex v2 is in the cache
            lc_pos_v = 4 + vertices[rot[2]]->in_cache;
          }
          else if (vertices[rot[3]]->in_cache != -1) // this is vertex f_visited[1]->next->target
          {
            // vertex v2 (rot) is in the cache
            lc_pos_v = 4 + vertices[rot[3]]->in_cache;
            // but have to rotate the tetrahedron 
            help = f_visited[0];
            f_visited[0] = f_visited[1];
            f_visited[1] = help;
            i = rot[0];
            rot[0] = rot[1];
            rot[1] = i;
            i = rot[2];
            rot[2] = rot[3];
            rot[3] = i;
          }
        
          if (lc_pos_e != -1)
          {
#ifdef PRINT_CONTROL_OUTPUT
            flip_hit_edge++;
#endif
            if (lc_pos_e == 1)
              assert(vertices[rot[1]]->in_cache != -1 && vertices[rot[2]]->in_cache != -1);
            else if (lc_pos_e == 2)
              assert(vertices[rot[0]]->in_cache != -1 && vertices[rot[2]]->in_cache != -1);
            else
              assert(false);
          }
          else if (lc_pos_v != -1)
          {
#ifdef PRINT_CONTROL_OUTPUT
            flip_hit_vertex++;
#endif
            if (lc_pos_v < 4)
              assert(vertices[rot[1]]->in_cache == lc_pos_v && f_visited[0]->next->next->target == vertices[rot[1]]);
            else
              assert(vertices[rot[2]]->in_cache == lc_pos_v-4 && f_visited[0]->next->target == vertices[rot[2]]);
          }
          else
          {
            assert(false);
          }
        }
        else
        {
#ifdef PRINT_CONTROL_OUTPUT
          flip_miss++;
#endif
        }
      }
      assert(f_visited[0]->target == vertices[rot[0]]);
      assert(f_visited[0]->next->target == vertices[rot[2]]);
      assert(f_visited[0]->next->next->target == vertices[rot[1]]);

      assert(f_visited[1]->target == vertices[rot[1]]);
      assert(f_visited[1]->next->target == vertices[rot[3]]);
      assert(f_visited[1]->next->next->target == vertices[rot[0]]);
    }
    else if (num_f_visited == 3)
    {
      // rotate such that the known faces are face 0, face 1, and face 2 with the vertices in the appropriate order 
      if (f_visited[0] && f_visited[1] && f_visited[2])
      {
      }
      else if (f_visited[0] && f_visited[1] && f_visited[3])
      {
        f_visited[2] = f_visited[3]->next;
        f_visited[3] = f_visited[0];
        f_visited[0] = f_visited[1];
        f_visited[1] = f_visited[3];
        f_visited[3] = 0;
        rot[0] = 1;
        rot[1] = 0;
        rot[2] = 3;
        rot[3] = 2;
      }
      else if (f_visited[0] && f_visited[2] && f_visited[3])
      {
        f_visited[0] = f_visited[0]->next->next;
        f_visited[1] = f_visited[2];
        f_visited[2] = f_visited[3];
        f_visited[3] = 0;
        rot[0] = 1;
        rot[1] = 2;
        rot[2] = 0;
        rot[3] = 3;
      }
      else
      {
        assert(f_visited[1] && f_visited[2] && f_visited[3]);

        f_visited[0] = f_visited[1]->next->next;
        f_visited[1] = f_visited[3]->next;
        f_visited[2] = f_visited[2]->next->next;
        f_visited[3] = 0;
        rot[0] = 0;
        rot[1] = 3;
        rot[2] = 1;
        rot[3] = 2;
      }

      assert(f_visited[0]->target == vertices[rot[0]]);
      assert(f_visited[0]->next->target == vertices[rot[2]]);
      assert(f_visited[0]->next->next->target == vertices[rot[1]]);

      assert(f_visited[1]->target == vertices[rot[1]]);
      assert(f_visited[1]->next->target == vertices[rot[3]]);
      assert(f_visited[1]->next->next->target == vertices[rot[0]]);

      assert(f_visited[2]->target == vertices[rot[2]]);
      assert(f_visited[2]->next->target == vertices[rot[3]]);
      assert(f_visited[2]->next->next->target == vertices[rot[1]]);

      // check whether an edge of one of the known faces is in the
      // face cache if yes, then rotate the face appropriately.

      if (num_v_cached == 3)
      {
        lc_pos_f = lc_f->pos(f_visited[0]);
        if (lc_pos_f != -1) // check the first edge of face 0
        {
          // great!!!
        }
        else
        {
          lc_pos_f = lc_f->pos(f_visited[1]->next);
          if (lc_pos_f != -1) // check the symmetrically first edge of face 1
          {
            // great!!!
            help = f_visited[0];
            f_visited[0] = f_visited[1]->next;
            f_visited[1] = f_visited[2]->next->next;
            f_visited[2] = help;
            i = rot[0];
            rot[0] = rot[3];
            rot[3] = rot[2];
            rot[2] = i;
          }
          else
          {
            lc_pos_f = lc_f->pos(f_visited[2]);
            if (lc_pos_f != -1) // check the symmetrically first edge of face 2
            {
              // great!!!
              help = f_visited[0];
              f_visited[0] = f_visited[2];
              f_visited[2] = f_visited[1]->next;
              f_visited[1] = help->next->next;
              i = rot[0];
              rot[0] = rot[2];
              rot[2] = rot[3];
              rot[3] = i;
            }
            else
            {
              lc_pos_f = lc_f->pos(f_visited[0]->next);
              if (lc_pos_f != -1) // check the second edge of face 0
              {
                // great!!!
                lc_pos_f += 4;
              }
              else
              {
                lc_pos_f = lc_f->pos(f_visited[1]->next->next);
                if (lc_pos_f != -1) // check the symmetrically second edge of face 1
                {
                  // great!!!
                  lc_pos_f += 4;
                  // but need to rotate tet
                  help = f_visited[0];
                  f_visited[0] = f_visited[1]->next;
                  f_visited[1] = f_visited[2]->next->next;
                  f_visited[2] = help;
                  i = rot[0];
                  rot[0] = rot[3];
                  rot[3] = rot[2];
                  rot[2] = i;
                }
                else
                {
                  lc_pos_f = lc_f->pos(f_visited[2]->next);
                  if (lc_pos_f != -1) // check the symmetrically second edge of face 2
                  {
                    // great!!!
                    lc_pos_f += 4;
                    // but need to rotate tet
                    help = f_visited[0];
                    f_visited[0] = f_visited[2];
                    f_visited[2] = f_visited[1]->next;
                    f_visited[1] = help->next->next;
                    i = rot[0];
                    rot[0] = rot[2];
                    rot[2] = rot[3];
                    rot[3] = i;
                  }
                  else
                  {
                    lc_pos_f = lc_f->pos(f_visited[0]->next->next);
                    if (lc_pos_f != -1) // check the third edge of face 0
                    {
                      // great!!!
                      lc_pos_f += 8;
                    }
                    else
                    {
                      lc_pos_f = lc_f->pos(f_visited[1]);
                      if (lc_pos_f != -1) // check the symmetrically third edge of face 1
                      {
                        // great!!!
                        lc_pos_f += 8;
                        // but need to rotate tet
                        help = f_visited[0];
                        f_visited[0] = f_visited[1]->next;
                        f_visited[1] = f_visited[2]->next->next;
                        f_visited[2] = help;
                        i = rot[0];
                        rot[0] = rot[3];
                        rot[3] = rot[2];
                        rot[2] = i;
                      }
                      else
                      {
                        lc_pos_f = lc_f->pos(f_visited[2]->next->next);
                        if (lc_pos_f != -1) // check the symmetrically third edge of face 2
                        {
                          // great!!!
                          lc_pos_f += 8;
                          // but need to rotate tet
                          help = f_visited[0];
                          f_visited[0] = f_visited[2];
                          f_visited[2] = f_visited[1]->next;
                          f_visited[1] = help->next->next;
                          i = rot[0];
                          rot[0] = rot[2];
                          rot[2] = rot[3];
                          rot[3] = i;
                        }
                      }
                    }
                  }
                }
              }
            }
          }
        }
      }

      if (lc_pos_f != -1)
      {
#ifdef PRINT_CONTROL_OUTPUT
        fill_hit_face++;
#endif
        if (lc_pos_f < 4)
          assert(lc_f->pos(f_visited[0]) == lc_pos_f);
        else if (lc_pos_f < 8)
          assert(lc_f->pos(f_visited[0]->next) == lc_pos_f-4);
        else 
          assert(lc_f->pos(f_visited[0]->next->next) == lc_pos_f-8);
      }
      else
      {
        // check whether one or two vertices of one of known faces are in
        // the vertex cache. if yes, then rotate the face appropriately

        if (num_v_cached)
        {
          if (vertices[rot[1]]->in_cache != -1) 
          {
            if (vertices[rot[2]]->in_cache != -1) 
            {
              // vertex v1 and v2 are in the cache
              lc_pos_e = 1;
            }
            else if (vertices[rot[3]]->in_cache != -1) 
            {
              // vertex v1 and v2 (rot) are in the cache
              lc_pos_e = 1;
              // but need to rotate face
              help = f_visited[0];
              f_visited[0] = f_visited[2];
              f_visited[2] = f_visited[1]->next;
              f_visited[1] = help->next->next;
              i = rot[0];
              rot[0] = rot[2];
              rot[2] = rot[3];
              rot[3] = i;
            }
            else if (vertices[rot[0]]->in_cache != -1) 
            {
              // vertex v1 and v2 (rot rot) are in the cache
              lc_pos_e = 1;
              // but need to rotate face
              help = f_visited[0];
              f_visited[0] = f_visited[1]->next;
              f_visited[1] = f_visited[2]->next->next;
              f_visited[2] = help;
              i = rot[0];
              rot[0] = rot[3];
              rot[3] = rot[2];
              rot[2] = i;
            }
            else
            {
              lc_pos_v = vertices[rot[1]]->in_cache;
            }
          }
          else if (vertices[rot[2]]->in_cache != -1)
          {
            if (vertices[rot[0]]->in_cache != -1) 
            {
              // vertex v0 and v2 are in the cache
              lc_pos_e = 2;
            }
            else
            {
              // great
              lc_pos_v = 4 + vertices[rot[2]]->in_cache;
            }
          }
          else if (vertices[rot[0]]->in_cache != -1) 
          {
            if (vertices[rot[3]]->in_cache != -1) 
            {
              // vertex v0 and v2 (rot rot) are in the cache
              lc_pos_e = 2;
            }
            else
            {
              // great
              lc_pos_v = 4 + vertices[rot[0]]->in_cache;
            }
            // but need to rotate face
            help = f_visited[0];
            f_visited[0] = f_visited[1]->next;
            f_visited[1] = f_visited[2]->next->next;
            f_visited[2] = help;
            i = rot[0];
            rot[0] = rot[3];
            rot[3] = rot[2];
            rot[2] = i;
          }
          else if (vertices[rot[3]]->in_cache != -1) 
          {
            if (vertices[rot[2]]->in_cache != -1) 
            {
              // vertex v0 and v2 (rot) are in the cache
              lc_pos_e = 2;
            }
            else
            {
              // great
              lc_pos_v = 4 + vertices[rot[3]]->in_cache;
            }
            // but need to rotate face
            help = f_visited[0];
            f_visited[0] = f_visited[2];
            f_visited[2] = f_visited[1]->next;
            f_visited[1] = help->next->next;
            i = rot[0];
            rot[0] = rot[2];
            rot[2] = rot[3];
            rot[3] = i;
          }
        }

        if (lc_pos_e != -1)
        {
#ifdef PRINT_CONTROL_OUTPUT
          fill_hit_edge++;
#endif
          if (lc_pos_e == 1)
            assert(vertices[rot[1]]->in_cache != -1 && vertices[rot[2]]->in_cache != -1);
          else
            assert(vertices[rot[0]]->in_cache != -1 && vertices[rot[2]]->in_cache != -1);
        }
        else if (lc_pos_v != -1)
        {
#ifdef PRINT_CONTROL_OUTPUT
          fill_hit_vertex++;
#endif
          if (lc_pos_v < 4)
            assert(vertices[rot[1]]->in_cache == lc_pos_v);
          else
            assert(vertices[rot[2]]->in_cache == lc_pos_v-4);
        }
        else
        {
#ifdef PRINT_CONTROL_OUTPUT
          fill_miss++;
#endif
        }
      }
      assert(f_visited[0]->target == vertices[rot[0]]);
      assert(f_visited[0]->next->target == vertices[rot[2]]);
      assert(f_visited[0]->next->next->target == vertices[rot[1]]);

      assert(f_visited[1]->target == vertices[rot[1]]);
      assert(f_visited[1]->next->target == vertices[rot[3]]);
      assert(f_visited[1]->next->next->target == vertices[rot[0]]);

      assert(f_visited[2]->target == vertices[rot[2]]);
      assert(f_visited[2]->next->target == vertices[rot[3]]);
      assert(f_visited[2]->next->next->target == vertices[rot[1]]);
    }
    else
    {
      assert(num_f_visited == 4);

      // check whether an edge of one of the four faces is in the
      // face cache if yes, then rotate the face appropriately.

      if (num_v_cached == 3)
      {
        lc_pos_f = lc_f->pos(f_visited[0]);
        if (lc_pos_f != -1) // check the first edge of face 0
        {
          // great!!!
        }
        else
        {
          lc_pos_f = lc_f->pos(f_visited[0]->next);
          if (lc_pos_f != -1) // check the second edge of face 0
          {
            // great!!!
            f_visited[0] = f_visited[0]->next;
            help = f_visited[1];
            f_visited[1] = f_visited[3];
            f_visited[3] = f_visited[2];
            f_visited[2] = help;
            i = rot[0];
            rot[0] = rot[2];
            rot[2] = rot[1];
            rot[1] = i;
          }
          else
          {
            lc_pos_f = lc_f->pos(f_visited[0]->next->next);
            if (lc_pos_f != -1) // check the second edge of face 0
            {
              // great!!!
              f_visited[0] = f_visited[0]->next->next;
              help = f_visited[1];
              f_visited[1] = f_visited[2];
              f_visited[2] = f_visited[3];
              f_visited[3] = help;
              i = rot[0];
              rot[0] = rot[1];
              rot[1] = rot[2];
              rot[2] = i;
            }
            else
            {
              lc_pos_f = lc_f->pos(f_visited[1]);
              if (lc_pos_f != -1) // check the first edge of face 1
              {
                // great!!!
                help = f_visited[0];
                f_visited[0] = f_visited[1];
                f_visited[1] = help;
                help = f_visited[2];
                f_visited[2] = f_visited[3]->next;
                f_visited[3] = help->next->next;
                i = rot[0];
                rot[0] = rot[1];
                rot[1] = i;
                i = rot[2];
                rot[2] = rot[3];
                rot[3] = i;
              }
              else
              {
                lc_pos_f = lc_f->pos(f_visited[1]->next);
                if (lc_pos_f != -1) // check the second edge of face 1
                {
                  // great!!!
                  help = f_visited[0];
                  f_visited[0] = f_visited[1]->next;
                  f_visited[1] = f_visited[2]->next->next;
                  f_visited[2] = help;
                  f_visited[3] = f_visited[3]->next;
                  i = rot[0];
                  rot[0] = rot[3];
                  rot[3] = rot[2];
                  rot[2] = i;
                }
                else
                {
                  lc_pos_f = lc_f->pos(f_visited[1]->next->next);
                  if (lc_pos_f != -1) // check the third edge of face 1
                  {
                    // great!!!
                    help = f_visited[0];
                    f_visited[0] = f_visited[1]->next->next;
                    f_visited[1] = f_visited[3]->next;
                    f_visited[2] = f_visited[2]->next->next;
                    f_visited[3] = help;
                    i = rot[1];
                    rot[1] = rot[3];
                    rot[3] = rot[2];
                    rot[2] = i;
                  }
                  else
                  {
                    lc_pos_f = lc_f->pos(f_visited[2]);
                    if (lc_pos_f != -1) // check the first edge of face 2
                    {
                      // great!!!
                      help = f_visited[0];
                      f_visited[0] = f_visited[2];
                      f_visited[2] = f_visited[1]->next;
                      f_visited[1] = help->next->next;
                      f_visited[3] = f_visited[3]->next->next;
                      i = rot[0];
                      rot[0] = rot[2];
                      rot[2] = rot[3];
                      rot[3] = i;
                    }
                    else
                    {
                      lc_pos_f = lc_f->pos(f_visited[2]->next);
                      if (lc_pos_f != -1) // check the second edge of face 2
                      {
                        // great!!!
                        help = f_visited[0];
                        f_visited[0] = f_visited[2]->next;
                        f_visited[2] = help->next->next;
                        help = f_visited[3];
                        f_visited[3] = f_visited[1]->next;
                        f_visited[1] = help->next->next;
                        i = rot[0];
                        rot[0] = rot[3];
                        rot[3] = i;
                        i = rot[1];
                        rot[1] = rot[2];
                        rot[2] = i;
                      }
                      else
                      {
                        lc_pos_f = lc_f->pos(f_visited[2]->next->next);
                        if (lc_pos_f != -1) // check the third edge of face 2
                        {
                          // great!!!
                          help = f_visited[0];
                          f_visited[0] = f_visited[2]->next->next;
                          f_visited[1] = f_visited[1]->next;
                          f_visited[2] = f_visited[3]->next->next;
                          f_visited[3] = help->next->next;
                          i = rot[0];
                          rot[0] = rot[1];
                          rot[1] = rot[3];
                          rot[3] = i;
                        }
                        else
                        {
                          lc_pos_f = lc_f->pos(f_visited[3]);
                          if (lc_pos_f != -1) // check the first edge of face 3
                          {
                            // great!!!
                            help = f_visited[0];
                            f_visited[0] = f_visited[3];
                            f_visited[2] = f_visited[2]->next;
                            f_visited[3] = f_visited[1]->next->next;
                            f_visited[1] = help->next;
                            i = rot[1];
                            rot[1] = rot[2];
                            rot[2] = rot[3];
                            rot[3] = i;
                          }
                          else
                          {
                            lc_pos_f = lc_f->pos(f_visited[3]->next);
                            if (lc_pos_f != -1) // check the second edge of face 3
                            {
                              // great!!!
                              help = f_visited[0];
                              f_visited[0] = f_visited[3]->next;
                              f_visited[1] = f_visited[1]->next->next;
                              f_visited[3] = f_visited[2]->next;
                              f_visited[2] = help->next;
                              i = rot[0];
                              rot[0] = rot[3];
                              rot[3] = rot[1];
                              rot[1] = i;
                            }
                            else
                            {
                              lc_pos_f = lc_f->pos(f_visited[3]->next->next);
                              if (lc_pos_f != -1) // check the third edge of face 3
                              {
                                // great!!!
                                help = f_visited[0];
                                f_visited[0] = f_visited[3]->next->next;
                                f_visited[3] = help->next;
                                help = f_visited[1];
                                f_visited[1] = f_visited[2]->next;
                                f_visited[2] = help->next->next;
                                i = rot[0];
                                rot[0] = rot[2];
                                rot[2] = i;
                                i = rot[1];
                                rot[1] = rot[3];
                                rot[3] = i;
                              }
                            }
                          }
                        }
                      }
                    }
                  }
                }
              }
            }
          }
        }
      }

      if (lc_pos_f != -1)
      {
#ifdef PRINT_CONTROL_OUTPUT
        end_hit_face++;
#endif
        assert(lc_f->pos(f_visited[0]) == lc_pos_f);
      }
      else
      {
        // check whether a vertex of one of the known faces is in the vertex cache
        if (num_v_cached)
        {
          if (vertices[rot[1]]->in_cache != -1) 
          {
            // great
            lc_pos_v = vertices[rot[1]]->in_cache;
          }
          else if (vertices[rot[2]]->in_cache != -1)
          {
            lc_pos_v = vertices[rot[2]]->in_cache;
            // need to rotate face
            f_visited[0] = f_visited[0]->next->next;
            help = f_visited[1];
            f_visited[1] = f_visited[2];
            f_visited[2] = f_visited[3];
            f_visited[3] = help;
            i = rot[0];
            rot[0] = rot[1];
            rot[1] = rot[2];
            rot[2] = i;
          }
          else if (vertices[rot[0]]->in_cache != -1)
          {
            lc_pos_v = vertices[rot[0]]->in_cache;
            // need to rotate face
            f_visited[0] = f_visited[0]->next;
            help = f_visited[1];
            f_visited[1] = f_visited[3];
            f_visited[3] = f_visited[2];
            f_visited[2] = help;
            i = rot[0];
            rot[0] = rot[2];
            rot[2] = rot[1];
            rot[1] = i;
          }
          else if (vertices[rot[3]]->in_cache != -1) 
          {
            lc_pos_v = vertices[rot[3]]->in_cache;
            // need to rotate face
            help = f_visited[0];
            f_visited[0] = f_visited[2]->next->next;
            f_visited[1] = f_visited[1]->next;
            f_visited[2] = f_visited[3]->next->next;
            f_visited[3] = help->next->next;
            i = rot[0];
            rot[0] = rot[1];
            rot[1] = rot[3];
            rot[3] = i;
          }
          else
          {
            assert(false);
          }
        }

#ifdef PRINT_CONTROL_OUTPUT
        if (lc_pos_v != -1)
        {
          end_hit_vertex++;
        }
        else
        {
          end_miss++;
        }
#endif
      }

      assert(f_visited[0]->target == vertices[rot[0]]);
      assert(f_visited[0]->next->target == vertices[rot[2]]);
      assert(f_visited[0]->next->next->target == vertices[rot[1]]);

      assert(f_visited[1]->target == vertices[rot[1]]);
      assert(f_visited[1]->next->target == vertices[rot[3]]);
      assert(f_visited[1]->next->next->target == vertices[rot[0]]);

      assert(f_visited[2]->target == vertices[rot[2]]);
      assert(f_visited[2]->next->target == vertices[rot[3]]);
      assert(f_visited[2]->next->next->target == vertices[rot[1]]);

      assert(f_visited[3]->target == vertices[rot[0]]);
      assert(f_visited[3]->next->target == vertices[rot[3]]);
      assert(f_visited[3]->next->next->target == vertices[rot[2]]);
    }

    // try to encode face 0 with the face cache

    if (lc_pos_f != -1)
    {
//      re_conn_cache->encode(rmFE_FaceCache[last_op],lc_pos_f);
      re_conn_cache->encode(rmFE_FaceCacheX[last_fin][last_op],lc_pos_f);

//      fprintf(stderr,"flip face %d last_op %d\n",lc_pos_f,last_op);
    }
    else
    {
//      re_conn_cache->encode(rmFE_FaceCache[last_op],12);
      re_conn_cache->encode(rmFE_FaceCacheX[last_fin][last_op],12);

      // try to encode an edge of face 0 with the edge cache
//      fprintf(stderr,"flip face %d last_op %d\n",12,last_op);
      if (lc_pos_e != -1)
      {
        candidate_count = 0;
        candidate = -1;
        if (lc_pos_e == 1)
        {
          // vertex v1 and vertex v2 are in the cache
          lc_pos_e = vertices[rot[1]]->in_cache * 4 + vertices[rot[2]]->in_cache - 1;
          re_conn_cache->encode(rmFE_EdgeCache[last_op],lc_pos_e);

          // which of the faces incident to v1 and v2 is f0 (use vertex with shorter list)
          if (vertices[rot[1]]->list_size < vertices[rot[2]]->list_size)
          {
            for (i = 0; i < vertices[rot[1]]->list_size; i++)
            {
              if (vertices[rot[1]]->list[i]->next->target == vertices[rot[2]])
              {
                if (vertices[rot[1]]->list[i] == f_visited[0])
                {
                  assert(candidate == -1);
                  candidate = candidate_count;
                }
                candidate_count++;
              }
            }
          }
          else
          {
            for (i = 0; i < vertices[rot[2]]->list_size; i++)
            {
              if (vertices[rot[2]]->list[i]->target == vertices[rot[1]])
              {
                if (vertices[rot[2]]->list[i]->next == f_visited[0])
                {
                  assert(candidate == -1);
                  candidate = candidate_count;
                }
                candidate_count++;
              }
            }
          }
        }
        else
        {
          // vertex v0 and vertex v2 are in the cache
          lc_pos_e = vertices[rot[0]]->in_cache * 4 + vertices[rot[2]]->in_cache - 1 + 14;
          re_conn_cache->encode(rmFE_EdgeCache[last_op],lc_pos_e);

          // which of the faces incident to v0 and v2 is f0 (use vertex with shorter list)
          if (vertices[rot[0]]->list_size < vertices[rot[2]]->list_size)
          {
            for (i = 0; i < vertices[rot[0]]->list_size; i++)
            {
              if (vertices[rot[0]]->list[i]->target == vertices[rot[2]])
              {
                if (vertices[rot[0]]->list[i]->next->next == f_visited[0])
                {
                  assert(candidate == -1);
                  candidate = candidate_count;
                }
                candidate_count++;
              }
            }
          }
          else
          {
            for (i = 0; i < vertices[rot[2]]->list_size; i++)
            {
              if (vertices[rot[2]]->list[i]->next->target == vertices[rot[0]])
              {
                if (vertices[rot[2]]->list[i]->next == f_visited[0])
                {
                  assert(candidate == -1);
                  candidate = candidate_count;
                }
                candidate_count++;
              }
            }
          }
        }
        assert(candidate_count != 0 && candidate != -1);
        if (candidate_count > 1)
        {
          if (candidate_count == 2)
          {
            re_conn_candi->encode(rm_FaceCandi,candidate);
          }
          else
          {
            re_conn_candi->writeRange(candidate_count,candidate);
          }
        }
      }
      else
      {
        re_conn_cache->encode(rmFE_EdgeCache[last_op],4);

        if (lc_pos_v != -1)
        {
          re_conn_cache->encode(rmFE_VertexCache[last_op],lc_pos_v);
//          fprintf(stderr,"flip vertex %d last_op %d\n",lc_pos_v,last_op);
        }
        else
        {
          re_conn_cache->encode(rmFE_VertexCache[last_op],8);
//          fprintf(stderr,"flip vertex %d last_op %d\n",8,last_op);
          dv_index = dv->getRelativeIndex(vertices[rot[1]]);
          re_conn_index->writeRange(dv->size(),dv_index);
//          fprintf(stderr,"flip index %d size %d\n",dv_index,dv->size());
        }

        if (lc_pos_v < 4)
        {
          // encode which of the faces around v1 is f0
          for (i = 0; i < vertices[rot[1]]->list_size; i++)
          {
            if (vertices[rot[1]]->list[i] == f_visited[0])
            {
              if (vertices[rot[1]]->list_size > 1)
              {
                if (vertices[rot[1]]->list_size < 14)
                {
                  re_conn_candi->encode(rm_VertexCandi[vertices[rot[1]]->list_size],i);
                }
                else
                {
                  re_conn_candi->writeRange(vertices[rot[1]]->list_size,i);
                }
              }
              break;
            }
          }
          assert(i < vertices[rot[1]]->list_size);
        }
        else
        {
          // encode which of the faces around v2 is f0
          for (i = 0; i < vertices[rot[2]]->list_size; i++)
          {
            if (vertices[rot[2]]->list[i]->next == f_visited[0])
            {
              if (vertices[rot[2]]->list_size > 1)
              {
                if (vertices[rot[2]]->list_size < 14)
                {
                  re_conn_candi->encode(rm_VertexCandi[vertices[rot[2]]->list_size],i);
                }
                else
                {
                  re_conn_candi->writeRange(vertices[rot[2]]->list_size,i);
                }
              }
              break;
            }
          }
          assert(i < vertices[rot[2]]->list_size);
        }
      }
    }

    // encode which of the faces incident to v0 and v1 is f1
    candidate_count = 0;
    candidate = -1;
    // how many potential candidates are around the vertex with the shorter face list
    if (vertices[rot[0]]->list_size < vertices[rot[1]]->list_size)
    {
      for (i = 0; i < vertices[rot[0]]->list_size; i++)
      {
        if (vertices[rot[0]]->list[i]->target == vertices[rot[1]])
        {
          if (vertices[rot[0]]->list[i] == f_visited[1])
          {
            assert(candidate == -1);
            candidate = candidate_count;
          }
          candidate_count++;
        }
      }
    }
    else
    {
      for (i = 0; i < vertices[rot[1]]->list_size; i++)
      {
        if (vertices[rot[1]]->list[i]->next->target == vertices[rot[0]])
        {
          if (vertices[rot[1]]->list[i]->next->next == f_visited[1])
          {
            assert(candidate == -1);
            candidate = candidate_count;
          }
          candidate_count++;
        }
      }
    }
    assert(candidate_count != 0 && candidate != -1);
    if (candidate_count > 1)
    {
      if (candidate_count == 2)
      {
        re_conn_candi->encode(rm_FaceCandi,candidate);
      }
      else
      {
        re_conn_candi->writeRange(candidate_count,candidate);
      }
    }
    // the decoder will need to figure this out (but we also should do it here)
    if (num_f_visited == 3)
    {
      op = SVC_FILL;
    }
    else if (num_f_visited == 4)
    {
      op = SVC_END;
    }
  }

//  fprintf(stderr,"op %d %d\n",op,c_count); 

  // statistic

//  fprintf(stderr,"reordered %d %d %d %d\n",vertices[rot[0]]->index, vertices[rot[1]]->index, vertices[rot[2]]->index, vertices[rot[3]]->index);

  // update data structures

  if (f_visited[0])
  {
    removeEdgeFromVertex(f_visited[0]->next->next, vertices[rot[2]]);
    removeEdgeFromVertex(f_visited[0]->next, vertices[rot[0]]);
    removeEdgeFromVertex(f_visited[0], vertices[rot[1]]);

    deallocEdge(f_visited[0]->next->next);
    deallocEdge(f_visited[0]->next);
    deallocEdge(f_visited[0]);

//    fprintf(stderr,"remove %d %d %d\n",f_visited[0]->target->index, f_visited[0]->next->target->index, f_visited[0]->next->next->target->index);
    f_visited[0] = 0;
  }
  else
  {
    SVedge* e0 = allocEdge();
    SVedge* e1 = allocEdge();
    SVedge* e2 = allocEdge();

    e0->next = e1;
    e1->next = e2;
    e2->next = e0;

    e0->target = vertices[rot[1]];
    e1->target = vertices[rot[2]];
    e2->target = vertices[rot[0]];

    addEdgeToVertex(e0, vertices[rot[0]]);
    addEdgeToVertex(e1, vertices[rot[1]]);
    addEdgeToVertex(e2, vertices[rot[2]]);

#if defined(SVC_DO_TETRAHEDRAL_ACROSS_PREDICTION) ||  defined(SVC_DO_FLIP_PREDICTION)
  #ifdef SVC_USE_LOCAL_COORDINATE_SYSTEM
    VecCopy3fv(e0->across, vertices[rot[3]]->pos);
    VecCopy3fv(e1->across, vertices[rot[3]]->pos);
    VecCopy3fv(e2->across, vertices[rot[3]]->pos);
  #else
    VecCopy3fv(e0->across, vertices[rot[3]]->prop_f);
    VecCopy3fv(e1->across, vertices[rot[3]]->prop_f);
    VecCopy3fv(e2->across, vertices[rot[3]]->prop_f);
  #endif
#endif
    //    fprintf(stderr,"add %d %d %d\n",e0->target->index, e0->next->target->index, e0->next->next->target->index);
    f_visited[0] = e0;
  }

  if (f_visited[1])
  {
    removeEdgeFromVertex(f_visited[1]->next->next, vertices[rot[3]]);
    removeEdgeFromVertex(f_visited[1]->next, vertices[rot[1]]);
    removeEdgeFromVertex(f_visited[1], vertices[rot[0]]);

    deallocEdge(f_visited[1]->next->next);
    deallocEdge(f_visited[1]->next);
    deallocEdge(f_visited[1]);

//    fprintf(stderr,"remove %d %d %d\n",f_visited[1]->target->index, f_visited[1]->next->target->index, f_visited[1]->next->next->target->index);
    f_visited[1] = 0;
  }
  else
  {
    SVedge* e00 = allocEdge();
    SVedge* e01 = allocEdge();
    SVedge* e02 = allocEdge();

    e00->next = e01;
    e01->next = e02;
    e02->next = e00;

    e00->target = vertices[rot[0]];
    e01->target = vertices[rot[3]];
    e02->target = vertices[rot[1]];

    addEdgeToVertex(e00, vertices[rot[1]]);
    addEdgeToVertex(e01, vertices[rot[0]]);
    addEdgeToVertex(e02, vertices[rot[3]]);

#if defined(SVC_DO_TETRAHEDRAL_ACROSS_PREDICTION) ||  defined(SVC_DO_FLIP_PREDICTION)
  #ifdef SVC_USE_LOCAL_COORDINATE_SYSTEM
    VecCopy3fv(e00->across, vertices[rot[2]]->pos);
    VecCopy3fv(e01->across, vertices[rot[2]]->pos);
    VecCopy3fv(e02->across, vertices[rot[2]]->pos);
  #else
    VecCopy3fv(e00->across, vertices[rot[2]]->prop_f);
    VecCopy3fv(e01->across, vertices[rot[2]]->prop_f);
    VecCopy3fv(e02->across, vertices[rot[2]]->prop_f);
  #endif
#endif
//    fprintf(stderr,"add %d %d %d\n",e00->target->index, e00->next->target->index, e00->next->next->target->index);
    f_visited[1] = e00;
  }

  if (f_visited[2])
  {
    removeEdgeFromVertex(f_visited[2]->next->next, vertices[rot[3]]);
    removeEdgeFromVertex(f_visited[2]->next, vertices[rot[2]]);
    removeEdgeFromVertex(f_visited[2], vertices[rot[1]]);

    deallocEdge(f_visited[2]->next->next);
    deallocEdge(f_visited[2]->next);
    deallocEdge(f_visited[2]);

//    fprintf(stderr,"remove %d %d %d\n",f_visited[2]->target->index, f_visited[2]->next->target->index, f_visited[2]->next->next->target->index);
    f_visited[2] = 0;
  }
  else
  {
    SVedge* e10 = allocEdge();
    SVedge* e11 = allocEdge();
    SVedge* e12 = allocEdge();

    e10->next = e11;
    e11->next = e12;
    e12->next = e10;

    e10->target = vertices[rot[1]];
    e11->target = vertices[rot[3]];
    e12->target = vertices[rot[2]];

    addEdgeToVertex(e10, vertices[rot[2]]);
    addEdgeToVertex(e11, vertices[rot[1]]);
    addEdgeToVertex(e12, vertices[rot[3]]);

#if defined(SVC_DO_TETRAHEDRAL_ACROSS_PREDICTION) ||  defined(SVC_DO_FLIP_PREDICTION)
  #ifdef SVC_USE_LOCAL_COORDINATE_SYSTEM
    VecCopy3fv(e10->across, vertices[rot[0]]->pos);
    VecCopy3fv(e11->across, vertices[rot[0]]->pos);
    VecCopy3fv(e12->across, vertices[rot[0]]->pos);
  #else
    VecCopy3fv(e10->across, vertices[rot[0]]->prop_f);
    VecCopy3fv(e11->across, vertices[rot[0]]->prop_f);
    VecCopy3fv(e12->across, vertices[rot[0]]->prop_f);
  #endif
#endif
//    fprintf(stderr,"add %d %d %d\n",e10->target->index, e10->next->target->index, e10->next->next->target->index);
    f_visited[2] = e10;
  }

  if (f_visited[3])
  {
    removeEdgeFromVertex(f_visited[3]->next->next, vertices[rot[3]]);
    removeEdgeFromVertex(f_visited[3]->next, vertices[rot[0]]);
    removeEdgeFromVertex(f_visited[3], vertices[rot[2]]);

    deallocEdge(f_visited[3]->next->next);
    deallocEdge(f_visited[3]->next);
    deallocEdge(f_visited[3]);

//    fprintf(stderr,"remove %d %d %d\n",f_visited[3]->target->index, f_visited[3]->next->target->index, f_visited[3]->next->next->target->index);
    f_visited[3] = 0;
  }
  else
  {
    SVedge* e20 = allocEdge();
    SVedge* e21 = allocEdge();
    SVedge* e22 = allocEdge();

    e20->next = e21;
    e21->next = e22;
    e22->next = e20;

    e20->target = vertices[rot[2]];
    e21->target = vertices[rot[3]];
    e22->target = vertices[rot[0]];

    addEdgeToVertex(e20, vertices[rot[0]]);
    addEdgeToVertex(e21, vertices[rot[2]]);
    addEdgeToVertex(e22, vertices[rot[3]]);

#if defined(SVC_DO_TETRAHEDRAL_ACROSS_PREDICTION) ||  defined(SVC_DO_FLIP_PREDICTION)
  #ifdef SVC_USE_LOCAL_COORDINATE_SYSTEM
    VecCopy3fv(e20->across, vertices[rot[1]]->pos);
    VecCopy3fv(e21->across, vertices[rot[1]]->pos);
    VecCopy3fv(e22->across, vertices[rot[1]]->pos);
  #else
    VecCopy3fv(e20->across, vertices[rot[1]]->prop_f);
    VecCopy3fv(e21->across, vertices[rot[1]]->prop_f);
    VecCopy3fv(e22->across, vertices[rot[1]]->prop_f);
  #endif
#endif

//    fprintf(stderr,"add %d %d %d\n",e20->target->index, e20->next->target->index, e20->next->next->target->index);
    f_visited[3] = e20;
  }

  // update last_op

  last_op = op;

  // increment vertex use_counts, create edges, and update edge degrees

  for (i = 0; i < 4; i++)
  {
    vertices[i]->use_count++;
  }

  last_fin = 0;

  // write finalization info and finalize vertices
  for (i = 0; i < 4; i++)
  {
    degree_one = vertices[rot[i]]->list_size;
    if (degree_one >= SVC_MAX_DEGREE_ONE)
    {
      degree_one = SVC_MAX_DEGREE_ONE-1;
    }
    use_count = vertices[rot[i]]->use_count;
    if (use_count >= SVC_MAX_USE_COUNT)
    {
      use_count = SVC_MAX_USE_COUNT-1;
    }

#if !defined(SVC_DO_NOT_FINALIZE) && !defined(SVC_GUESS_FINALIZATION)
    re_conn_final->encode(rmFinalized[degree_one][use_count],vertices[rot[i]]->finalize);

//    fprintf(stderr, "fin %d %d %d %d\n", degree_one, use_count, vertices[rot[i]]->finalize, last_op);

    if (vertices[rot[i]]->finalize)
    {
      if (i < 2) last_fin = last_fin | (1<<i);

      SVvertex* vertex = vertices[rot[i]];
      vertices[rot[i]] = 0;
      vertex_hash->erase(hash_elements[rot[i]]);
//      fprintf(stderr,"remove vertex %d\n",vertex->index);
      dv->removeElement(vertex);
      for (j = 0; j < vertex->list_size; j++)
      {
        SVedge* edge = vertex->list[j];
        removeEdgeFromVertex(edge->next->next, edge->next->target);
        removeEdgeFromVertex(edge->next, edge->target);
//        removeEdgeFromVertex(edge, edge->next->next->target);
        deallocEdge(edge->next->next);
        deallocEdge(edge->next);
        deallocEdge(edge);
      }
      deallocVertex(vertex);
    }
#elif defined(SVC_GUESS_FINALIZATION)

    if (vertices[rot[i]]->list_size == 0)
    {
      if (i < 2) last_fin = last_fin | (1<<i);

      SVvertex* vertex = vertices[rot[i]];
      vertices[rot[i]] = 0;
      vertex_hash->erase(hash_elements[rot[i]]);
//      fprintf(stderr,"remove vertex %d\n",vertex->index);
      dv->removeElement(vertex);
      for (j = 0; j < vertex->list_size; j++)
      {
        SVedge* edge = vertex->list[j];
        removeEdgeFromVertex(edge->next->next, edge->next->target);
        removeEdgeFromVertex(edge->next, edge->target);
//        removeEdgeFromVertex(edge, edge->next->next->target);
        deallocEdge(edge->next->next);
        deallocEdge(edge->next);
        deallocEdge(edge);
      }
      deallocVertex(vertex);
    }
#endif
  }

  // clean cache flag of current elements in cache
  for (i = 0; i < 4; i++)
  {
    SVvertex* vertex = (SVvertex*)lc->get(i);
    if (vertex) vertex->in_cache = -1;
  }

  // update cache elements
  lc->put(vertices[rot[0]], vertices[rot[1]], vertices[rot[2]], vertices[rot[3]]);
  lc_f->put(f_visited[0], f_visited[1], f_visited[2], f_visited[3]);

  // set cache flag of current elements in cache
  for (i = 0; i < 4; i++)
  {
    if (vertices[rot[i]]) vertices[rot[i]]->in_cache = i;
  }
}

void SVwriter_svc::write_tetrahedron(const I32* c_idx, const bool* c_final)
{
  if (!c_final)
  {
    fprintf(stderr, "ERROR: delayed finalization not supported by SVwriter_svc\n");
    exit(0);
  }

  compress_tetrahedron(c_idx, 0, c_final);

  c_count++;
}

void SVwriter_svc::write_tetrahedron(const I32* c_idx, const float* c_prop_f, const bool* c_final)
{
  if (!c_final)
  {
    fprintf(stderr, "ERROR: delayed finalization not supported by SVwriter_svc\n");
    exit(0);
  }

  compress_tetrahedron(c_idx, c_prop_f, c_final);

  c_count++;
}

void SVwriter_svc::write_hexahedron(const I32* c_idx, const bool* c_final)
{
    fprintf(stderr, "WARNING: SVwriter_svc does not implement write_hexahedron()\n");
}

void SVwriter_svc::write_hexahedron(const I32* c_idx, const float* c_prop_f, const bool* c_final)
{
    fprintf(stderr, "WARNING: SVwriter_svc does not implement write_hexahedron()\n");
}

void SVwriter_svc::write_finalized(I32 final_idx)
{
  fprintf(stderr, "ERROR: delayed finalization not supported by SVwriter_svc\n");
  exit(0);
}

void SVwriter_svc::close(bool close_file, bool update_header)
{
  // close of SVwriter_svc
  if (edge_buffer_size)
  {
//    re_conn_op->encode(rmOp[last_op], SVC_START);
    re_conn_op->encode(rmOpX[last_fin][last_op], SVC_START);
    re_conn_op->encode(rmS_Old, 5);
  }
  else
  {
    assert(dv->size() == 0);
    re_conn_op->encode(rmDone, 1); // done
  }

#ifdef SVC_CALCULATE_PREDICTION_ERROR 
  if (total_pred_err_normal)
  {
    fprintf(stdout,"err %.1f %.1f %.1f ",(float)(total_pred_err/(predict_inside+predict_across)),(float)(total_pred_err_normal/(predict_inside+predict_across)),(float)(total_pred_err_tangential/(predict_inside+predict_across)));
  }
  else
  {
    fprintf(stdout,"err %.1f ",(float)(total_pred_err/(predict_inside+predict_across)));
  }
#endif

#ifdef SVC_CALCULATE_QUANTIZATION_ERROR 
  fprintf(stdout,"qnt %g ",max_quant_err);
#endif

  if (normal_not_valid_counter) fprintf(stderr,"normal_not_valid_counter %d\n",normal_not_valid_counter);

#ifdef PRINT_CONTROL_OUTPUT
  fprintf(stderr,"edge_buffer_size %d edge_buffer_maxsize %d\n",edge_buffer_size,edge_buffer_maxsize);
  fprintf(stderr,"vertex_buffer_size %d vertex_buffer_maxsize %d\n",vertex_buffer_size,vertex_buffer_maxsize);
#else
//  fprintf(stderr,"edge_buffer_size %d\n",edge_buffer_size);
//  fprintf(stderr,"vertex_buffer_size %d\n",vertex_buffer_size);
#endif

  finishEncoder(v_count, c_count);
  if (v_pq)
  {
    global_ic[0]->FinishCompressor();
    global_ic[1]->FinishCompressor();
    global_ic[2]->FinishCompressor();
    delete global_ic[0];
    delete global_ic[1];
    delete global_ic[2];
    delete v_pq;
  }
  else
  {
    v_fc[0]->FinishCompressor();
    v_fc[1]->FinishCompressor();
    v_fc[2]->FinishCompressor();
    delete v_fc[0];
    delete v_fc[1];
    delete v_fc[2];
  }
  finishModels();

  destroyVertexBuffer();
  destroyEdgeBuffer();

  if (dv->size()) fprintf(stderr,"WARNING: there are %d unfinalized vertices in dynamic vector\n",dv->size());
  if (vertex_hash->size()) fprintf(stderr,"WARNING: there are %d unfinalized vertices in vertex hash\n",(int)vertex_hash->size());

  delete dv;
  delete vertex_hash;
  delete lc;
  delete lc_f;

#ifdef SVC_STORE_QUANTIZED_COORDINATES
  fclose(quantized_coordinates_file);
#endif

  // close of SVwriter interface

  if (nverts != -1) if (nverts != v_count) fprintf(stderr,"WARNING: set nverts %d but v_count %d\n",nverts,v_count);
  if (ncells != -1) if (ncells != c_count) fprintf(stderr,"WARNING: set ncells %d but c_count %d\n",ncells,c_count);
  nverts = v_count;
  ncells = c_count;
  v_count = -1;
  c_count = -1;
  if (close_file && file && file != stdout) fclose(file);
  file = 0;
}

void SVwriter_svc::write_header()
{
  int i;

  // initialize internal memory management
  initVertexBuffer(1024, v_pnum_f);
  initEdgeBuffer(1024);

  // write nverts
  if (nverts == -1)
  {
    re_conn->writeBit(0);
  }
  else
  {
    re_conn->writeBit(1);
    re_conn->writeInt(nverts);
  }
  // write ncells
  if (ncells == -1)
  {
    re_conn->writeBit(0);
  }
  else
  {
    re_conn->writeBit(1);
    re_conn->writeInt(ncells);
  }
  // write which properties are there
  re_geom->writeByte((!!v_pnum_f) | ((!!c_pnum_f) << 2));
#ifdef PRINT_CONTROL_OUTPUT
  printf("property mapping is %d\n",((!!v_pnum_f) | ((!!c_pnum_f) << 2)));
#endif

  if (v_pnum_f)
  {
    re_geom->writeByte(v_pnum_f);
#ifdef PRINT_CONTROL_OUTPUT
      printf("have %d float vertex attributes\n",v_pnum_f);
#endif
    if (v_pmin_f && v_pmax_f)
    {
      re_geom->writeBit(1);
#ifdef PRINT_CONTROL_OUTPUT
      printf("have bounding box\n");
#endif
      for (i=0; i < v_pnum_f; i++)
      {
        re_geom->writeFloat(v_pmin_f[i]);
        re_geom->writeFloat(v_pmax_f[i]);
      }
    }
    else
    {
      re_geom->writeBit(0);
#ifdef PRINT_CONTROL_OUTPUT
      printf("have no bounding box\n");
#endif
    }
  }
  if (0) // v_pnum_i
  {
    re_geom->writeByte(0);
    if (0) 
    {
    }
    else
    {
      re_geom->writeBit(0);
    }
  }
  if (c_pnum_f)
  {
    re_geom->writeByte(c_pnum_f);
    if (c_pmin_f && c_pmax_f)
    {
      re_geom->writeBit(1);
      for (i=0; i < c_pnum_f; i++)
      {
        re_geom->writeFloat(c_pmin_f[i]);
        re_geom->writeFloat(c_pmax_f[i]);
      }
    }
    else
    {
      re_geom->writeBit(0);
#ifdef PRINT_CONTROL_OUTPUT
      printf("have no bounding box\n");
#endif
    }
  }
  if (0)
  {
    re_geom->writeByte(0);
    if (0)
    {
      re_geom->writeBit(1);
    }
    else
    {
      re_geom->writeBit(0);
    }
  }

  if (true) // if have no comments
  {
    re_conn->writeBit(0);
#ifdef PRINT_CONTROL_OUTPUT
    printf("have no comments\n");
#endif
  }
  else
  {
    /* yet to be implemented */
  }

  // setup float quantizers and integer compressors

  if (v_pnum_f < 3)
  {
    fprintf(stderr,"WARNING: v_pnum_f is %d ... not implemented yet\n", v_pnum_f);
    exit(1);
  }

  re_geom->writeRange(17,high_bits);
#ifdef PRINT_CONTROL_OUTPUT
    printf("high_bits %d\n", high_bits);
#endif

  if (v_pmin_f && v_pmax_f && nbits)
  {
    fprintf(stderr,"quantizing positions within bounding box to %d bits\n",nbits);
#ifdef PRINT_CONTROL_OUTPUT
    fprintf(stderr,"quantizing v_prop_f with %d bits\n",nbits);
#endif
    v_pq = new PositionQuantizerNew();
    v_pq->SetMinMax(v_pmin_f, v_pmax_f);
    v_pq->SetPrecision(nbits);

    if (cbm_precision)
    {
      re_geom->writeBit(1);
      float diagRange = VecDistance3fv(v_pmin_f, v_pmax_f);
      float tmpRanges[3];
      VecSubtract3fv(tmpRanges, v_pmax_f, v_pmin_f);
      float axisRange = VecMax3fv(tmpRanges);
      int cbmPrecisionRange = (int)(axisRange/diagRange * (1<<(nbits-1)));
      v_pq->SetRange(cbmPrecisionRange);
      fprintf(stderr,"CBM precision simulation on ... using partial range of %d instead of %d\n",cbmPrecisionRange,1<<nbits);
    }
    else
    {
      re_geom->writeBit(0);
      fprintf(stderr,"CBM precision simulation off ... using full range of %d\n",1<<nbits);
    }
    v_pq->SetupQuantizer();

#ifdef PRINT_CONTROL_OUTPUT
    printf("aiQuantRange %d %d %d\n", v_pq->m_aiQuantRange[0], v_pq->m_aiQuantRange[1], v_pq->m_aiQuantRange[2]);
#endif

    global_ic[0] = new RangeIntegerCompressor();
    global_ic[0]->SetRange(v_pq->m_aiQuantRange[0]);
    global_ic[0]->SetupCompressor(re_geom, 3, high_bits);

    global_ic[1] = new RangeIntegerCompressor();
    global_ic[1]->SetRange(v_pq->m_aiQuantRange[1]);
    global_ic[1]->SetupCompressor(re_geom, 3, high_bits);

    global_ic[2] = new RangeIntegerCompressor();
    global_ic[2]->SetRange(v_pq->m_aiQuantRange[2]);
    global_ic[2]->SetupCompressor(re_geom, 3, high_bits);
  }
  else
  {
    if (nbits)
    {
      fprintf(stderr,"no bounding box. quantizing positions adaptively to %d bits\n",nbits);
    }
    else
    {
      fprintf(stderr,"no quantizing. lossless floating-point compression.\n");
    }
#ifdef PRINT_CONTROL_OUTPUT
    fprintf(stderr,"no bounding box. quantizing v_prop_f adaptively to %d bits\n",nbits);
#endif
    v_pq = 0;

    v_fc[0] = new RangeFloatCompressor();
    v_fc[0]->SetPrecision(nbits);
    v_fc[0]->SetupCompressor(re_geom, 3);

    v_fc[1] = new RangeFloatCompressor();
    v_fc[1]->SetPrecision(nbits);
    v_fc[1]->SetupCompressor(re_geom, 3);

    v_fc[2] = new RangeFloatCompressor();
    v_fc[2]->SetPrecision(nbits);
    v_fc[2]->SetupCompressor(re_geom, 3);
  }

#if defined(SVC_USE_LOCAL_COORDINATE_SYSTEM)

  fprintf(stderr,"prediction correctors expressed in local coordinates\n");

  if (v_pmin_f == 0 || v_pmax_f == 0)
  {
    fprintf(stderr,"ERROR: needs advance knowledge of bounding box for compressing with local coordinates\n");
    exit(0);
  }

  if (cbm_precision)
  {
    maxAbso = (1 << (nbits-1)) - 1;
    maxDist = VecDistance3fv(v_pmin_f, v_pmax_f);
    re_geom->writeFloat(maxDist);

    local_ic[0] = new RangeIntegerCompressor();
#if defined(SVC_DO_TETRAHEDRAL_ACROSS_PREDICTION)
    fprintf(stderr,"performing mid-point + height flip prediction\n");
    local_ic[0]->SetPrecision(nbits);
#elif defined(SVC_DO_FLIP_PREDICTION)
    fprintf(stderr,"performing flip across through mid-point prediction\n");
    local_ic[0]->SetPrecision(nbits);
#elif defined(SVC_ADD_NORMAL_COMPONENT)
    fprintf(stderr,"performing add %g base normal to mid-point prediction\n",SVC_NORMAL_SCALE);
    local_ic[0]->SetPrecision(nbits);
#else
    fprintf(stderr,"performing pure mid-point prediction\n");
    local_ic[0]->SetPrecision(nbits-1);
#endif
    local_ic[0]->SetupCompressor(re_geom, 3, high_bits);

    local_ic[1] = new RangeIntegerCompressor();
    local_ic[1]->SetPrecision(nbits);
    local_ic[1]->SetupCompressor(re_geom, 3, high_bits);

    local_ic[2] = new RangeIntegerCompressor();
    local_ic[2]->SetPrecision(nbits);
    local_ic[2]->SetupCompressor(re_geom, 3, high_bits);
  }
  else
  {
    float tmpRanges[3];
    VecSubtract3fv(tmpRanges, v_pmax_f, v_pmin_f);
    maxAbso = (1 << nbits) - 1;
    maxDist = VecMax3fv(tmpRanges);
    re_geom->writeFloat(maxDist);

    local_ic[0] = new RangeIntegerCompressor();
#if defined(SVC_DO_TETRAHEDRAL_ACROSS_PREDICTION)
    fprintf(stderr,"performing mid-point + height prediction\n");
    local_ic[0]->SetPrecision(nbits+1);
#elif defined(SVC_DO_FLIP_PREDICTION)
    fprintf(stderr,"performing flip across through mid-point prediction\n");
    local_ic[0]->SetPrecision(nbits+1);
#elif defined(SVC_ADD_NORMAL_COMPONENT)
    fprintf(stderr,"performing add %g base normal to mid-point prediction\n",SVC_NORMAL_SCALE);
    local_ic[0]->SetPrecision(nbits+1);
#else
    fprintf(stderr,"performing pure mid-point prediction\n");
    local_ic[0]->SetPrecision(nbits);
#endif
    local_ic[0]->SetupCompressor(re_geom, 3, high_bits);

    local_ic[1] = new RangeIntegerCompressor();
    local_ic[1]->SetPrecision(nbits+1);
    local_ic[1]->SetupCompressor(re_geom, 3, high_bits);

    local_ic[2] = new RangeIntegerCompressor();
    local_ic[2]->SetPrecision(nbits+1);
    local_ic[2]->SetupCompressor(re_geom, 3, high_bits);
  }

#else

  fprintf(stderr,"prediction correctors expressed in global coordinates\n");

#if defined(SVC_DO_TETRAHEDRAL_ACROSS_PREDICTION)
    fprintf(stderr,"performing mid-point + height prediction\n");
#elif defined(SVC_DO_FLIP_PREDICTION)
    fprintf(stderr,"performing flip across through mid-point prediction\n");
#elif defined(SVC_ADD_NORMAL_COMPONENT)
    fprintf(stderr,"performing add %g base normal to mid-point prediction\n",SVC_NORMAL_SCALE);
#else
    fprintf(stderr,"performing pure mid-point prediction\n");
#endif

#endif

}

SVwriter_svc::SVwriter_svc()
{
  // init of SVwriter_svc
}

SVwriter_svc::~SVwriter_svc()
{
  // clean-up for SVwriter_svc
}
