/*
===============================================================================

  FILE:  sv_viewer.cpp
  
  CONTENTS:
  
    This little tool visualizes a streaming volume mesh.
  
  PROGRAMMERS:
  
    martin isenburg@cs.unc.edu
  
  COPYRIGHT:
  
    copyright (C) 2003  martin isenburg@cs.unc.edu
    
    This software is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
  
  CHANGE HISTORY:
  
    13 August 2008 -- added support for visualizing hexahedral elements
    16 April 2008 -- added option to visualize badly shaped elements
    01 October 2007 -- using svreadopener for file IO 
    06 April 2005 -- created after bayern lost 2:4 against chelsea
  
===============================================================================
*/

#include <stdlib.h>
#include <stdio.h>
#include <string.h>
#include <GL/glut.h>
#include <assert.h>

//this is for compatibility with the new g++
#ifdef __GNUC__
	#include <ext/hash_map>
	namespace std { using namespace __gnu_cxx; }
	using namespace std;
#else
	#include <hash_map.h>
#endif

#include "vec3fv.h"

#include "svreadopener.h"

// MOUSE INTERFACE

static int LeftButtonDown=0;
static int MiddleButtonDown=0;
static int RightButtonDown=0;
static int OldX,OldY;
static float Elevation=0;
static float Azimuth=0;
static float DistX=0;
static float DistY=0;
static float DistZ=2;

// VISUALIZATION SETTINGS

static float boundingBoxScale=1.0f;
static float boundingBoxTranslateX = 0.0f;
static float boundingBoxTranslateY = 0.0f;
static float boundingBoxTranslateZ = 0.0f;

static float currentBoundingBoxScale=1.0f;
static float currentBoundingBoxTranslateX = 0.0f;
static float currentBoundingBoxTranslateY = 0.0f;
static float currentBoundingBoxTranslateZ = 0.0f;

static float currentBoundingBoxMin[3];
static float currentBoundingBoxMax[3];

// GLOBAL CONTROL VARIABLES

static int WindowW=1024, WindowH=768;
//static int WindowW=800, WindowH=600;
//static int WindowW=1280, WindowH=960;
static int InteractionMode=0;
static int AnimationOn=0;
static int WireFrameOn=0;
static int CurrentBoundingBoxOn=0;
static int PrecomputeBoundingBoxOn=0;
static int WorkingOn=0;
static int PrintToFileOn=0;
static float BadInverseMeanRatio=0;
static bool ReorientTetrahedra=false;

// COLORS

static bool switch_colors = true;

static float colours_diffuse[10][4];
static float colours_white[4];
static float colours_light_blue[4];
static float colours_red_white[4];
static float colours_stripe_white[4];
static float colours_stripe_black[4];
static float colours_slow_change[4];

// DATA STORAGE FOR STREAM VISUALIZER OUTPUT

static SVreadOpener* svreadopener = 0;
static SVreader* svreader = 0;

static I32 nverts;
static I32 ncells;

static int pos_vol = 0;
static int neg_vol = 0;

static int RUN_UNTIL = 0;
static int EXACTLY_N_STEPS = 50;
static int EVERY_NTH_STEP = -1;
static int NEXT_STEP;

static int NUMBER_STRIPES=20;
static int STRIPE_WHITE=0;
static int RENDER_MODE=0;
static int STREAM_COLORING=0;

static unsigned char* Framebuffer = 0;
static const char* PrintFileName = "frame";
static int Time=0;

static const char* cameraSaveFilename = "cameraSave";

//to color the last added hexa
static int latestAddedFaces[]={-1,-1,-1,-1,-1,-1};

// efficient memory allocation

typedef struct SVconnectivityVertex
{
  SVconnectivityVertex* buffer_next;
  I32 index;
  I32 use_count;
  int list_size;
  int list_alloc;
  int* list;
} SVconnectivityVertex;

typedef struct SVrenderVertex
{
  int flag;
  I32 index;
  float v[3];
} SVrenderVertex;

typedef struct SVrenderTriangle
{
  int flag;
  float n[3];
  I32 indices[3];
} SVrenderTriangle;

typedef struct SVrenderQuadrangle
{
  int flag;
  float n[3];
  I32 indices[4];
} SVrenderQuadrangle;

typedef struct SVrenderTetrahedron
{
  int flag;
  I32 indices[4];
} SVrenderTetrahedron;

typedef struct SVrenderHexahedron
{
  int flag;
  I32 indices[8];
} SVrenderHexahedron;

static SVconnectivityVertex* connectivity_vertex_buffer_next = 0;
static int connectivity_vertex_buffer_size = 0;
static int connectivity_vertex_buffer_alloc = 1024;

static SVconnectivityVertex* allocConnectivityVertex()
{
  // allocate new vertices if necessary
  if (connectivity_vertex_buffer_next == 0)
  {
    connectivity_vertex_buffer_alloc *= 2;
    connectivity_vertex_buffer_next = (SVconnectivityVertex*)malloc(sizeof(SVconnectivityVertex)*connectivity_vertex_buffer_alloc);
    if (connectivity_vertex_buffer_next == 0)
    {
      fprintf(stderr,"FATAL ERROR: malloc for connectivity vertex buffer with %d failed\n",connectivity_vertex_buffer_alloc);
      exit(0);
    }
    for (int i = 0; i < connectivity_vertex_buffer_alloc; i++)
    {
      connectivity_vertex_buffer_next[i].buffer_next = &(connectivity_vertex_buffer_next[i+1]);
      connectivity_vertex_buffer_next[i].list = 0;
    }
    connectivity_vertex_buffer_next[connectivity_vertex_buffer_alloc-1].buffer_next = 0;
  }

  // get pointer to next available vertex
  SVconnectivityVertex* vertex = connectivity_vertex_buffer_next;
  connectivity_vertex_buffer_next = vertex->buffer_next;

  if (vertex->list == 0)
  {
    vertex->list = (int*)malloc(sizeof(int)*12);
    vertex->list_alloc = 12;
  }
  vertex->use_count = 0;
  vertex->list_size = 0;

  connectivity_vertex_buffer_size++; 

  return vertex;
}

static void deallocConnectivityVertex(SVconnectivityVertex* vertex)
{
  vertex->buffer_next = connectivity_vertex_buffer_next;
  connectivity_vertex_buffer_next = vertex;
  connectivity_vertex_buffer_size--;
}

static SVrenderVertex* render_vertex_buffer = 0;
static int render_vertex_buffer_next = 0;
static int render_vertex_buffer_size = 0;
static int render_vertex_buffer_alloc = 0;

static void initRenderVertexBuffer(int alloc)
{
  render_vertex_buffer_next = 0;
  render_vertex_buffer_size = 0;
  if (render_vertex_buffer)
  {
    free(render_vertex_buffer);
  }
  render_vertex_buffer = (SVrenderVertex*)malloc(sizeof(SVrenderVertex)*alloc);
  if (!render_vertex_buffer)
  {
    fprintf(stderr,"FATAL ERROR: alloc render_vertex_buffer with %d failed.\n",alloc);
    exit(0);
  }
  render_vertex_buffer_alloc = alloc;
  for (int i = 0; i < render_vertex_buffer_alloc; i++)
  {
    render_vertex_buffer[i].flag = i+1;
  }
}

static int allocRenderVertex()
{
  if (render_vertex_buffer_size == render_vertex_buffer_alloc)
  {
    render_vertex_buffer = (SVrenderVertex*)realloc(render_vertex_buffer,sizeof(SVrenderVertex)*render_vertex_buffer_alloc*2);
    if (!render_vertex_buffer)
    {
      fprintf(stderr,"FATAL ERROR: realloc render_vertex_buffer with %d failed.\n",render_vertex_buffer_alloc*2);
      exit(0);
    }
    render_vertex_buffer_alloc *= 2;
    for (int i = render_vertex_buffer_size; i < render_vertex_buffer_alloc; i++)
    {
      render_vertex_buffer[i].flag = i+1;
    }
  }
  int index = render_vertex_buffer_next;
  render_vertex_buffer_next = render_vertex_buffer[index].flag;
  render_vertex_buffer[index].flag = -1;
  render_vertex_buffer_size++;
  return index;
}

static void deallocRenderVertex(int index)
{
  render_vertex_buffer[index].flag = render_vertex_buffer_next;
  render_vertex_buffer_next = index;
  render_vertex_buffer_size--;
}

static void destroyRenderVertexBuffer()
{
  render_vertex_buffer_next = 0;
  render_vertex_buffer_size = 0;
  render_vertex_buffer_alloc = 0;
  if (render_vertex_buffer)
  {
    free(render_vertex_buffer);
  }
  render_vertex_buffer = 0;
}

static SVrenderTriangle* triangle_buffer = 0;
static int triangle_buffer_next = 0;
static int triangle_buffer_size = 0;
static int triangle_buffer_alloc = 0;

static void initTriangleBuffer(int alloc)
{
  triangle_buffer_next = 0;
  triangle_buffer_size = 0;
  if (triangle_buffer)
  {
    free(triangle_buffer);
  }
  triangle_buffer = (SVrenderTriangle*)malloc(sizeof(SVrenderTriangle)*alloc);
  if (!triangle_buffer)
  {
    fprintf(stderr,"FATAL ERROR: alloc triangle_buffer with %d failed.\n",alloc);
    exit(0);
  }
  triangle_buffer_alloc = alloc;
  for (int i = 0; i < triangle_buffer_alloc; i++)
  {
    triangle_buffer[i].flag = i+1;
  }
}

static int allocTriangle()
{
  if (triangle_buffer_size == triangle_buffer_alloc)
  {
    triangle_buffer = (SVrenderTriangle*)realloc(triangle_buffer,sizeof(SVrenderTriangle)*triangle_buffer_alloc*2);
    if (!triangle_buffer)
    {
      fprintf(stderr,"FATAL ERROR: realloc triangle_buffer with %d failed.\n",triangle_buffer_alloc*2);
      exit(0);
    }
    triangle_buffer_alloc *= 2;
    for (int i = triangle_buffer_size; i < triangle_buffer_alloc; i++)
    {
      triangle_buffer[i].flag = i+1;
    }
  }
  int index = triangle_buffer_next;
  triangle_buffer_next = triangle_buffer[index].flag;
  triangle_buffer_size++;
  return index;
}

static void deallocTriangle(int index)
{
  triangle_buffer[index].flag = triangle_buffer_next;
  triangle_buffer_next = index;
  triangle_buffer_size--;
}

static void destroyTriangleBuffer()
{
  triangle_buffer_next = 0;
  triangle_buffer_size = 0;
  triangle_buffer_alloc = 0;
  if (triangle_buffer)
  {
    free(triangle_buffer);
  }
  triangle_buffer = 0;
}

static SVrenderQuadrangle* quadrangle_buffer = 0;
static int quadrangle_buffer_next = 0;
static int quadrangle_buffer_size = 0;
static int quadrangle_buffer_alloc = 0;

static void initQuadrangleBuffer(int alloc)
{
  quadrangle_buffer_next = 0;
  quadrangle_buffer_size = 0;
  if (quadrangle_buffer)
  {
    free(quadrangle_buffer);
  }
  quadrangle_buffer = (SVrenderQuadrangle*)malloc(sizeof(SVrenderQuadrangle)*alloc);
  if (!quadrangle_buffer)
  {
    fprintf(stderr,"FATAL ERROR: alloc quadrangle_buffer with %d failed.\n",alloc);
    exit(0);
  }
  quadrangle_buffer_alloc = alloc;
  for (int i = 0; i < quadrangle_buffer_alloc; i++)
  {
    quadrangle_buffer[i].flag = i+1;
  }
}

static int allocQuadrangle()
{
  if (quadrangle_buffer_size == quadrangle_buffer_alloc)
  {
    quadrangle_buffer = (SVrenderQuadrangle*)realloc(quadrangle_buffer,sizeof(SVrenderQuadrangle)*quadrangle_buffer_alloc*2);
    if (!quadrangle_buffer)
    {
      fprintf(stderr,"FATAL ERROR: realloc quadrangle_buffer with %d failed.\n",quadrangle_buffer_alloc*2);
      exit(0);
    }
    quadrangle_buffer_alloc *= 2;
    for (int i = quadrangle_buffer_size; i < quadrangle_buffer_alloc; i++)
    {
      quadrangle_buffer[i].flag = i+1;
    }
  }
  int index = quadrangle_buffer_next;
  quadrangle_buffer_next = quadrangle_buffer[index].flag;
  quadrangle_buffer_size++;
  return index;
}

static void deallocQuadrangle(int index)
{
  quadrangle_buffer[index].flag = quadrangle_buffer_next;
  quadrangle_buffer_next = index;
  quadrangle_buffer_size--;
}

static void destroyQuadrangleBuffer()
{
  quadrangle_buffer_next = 0;
  quadrangle_buffer_size = 0;
  quadrangle_buffer_alloc = 0;
  if (quadrangle_buffer)
  {
    free(quadrangle_buffer);
  }
  quadrangle_buffer = 0;
}

static SVrenderTriangle* boundary_triangle_buffer = 0;
static int boundary_triangle_buffer_next = 0;
static int boundary_triangle_buffer_size = 0;
static int boundary_triangle_buffer_alloc = 0;

static void initBoundaryTriangleBuffer(int alloc)
{
  boundary_triangle_buffer_next = 0;
  boundary_triangle_buffer_size = 0;
  if (boundary_triangle_buffer)
  {
    free(boundary_triangle_buffer);
  }
  boundary_triangle_buffer = (SVrenderTriangle*)malloc(sizeof(SVrenderTriangle)*alloc);
  if (!boundary_triangle_buffer)
  {
    fprintf(stderr,"FATAL ERROR: alloc boundary_triangle_buffer with %d failed.\n",alloc);
    exit(0);
  }
  boundary_triangle_buffer_alloc = alloc;
  for (int i = 0; i < boundary_triangle_buffer_alloc; i++)
  {
    boundary_triangle_buffer[i].flag = i+1;
  }
}

static int allocBoundaryTriangle()
{
  if (boundary_triangle_buffer_size == boundary_triangle_buffer_alloc)
  {
    boundary_triangle_buffer = (SVrenderTriangle*)realloc(boundary_triangle_buffer,sizeof(SVrenderTriangle)*boundary_triangle_buffer_alloc*2);
    if (!boundary_triangle_buffer)
    {
      fprintf(stderr,"FATAL ERROR: realloc boundary_triangle_buffer with %d failed.\n",boundary_triangle_buffer_alloc*2);
      exit(0);
    }
    boundary_triangle_buffer_alloc *= 2;
    for (int i = boundary_triangle_buffer_size; i < boundary_triangle_buffer_alloc; i++)
    {
      boundary_triangle_buffer[i].flag = i+1;
    }
  }
  int index = boundary_triangle_buffer_next;
  boundary_triangle_buffer_next = boundary_triangle_buffer[index].flag;
  boundary_triangle_buffer_size++;
  return index;
}

static void deallocBoundaryTriangle(int index)
{
  boundary_triangle_buffer[index].flag = boundary_triangle_buffer_next;
  boundary_triangle_buffer_next = index;
  boundary_triangle_buffer_size--;
}

static void destroyBoundaryTriangleBuffer()
{
  boundary_triangle_buffer_next = 0;
  boundary_triangle_buffer_size = 0;
  boundary_triangle_buffer_alloc = 0;
  if (boundary_triangle_buffer)
  {
    free(boundary_triangle_buffer);
  }
  boundary_triangle_buffer = 0;
}

static SVrenderQuadrangle* boundary_quadrangle_buffer = 0;
static int boundary_quadrangle_buffer_next = 0;
static int boundary_quadrangle_buffer_size = 0;
static int boundary_quadrangle_buffer_alloc = 0;

static void initBoundaryQuadrangleBuffer(int alloc)
{
  boundary_quadrangle_buffer_next = 0;
  boundary_quadrangle_buffer_size = 0;
  if (boundary_quadrangle_buffer)
  {
    free(boundary_quadrangle_buffer);
  }
  boundary_quadrangle_buffer = (SVrenderQuadrangle*)malloc(sizeof(SVrenderQuadrangle)*alloc);
  if (!boundary_quadrangle_buffer)
  {
    fprintf(stderr,"FATAL ERROR: alloc boundary_quadrangle_buffer with %d failed.\n",alloc);
    exit(0);
  }
  boundary_quadrangle_buffer_alloc = alloc;
  for (int i = 0; i < boundary_quadrangle_buffer_alloc; i++)
  {
    boundary_quadrangle_buffer[i].flag = i+1;
  }
}

static int allocBoundaryQuadrangle()
{
  if (boundary_quadrangle_buffer_size == boundary_quadrangle_buffer_alloc)
  {
    boundary_quadrangle_buffer = (SVrenderQuadrangle*)realloc(boundary_quadrangle_buffer,sizeof(SVrenderQuadrangle)*boundary_quadrangle_buffer_alloc*2);
    if (!boundary_quadrangle_buffer)
    {
      fprintf(stderr,"FATAL ERROR: realloc boundary_quadrangle_buffer with %d failed.\n",boundary_quadrangle_buffer_alloc*2);
      exit(0);
    }
    boundary_quadrangle_buffer_alloc *= 2;
    for (int i = boundary_quadrangle_buffer_size; i < boundary_quadrangle_buffer_alloc; i++)
    {
      boundary_quadrangle_buffer[i].flag = i+1;
    }
  }
  int index = boundary_quadrangle_buffer_next;
  boundary_quadrangle_buffer_next = boundary_quadrangle_buffer[index].flag;
  boundary_quadrangle_buffer_size++;
  return index;
}

static void deallocBoundaryQuadrangle(int index)
{
  boundary_quadrangle_buffer[index].flag = boundary_quadrangle_buffer_next;
  boundary_quadrangle_buffer_next = index;
  boundary_quadrangle_buffer_size--;
}

static void destroyBoundaryQuadrangleBuffer()
{
  boundary_quadrangle_buffer_next = 0;
  boundary_quadrangle_buffer_size = 0;
  boundary_quadrangle_buffer_alloc = 0;
  if (boundary_quadrangle_buffer)
  {
    free(boundary_quadrangle_buffer);
  }
  boundary_quadrangle_buffer = 0;
}

static SVrenderTetrahedron* tetrahedron_buffer = 0;
static int tetrahedron_buffer_next = 0;
static int tetrahedron_buffer_size = 0;
static int tetrahedron_buffer_alloc = 0;

static void initTetrahedronBuffer(int alloc)
{
  tetrahedron_buffer_next = 0;
  tetrahedron_buffer_size = 0;
  if (tetrahedron_buffer)
  {
    free(tetrahedron_buffer);
  }
  tetrahedron_buffer = (SVrenderTetrahedron*)malloc(sizeof(SVrenderTetrahedron)*alloc);
  if (!tetrahedron_buffer)
  {
    fprintf(stderr,"FATAL ERROR: alloc tetrahedron_buffer with %d failed.\n",alloc);
    exit(0);
  }
  tetrahedron_buffer_alloc = alloc;
  for (int i = 0; i < tetrahedron_buffer_alloc; i++)
  {
    tetrahedron_buffer[i].flag = i+1;
  }
}

static int allocTetrahedron()
{
  if (tetrahedron_buffer_size == tetrahedron_buffer_alloc)
  {
    tetrahedron_buffer = (SVrenderTetrahedron*)realloc(tetrahedron_buffer,sizeof(SVrenderTetrahedron)*tetrahedron_buffer_alloc*2);
    if (!tetrahedron_buffer)
    {
      fprintf(stderr,"FATAL ERROR: realloc tetrahedron_buffer with %d failed.\n",tetrahedron_buffer_alloc*2);
      exit(0);
    }
    tetrahedron_buffer_alloc *= 2;
    for (int i = tetrahedron_buffer_size; i < tetrahedron_buffer_alloc; i++)
    {
      tetrahedron_buffer[i].flag = i+1;
    }
  }
  int index = tetrahedron_buffer_next;
  tetrahedron_buffer_next = tetrahedron_buffer[index].flag;
  tetrahedron_buffer_size++;
  return index;
}

static void deallocTetrahedron(int index)
{
  tetrahedron_buffer[index].flag = tetrahedron_buffer_next;
  tetrahedron_buffer_next = index;
  tetrahedron_buffer_size--;
}

static void destroyTetrahedronBuffer()
{
  tetrahedron_buffer_next = 0;
  tetrahedron_buffer_size = 0;
  tetrahedron_buffer_alloc = 0;
  if (tetrahedron_buffer)
  {
    free(tetrahedron_buffer);
  }
  tetrahedron_buffer = 0;
}

static SVrenderHexahedron* hexahedron_buffer = 0;
static int hexahedron_buffer_next = 0;
static int hexahedron_buffer_size = 0;
static int hexahedron_buffer_alloc = 0;

static void initHexahedronBuffer(int alloc)
{
  hexahedron_buffer_next = 0;
  hexahedron_buffer_size = 0;
  if (hexahedron_buffer)
  {
    free(hexahedron_buffer);
  }
  hexahedron_buffer = (SVrenderHexahedron*)malloc(sizeof(SVrenderHexahedron)*alloc);
  if (!hexahedron_buffer)
  {
    fprintf(stderr,"FATAL ERROR: alloc hexahedron_buffer with %d failed.\n",alloc);
    exit(0);
  }
  hexahedron_buffer_alloc = alloc;
  for (int i = 0; i < hexahedron_buffer_alloc; i++)
  {
    hexahedron_buffer[i].flag = i+1;
  }
}

static int allocHexahedron()
{
  if (hexahedron_buffer_size == hexahedron_buffer_alloc)
  {
    hexahedron_buffer = (SVrenderHexahedron*)realloc(hexahedron_buffer,sizeof(SVrenderHexahedron)*hexahedron_buffer_alloc*2);
    if (!hexahedron_buffer)
    {
      fprintf(stderr,"FATAL ERROR: realloc hexahedron_buffer with %d failed.\n",hexahedron_buffer_alloc*2);
      exit(0);
    }
    hexahedron_buffer_alloc *= 2;
    for (int i = hexahedron_buffer_size; i < hexahedron_buffer_alloc; i++)
    {
      hexahedron_buffer[i].flag = i+1;
    }
  }
  int index = hexahedron_buffer_next;
  hexahedron_buffer_next = hexahedron_buffer[index].flag;
  hexahedron_buffer_size++;
  return index;
}

static void deallocHexahedron(int index)
{
  hexahedron_buffer[index].flag = hexahedron_buffer_next;
  hexahedron_buffer_next = index;
  hexahedron_buffer_size--;
}

static void destroyHexahedronBuffer()
{
  hexahedron_buffer_next = 0;
  hexahedron_buffer_size = 0;
  hexahedron_buffer_alloc = 0;
  if (hexahedron_buffer)
  {
    free(hexahedron_buffer);
  }
  hexahedron_buffer = 0;
}

typedef hash_map<I32, SVconnectivityVertex*> my_hash;

static my_hash* map_hash;
static int index_map_size = 0;
static int index_map_maxsize = 0;

static void SavePPM(char *FileName, unsigned char* Colour, int Width, int Height)
{
  FILE *fp = fopen(FileName, "wb");
  fprintf(fp, "P6\n%d %d\n255\n", Width, Height);
  int NumRowPixels = Width*3;
  for (int i=(Height-1)*Width*3; i>=0; i-=(Width*3))
  {
    fwrite(&(Colour[i]),1,NumRowPixels,fp);
  }
  fclose(fp);
}

static int pick_vertex(int x, int y)
{
  // small viewport for picking

  y = WindowH-y;
  glViewport(x-5, y-5, 10, 10);

  float vertex_size;

  // set projection matrix

  GLint vp[4];
  vp[0] = 0;
  vp[1] = 0;
  vp[2] = WindowW;
  vp[3] = WindowH;

  glMatrixMode(GL_PROJECTION);
  glLoadIdentity();
  gluPickMatrix(x, y, 10, 10, vp); 
  gluPerspective(30.0f,(float)WindowW/WindowH,0.0625f,5.0f);

  glMatrixMode(GL_MODELVIEW);
  glLoadIdentity();
  gluLookAt(DistX,DistY,DistZ, DistX,DistY,0, 0,1,0);
  glRotatef(Elevation,1,0,0);
  glRotatef(Azimuth,0,1,0);

  if (CurrentBoundingBoxOn)
  {
    if ((currentBoundingBoxMax[1]-currentBoundingBoxMin[1]) > (currentBoundingBoxMax[0]-currentBoundingBoxMin[0]))
    {
      if ((currentBoundingBoxMax[1]-currentBoundingBoxMin[1]) > (currentBoundingBoxMax[2]-currentBoundingBoxMin[2]))
      {
        currentBoundingBoxScale = 1.0f/(currentBoundingBoxMax[1]-currentBoundingBoxMin[1]);
      }
      else
      {
        currentBoundingBoxScale = 1.0f/(currentBoundingBoxMax[2]-currentBoundingBoxMin[2]);
      }
    }
    else
    {
      if ((currentBoundingBoxMax[0]-currentBoundingBoxMin[0]) > (currentBoundingBoxMax[2]-currentBoundingBoxMin[2]))
      {
        currentBoundingBoxScale = 1.0f/(currentBoundingBoxMax[0]-currentBoundingBoxMin[0]);
      }
      else
      {
        currentBoundingBoxScale = 1.0f/(currentBoundingBoxMax[2]-currentBoundingBoxMin[2]);
      }
    }
    currentBoundingBoxTranslateX = - currentBoundingBoxScale * (currentBoundingBoxMin[0] + 0.5f * (currentBoundingBoxMax[0]-currentBoundingBoxMin[0]));
    currentBoundingBoxTranslateY = - currentBoundingBoxScale * (currentBoundingBoxMin[1] + 0.5f * (currentBoundingBoxMax[1]-currentBoundingBoxMin[1]));
    currentBoundingBoxTranslateZ = - currentBoundingBoxScale * (currentBoundingBoxMin[2] + 0.5f * (currentBoundingBoxMax[2]-currentBoundingBoxMin[2]));

    glTranslatef(currentBoundingBoxTranslateX,currentBoundingBoxTranslateY,currentBoundingBoxTranslateZ);
    glScalef(currentBoundingBoxScale,currentBoundingBoxScale,currentBoundingBoxScale);
    vertex_size = 0.01f / currentBoundingBoxScale;
  }
  else
  {
    glTranslatef(boundingBoxTranslateX,boundingBoxTranslateY,boundingBoxTranslateZ);
    glScalef(boundingBoxScale,boundingBoxScale,boundingBoxScale);
    vertex_size = 0.01f / boundingBoxScale;
  }

  glClearColor(0,0,0,0);
  glClear(GL_COLOR_BUFFER_BIT | GL_DEPTH_BUFFER_BIT);

  glEnable(GL_DEPTH_TEST);
  glDisable(GL_LIGHTING);
  glDisable(GL_NORMALIZE);

  glPolygonMode(GL_FRONT_AND_BACK, GL_FILL);
  glBegin(GL_QUADS);

  for (int i = 0; i < render_vertex_buffer_alloc; i++)
  {
    if (render_vertex_buffer[i].flag < 0)
    {
      int idx = i + 1;
      glColor4ubv((const GLubyte *)&idx);

      // draw
      glVertex3f(render_vertex_buffer[i].v[0]-vertex_size, render_vertex_buffer[i].v[1]-vertex_size, render_vertex_buffer[i].v[2]-vertex_size);
      glVertex3f(render_vertex_buffer[i].v[0]-vertex_size, render_vertex_buffer[i].v[1]-vertex_size, render_vertex_buffer[i].v[2]+vertex_size);
      glVertex3f(render_vertex_buffer[i].v[0]-vertex_size, render_vertex_buffer[i].v[1]+vertex_size, render_vertex_buffer[i].v[2]-vertex_size);
      glVertex3f(render_vertex_buffer[i].v[0]-vertex_size, render_vertex_buffer[i].v[1]+vertex_size, render_vertex_buffer[i].v[2]+vertex_size);

      glVertex3f(render_vertex_buffer[i].v[0]+vertex_size, render_vertex_buffer[i].v[1]-vertex_size, render_vertex_buffer[i].v[2]-vertex_size);
      glVertex3f(render_vertex_buffer[i].v[0]+vertex_size, render_vertex_buffer[i].v[1]-vertex_size, render_vertex_buffer[i].v[2]+vertex_size);
      glVertex3f(render_vertex_buffer[i].v[0]+vertex_size, render_vertex_buffer[i].v[1]+vertex_size, render_vertex_buffer[i].v[2]-vertex_size);
      glVertex3f(render_vertex_buffer[i].v[0]+vertex_size, render_vertex_buffer[i].v[1]+vertex_size, render_vertex_buffer[i].v[2]+vertex_size);

      glVertex3f(render_vertex_buffer[i].v[0]-vertex_size, render_vertex_buffer[i].v[1]-vertex_size, render_vertex_buffer[i].v[2]-vertex_size);
      glVertex3f(render_vertex_buffer[i].v[0]-vertex_size, render_vertex_buffer[i].v[1]-vertex_size, render_vertex_buffer[i].v[2]+vertex_size);
      glVertex3f(render_vertex_buffer[i].v[0]+vertex_size, render_vertex_buffer[i].v[1]-vertex_size, render_vertex_buffer[i].v[2]-vertex_size);
      glVertex3f(render_vertex_buffer[i].v[0]+vertex_size, render_vertex_buffer[i].v[1]-vertex_size, render_vertex_buffer[i].v[2]+vertex_size);

      glVertex3f(render_vertex_buffer[i].v[0]-vertex_size, render_vertex_buffer[i].v[1]+vertex_size, render_vertex_buffer[i].v[2]-vertex_size);
      glVertex3f(render_vertex_buffer[i].v[0]-vertex_size, render_vertex_buffer[i].v[1]+vertex_size, render_vertex_buffer[i].v[2]+vertex_size);
      glVertex3f(render_vertex_buffer[i].v[0]+vertex_size, render_vertex_buffer[i].v[1]+vertex_size, render_vertex_buffer[i].v[2]-vertex_size);
      glVertex3f(render_vertex_buffer[i].v[0]+vertex_size, render_vertex_buffer[i].v[1]+vertex_size, render_vertex_buffer[i].v[2]+vertex_size);

      glVertex3f(render_vertex_buffer[i].v[0]-vertex_size, render_vertex_buffer[i].v[1]-vertex_size, render_vertex_buffer[i].v[2]-vertex_size);
      glVertex3f(render_vertex_buffer[i].v[0]-vertex_size, render_vertex_buffer[i].v[1]+vertex_size, render_vertex_buffer[i].v[2]-vertex_size);
      glVertex3f(render_vertex_buffer[i].v[0]+vertex_size, render_vertex_buffer[i].v[1]-vertex_size, render_vertex_buffer[i].v[2]-vertex_size);
      glVertex3f(render_vertex_buffer[i].v[0]+vertex_size, render_vertex_buffer[i].v[1]+vertex_size, render_vertex_buffer[i].v[2]-vertex_size);

      glVertex3f(render_vertex_buffer[i].v[0]-vertex_size, render_vertex_buffer[i].v[1]-vertex_size, render_vertex_buffer[i].v[2]+vertex_size);
      glVertex3f(render_vertex_buffer[i].v[0]-vertex_size, render_vertex_buffer[i].v[1]+vertex_size, render_vertex_buffer[i].v[2]+vertex_size);
      glVertex3f(render_vertex_buffer[i].v[0]+vertex_size, render_vertex_buffer[i].v[1]-vertex_size, render_vertex_buffer[i].v[2]+vertex_size);
      glVertex3f(render_vertex_buffer[i].v[0]+vertex_size, render_vertex_buffer[i].v[1]+vertex_size, render_vertex_buffer[i].v[2]+vertex_size);
    }
  }
  glEnd();

  glDisable(GL_DEPTH_TEST);

  unsigned int id;

  glReadPixels(x,y,1,1,GL_RGB,GL_UNSIGNED_BYTE,&id);

  id = 0x00FFFFFF & id;

  fprintf(stderr, "at %d %d picked color %u with index %d and pos %g %g %g\n", x, y, id, render_vertex_buffer[id-1].index, render_vertex_buffer[id-1].v[0], render_vertex_buffer[id-1].v[1], render_vertex_buffer[id-1].v[2]);

  if (id == 0)
  {
    return -1;
  }
  else 
  {
    return id - 1;
  }
}

static void InitColors()
{
  colours_diffuse[0][0] = 0.0f; colours_diffuse[0][1] = 0.0f; colours_diffuse[0][2] = 0.0f; colours_diffuse[0][3] = 1.0f; // black
  colours_diffuse[1][0] = 0.6f; colours_diffuse[1][1] = 0.0f; colours_diffuse[1][2] = 0.0f; colours_diffuse[1][3] = 1.0f; // red
  colours_diffuse[2][0] = 0.0f; colours_diffuse[2][1] = 0.8f; colours_diffuse[2][2] = 0.0f; colours_diffuse[2][3] = 1.0f; // green
  colours_diffuse[3][0] = 0.0f; colours_diffuse[3][1] = 0.0f; colours_diffuse[3][2] = 0.6f; colours_diffuse[3][3] = 1.0f; // blue
  colours_diffuse[4][0] = 0.6f; colours_diffuse[4][1] = 0.6f; colours_diffuse[4][2] = 0.0f; colours_diffuse[4][3] = 1.0f; // yellow
  colours_diffuse[5][0] = 0.6f; colours_diffuse[5][1] = 0.0f; colours_diffuse[5][2] = 0.6f; colours_diffuse[5][3] = 1.0f; // purple
  colours_diffuse[6][0] = 0.0f; colours_diffuse[6][1] = 0.6f; colours_diffuse[6][2] = 0.6f; colours_diffuse[6][3] = 1.0f; // cyan
  colours_diffuse[7][0] = 0.9f; colours_diffuse[7][1] = 0.9f; colours_diffuse[7][2] = 0.9f; colours_diffuse[7][3] = 1.0f; // even ligher white
  colours_diffuse[8][0] = 0.5f; colours_diffuse[8][1] = 0.5f; colours_diffuse[8][2] = 0.9f; colours_diffuse[8][3] = 1.0f; // even lighter blue
  colours_diffuse[9][0] = 0.9f; colours_diffuse[9][1] = 0.4f; colours_diffuse[9][2] = 0.7f; colours_diffuse[9][3] = 1.0f; // violett
  
  colours_white[0] = 0.7f; colours_white[1] = 0.7f; colours_white[2] = 0.7f; colours_white[3] = 1.0f; // white
  colours_light_blue[0] = 0.2f; colours_light_blue[1] = 0.2f; colours_light_blue[2] = 0.6f; colours_light_blue[3] = 1.0f; // blue
  colours_red_white[0] = 0.7f; colours_red_white[1] = 0.7f; colours_red_white[2] = 0.7f; colours_red_white[3] = 1.0f; // white or red

  colours_stripe_white[0] = 0.8f; colours_stripe_white[1] = 0.8f; colours_stripe_white[2] = 0.8f; colours_stripe_white[3] = 1.0f; // white
  colours_stripe_black[0] = 0.5f; colours_stripe_black[1] = 0.5f; colours_stripe_black[2] = 0.5f; colours_stripe_black[3] = 1.0f; // black

  colours_slow_change[0] = 0.5f; colours_slow_change[1] = 0.5f; colours_slow_change[2] = 0.5f; colours_slow_change[3] = 1.0f; // black
} 

static void InitLight()
{
  float intensity[] = {1,1,1,1};
  float position[] = {1,1,5,0}; // directional behind the viewer
  glLightfv(GL_LIGHT0,GL_DIFFUSE,intensity);
  glLightfv(GL_LIGHT0,GL_SPECULAR,intensity);
  glLightfv(GL_LIGHT0,GL_POSITION,position);
  glLightModeli(GL_LIGHT_MODEL_LOCAL_VIEWER,GL_FALSE);
  glLightModeli(GL_LIGHT_MODEL_TWO_SIDE,GL_TRUE);
}

/*****************************************************************************/
/*  Calculate ideal weight inverse mean ratio of tetrahedral element         */
/*****************************************************************************/
#define isqrt3   5.77350269189625797959429519858e-01        /*  1.0/sqrt(3.0)*/
#define isqrt6   4.08248290463863052509822647505e-01        /*  1.0/sqrt(6.0)*/
static float inverse_mean_ratio(const float* x0, const float* x1, const float* x2, const float* x3)
{
  double matr[9];
  double f, g;

  /* Calculate M = A*inv(W). */
  f       = x1[0] + x0[0];
  matr[0] = x1[0] - x0[0];
  matr[1] = (2.0*x2[0] - f)*isqrt3;
  matr[2] = (3.0*x3[0] - x2[0] - f)*isqrt6;

  f       = x1[1] + x0[1];
  matr[3] = x1[1] - x0[1];
  matr[4] = (2.0*x2[1] - f)*isqrt3;
  matr[5] = (3.0*x3[1] - x2[1] - f)*isqrt6;

  f       = x1[2] + x0[2];
  matr[6] = x1[2] - x0[2];
  matr[7] = (2.0*x2[2] - f)*isqrt3;
  matr[8] = (3.0*x3[2] - x2[2] - f)*isqrt6;

  /* Calculate det(M). */
  g = matr[0]*(matr[4]*matr[8] - matr[5]*matr[7]) +
      matr[1]*(matr[5]*matr[6] - matr[3]*matr[8]) +
      matr[2]*(matr[3]*matr[7] - matr[4]*matr[6]);

  if (g < 1e-30) return 0; // maybe should return -1 to signal error

  /* Calculate norm(M). */
  f = matr[0]*matr[0] + matr[1]*matr[1] + matr[2]*matr[2] +
      matr[3]*matr[3] + matr[4]*matr[4] + matr[5]*matr[5] +
      matr[6]*matr[6] + matr[7]*matr[7] + matr[8]*matr[8];

  /* Calculate and Return objective function. */
  return (float)(1.0/3.0 * f * pow( g, -2.0/3.0 ));
}

/*****************************************************************************/
/*  Calculate ideal weight inverse mean ratio of hexahedral element         */
/*****************************************************************************/
static float inverse_mean_ratio(const float* x0, const float* x1, const float* x2, const float* x3, const float* x4, const float* x5, const float* x6, const float* x7)
{
  // Hex element descriptions
  static const int locs_hex[8][4] = {{0, 1, 3, 4},
                                     {1, 2, 0, 5},
                                     {2, 3, 1, 6},
                                     {3, 0, 2, 7},
                                     {4, 7, 5, 0},
                                     {5, 4, 6, 1},
                                     {6, 5, 7, 2},
                                     {7, 6, 4, 3}};

  const float* vertices[8];
  vertices[0] = x0;
  vertices[1] = x1;
  vertices[2] = x2;
  vertices[3] = x3;
  vertices[4] = x4;
  vertices[5] = x5;
  vertices[6] = x6;
  vertices[7] = x7;

  int i;
  double m = 0.0;
  for (i = 0; i < 8; i++) {
    m += inverse_mean_ratio(vertices[locs_hex[i][0]],vertices[locs_hex[i][1]],vertices[locs_hex[i][2]],vertices[locs_hex[i][3]]);
  }
  
  m = m / 8.0;

  return (float)m;
}

static void usage()
{
  fprintf(stderr,"usage:\n");
  fprintf(stderr,"sv_viewer mesh.sva\n");
  fprintf(stderr,"sv_viewer -win 640 480 -i mesh.svc\n");
  fprintf(stderr,"sv_viewer -isvc -every 1000 < mesh.svc\n");
  fprintf(stderr,"sv_viewer -steps 10 -i mesh.off\n");
  fprintf(stderr,"sv_viewer -i mesh.svc -current_bb\n");
  fprintf(stderr,"sv_viewer -i mesh.off.gz\n");
  fprintf(stderr,"sv_viewer -precomp_bb -i mesh.off.gz\n");
}

static void removeTriangleFromVertex(int i, SVconnectivityVertex* v)
{
  v->list_size--;
  v->list[i] = v->list[v->list_size];
}

static bool search_and_remove_triangle(SVconnectivityVertex* vertex0, SVconnectivityVertex* vertex1, SVconnectivityVertex* vertex2)
{
  int i, t_idx;
  SVrenderTriangle* triangle;

  for (i = 0; i < vertex0->list_size; i++)
  {
    t_idx = vertex0->list[i];
    triangle = &(triangle_buffer[t_idx]);
    assert(triangle->indices[0] == vertex0->index);
    if (triangle->indices[1] == vertex2->index && triangle->indices[2] == vertex1->index)
    {
      removeTriangleFromVertex(i, vertex0);
      deallocTriangle(t_idx);
      vertex0->use_count--;
      vertex1->use_count--;
      vertex2->use_count--;
      return true;
    }
  }
  for (i = 0; i < vertex1->list_size; i++)
  {
    t_idx = vertex1->list[i];
    triangle = &(triangle_buffer[t_idx]);
    assert(triangle->indices[0] == vertex1->index);
    if (triangle->indices[1] == vertex0->index && triangle->indices[2] == vertex2->index)
    {
      removeTriangleFromVertex(i, vertex1);
      deallocTriangle(t_idx);
      vertex0->use_count--;
      vertex1->use_count--;
      vertex2->use_count--;
      return true;
    }
  }
  for (i = 0; i < vertex2->list_size; i++)
  {
    t_idx = vertex2->list[i];
    triangle = &(triangle_buffer[t_idx]);
    assert(triangle->indices[0] == vertex2->index);
    if (triangle->indices[1] == vertex1->index && triangle->indices[2] == vertex0->index)
    {
      removeTriangleFromVertex(i, vertex2);
      deallocTriangle(t_idx);
      vertex0->use_count--;
      vertex1->use_count--;
      vertex2->use_count--;
      return true;
    }
  }
  return false;
}

static void removeQuadrangleFromVertex(int i, SVconnectivityVertex* v)
{
  v->list_size--;
  v->list[i] = v->list[v->list_size];
}

static bool search_and_remove_quadrangle(SVconnectivityVertex* vertex0, SVconnectivityVertex* vertex1, SVconnectivityVertex* vertex2,  SVconnectivityVertex* vertex3)
{
  int i, q_idx;
  SVrenderQuadrangle* quadrangle;

  for (i = 0; i < vertex0->list_size; i++)
  {
    q_idx = vertex0->list[i];
    quadrangle = &(quadrangle_buffer[q_idx]);
    assert(quadrangle->indices[0] == vertex0->index);
    if (quadrangle->indices[1] == vertex3->index && quadrangle->indices[2] == vertex2->index && quadrangle->indices[3] == vertex1->index)
    {
      removeQuadrangleFromVertex(i, vertex0);
      deallocQuadrangle(q_idx);
      vertex0->use_count--;
      vertex1->use_count--;
      vertex2->use_count--;
      vertex3->use_count--;
      return true;
    }
  }
  for (i = 0; i < vertex1->list_size; i++)
  {
    q_idx = vertex1->list[i];
    quadrangle = &(quadrangle_buffer[q_idx]);
    assert(quadrangle->indices[0] == vertex1->index);
    if (quadrangle->indices[1] == vertex0->index && quadrangle->indices[2] == vertex3->index && quadrangle->indices[3] == vertex2->index)
    {
      removeQuadrangleFromVertex(i, vertex1);
      deallocQuadrangle(q_idx);
      vertex0->use_count--;
      vertex1->use_count--;
      vertex2->use_count--;
      vertex3->use_count--;
      return true;
    }
  }
  for (i = 0; i < vertex2->list_size; i++)
  {
    q_idx = vertex2->list[i];
    quadrangle = &(quadrangle_buffer[q_idx]);
    assert(quadrangle->indices[0] == vertex2->index);
    if (quadrangle->indices[1] == vertex1->index && quadrangle->indices[2] == vertex0->index && quadrangle->indices[3] == vertex3->index)
    {
      removeQuadrangleFromVertex(i, vertex2);
      deallocQuadrangle(q_idx);
      vertex0->use_count--;
      vertex1->use_count--;
      vertex2->use_count--;
      vertex3->use_count--;
      return true;
    }
  }
  for (i = 0; i < vertex3->list_size; i++)
  {
    q_idx = vertex3->list[i];
    quadrangle = &(quadrangle_buffer[q_idx]);
    assert(quadrangle->indices[0] == vertex3->index);
    if (quadrangle->indices[1] == vertex2->index && quadrangle->indices[2] == vertex1->index && quadrangle->indices[3] == vertex0->index)
    {
      removeQuadrangleFromVertex(i, vertex3);
      deallocQuadrangle(q_idx);
      vertex0->use_count--;
      vertex1->use_count--;
      vertex2->use_count--;
      vertex3->use_count--;
      return true;
    }
  }
  return false;
}

static void addTriangleToVertex(int t, SVconnectivityVertex* v)
{
  if (v->list_size == v->list_alloc)
  {
    v->list_alloc += 4;
    v->list = (int*)realloc(v->list,sizeof(int)*v->list_alloc);
  }
  v->list[v->list_size] = t;
  v->list_size++;
}

static void create_triangle(SVconnectivityVertex* vertex0, SVconnectivityVertex* vertex1, SVconnectivityVertex* vertex2)
{
  int t_idx = allocTriangle();
  addTriangleToVertex(t_idx, vertex0);

  SVrenderTriangle* triangle = &(triangle_buffer[t_idx]);
  triangle->flag = -1;
  VecCcwNormal3fv(triangle->n, render_vertex_buffer[vertex0->index].v, render_vertex_buffer[vertex1->index].v, render_vertex_buffer[vertex2->index].v);
  triangle->indices[0] = vertex0->index;
  triangle->indices[1] = vertex1->index;
  triangle->indices[2] = vertex2->index;

  vertex0->use_count++;
  vertex1->use_count++;
  vertex2->use_count++;
}

static void addQuadrangleToVertex(int q, SVconnectivityVertex* v)
{
  if (v->list_size == v->list_alloc)
  {
    v->list_alloc += 4;
    v->list = (int*)realloc(v->list,sizeof(int)*v->list_alloc);
  }
  v->list[v->list_size] = q;
  v->list_size++;
}

static int create_quadrangle(SVconnectivityVertex* vertex0, SVconnectivityVertex* vertex1, SVconnectivityVertex* vertex2, SVconnectivityVertex* vertex3)
{
  int q_idx = allocQuadrangle();
  addQuadrangleToVertex(q_idx, vertex0);

  SVrenderQuadrangle* quadrangle = &(quadrangle_buffer[q_idx]);
  quadrangle->flag = -1;

  float n[3];
  VecCcwNormal3fv(n, render_vertex_buffer[vertex0->index].v, render_vertex_buffer[vertex1->index].v, render_vertex_buffer[vertex2->index].v);
  VecCopy3fv(quadrangle->n, n);
  VecCcwNormal3fv(n, render_vertex_buffer[vertex2->index].v, render_vertex_buffer[vertex3->index].v, render_vertex_buffer[vertex0->index].v);
  VecSelfAdd3fv(quadrangle->n, n);
  VecSelfNormalize3fv(quadrangle->n);

  quadrangle->indices[0] = vertex0->index;
  quadrangle->indices[1] = vertex1->index;
  quadrangle->indices[2] = vertex2->index;
  quadrangle->indices[3] = vertex3->index;

  vertex0->use_count++;
  vertex1->use_count++;
  vertex2->use_count++;
  vertex3->use_count++;

  return q_idx;
}

static void create_tetrahedron(SVconnectivityVertex** vertices)
{
  int tet_idx = allocTetrahedron();
  SVrenderTetrahedron* tet = tetrahedron_buffer + tet_idx;
  tet->flag = -1;
  tet->indices[0] = vertices[0]->index; vertices[0]->use_count++;
  tet->indices[1] = vertices[1]->index; vertices[1]->use_count++;
  tet->indices[2] = vertices[2]->index; vertices[2]->use_count++;
  tet->indices[3] = vertices[3]->index; vertices[3]->use_count++;

//   for (int i = 0; i < 4; i++) {
//     SVrenderVertex* v = &(render_vertex_buffer[vertices[i]->index]);
//   }
}

static void create_hexahedron(SVconnectivityVertex** vertices)
{
  int hex_idx = allocHexahedron();
  SVrenderHexahedron* hex = hexahedron_buffer + hex_idx;
  hex->flag = -1;
  hex->indices[0] = vertices[0]->index; vertices[0]->use_count++;
  hex->indices[1] = vertices[1]->index; vertices[1]->use_count++;
  hex->indices[2] = vertices[2]->index; vertices[2]->use_count++;
  hex->indices[3] = vertices[3]->index; vertices[3]->use_count++;
  hex->indices[4] = vertices[4]->index; vertices[4]->use_count++;
  hex->indices[5] = vertices[5]->index; vertices[5]->use_count++;
  hex->indices[6] = vertices[6]->index; vertices[6]->use_count++;
  hex->indices[7] = vertices[7]->index; vertices[7]->use_count++;
}

static void vizBegin()
{
  if (svreadopener->file_name)
  {
    fprintf(stderr,"loading '%s' ...\n", svreadopener->file_name);
  }

  if (svreader) // if we have a mesh reader then just re-open the file
  {
    if (!(svreadopener->reopen(svreader)))
    {
      fprintf(stderr,"ERROR: cannot reopen svreader\n");
      usage();
    }
  }
  else
  {
    if (!(svreader = svreadopener->open()))
    {
      fprintf(stderr,"ERROR: cannot open the mesh reader\n");
      usage();
    }
  }

  if (PrecomputeBoundingBoxOn)
  {
    if (svreadopener->file_name)
    {
      fprintf(stderr,"performing additional pass to pre-compute bounding box\n");

      if (!svreader->compute_bounding_box())
      {
        if (!svreadopener->reopen(svreader))
        {
          fprintf(stderr,"ERROR: additional pass to pre-compute bounding box\n");
        }
      }
    }
    else
    {
      fprintf(stderr,"WARNING: cannot perform additional pass to pre-compute bounding box on stdin input\n");
    }
  }

  if (svreader->post_order)
  {
/*
    fprintf(stderr,"WARNING: input streaming mesh is post-order\n");
    fprintf(stderr,"         using 'readPostAsCompactPre' ...\n");
    SVreadPostAsCompactPre* svreadpostascompactpre = new SVreadPostAsCompactPre();
    svreadpostascompactpre->open(svreader);
    svreader = svreadpostascompactpre;
*/
  }

  if (CurrentBoundingBoxOn || svreader->v_pmin_f == 0 || svreader->v_pmax_f == 0)
  {
    CurrentBoundingBoxOn = 1;
  }
  else
  {
    if ((svreader->v_pmax_f[1]-svreader->v_pmin_f[1]) > (svreader->v_pmax_f[0]-svreader->v_pmin_f[0]))
    {
      if ((svreader->v_pmax_f[1]-svreader->v_pmin_f[1]) > (svreader->v_pmax_f[2]-svreader->v_pmin_f[2]))
      {
        boundingBoxScale = 1.0f/(svreader->v_pmax_f[1]-svreader->v_pmin_f[1]);
      }
      else
      {
        boundingBoxScale = 1.0f/(svreader->v_pmax_f[2]-svreader->v_pmin_f[2]);
      }
    }
    else
    {
      if ((svreader->v_pmax_f[0]-svreader->v_pmin_f[0]) > (svreader->v_pmax_f[2]-svreader->v_pmin_f[2]))
      {
        boundingBoxScale = 1.0f/(svreader->v_pmax_f[0]-svreader->v_pmin_f[0]);
      }
      else
      {
        boundingBoxScale = 1.0f/(svreader->v_pmax_f[2]-svreader->v_pmin_f[2]);
      }
    }
    boundingBoxTranslateX = - boundingBoxScale * (svreader->v_pmin_f[0] + 0.5f * (svreader->v_pmax_f[0]-svreader->v_pmin_f[0]));
    boundingBoxTranslateY = - boundingBoxScale * (svreader->v_pmin_f[1] + 0.5f * (svreader->v_pmax_f[1]-svreader->v_pmin_f[1]));
    boundingBoxTranslateZ = - boundingBoxScale * (svreader->v_pmin_f[2] + 0.5f * (svreader->v_pmax_f[2]-svreader->v_pmin_f[2]));
  }

  nverts = svreader->nverts;
  ncells = svreader->ncells;

  pos_vol = 0;
  neg_vol = 0;

  fprintf(stderr,"nverts %d ncells %d\n",nverts,ncells);

  if (EVERY_NTH_STEP == -1)
  {
    if (ncells == -1)
    {
      EVERY_NTH_STEP = 10000;
    }
    else
    {
      EVERY_NTH_STEP = ncells / EXACTLY_N_STEPS;
    }
  }
  if (EVERY_NTH_STEP == 0)
  {
    EVERY_NTH_STEP = 1;
  }
  NEXT_STEP = EVERY_NTH_STEP;

  if (RUN_UNTIL)
  {
    NEXT_STEP = RUN_UNTIL;
  }

  initRenderVertexBuffer(64);
  initTriangleBuffer(64);
  initQuadrangleBuffer(64);
  initBoundaryTriangleBuffer(64);
  initBoundaryQuadrangleBuffer(64);
  initTetrahedronBuffer(64);
  initHexahedronBuffer(64);

  map_hash->clear();
  index_map_maxsize = 0;
  index_map_size = 0;
}

static void vizEnd()
{
  fprintf(stderr,"  number boundary vertices: %d\n", render_vertex_buffer_size);
  fprintf(stderr,"  number boundary triangles: %d\n", boundary_triangle_buffer_size);
  fprintf(stderr,"  pos_vol: %d neg_vol: %d\n", pos_vol, neg_vol);

  fprintf(stderr,"maxsize of index hash_map: %d (i.e. front width %2.2f %%)\n", index_map_maxsize, ((float)index_map_maxsize)*100/svreader->v_count);
  index_map_maxsize = 0;

  if (index_map_size) fprintf(stderr,"EFFICIENCY WARNING: there were %d unfinalized vertices\n", index_map_size);
  index_map_size = 0;

  svreader->close(true);
  delete svreader;
  svreader = 0;
}

int vizContinue()
{
  SVevent element = SV_ERROR;

  while (element)
  {
    element = svreader->read_element();

    if (element == SV_TETRAHEDRON)
    {
      int i;
      SVconnectivityVertex* vertices[4];
      bool finalize[4];
      my_hash::iterator hash_element;
      // get vertices from hash
      for (i = 0; i < 4; i++)
      {
        hash_element = map_hash->find(svreader->c_idx[i]);
        if (hash_element == map_hash->end())
        {
          fprintf(stderr,"FATAL ERROR: index %d not in index map hash.\n",svreader->c_idx[i]);
          fprintf(stderr,"             either the streaming mesh is not in\n");
          fprintf(stderr,"             pre-order or it is corrupt.\n");
          exit(0);
        }
        else
        {
          vertices[i] = (*hash_element).second;
        }
        if (svreader->c_final[i])
        {
          map_hash->erase(hash_element);
          index_map_size--;
          finalize[i] = true;
        }
        else
        {
          finalize[i] = false;
        }
      }

      // check the orientation of tetrahedron

      float u1[3], u2[3], u3[3], u1u2[3];
      VecSubtract3fv(u1, render_vertex_buffer[vertices[1]->index].v, render_vertex_buffer[vertices[0]->index].v);
      VecSubtract3fv(u2, render_vertex_buffer[vertices[2]->index].v, render_vertex_buffer[vertices[0]->index].v);
      VecSubtract3fv(u3, render_vertex_buffer[vertices[3]->index].v, render_vertex_buffer[vertices[0]->index].v);
      VecCrossProd3fv(u1u2, u1, u2);
      if (VecDotProd3fv(u1u2, u3) > 0.0f)
      {
        pos_vol++;
      }
      else
      {
        neg_vol++;
	if (ReorientTetrahedra)
	{
	  SVconnectivityVertex* v = vertices[2];
	  vertices[2] = vertices[3];
	  vertices[3] = v;
	  bool f = finalize[2];
	  finalize[2] = finalize[3];
	  finalize[3] = f;
	}
	else
	{
	  create_tetrahedron(vertices);
	}
      }

      if (BadInverseMeanRatio)
      {
	if (inverse_mean_ratio(render_vertex_buffer[vertices[0]->index].v, render_vertex_buffer[vertices[1]->index].v, render_vertex_buffer[vertices[2]->index].v, render_vertex_buffer[vertices[3]->index].v) > BadInverseMeanRatio)
	{
	  create_tetrahedron(vertices);
	} 
      }

      // search for existing triangles and remove them or otherwise create them
      bool removed[4];

      if (vertices[0]->use_count && vertices[1]->use_count && vertices[2]->use_count)
      {
        removed[0] = search_and_remove_triangle(vertices[0], vertices[1], vertices[2]);
      }
      else
      {
        removed[0] = false;
      }

      if (vertices[1]->use_count && vertices[0]->use_count && vertices[3]->use_count)
      {
        removed[1] = search_and_remove_triangle(vertices[1], vertices[0], vertices[3]);
      }
      else
      {
        removed[1] = false;
      }
      
      if (vertices[2]->use_count && vertices[1]->use_count && vertices[3]->use_count)
      {
        removed[2] = search_and_remove_triangle(vertices[2], vertices[1], vertices[3]);
      }
      else
      {
        removed[2] = false;
      }

      if (vertices[0]->use_count && vertices[2]->use_count && vertices[3]->use_count)
      {
        removed[3] = search_and_remove_triangle(vertices[0], vertices[2], vertices[3]);
      }
      else
      {
        removed[3] = false;
      }

      if (!removed[0]) create_triangle(vertices[0], vertices[1], vertices[2]);
      if (!removed[1]) create_triangle(vertices[1], vertices[0], vertices[3]);
      if (!removed[2]) create_triangle(vertices[2], vertices[1], vertices[3]);
      if (!removed[3]) create_triangle(vertices[0], vertices[2], vertices[3]);

      for (i = 0; i < 4; i++)
      {
        if (finalize[i])
        {
          if (vertices[i]->use_count == 0)
          {
            deallocRenderVertex(vertices[i]->index);
          }
          for (int j = 0; j < vertices[i]->list_size; j++)
          {
            int t_idx = vertices[i]->list[j];
            int bt_idx = allocBoundaryTriangle();
            memcpy(&(boundary_triangle_buffer[bt_idx]), &(triangle_buffer[t_idx]), sizeof(SVrenderTriangle));
            deallocTriangle(t_idx);
          }
          deallocConnectivityVertex(vertices[i]);
        }
      }
    }
    else if (element == SV_VERTEX)
    {
      if (CurrentBoundingBoxOn)
      {
        if (render_vertex_buffer_size)
        {
          VecUpdateMinMax3fv(currentBoundingBoxMin, currentBoundingBoxMax, svreader->v_prop_f);
        }
        else
        {
          VecCopy3fv(currentBoundingBoxMin, svreader->v_prop_f);
          VecCopy3fv(currentBoundingBoxMax, svreader->v_prop_f);
        }
      }
      // alloc render vertex
      int vertex_idx = allocRenderVertex();
      render_vertex_buffer[vertex_idx].index = svreader->v_idx;
      VecCopy3fv(render_vertex_buffer[vertex_idx].v, svreader->v_prop_f);
      // alloc connectivity vertex
      SVconnectivityVertex* vertex = allocConnectivityVertex();
      vertex->index = vertex_idx;
      // all following triangles can find the connectivity vertex in the map_hash
      map_hash->insert(my_hash::value_type(svreader->v_idx, vertex));
      index_map_size++;
      if (index_map_size > index_map_maxsize) index_map_maxsize = index_map_size;
    }
    else if (element == SV_FINALIZED)
    {
      my_hash::iterator hash_element = map_hash->find(svreader->final_idx);
      if (hash_element == map_hash->end())
      {
        fprintf(stderr,"FATAL ERROR: non-existing vertex was explicitely finalized.\n");
        fprintf(stderr,"             the streaming mesh is corrupt.\n");
        exit(0);
      }
      else
      {
        SVconnectivityVertex* vertex = (*hash_element).second;
        if (vertex->use_count == 0)
        {
          deallocRenderVertex(vertex->index);
        }
        deallocConnectivityVertex(vertex);
        map_hash->erase(hash_element);
        index_map_size--;
      }
    }
    else if (element == SV_HEXAHEDRON)
    {
      int i;
      SVconnectivityVertex* vertices[8];
      bool finalize[8];
      my_hash::iterator hash_element;
      // get vertices from hash
      for (i = 0; i < 8; i++)
      {
        hash_element = map_hash->find(svreader->c_idx[i]);
        if (hash_element == map_hash->end())
        {
          fprintf(stderr,"FATAL ERROR: index %d not in index map hash.\n",svreader->c_idx[i]);
          fprintf(stderr,"             either the streaming mesh is not in\n");
          fprintf(stderr,"             pre-order or it is corrupt.\n");
          exit(0);
        }
        else
        {
          vertices[i] = (*hash_element).second;
        }
        if (svreader->c_final[i])
        {
          map_hash->erase(hash_element);
          index_map_size--;
          finalize[i] = true;
        }
        else
        {
          finalize[i] = false;
        }
      }

      if (BadInverseMeanRatio)
      {
	if (inverse_mean_ratio(render_vertex_buffer[vertices[0]->index].v,
			       render_vertex_buffer[vertices[1]->index].v,
			       render_vertex_buffer[vertices[2]->index].v,
			       render_vertex_buffer[vertices[3]->index].v,
			       render_vertex_buffer[vertices[4]->index].v,
			       render_vertex_buffer[vertices[5]->index].v,
			       render_vertex_buffer[vertices[6]->index].v,
			       render_vertex_buffer[vertices[7]->index].v) > BadInverseMeanRatio)
	{
	  create_hexahedron(vertices);
	}
      }

      // search for existing quadtrangles and remove them or otherwise create them
      bool removed[6];

      if (vertices[0]->use_count && vertices[1]->use_count && vertices[2]->use_count && vertices[3]->use_count)
      {
        removed[0] = search_and_remove_quadrangle(vertices[0], vertices[1], vertices[2], vertices[3]);
      }
      else
      {
        removed[0] = false;
      }

      if (vertices[7]->use_count && vertices[6]->use_count && vertices[5]->use_count && vertices[4]->use_count)
      {
        removed[1] = search_and_remove_quadrangle(vertices[7], vertices[6], vertices[5], vertices[4]);
      }
      else
      {
        removed[1] = false;
      }
      
      if (vertices[4]->use_count && vertices[5]->use_count && vertices[1]->use_count && vertices[0]->use_count)
      {
        removed[2] = search_and_remove_quadrangle(vertices[4], vertices[5], vertices[1], vertices[0]);
      }
      else
      {
        removed[2] = false;
      }

      if (vertices[5]->use_count && vertices[6]->use_count && vertices[2]->use_count && vertices[1]->use_count)
      {
        removed[3] = search_and_remove_quadrangle(vertices[5], vertices[6], vertices[2], vertices[1]);
      }
      else
      {
        removed[3] = false;
      }

      if (vertices[6]->use_count && vertices[7]->use_count && vertices[3]->use_count && vertices[2]->use_count)
      {
        removed[4] = search_and_remove_quadrangle(vertices[6], vertices[7], vertices[3], vertices[2]);
      }
      else
      {
        removed[4] = false;
      }
      
      if (vertices[7]->use_count && vertices[4]->use_count && vertices[0]->use_count && vertices[3]->use_count)
      {
        removed[5] = search_and_remove_quadrangle(vertices[7], vertices[4], vertices[0], vertices[3]);
      }
      else
      {
        removed[5] = false;
      }
      
      if (!removed[0]) 
		latestAddedFaces[0] = create_quadrangle(vertices[0], vertices[1], vertices[2], vertices[3]);
      else
		latestAddedFaces[0] = -1;
      if (!removed[1]) 
		latestAddedFaces[1] = create_quadrangle(vertices[7], vertices[6], vertices[5], vertices[4]);
      else
		latestAddedFaces[1] = -1;
      if (!removed[2]) 
		latestAddedFaces[2] = create_quadrangle(vertices[4], vertices[5], vertices[1], vertices[0]);
      else
		latestAddedFaces[2] = -1;
      if (!removed[3]) 
		latestAddedFaces[3] = create_quadrangle(vertices[5], vertices[6], vertices[2], vertices[1]);
      else
		latestAddedFaces[3] = -1;
      if (!removed[4]) 
		latestAddedFaces[4] = create_quadrangle(vertices[6], vertices[7], vertices[3], vertices[2]);
      else
		latestAddedFaces[1] = -1;
      if (!removed[5]) 
		latestAddedFaces[5] = create_quadrangle(vertices[7], vertices[4], vertices[0], vertices[3]);
      else
		latestAddedFaces[5] = -1;

      for (i = 0; i < 8; i++)
      {
        if (finalize[i])
        {
          if (vertices[i]->use_count == 0)
          {
            deallocRenderVertex(vertices[i]->index);
          }
          for (int j = 0; j < vertices[i]->list_size; j++)
          {
            int q_idx = vertices[i]->list[j];
            int bq_idx = allocBoundaryQuadrangle();
            memcpy(&(boundary_quadrangle_buffer[bq_idx]), &(quadrangle_buffer[q_idx]), sizeof(SVrenderQuadrangle));
            deallocQuadrangle(q_idx);
          }
          deallocConnectivityVertex(vertices[i]);
        }
      }
    }
    else if (element == SV_EOF)
    {
      index_map_size = 0;
      map_hash->clear();
    }

    if (svreader->c_count >= NEXT_STEP)
    {
      NEXT_STEP += EVERY_NTH_STEP;
      break;
    }
  }

  if (element)
  {
    return 1;
  }
  else
  {
    return 0;
  }
}

static void myReshape(int w, int h)
{
  glutReshapeWindow(WindowW,WindowH);
}

static void myIdle()
{
  if (AnimationOn)
  {
    AnimationOn = vizContinue();
    if (!AnimationOn)
    {
      WorkingOn = 0;
      vizEnd();
    }
    glutPostRedisplay();
  }
}

static void myMouseFunc(int button, int state, int x, int y)
{
  OldX=x;
  OldY=y;    
  if (button == GLUT_LEFT_BUTTON)
  {
    LeftButtonDown = !state;
    MiddleButtonDown = 0;
    RightButtonDown = 0;
  }
  else if (button == GLUT_RIGHT_BUTTON)
  {
    LeftButtonDown = 0;
    MiddleButtonDown = 0;
    RightButtonDown = !state;
  }
  else if (button == GLUT_MIDDLE_BUTTON)
  {
    LeftButtonDown = 0;
    MiddleButtonDown = !state;
    RightButtonDown = 0;
  }
}

static void myMotionFunc(int x, int y)
{  
  float RelX = (x-OldX) / (float)glutGet((GLenum)GLUT_WINDOW_WIDTH);
  float RelY = (y-OldY) / (float)glutGet((GLenum)GLUT_WINDOW_HEIGHT);
  OldX=x;
  OldY=y;
  if (LeftButtonDown) 
  { 
    if (InteractionMode == 0)
    {
      Azimuth += (RelX*180);
      Elevation += (RelY*180);
    }
    else if (InteractionMode == 1)
    {
      DistX-=RelX;
      DistY+=RelY;
    }
    else if (InteractionMode == 2)
    {
      DistZ-=RelY*DistZ;
    }
  }
  else if (MiddleButtonDown)
  {
    DistX-=RelX*1.0f;
    DistY+=RelY*1.0f;
  }
  glutPostRedisplay();
}

void MyMenuFunc(int value);

void myKeyboard(unsigned char Key, int x, int y)
{
  FILE* hf;

  switch(Key)
  {
  case 'Q':
  case 'q':
  case 27:
    exit(0);
    break;
  case ' ': // rotate, translate, or zoom
    if (InteractionMode == 2)
    {
      InteractionMode = 0;
    }
    else
    {
      InteractionMode += 1;
    }
    glutPostRedisplay();
    break;
  case '>': //zoom in
    DistZ-=0.1f;
    break;
  case '<': //zoom out
    DistZ+=0.1f;
    break;
  case '-':
    break;
  case '=':
    break;
  case 'C':
  case 'c':
    STREAM_COLORING = (STREAM_COLORING+1)%4;
    fprintf(stderr,"STREAM_COLORING %d\n",STREAM_COLORING);
    glutPostRedisplay();
    break;
  case 'M':
  case 'm':
    RENDER_MODE = (RENDER_MODE+1)%3;
    fprintf(stderr,"RENDER_MODE %d\n",RENDER_MODE);
    glutPostRedisplay();
    break;
  case 'O':
  case 'o':
    AnimationOn = 0;
    break;
  case 'V':
  case 'v':
    pick_vertex(x,y);
    break;
  case 'B':
  case 'b':
    CurrentBoundingBoxOn = !CurrentBoundingBoxOn;
    glutPostRedisplay();
    break;
  case 'R':
  case 'r':
    break;
  case 'W':
    WireFrameOn = !WireFrameOn;
    glutPostRedisplay();
    break;
  case 'w':
    switch_colors = !switch_colors;
    glutPostRedisplay();
    break;
  case 'D':
  case 'd':
    PrintToFileOn=1;
    if (Framebuffer) delete [] Framebuffer;
    Framebuffer = new unsigned char[WindowW*WindowH*3];
    fprintf(stderr,"print_to_file %d\n",PrintToFileOn);
    break;
  case 'P':
  case 'p':
    AnimationOn = !AnimationOn;
  case 'S':
  case 's':
    if (WorkingOn == 0)
    {
      vizBegin();
      WorkingOn = vizContinue();
    }
    else
    {
      WorkingOn = vizContinue();
    }
    if (WorkingOn == 0)
    {
      vizEnd();
      AnimationOn = 0;
      PrintToFileOn = 0;
    }
    glutPostRedisplay();
    break;
  case 'T':
  case 't':
    if (WorkingOn == 0)
    {
      vizBegin();
    }
    y = NEXT_STEP;
    NEXT_STEP = 0;
    WorkingOn = vizContinue();
    NEXT_STEP = y;
    if (WorkingOn == 0)
    {
      vizEnd();
      AnimationOn = 0;
      PrintToFileOn = 0;
    }
    glutPostRedisplay();
    break;
  case 'K':
  case 'k':
    printf("-kamera %g %g %g %g %g\n", Azimuth, Elevation, DistX,DistY,DistZ);
    hf = fopen(cameraSaveFilename,"w");
    fprintf(hf, "%g %g %g %g %g\n", Azimuth, Elevation, DistX,DistY,DistZ);
    fclose(hf);
    break;
  case 'L':
  case 'l':
    hf = fopen(cameraSaveFilename,"r");
    fscanf(hf, "%g %g %g %g %g", &Azimuth, &Elevation, &DistX,&DistY,&DistZ);
    fclose(hf);
    printf("+kamera %g %g %g %g %g\n", Azimuth, Elevation, DistX,DistY,DistZ);
    glutPostRedisplay();
    break;
  case '1':
    MyMenuFunc(1);
    break;
  case '2':
    MyMenuFunc(2);
    break;
  case '3':
    MyMenuFunc(3);
    break;
  case '4':
    MyMenuFunc(4);
    break;
  case '5':
    MyMenuFunc(5);
    break;
  case '6':
    MyMenuFunc(6);
    break;
  case '7':
    MyMenuFunc(7);
    break;
  case '8':
    MyMenuFunc(8);
    break;
  case '9':
    MyMenuFunc(9);
    break;
  case '0':
    MyMenuFunc(10);
    break;
  };
}

void MyMenuFunc(int value)
{
  if (value >= 100)
  {
    if (value <= 102)
    {
      InteractionMode = value - 100;
      glutPostRedisplay();
    }
    else if (value == 103)
    {
      myKeyboard('s',0,0);
    }
    else if (value == 104)
    {
      myKeyboard('p',0,0);
    }
    else if (value == 105)
    {
      myKeyboard('o',0,0);
    }
    else if (value == 109)
    {
      myKeyboard('q',0,0);
    }
    else if (value == 151)
    {
      myKeyboard('l',0,0);
    }
    else if (value == 152)
    {
      myKeyboard('c',0,0);
    }
    else if (value == 153)
    {
      myKeyboard('m',0,0);
    }
  }
  else if (value == 40)
  {
    EXACTLY_N_STEPS = 5;
    if (ncells != -1) EVERY_NTH_STEP = ncells / EXACTLY_N_STEPS;
  }
  else if (value == 41)
  {
    EXACTLY_N_STEPS = 10;
    if (ncells != -1) EVERY_NTH_STEP = ncells / EXACTLY_N_STEPS;
  }
  else if (value == 42)
  {
    EXACTLY_N_STEPS = 25;
    if (ncells != -1) EVERY_NTH_STEP = ncells / EXACTLY_N_STEPS;
  }
  else if (value == 43)
  {
    EXACTLY_N_STEPS = 50;
    if (ncells != -1) EVERY_NTH_STEP = ncells / EXACTLY_N_STEPS;
  }
  else if (value == 44)
  {
    EXACTLY_N_STEPS = 100;
    if (ncells != -1) EVERY_NTH_STEP = ncells / EXACTLY_N_STEPS;
  }
  else if (value == 45)
  {
    EXACTLY_N_STEPS = 250;
    if (ncells != -1) EVERY_NTH_STEP = ncells / EXACTLY_N_STEPS;
  }
  else if (value == 46)
  {
    EXACTLY_N_STEPS = 500;
    if (ncells != -1) EVERY_NTH_STEP = ncells / EXACTLY_N_STEPS;
  }
  else if (value == 47)
  {
    EXACTLY_N_STEPS = 1000;
    if (ncells != -1) EVERY_NTH_STEP = ncells / EXACTLY_N_STEPS;
  }
  else if (value == 48)
  {
    NEXT_STEP -= EVERY_NTH_STEP-1;
    EXACTLY_N_STEPS = ncells;
    EVERY_NTH_STEP = 1; 
  }
}

void displayMessage( )
{
  glColor3f( 0.7f, 0.7f, 0.7f );  // Set colour to grey
  glMatrixMode( GL_PROJECTION );
  glPushMatrix();
  glLoadIdentity();
  gluOrtho2D( 0.0f, 1.0f, 0.0f, 1.0f );
  glMatrixMode( GL_MODELVIEW );
  glPushMatrix();
  glLoadIdentity();
  glRasterPos2f( 0.03f, 0.95f );
  
  if( InteractionMode == 0 )
  {
    glutBitmapCharacter(GLUT_BITMAP_8_BY_13, 'r');
    glutBitmapCharacter(GLUT_BITMAP_8_BY_13, 'o');
    glutBitmapCharacter(GLUT_BITMAP_8_BY_13, 't');
    glutBitmapCharacter(GLUT_BITMAP_8_BY_13, 'a');
    glutBitmapCharacter(GLUT_BITMAP_8_BY_13, 't');
    glutBitmapCharacter(GLUT_BITMAP_8_BY_13, 'e');
  }
  else if( InteractionMode == 1 )
  {    
    glutBitmapCharacter(GLUT_BITMAP_8_BY_13, 't');
    glutBitmapCharacter(GLUT_BITMAP_8_BY_13, 'r');
    glutBitmapCharacter(GLUT_BITMAP_8_BY_13, 'a');
    glutBitmapCharacter(GLUT_BITMAP_8_BY_13, 'n');
    glutBitmapCharacter(GLUT_BITMAP_8_BY_13, 's');
    glutBitmapCharacter(GLUT_BITMAP_8_BY_13, 'l');
    glutBitmapCharacter(GLUT_BITMAP_8_BY_13, 'a');
    glutBitmapCharacter(GLUT_BITMAP_8_BY_13, 't');
    glutBitmapCharacter(GLUT_BITMAP_8_BY_13, 'e');
  }
  else
  {
    glutBitmapCharacter(GLUT_BITMAP_8_BY_13, 'z');
    glutBitmapCharacter(GLUT_BITMAP_8_BY_13, 'o');
    glutBitmapCharacter(GLUT_BITMAP_8_BY_13, 'o');
    glutBitmapCharacter(GLUT_BITMAP_8_BY_13, 'm');
  }
  
  glPopMatrix();
  glMatrixMode( GL_PROJECTION );
  glPopMatrix();
}

void render_triangle(SVrenderTriangle* triangle)
{
  glNormal3fv(triangle->n);
  glVertex3fv(render_vertex_buffer[triangle->indices[0]].v);
  glVertex3fv(render_vertex_buffer[triangle->indices[1]].v);
  glVertex3fv(render_vertex_buffer[triangle->indices[2]].v);
}

void render_quadrangle(SVrenderQuadrangle* quadrangle)
{
  glNormal3fv(quadrangle->n);
  glVertex3fv(render_vertex_buffer[quadrangle->indices[0]].v);
  glVertex3fv(render_vertex_buffer[quadrangle->indices[1]].v);
  glVertex3fv(render_vertex_buffer[quadrangle->indices[2]].v);
  glVertex3fv(render_vertex_buffer[quadrangle->indices[3]].v);
}

void render_tetrahedron(SVrenderTetrahedron* tetrahedron)
{
  float n[3];
  VecCcwNormal3fv(n, render_vertex_buffer[tetrahedron->indices[0]].v, render_vertex_buffer[tetrahedron->indices[1]].v, render_vertex_buffer[tetrahedron->indices[2]].v);
  glNormal3fv(n);
  glVertex3fv(render_vertex_buffer[tetrahedron->indices[0]].v);
  glVertex3fv(render_vertex_buffer[tetrahedron->indices[1]].v);
  glVertex3fv(render_vertex_buffer[tetrahedron->indices[2]].v);
  VecCcwNormal3fv(n, render_vertex_buffer[tetrahedron->indices[0]].v, render_vertex_buffer[tetrahedron->indices[1]].v, render_vertex_buffer[tetrahedron->indices[3]].v);
  glNormal3fv(n);
  glVertex3fv(render_vertex_buffer[tetrahedron->indices[0]].v);
  glVertex3fv(render_vertex_buffer[tetrahedron->indices[1]].v);
  glVertex3fv(render_vertex_buffer[tetrahedron->indices[3]].v);
  VecCcwNormal3fv(n, render_vertex_buffer[tetrahedron->indices[0]].v, render_vertex_buffer[tetrahedron->indices[2]].v, render_vertex_buffer[tetrahedron->indices[3]].v);
  glNormal3fv(n);
  glVertex3fv(render_vertex_buffer[tetrahedron->indices[0]].v);
  glVertex3fv(render_vertex_buffer[tetrahedron->indices[2]].v);
  glVertex3fv(render_vertex_buffer[tetrahedron->indices[3]].v);
  VecCcwNormal3fv(n, render_vertex_buffer[tetrahedron->indices[1]].v, render_vertex_buffer[tetrahedron->indices[2]].v, render_vertex_buffer[tetrahedron->indices[3]].v);
  glNormal3fv(n);
  glVertex3fv(render_vertex_buffer[tetrahedron->indices[1]].v);
  glVertex3fv(render_vertex_buffer[tetrahedron->indices[2]].v);
  glVertex3fv(render_vertex_buffer[tetrahedron->indices[3]].v);
}


void render_hexahedron(SVrenderHexahedron* hexahedron)
{
  float n[3];
  VecCcwNormal3fv(n, render_vertex_buffer[hexahedron->indices[0]].v, render_vertex_buffer[hexahedron->indices[1]].v, render_vertex_buffer[hexahedron->indices[2]].v);
  glNormal3fv(n);
  glVertex3fv(render_vertex_buffer[hexahedron->indices[0]].v);
  glVertex3fv(render_vertex_buffer[hexahedron->indices[1]].v);
  glVertex3fv(render_vertex_buffer[hexahedron->indices[2]].v);
  glVertex3fv(render_vertex_buffer[hexahedron->indices[3]].v);
  VecCcwNormal3fv(n, render_vertex_buffer[hexahedron->indices[7]].v, render_vertex_buffer[hexahedron->indices[6]].v, render_vertex_buffer[hexahedron->indices[5]].v);
  glNormal3fv(n);
  glVertex3fv(render_vertex_buffer[hexahedron->indices[7]].v);
  glVertex3fv(render_vertex_buffer[hexahedron->indices[6]].v);
  glVertex3fv(render_vertex_buffer[hexahedron->indices[5]].v);
  glVertex3fv(render_vertex_buffer[hexahedron->indices[4]].v);
  VecCcwNormal3fv(n, render_vertex_buffer[hexahedron->indices[4]].v, render_vertex_buffer[hexahedron->indices[5]].v, render_vertex_buffer[hexahedron->indices[1]].v);
  glNormal3fv(n);
  glVertex3fv(render_vertex_buffer[hexahedron->indices[4]].v);
  glVertex3fv(render_vertex_buffer[hexahedron->indices[5]].v);
  glVertex3fv(render_vertex_buffer[hexahedron->indices[1]].v);
  glVertex3fv(render_vertex_buffer[hexahedron->indices[0]].v);
  VecCcwNormal3fv(n, render_vertex_buffer[hexahedron->indices[5]].v, render_vertex_buffer[hexahedron->indices[6]].v, render_vertex_buffer[hexahedron->indices[2]].v);
  glNormal3fv(n);
  glVertex3fv(render_vertex_buffer[hexahedron->indices[5]].v);
  glVertex3fv(render_vertex_buffer[hexahedron->indices[6]].v);
  glVertex3fv(render_vertex_buffer[hexahedron->indices[2]].v);
  glVertex3fv(render_vertex_buffer[hexahedron->indices[1]].v);
  VecCcwNormal3fv(n, render_vertex_buffer[hexahedron->indices[6]].v, render_vertex_buffer[hexahedron->indices[7]].v, render_vertex_buffer[hexahedron->indices[3]].v);
  glNormal3fv(n);
  glVertex3fv(render_vertex_buffer[hexahedron->indices[6]].v);
  glVertex3fv(render_vertex_buffer[hexahedron->indices[7]].v);
  glVertex3fv(render_vertex_buffer[hexahedron->indices[3]].v);
  glVertex3fv(render_vertex_buffer[hexahedron->indices[2]].v);
  VecCcwNormal3fv(n, render_vertex_buffer[hexahedron->indices[7]].v, render_vertex_buffer[hexahedron->indices[4]].v, render_vertex_buffer[hexahedron->indices[0]].v);
  glNormal3fv(n);
  glVertex3fv(render_vertex_buffer[hexahedron->indices[7]].v);
  glVertex3fv(render_vertex_buffer[hexahedron->indices[4]].v);
  glVertex3fv(render_vertex_buffer[hexahedron->indices[0]].v);
  glVertex3fv(render_vertex_buffer[hexahedron->indices[3]].v);
}

void myDisplay()
{
  glClearColor(1.0f, 1.0f, 1.0f, 1.0f);

  glClear(GL_COLOR_BUFFER_BIT | GL_DEPTH_BUFFER_BIT);
  
  glViewport(0,0,WindowW,WindowH);

  glMatrixMode(GL_PROJECTION);
  glLoadIdentity();
  gluPerspective(30.0f,(float)WindowW/WindowH,0.0625f,5.0f);

  glMatrixMode(GL_MODELVIEW);
  glLoadIdentity();
  gluLookAt(DistX,DistY,DistZ, DistX,DistY,0, 0,1,0);

//  glRotatef(-90,0,0,1);
  glRotatef(Elevation,1,0,0);
  glRotatef(Azimuth,0,1,0);

  if (CurrentBoundingBoxOn)
  {
    if ((currentBoundingBoxMax[1]-currentBoundingBoxMin[1]) > (currentBoundingBoxMax[0]-currentBoundingBoxMin[0]))
    {
      if ((currentBoundingBoxMax[1]-currentBoundingBoxMin[1]) > (currentBoundingBoxMax[2]-currentBoundingBoxMin[2]))
      {
        currentBoundingBoxScale = 1.0f/(currentBoundingBoxMax[1]-currentBoundingBoxMin[1]);
      }
      else
      {
        currentBoundingBoxScale = 1.0f/(currentBoundingBoxMax[2]-currentBoundingBoxMin[2]);
      }
    }
    else
    {
      if ((currentBoundingBoxMax[0]-currentBoundingBoxMin[0]) > (currentBoundingBoxMax[2]-currentBoundingBoxMin[2]))
      {
        currentBoundingBoxScale = 1.0f/(currentBoundingBoxMax[0]-currentBoundingBoxMin[0]);
      }
      else
      {
        currentBoundingBoxScale = 1.0f/(currentBoundingBoxMax[2]-currentBoundingBoxMin[2]);
      }
    }
    currentBoundingBoxTranslateX = - currentBoundingBoxScale * (currentBoundingBoxMin[0] + 0.5f * (currentBoundingBoxMax[0]-currentBoundingBoxMin[0]));
    currentBoundingBoxTranslateY = - currentBoundingBoxScale * (currentBoundingBoxMin[1] + 0.5f * (currentBoundingBoxMax[1]-currentBoundingBoxMin[1]));
    currentBoundingBoxTranslateZ = - currentBoundingBoxScale * (currentBoundingBoxMin[2] + 0.5f * (currentBoundingBoxMax[2]-currentBoundingBoxMin[2]));

    glTranslatef(currentBoundingBoxTranslateX,currentBoundingBoxTranslateY,currentBoundingBoxTranslateZ);
    glScalef(currentBoundingBoxScale,currentBoundingBoxScale,currentBoundingBoxScale);
  }
  else
  {
    glTranslatef(boundingBoxTranslateX,boundingBoxTranslateY,boundingBoxTranslateZ);
    glScalef(boundingBoxScale,boundingBoxScale,boundingBoxScale);
  }

  glEnable(GL_DEPTH_TEST);
  glEnable(GL_LIGHTING);
  glEnable(GL_LIGHT0);
  glEnable(GL_NORMALIZE);

  if (RENDER_MODE == 0)
  {
    glPolygonMode(GL_FRONT_AND_BACK, GL_FILL);
  }
  else if (RENDER_MODE == 1)
  {
    glPolygonMode(GL_FRONT_AND_BACK, GL_LINE);
  }
  else if (RENDER_MODE == 2)
  {
    glPolygonMode(GL_FRONT_AND_BACK, GL_POINT);
  }

  if (switch_colors)
  {
    glMaterialfv(GL_FRONT, GL_DIFFUSE, colours_diffuse[7]);
    glMaterialfv(GL_BACK, GL_DIFFUSE, colours_diffuse[8]);
  }
  else
  {
    glMaterialfv(GL_FRONT, GL_DIFFUSE, colours_diffuse[8]);
    glMaterialfv(GL_BACK, GL_DIFFUSE, colours_diffuse[7]);
  }

  // draw interior triangles

  if (triangle_buffer_size)
  {
    int i = 0;
    int t = 0;
    glBegin(GL_TRIANGLES);
    while (t < triangle_buffer_size)
    {
      if (triangle_buffer[i].flag < 0)
      {
        render_triangle(&(triangle_buffer[i]));
        t++;
      }
      i++;
    }
    glEnd();
  }

  // draw interior quadrangles

  if (quadrangle_buffer_size)
  {
    int i = 0;
    int q = 0;
    glBegin(GL_QUADS);
    while (q < quadrangle_buffer_size)
    {
      if (quadrangle_buffer[i].flag < 0)
      {
	if( (i == latestAddedFaces[0]) || (i == latestAddedFaces[1]) ||(i == latestAddedFaces[2]) ||(i == latestAddedFaces[3]) ||(i == latestAddedFaces[4]) ||(i == latestAddedFaces[5]))
	{
		glMaterialfv(GL_FRONT_AND_BACK, GL_DIFFUSE, colours_diffuse[1]);
		render_quadrangle(&(quadrangle_buffer[i]));
		if (switch_colors)
		{
			glMaterialfv(GL_FRONT, GL_DIFFUSE, colours_diffuse[7]);
			glMaterialfv(GL_BACK, GL_DIFFUSE, colours_diffuse[8]);
		}
		else
		{
			glMaterialfv(GL_FRONT, GL_DIFFUSE, colours_diffuse[8]);
			glMaterialfv(GL_BACK, GL_DIFFUSE, colours_diffuse[7]);
		}
	}
	else
        	render_quadrangle(&(quadrangle_buffer[i]));
        q++;
      }
      i++;
    }
    glEnd();
  }

  if (switch_colors)
  {
    glMaterialfv(GL_FRONT, GL_DIFFUSE, colours_white);
    glMaterialfv(GL_BACK, GL_DIFFUSE, colours_light_blue);
  }
  else
  {
    glMaterialfv(GL_FRONT, GL_DIFFUSE, colours_light_blue);
    glMaterialfv(GL_BACK, GL_DIFFUSE, colours_white);
  }

  // draw boundary triangles

  if (boundary_triangle_buffer_size)
  {
    STRIPE_WHITE = 0;
    glBegin(GL_TRIANGLES);
    for (int i = 0; i < boundary_triangle_buffer_size; i++)
    {
      if (STREAM_COLORING)
      {
        if (STREAM_COLORING == 1)
        {
          if ((i % (boundary_triangle_buffer_size/NUMBER_STRIPES)) == 0)
          {
            if (STRIPE_WHITE)
            {
              glMaterialfv(GL_FRONT_AND_BACK, GL_DIFFUSE,colours_stripe_white);
            }
            else
            {
              glMaterialfv(GL_FRONT_AND_BACK, GL_DIFFUSE, colours_stripe_black);
            }
            STRIPE_WHITE = !STRIPE_WHITE;
          }
        }
        else if (STREAM_COLORING == 2)
        {
          colours_slow_change[0] = 0.1f+0.8f*i/boundary_triangle_buffer_size;
          colours_slow_change[1] = colours_slow_change[0];
          colours_slow_change[2] = colours_slow_change[0];
          glMaterialfv(GL_FRONT_AND_BACK, GL_DIFFUSE, colours_slow_change);
        }
        else if (STREAM_COLORING == 3)
        {
          if (i < boundary_triangle_buffer_size/3)
          {
            colours_slow_change[0] = 0.1f+0.8f*i/(boundary_triangle_buffer_size/3);
            colours_slow_change[1] = 0.1f;
            colours_slow_change[2] = 0.1f;
          }
          else if (i < 2*(boundary_triangle_buffer_size/3))
          {
            colours_slow_change[0] = 0.9f;
            colours_slow_change[1] = 0.1f+0.8f*(i-(boundary_triangle_buffer_size/3))/(boundary_triangle_buffer_size/3);
            colours_slow_change[2] = 0.1f;
          }
          else
          {
            colours_slow_change[0] = 0.9f;
            colours_slow_change[1] = 0.9f;
            colours_slow_change[2] = 0.1f+0.8f*(i-2*(boundary_triangle_buffer_size/3))/(boundary_triangle_buffer_size/3);
          }
          glMaterialfv(GL_FRONT_AND_BACK, GL_DIFFUSE, colours_slow_change);
        }
      }
      render_triangle(&(boundary_triangle_buffer[i]));
    }
    glEnd();
  }

  // draw boundary quadrangles

  if (boundary_quadrangle_buffer_size)
  {
    STRIPE_WHITE = 0;
    glBegin(GL_QUADS);
    for (int i = 0; i < boundary_quadrangle_buffer_size; i++)
    {
      if (STREAM_COLORING)
      {
        if (STREAM_COLORING == 1)
        {
          if ((i % (boundary_quadrangle_buffer_size/NUMBER_STRIPES)) == 0)
          {
            if (STRIPE_WHITE)
            {
              glMaterialfv(GL_FRONT_AND_BACK, GL_DIFFUSE,colours_stripe_white);
            }
            else
            {
              glMaterialfv(GL_FRONT_AND_BACK, GL_DIFFUSE, colours_stripe_black);
            }
            STRIPE_WHITE = !STRIPE_WHITE;
          }
        }
        else if (STREAM_COLORING == 2)
        {
          colours_slow_change[0] = 0.1f+0.8f*i/boundary_quadrangle_buffer_size;
          colours_slow_change[1] = colours_slow_change[0];
          colours_slow_change[2] = colours_slow_change[0];
          glMaterialfv(GL_FRONT_AND_BACK, GL_DIFFUSE, colours_slow_change);
        }
        else if (STREAM_COLORING == 3)
        {
          if (i < boundary_quadrangle_buffer_size/3)
          {
            colours_slow_change[0] = 0.1f+0.8f*i/(boundary_quadrangle_buffer_size/3);
            colours_slow_change[1] = 0.1f;
            colours_slow_change[2] = 0.1f;
          }
          else if (i < 2*(boundary_quadrangle_buffer_size/3))
          {
            colours_slow_change[0] = 0.9f;
            colours_slow_change[1] = 0.1f+0.8f*(i-(boundary_quadrangle_buffer_size/3))/(boundary_quadrangle_buffer_size/3);
            colours_slow_change[2] = 0.1f;
          }
          else
          {
            colours_slow_change[0] = 0.9f;
            colours_slow_change[1] = 0.9f;
            colours_slow_change[2] = 0.1f+0.8f*(i-2*(boundary_quadrangle_buffer_size/3))/(boundary_quadrangle_buffer_size/3);
          }
          glMaterialfv(GL_FRONT_AND_BACK, GL_DIFFUSE, colours_slow_change);
        }
      }
      render_quadrangle(&(boundary_quadrangle_buffer[i]));
    }
    glEnd();
  }

  // draw special tetrahedra

  if (tetrahedron_buffer_size)
  {
    int i = 0;
    int t = 0;
    glPolygonMode(GL_FRONT_AND_BACK, GL_FILL);
    glMaterialfv(GL_FRONT_AND_BACK, GL_DIFFUSE,colours_diffuse[5]);
    glBegin(GL_TRIANGLES);
    while (t < tetrahedron_buffer_size)
    {
      if (tetrahedron_buffer[i].flag < 0)
      {
	render_tetrahedron(&(tetrahedron_buffer[i]));
	t++;
      }
      i++;
    }
    glEnd();
  }

  // draw special hexahedra

  if (hexahedron_buffer_size)
  {
    int i = 0;
    int q = 0;
    glPolygonMode(GL_FRONT_AND_BACK, GL_FILL);
    glMaterialfv(GL_FRONT_AND_BACK, GL_DIFFUSE,colours_diffuse[5]);
    glBegin(GL_QUADS);
    while (q < hexahedron_buffer_size)
    {
      if (hexahedron_buffer[i].flag < 0)
      {
	render_hexahedron(&(hexahedron_buffer[i]));
	q++;
      }
      i++;
    }
    glEnd();
  }

  glDisable(GL_LIGHTING);
  glDisable(GL_LIGHT0);
  glDisable(GL_NORMALIZE);

  if (WireFrameOn)
  {
    glColor3f(1.0f,1.0f,1.0f);
    glLineWidth(2.0f);
    glPolygonMode(GL_FRONT_AND_BACK, GL_LINE);
    glBegin(GL_TRIANGLES);
    int i = 0;
    int t = 0;
    while (t < triangle_buffer_size)
    {
      if (triangle_buffer[i].flag < 0)
      {
        render_triangle(&(triangle_buffer[i]));
        t++;
      }
      i++;
    }
    i = 0;
    t = 0;
    while (t < boundary_triangle_buffer_size)
    {
      if (boundary_triangle_buffer[i].flag < 0)
      {
        render_triangle(&(boundary_triangle_buffer[i]));
        t++;
      }
      i++;
    }
    glEnd();
    glBegin(GL_QUADS);
    i = 0;
    t = 0;
    while (t < quadrangle_buffer_size)
    {
      if (quadrangle_buffer[i].flag < 0)
      {
        render_quadrangle(&(quadrangle_buffer[i]));
        t++;
      }
      i++;
    }
    i = 0;
    t = 0;
    while (t < boundary_quadrangle_buffer_size)
    {
      if (boundary_quadrangle_buffer[i].flag < 0)
      {
        render_quadrangle(&(boundary_quadrangle_buffer[i]));
        t++;
      }
      i++;
    }
    glEnd();
    glLineWidth(1.0f);
  }
  glDisable(GL_DEPTH_TEST);
  
  if (!PrintToFileOn)
  {
    displayMessage();
  }
  else
  {
    char FileName[256], Command[256];
    sprintf(FileName, "./temp.ppm");
    glPixelStorei(GL_PACK_ALIGNMENT, 1);
    #ifdef _WIN32
    glReadPixels( 0, 0, WindowW, WindowH, GL_RGB, GL_UNSIGNED_BYTE, Framebuffer );
    #else
    glReadPixels( (1024-WindowW)/2, (768-WindowH)/2, WindowW, WindowH, GL_RGB, GL_UNSIGNED_BYTE, Framebuffer );
    #endif
    SavePPM(FileName, Framebuffer, WindowW, WindowH);
    #ifdef _WIN32
      sprintf(Command, "i_view32.exe temp.ppm /convert=%s%d%d%d%d.jpg",PrintFileName,((Time/1000)%10),((Time/100)%10),((Time/10)%10),(Time%10));
    #else
      sprintf(Command, "convert ./temp.ppm ./%s%d%d%d%d.png",PrintFileName,((Time/1000)%10),((Time/100)%10),((Time/10)%10),(Time%10));
//      sprintf(Command, "imgcopy -q%d ./temp.ppm ./pics/%s%d%d%d%d.jpg",PrintQuality,PrintFileName,((Time/1000)%10),((Time/100)%10),((Time/10)%10),(Time%10));
    #endif
    printf("performing: '%s'\n", Command);
    system(Command);
    Time++;
  }
  glutSwapBuffers();
}

int main(int argc, char *argv[])
{
  int i;

  glutInit(&argc, argv);

  // first look over the command line arguments for input
  for (i = 1; i < argc; i++)
  {
    if (strncmp(argv[i],"-i",2) == 0)
    {
      if (svreadopener == 0)
      {
        // apparently there is a point reader to be opened
        svreadopener = new SVreadOpener();
      }
      if (strcmp(argv[i],"-i") == 0) // the commandline '-i' followed by the input filename
      {
        i++;
        svreadopener->set_file_name(argv[i]);
      }
      else // the commandline '-ixxx' specifies the input format 'xxx'
      {
        svreadopener->set_file_format(&(argv[i][2]));
      }
    }
  }
  // second look over the command line arguments for options
  for (i = 1; i < argc; i++)
  {
    if (strncmp(argv[i],"-i",2) == 0)
    {
      if (strcmp(argv[i],"-i") == 0) i++;
      // already processed
    }
    else if (strcmp(argv[1],"-h") == 0 || strcmp(argv[1],"-help") == 0) 
    {
      usage();
    }
    else if (strcmp(argv[i],"-win") == 0)
    {
      i++;
      WindowW = atoi(argv[i]);
      i++;
      WindowH = atoi(argv[i]);
    }
    else if (strcmp(argv[i],"-every") == 0)
    {
      i++;
      EVERY_NTH_STEP = atoi(argv[i]);
    }
    else if (strcmp(argv[i],"-until") == 0)
    {
      i++;
      RUN_UNTIL = atoi(argv[i]);;
    }
    else if (strcmp(argv[i],"-steps") == 0)
    {
      i++;
      EXACTLY_N_STEPS = atoi(argv[i]);
    }
    else if (strcmp(argv[i],"-precomp_bb") == 0)
    {
      PrecomputeBoundingBoxOn=1;
    }
    else if (strcmp(argv[i],"-current_bb") == 0)
    {
      CurrentBoundingBoxOn=1;
    }
    else if (strcmp(argv[i],"-bad_imr") == 0)
    {
      i++;
      BadInverseMeanRatio = atof(argv[i]);
    }
    else if (strcmp(argv[i],"-reorient") == 0)
    {
      ReorientTetrahedra = true;
    }
    else if (strcmp(argv[i],"-kamera") == 0)
    {
      i++;
      sscanf(argv[i], "%f", &Azimuth);
      i++;
      sscanf(argv[i], "%f", &Elevation);
      i++;
      sscanf(argv[i], "%f", &DistX);
      i++;
      sscanf(argv[i], "%f", &DistY);
      i++;
      sscanf(argv[i], "%f", &DistZ);
    }
    else if ((i == argc-1) && svreadopener == 0)
    {
      // last argument unknown and no input specified yet ... try it
      svreadopener = new SVreadOpener();
      svreadopener->set_file_name(argv[i]);
    }
    else
    {
      usage();
    }
  }

  // if we had no input try using this as the default input
  if (svreadopener == 0)
  {
    svreadopener = new SVreadOpener();
    svreadopener->set_file_name("../data/torso_z.svb.gz");
  }

  map_hash = new my_hash;

  glutInitDisplayMode(GLUT_DOUBLE | GLUT_RGB | GLUT_DEPTH);
  glutInitWindowSize(WindowW,WindowH);
  glutInitWindowPosition(0,0);
  glutCreateWindow("Streaming Volume Mesh Viewer (sv_viewer.exe)");
  
  glShadeModel(GL_FLAT);
  
  InitColors();
  InitLight();
  
  glutDisplayFunc(myDisplay);
  glutReshapeFunc(myReshape);
  glutIdleFunc(myIdle);
  
  glutMouseFunc(myMouseFunc);
  glutMotionFunc(myMotionFunc);
  glutKeyboardFunc(myKeyboard);
    
  // steps sub menu
  int menuSteps = glutCreateMenu(MyMenuFunc);
  glutAddMenuEntry("in 5 steps", 40);
  glutAddMenuEntry("in 10 steps", 41);
  glutAddMenuEntry("in 25 steps", 42);
  glutAddMenuEntry("in 50 steps", 43);
  glutAddMenuEntry("in 100 steps", 44);
  glutAddMenuEntry("in 250 steps", 45);
  glutAddMenuEntry("in 500 steps", 46);
  glutAddMenuEntry("in 1000 steps", 47);
  glutAddMenuEntry("in mesh size steps", 48);

  // main menu
  glutCreateMenu(MyMenuFunc);
  glutAddSubMenu("steps ...", menuSteps);
  glutAddMenuEntry("", 0);
  glutAddMenuEntry("rotate <SPACE>", 100);
  glutAddMenuEntry("translate <SPACE>", 101);
  glutAddMenuEntry("zoom <SPACE>", 102);
  glutAddMenuEntry("", 0);
  glutAddMenuEntry("<s>tep", 103);
  glutAddMenuEntry("<p>lay", 104);
  glutAddMenuEntry("st<o>p", 105);
  glutAddMenuEntry("", 0);
  glutAddMenuEntry("stream <c>oloring", 152);
  glutAddMenuEntry("render <m>ode", 153);
  glutAddMenuEntry("", 0);
  glutAddMenuEntry("<Q>UIT", 109);
  glutAttachMenu(GLUT_RIGHT_BUTTON);

  if (svreadopener->file_name == 0) myKeyboard('p',0,0);

  glutMainLoop();

  return 0;
}
