/*
===============================================================================

  FILE:  rangefloatcompressor.h
  
  CONTENTS:
  
    This floating point compressor can:

    * compress floating point numbers in a lossless manner while exploiting
      the benefits of predictive coding (set precision to be 0)

    * uniformly quantize and compress floating point numbers without a-priori
      informtion about the bounding box while guaranteeing k bits of precision
  
  PROGRAMMERS:
  
    martin isenburg@cs.unc.edu
  
  COPYRIGHT:
  
    Copyright (C) 2000  Martin Isenburg (isenburg@cs.unc.edu)
    
    This software is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
  
  CHANGE HISTORY:
  
    20 January 2010 -- completely re-done the way contexts are handled
    30 September 2003 -- created initial version after Ajith's good-bye lunch
  
===============================================================================
*/
#ifndef RANGE_FLOAT_COMPRESSOR_H
#define RANGE_FLOAT_COMPRESSOR_H

#include "rangeencoder.h"
#include "rangedecoder.h"

class RangeFloatCompressor
{
public:

  // SetPrecision:
  inline void SetPrecision(U32 bits) {this->bits = bits;};
  // GetPrecision:
  inline U32 GetPrecision() const {return bits;};

  // SetupCompressor:
  void SetupCompressor(RangeEncoder* re, U32 contexts=1);
  void FinishCompressor();

  // Compress:
  F32 Compress(F32 fPred, F32 fReal, U32 context=0);

  // Decompress:
  void SetupDecompressor(RangeDecoder* rd, U32 contexts=1);
  void FinishDecompressor();

  // Compress:
  F32 Decompress(F32 fPred, U32 context=0);

  // Constructor:
  RangeFloatCompressor();
  // Destructor:
  ~RangeFloatCompressor();

private:
  // Private Functions
  void alloc_range_tables();
  void dealloc_range_tables();

  void reset_precision();
  void update_precision(F32 number);

  void compress_exponent(I32 exponentPred, I32 exponentReal, RangeEncoder* re_exponent, RangeModel** rmExponent);
  I32 decompress_exponent(I32 exponentPred, RangeDecoder* rd_exponent, RangeModel** rmExponent);

  void compress_sign(I32 exponent, I32 sign, RangeEncoder* re_sign, RangeModel** rmSign);
  I32 decompress_sign(I32 exponent, RangeDecoder* rd_sign, RangeModel** rmSign);

  I32 compress_mantissa(I32 exponent, I32 mantissaPred, I32 mantissaReal, RangeEncoder* re_mantissa, RangeModel** rmMantissaHigh, RangeModel** rmMantissaLow);
  I32 decompress_mantissa(I32 exponent, I32 mantissaPred, RangeDecoder* rd_mantissa, RangeModel** rmMantissaHigh, RangeModel** rmMantissaLow);

  // Private Variables
  U32 contexts;

  RangeEncoder* re;
  RangeDecoder* rd;

  RangeModel*** rmSign;
  RangeModel*** rmExponent;
  RangeModel*** rmMantissaLow;
  RangeModel*** rmMantissaHigh;

  RangeModel** rmZeroOrExtremeValue;
  RangeModel** rmPositiveOrNegativeValue;

  I32* bits_for_mantissa;
  I32 range_exponent;
  F32 min_float;
  F32 max_float;
  U32 bits;
};

#endif
